package ExSite::Attribute;
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

# ExSite::Attribute - a general-purpose meta-data tool for databases
#
# Allows you to define attributes (name-value pairs) and link them 
# to any ExSite database record.  All attributes are stored in their
# own table.
#
# This allows the end user to extend the effective sets of values 
# in a record, without modifying the database schema.  A common
# usage is to add customized membership information.

### TODO
#
# attribute datatypes
# autogenerate attribute input tags
# autogenerate a form with all attributes
# bulk update of attributes

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Base;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

# create an attribute object, and tie it to a record

sub new {
    my ($this,$table,$id) = @_;
    return undef if (! $table || ! $id);
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    if (! $table) { return undef; }
    my $key = $share{DB}->get_key($table);
    if (! $key) { return undef; }
    $obj->{table} = $table;
    $obj->{key} = $key;
    $obj->{id} = $id;
    $obj->{attributes} = {};
    return $obj;
}

# change to a new record

sub id {
    my ($this,$id) = @_;
    $this->{attributes} = {};
    $this->{id} = $id;
}

# fetch the current attribute record

sub fetch {
    my ($this,$name) = @_;
    my @attr = $share{DB}->fetch_match("attribute",{name=>$name,
						    tablename=>$this->{table},
						    id=>$this->{id}},
				       "attribute_id");
    return $attr[-1];
}

# fetch a particular attribute value

sub get {
    my ($this,$name) = @_;
    if (exists $this->{attributes}{$name}) {
	return $this->{attributes}{$name};
    }
    return $this->fetch($name)->{value};
}

# set/save an attribute value

sub set {
    my ($this,$name,$value) = @_;
    my @allowed = $this->get_allowed();
    return 0 if (@allowed > 0 && ! grep /^$name$/, @allowed);
    $this->{attributes}{$name} = $value;

    # check for existing attribute value
    my $attr = $this->fetch($name);
    if ($attr) {
	# already exists, update it
	$share{DB}->update("attribute",
			   {value=>$value},
			   {attribute_id=>$attr->{attribute_id}});
	return $attr->{attribute_id};
    }
    else {
	# no such attribute, add it
	return $share{DB}->insert("attribute",{name=>$name,
					       value=>$value,
					       tablename=>$this->{table},
					       id=>$this->{id}},
				  );
    }
}

# set an attribute value temporarily (do not save)

sub set_nosave {
    my ($this,$name,$value) = @_;
    my @allowed = $this->get_allowed();
    return 0 if (@allowed > 0 && ! grep /^$name$/, @allowed);
    $this->{attributes}{$name} = $value;
    return 1;
}

# get all attributes for this record

sub get_all {
    my ($this) = @_;
    my @attr = $share{DB}->fetch_match("attribute",{tablename=>$this->{table},
						    id=>$this->{id}},
				       "attribute_id");
    # convert to a single attribute hash
    my %attr;
    foreach my $a (@attr) {
	$attr{$a->{name}} = $a->{value};
    }
    return wantarray ? %attr : \%attr;
}

# restore previous attribute value

sub rollback {
    my ($this,$name) = @_;
    my @attr = $share{DB}->fetch_match("attribute",{name=>$name,
						    tablename=>$this->{table},
						    id=>$this->{id}},
				       "attribute_id");
    if (@attr > 0) {
	$share{DB}->trash_key("attribute",$attr[-1]{attribute_id});
	# reset the attribute
	if (@attr == 1) {
	    $this->{attributes}{$name} = undef;
	}
	else {
	    $this->{attributes}{$name} = $attr[-2]{value};
	}
	return 1;
    }
    return 0;
}

# remove all attribute values

# FIXME: trash these instead of delete them.

sub clear {
    my ($this,$name) = @_;
    if ($name) { 
	$share{DB}->delete("attribute",{name=>$name,
				       tablename=>$this->{table},
				       id=>$this->{id}});
    }
    else {
	# delete all attributes
	$share{DB}->delete("attribute",{tablename=>$this->{table},
				       id=>$this->{id}});
    }
}

# remove previous attribute settings

sub purge {
    my ($this,$name) = @_;
    if ($name) { 
	my @attr = $share{DB}->fetch_match("attribute",{name=>$name,
							tablename=>$this->{table},
							id=>$this->{id}},
					   "attribute_id");
	pop @attr;  # do not remove last one
	foreach my $attr (@attr) {
	    $share{DB}->trash_key("attribute",$attr->{attribute_id});
	}
    }
    # no option to purge all
}

# get a list of allowed attribute names
# (allowed names are those in the attribute table that have no record ID)
# datatypes are stored in $this->{datatype}

sub get_allowed {
    my ($this) = @_;
    if ($this->{allowed}) { return @{$this->{allowed}}; }
    my @attr = $share{DB}->fetch_match("attribute",{id=>0,tablename=>$this->{table}});
    # convert to attribute name list
    my @name;
    foreach my $a (@attr) {
	push @name, $a->{name};
	$this->{datatype}{$a->{name}} = $a->{value};
    }
    $this->{allowed} = \@name;    
    return @name;
}

# get the preferred datatype for an attribute

sub get_datatype {
    my ($this,$name) = @_;
    my @attr = $share{DB}->fetch_match("attribute",{id=>0,
						    tablename=>$this->{table},
						    name=>$name, });
    return $attr[-1]->{value};
}

# add an allowed attribute (admin users only)

sub add_allowed {
    my ($this,$name,$datatype) = @_;
    if ($share{DB}->is_admin) {
	$share{DB}->insert("attribute",{name=>$name,
					value=>$datatype,
					tablename=>$this->{table},
					id=>$this->{id}},
			   );
	push @{$this->{allowed}},$name;
	return 1;
    }
    else {
	$this->error("ExSite::Attribute::add_allowed: permission denied");
    }
    return 0;
}

sub input {
    my ($this,$input_name,$name,$value) = @_;
    if (grep /^$name$/,$this->get_allowed) {
	return $share{DB}->input_exsite(datatype=>$this->get_datatype($name),
					name=>($input_name || $name),
					value=>($value || $this->get($name)));
    }
    else {
	return $share{DB}->error("Cannot generate input field for invalid attribute \"$name\"");
    }
}

1;


