#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Auth;

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Base;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Input;
use ExSite::Mail;
use ExSite::Crypt;
use Digest::MD5 qw(md5_hex);

our @ISA = qw(ExSite::Base);


#========================================================================

=pod

=head1 ExSite::Auth -- authentication and security tools

Authorization is performed before any standard operation on
the database, when the operation is performed by one of ExSite::DB's
descendant classes.  The authorization process has the following
steps:

=over 4

=item 1. identify - determine who the user claims to be

=item 2. lookup - determine whom we understand that username to represent

=item 3. authenticate - determine if the user's claim matches our knowledge

=item 4. authorize - grant a certain level of access to the database

=item 5. approve - approve a specific operation

=back

Each of the steps in the authorization process may be replaced with
a custom handler if the procedure differs from the default ExSite logic.

=head2 identify() - determine who the user claims to be

Determines the user's claimed identity (such as by checking the
identity cookie), and stores the result so that authorization methods
can check it.  This routine does not authenticate the user's identity
claim, and so is usually not called directly.

=head2 lookup() - look up the user to see if we know who they are

Consult our own database to see if we have a record corresponding to
this user's identity (determined from C<identify()>).  Returns the
user record in a hash, if found.  The default logic for looking up a
user record is to look up the login name, which implies that unique
login names are required.  Once this lookup has occurred, the
following identity query functions can be used:

=head2 my_name() - return the login name of the current user

=head2 my_uid() - return the primary key of the current user's record

=head2 my_gid() - return the primary key of the current user's group record

=head2 my_user_record() - return the user's identity hash

=head2 authenticate() - verify the user's identity claim

Returns true (1) if the user's identity matches the information
returned by lookup(), false (0) otherwise.  This serves as a simple
check that the user is logged in, and that their identity is valid.

=head2 authorize() - grant the user a specific level of access

Returns the user's access level.  Access level is a value of 0 or
higher.  0 corresponds to "nobody", or the general public.  Values of
1 or more correspond to authenticated users, with higher values giving
greater permissions to the database.  The maximum value is
C<$config{auth}-E<gt>{root_auth_level}>, which grants unrestricted
"super-user" access to the data.  (This level of access can be
dangerous, as it allows one to mess with sensitive data fields and
referential integrity.)

The default authorize routine assumes that the user's access level
is stored in the user table.  This value is simply returned without
any sanity checks.

=head2 approve() - approve a specific database operation

Usage: C<$db-E<gt>approve( OP, TABLE, RECORD );>

OP is a low-level database operation (one of "read", "update",
"insert", or "delete").  TABLE is the table that this operation is
being performed on. RECORD is the actual record datahash that this
operation is being performed on (only used for "update" and "delete").

The return value is true if the operation is permitted, or false if:

=over 4

=item * the user does not have permission to read or write to the table

=item * the user is referencing records owned by a different user or group.

=back

=cut 

#----------------------------------------------------------------------
# identify : determine whom the user claims to be

sub identify {
    my ($this) = shift;
    my ($name,$pwd,$uid,$gid);
    my $stat = $this->run_handler("identify");
    my $authtype = $config{auth}{type};
    if (defined $stat) {
	# application has its own identify algorithm
	($name,$pwd) = @$stat;
    }
    elsif ($authtype eq "cookie") {
	($name,$pwd) = &get_cookie($config{site}{identity_cookie},":",2);
    }
    elsif ($authtype eq "fixed") {
	$name = $config{auth}->{default_identity}->{name};
    }
    elsif ($authtype eq "session") {
	# User identity claim is implicit in the session ID.
	# Therefore this call is kind of meaningless in the case of sessions.
	# Let's return the session ID, and see what happens.
	my $session_cookie = $config{site}->{id}."_sessionID";
	return &get_cookie($session_cookie);
    }
    elsif ($authtype eq "none") {
	$name = "nobody";
    }
    $share{identity} = { password => $pwd };
    $share{identity}{name} = $name;
    return $name;
}

#----------------------------------------------------------------------
# lookup : look up the user, to see if we know who they are

sub lookup {
    my ($this,$login) = @_;
    if ($share{identity}{lookup}) {
	return $share{identity}{lookup};
    }
    # check for custom lookup algorithm
    my $stat = $this->run_handler("lookup");
    return $stat if (defined $stat);
    #
    return undef if ($this->{authentication} eq "none");
    my $udata = $share{DB}->fetch_match($config{auth}->{user_table},
			    {$config{auth}->{user_login_column}=>$login});
    if (@$udata > 0) {
	$share{identity}{lookup} = $udata->[0];
	$share{identity}{uid} = $udata->[0]{$config{auth}{user_uid_column}};
	$share{identity}{gid} = $udata->[0]{$config{auth}{user_gid_column}};
	$share{identity}{access} = $udata->[0]{$config{auth}{user_access_column}};
	return $udata->[0];
    }
    return undef;
}

#----------------------------------------------------------------------
# authenticate : verify the user's identity claim
#
# Returns true (1) if the user's identity matches the information
# returned by lookup(), false (0) otherwise.

sub authenticate {
    my ($this) = shift;
    if (exists $this->{authenticated}) {
	return $this->{authenticated};
    }
    # check for custom authentication algorithm
    my $stat = $this->run_handler("authenticate");
    return $stat if (defined $stat);
    
    if ($config{auth}{type} eq "none") {
	$this->{authenticated} = 1;
	return 1;
    }
    elsif ($config{auth}{type} eq "fixed") {
	$this->{authenticated} = 1;
	$share{identity} = $config{auth}{default_identity};
	return 1;
    }
    elsif ($config{auth}{type} eq "session") {
	if (exists $session{Auth_identity}) {
	    # we authenticated on a previous visit
	    # copy identity info from session
	    if (! $share{identity}) {
		my ($name,$uid,$gid,$access) = 
		    split /,/,$session{Auth_identity};
		$this->info("restoring identity \"$name\" from session");
		$share{identity} = {
		    name   => $name,
		    uid    => $uid,
		    gid    => $gid,
		    access => $access,
		};
	    }
	    $this->{authenticated} = 1;
	    return 1;
	}
	else {
	    return 0;
	}
    }
    else {   # this->{authentication} eq "cookie"

	# check identity
	if (!$share{identity}) { &identify($this); }
	if (!$share{identity}{name}) {
	    # nobody logged in - not necessarily an error
	    return 0;
	}
	# look up the user
	$this->info("authenticating user \"$this->{identity}->{name}\"");
	my $user = &lookup($this,$share{identity}{name});
	if (!$user) {
	    # unknown username
	    $this->warn("no such user \"$this->{identity}->{name}\"");
	    return 0;
	}
	elsif (! $user->{$config{auth}->{user_pwd_column}}) {
	    # password field is blank - deny access
	    $this->warn("blank password for user \"$user->{$config{auth}->{user_login_column}}\"");
	    return 0;
	}

	# compare entered and stored passwords

	my $salt = substr($share{identity}{password},0,2);
	if ($share{identity}{password} eq
	    crypt($user->{$config{auth}{user_pwd_column}},$salt)) {
	    # passwords match
	    $this->{authenticated} = 1;
	    # store extra user info in identity
	    $share{identity}{gid} = $user->{$config{auth}->{user_gid_column}};
	    ####?
	    $share{identity}{access} = $user->{$config{auth}->{user_access_column}};
	    return 1;
	}
	else {
	    # bad password
	    $this->warn("incorrect password for user \"$user->{$config{auth}->{user_login_column}}\" - access denied");
	    $this->{authenticated} = 0;
	    return 0;
	}
    }
}

#----------------------------------------------------------------------
# authorize : grant the user a specific level of access
#
# Access level is a value of 0 or higher.  0 corresponds to "nobody",
# or the general public.  Values of 1 or more correspond to
# authenticated users, with higher values giving greater permissions
# to the database.  The maximum value is $config::root_auth_level,
# which grants unrestricted "super-user" access to the data.  (This
# level of access can be dangerous, as it allows one to mess with
# sensitive data fields and referential integrity.)
#
# The default authorize routine assumes that the user's access level
# is stored in the user table.  This value is simply returned without
# any sanity checks.

sub authorize {
    my ($this) = shift;
    my $stat = $this->run_handler("authorize");
    return $stat if (defined $stat);
    if (! exists $this->{authenticated}) {
	&authenticate($this);
    }
    if ($this->{authenticated}) {
	return $share{identity}{access};
    }
    # give public access by default
    return 0;
}

#----------------------------------------------------------------------
# approve : approve a specific database operation
#
# Arguments:
# op - "read", "update", "insert", "delete"
# table - the table being operated on
# [record] - reference to a datahash containing the record being
#            operated on (used for update, delete only)
#
# Returns true if the operation is permitted.
# Returns false if:
#   - the user does not have permission to read or write to the table
#   - the user is referencing records owned by a different user or group.
#   - the user does not have the required level of access for this CGI script

sub approve {
    my ($this,$op,$table,$record) = @_;
    my $stat = $this->run_handler("approve",$op,$table,$record);
    return $stat if (defined $stat);
    my $level = $this->authorize();
    my $tablemap = $share{DB}{map}->get_table($table);
    # script okay, now check table permissions
    if (exists $tablemap->{$op}) {
	if ($op =~ /read|insert|search/) {
	    return ($level >= $tablemap->{$op});
	}
	else {
	    # for updates & deletions, must check row ownership
	    if (!$record) {
		if ($level >= $config{auth}->{root_auth_level} - 1) { 
		    return 1; 
		}
		else { return 0; }
	    }
	    elsif ($share{DB}->user_owns($table,$record)) {
		return ($level >= $tablemap->{$op});
	    }
	    else {
		return 0;
	    }
	}
    }
    elsif ($this->{dbmap}) {
	# dbmap kludge - no tablemap or authentication when editing dbmaps
	return 1;
    }
    else {
	# unknown operation
	return 0;
    }
}

# some identity queries, for convenience:

sub my_name {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{name};
    }
    return undef;
}

sub my_email {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{lookup}{email};
    }
    return undef;
}

sub my_uid {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{uid};
    }
    return undef;
}

sub my_gid {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{gid};
    }
    return undef;
}

sub my_user_record {
    my ($this) = shift;
    my $u;
    if ($this->authenticate) {
	if (exists $share{identity}{lookup}) {
	    $u = $share{identity}{lookup};
	}
	elsif ($this->my_uid) {
	    $u = $share{identity}{lookup} = 
		$share{DB}->fetch($config{auth}{user_table},$this->my_uid);
	}
	else {
	    $u = {};
	}
    }
    else {
	$u = {};
    }
    return (wantarray && defined $u) ? %$u : $u;
}

# user permission tests...

sub is_admin {
    my $this = shift;
    return ($this->authorize >= ($config{auth}{root_auth_level} - 1));
}

sub is_manager {
    my ($this,$site) = @_;
    # false if insufficient level
    return 0 if ($this->authorize < ($config{auth}{root_auth_level} - 2));
    # true if a system admin
    return 1 if ($this->authorize >= ($config{auth}{root_auth_level} - 1));
    # otherwise, check for admin key
    return $this->owns($site);
}

sub is_member {
    my ($this,$site) = @_;
    return 0 if (! $this->authorize);
    return 1 if ($this->is_manager);
    my $u = $this->my_user_record();
    if ($u->{section_id}) {
	# the user is tied to a particular site
	if (!$site) { $site = $this->this_site; }
	return ($site->{section_id} == $u->{section_id});
    }
    else {
	# the user belongs to all sites
	return 1;
    }
}

# return the effective access level the user is operating at

sub level {
    my $this = shift;
    if (! exists $this->{effective_level}) {
	$this->{effective_level} = $this->real_level;
    }
    return $this->{effective_level};
}

# return the real access level of the user

sub real_level {
    my $this = shift;
    if (! exists $this->{level}) {
	$this->{level} = $this->authorize;
    }
    return $this->{level};
}

# change to a different (lower) effective access level
#
# This is used when publishing, to make pages render as if being viewed
# by the public.

sub set_level {
    my ($this,$level) = @_;
    $this->{effective_level} = ($level <= $this->real_level) ? $level : 0;
    return $this->{effective_level};
}

# return to our real access level.

sub reset_level {
    my $this = shift;
    delete $this->{effective_level};
}

# remove all traces of current user login

sub clear_login {
    my $this = shift;
    delete $this->{authenticated};  # already authenticated flag
    delete $share{identity};        # user identity information
    delete $session{Auth_identity}; # preserved identity data in Session
    return;
}

# return a standard login form

sub login_form {
    my ($this,$action,$goto) = @_;
    my $out;
    my $in = new ExSite::Input;
    my $data = $in->combine;  # get any input data
    if ($data->{login_action} eq "forgot") {
	return $this->login_reminder_form;
    }
    elsif ($data->{login} && $data->{password} && $this->do_login_form($data)) {
	# successful login
	return undef;
    }
    elsif ($data->{login} || $data->{password}) {
	# failed login
	$out .= "<p class=error>Login failed - please try again.</p>\n";
    }

    # $action determines who process this form; 
    # default is login.cgi - used if $action is undef
    # if $action is defined but blank/false, no action is used (use this
    # to submit the login form to itself)
    $action = defined $action ? 
	($action ? " ACTION=\"$action\"" : "") : 
	" ACTION=\"$config{server}->{CGIpath}/login.cgi\"";
    # $goto determines where the user logs in to; default is login_destination
    $goto = $goto || $config{auth}->{login_destination}[$this->authorize()];
    $out .= "<FORM METHOD=POST$action>
<TABLE CLASS=\"LoginForm\">
<TR><TD>Login:</TD><TD><INPUT TYPE=text NAME=login></TD></TR>
<TR><TD>Password:</TD><TD><INPUT TYPE=password NAME=password></TD></TR>
</TABLE>\n";
    if ($goto) {
	$out .= "<INPUT TYPE=hidden NAME=goto VALUE=\"$goto\">\n";
    }
    my $forgoturl = &relink(login_action=>"forgot");
    $out .= "<P><INPUT TYPE=submit VALUE=\" Log in \">&nbsp;<INPUT TYPE=button VALUE=\" Cancel \" ONCLICK=\"history.back()\"></P>
</form>
<p><a href=\"$forgoturl\">Forgot your login/password?</a></p>\n
";
    return $out;
}

# process a standard login form - return error message, or redirect

sub do_login_form {
    my ($this,$data) = @_;
    if ($this->login($data->{login},$data->{password})) {
	# success!
	my $goto = $data->{goto};
	if (! $goto) {
	    # no place to go;  redirect to default location specified
	    # in config file
	    $goto = $config{auth}{login_destination}[$this->level];
	}
	&redirect($goto);
	return 1;
    }
    return 0;
}

# standard form to send login instructions and password reminders
#
# TODO: assumes cleartext password storage; expand to other cases

sub login_reminder_form {
    my $this = shift;
    my $out;
    my $in = new ExSite::Input;
    my $data = $in->combine;  # get any input data
    if ($data->{login} || $data->{email}) {
	# reminder request
	my $db = new ExSite::DB;
	my $admin = $this->owner_email;
	my $user_list;
	if ($data->{login}) {
	    $user_list = $db->fetch_match($config{auth}{user_table},
					  {$config{auth}{user_login_column}=>$data->{login}});
	}
	elsif ($data->{email}) {
	    $user_list = $db->fetch_match($config{auth}{user_table},
					  {email=>$data->{email}});
	}
	
	my $num_matches = scalar(@$user_list);
	my $url = &relink(login_action=>undef);
	my $forgoturl = &relink(login_action=>"forgot");
	if ($num_matches==0) {
	    if ($data->{login}) {
		$out .= "<p>Sorry, there is no user named <b>$data->{login}</b>.</p>\n";
	    }
	    elsif ($data->{email}) {
		$out .= "<p>Sorry, that email address was not found.</p>\n";
	    }
	    $out .= "<P><A HREF=\"$forgoturl\">Try again?</A></P>
<P><A HREF=\"$url\">Return to login screen</A></P>\n";
	} 
	elsif ($num_matches>1 && $data->{login}) {
	    $out .= "<p>ERROR: there is more than one user named <b>$data->{login}</b>.<br>
Please contact the site administrator at <a href=\"mailto:$admin\">$admin</a> for assistance.</p>
<P><A HREF=\"$forgoturl\">Try again?</A></P>
<P><A HREF=\"$url\">Return to login screen</A></P>\n";
	} 
	else {
	    my $email = $data->{email} || $user_list->[0]{email};
	    if (not $email) {
		$out .= "<p>Sorry, there is no email address recorded for the user <b>$data->{login}</b>.<br>
Please contact the site administrator at <a href=\"mailto:$admin\">$admin</a> for assistance.</p>
<P><A HREF=\"$forgoturl\">Try again?</A></P>
<P><A HREF=\"$url\">Return to login screen</A></P>\n";
	    }
	    else {
		my $site_title = 
		    $this->this_site()->{title} || 
		    $config{site}{name};
		my $subject = "Login instructions for $site_title";
		my $body = "Here are login instructions for ";
		if ($data->{login}) {
		    $body .= "user $data->{login}";
		}
		else {
		    $body .= "$email";
		}
		$body .= " for the website $site_title\n";
		foreach my $u (@$user_list) {
		    $body .= "\nLogin: $u->{login}\nPassword: $u->{password}\n";
		}
		if (&SendMail($email,$admin,$subject,$body)) {
		    $out .= "<p>ERROR: unable to send email.<br>
Please notify your <a href=\"mailto:$admin\">site administrator</a> of this error.</p>
<P><A HREF=\"$forgoturl\">Try again?</A></P>
<P><A HREF=\"$url\">Return to login screen</A></P>\n";
		} 
		else {
		    $out .= "<p>Thank you.</p>
<p>Login instructions have been emailed to $email.
If this is not your current e-mail address, then please contact the 
site administrator at <a href=\"mailto:$admin\">$admin</a> for assistance.</p>
<P><A HREF=\"$url\">Return to login screen</A></P>\n";
		}
	    }
	}
    }
    else {
	$out .= "<FORM METHOD=POST>
<p>To receive login instructions by e-mail, please provide your Login ID 
if you know it, or your e-mail address if you do not.</p>
<table cellspacing=15>
<TR><TD>Your login ID:</TD><TD><INPUT TYPE=text NAME=login></TD></TR>
<TR><TD>OR, your e-mail address:</TD><TD><INPUT TYPE=text NAME=email></TD></TR>
</TABLE>
<P><INPUT TYPE=submit VALUE=\" Submit \">&nbsp;<INPUT TYPE=button VALUE=\" Cancel \" ONCLICK=\"history.back()\"></P>
</form>\n";
    }
    return $out;
}

# validate a login/password pair, and set an identity cookie

sub login {
    my ($this,$login,$passwd) = @_;
    # check for custom login algorithm
    my $stat = $this->run_handler("login");
    return $stat if (defined $stat);

    my $cookie;
    my $user = $this->lookup($login);

    if (ref $user eq "HASH") {
	if ($this->validate_password($passwd,$user)) {
	    return $this->do_login($user);
	}
	else {
	    $this->warn("login: invalid password");
	    return 0;
	}
    }
    else {
	$this->warn("login: no such user");
	return 0;
    }
}	    

# do_login : set login identity, after the user has been authenticated

sub do_login {
    my ($this,$user) = @_;
    # check for custom login algorithm
    my $stat = $this->run_handler("do_login");
    return $stat if (defined $stat);

    $this->info("login: $user");

    my $cookie;
    my $u = (ref $user eq "HASH" ? $user : $this->lookup($user));

    if (ref $u eq "HASH") {
	$this->clear_login;
	if ($config{auth}{type} eq "session") {
	    
	    # set identity for this request
	    $share{identity} = {
		name   => $u->{$config{auth}{user_login_column}},
		uid    => $u->{$config{auth}{user_uid_column}},
		gid    => $u->{$config{auth}{user_gid_column}},
		access => $u->{$config{auth}{user_access_column}},
	    };
	    # set identity for subsequent requests
	    $session{Auth_identity} = join(",",
			      $u->{$config{auth}{user_login_column}},
			      $u->{$config{auth}{user_uid_column}},
		      	      $u->{$config{auth}{user_gid_column}},
			      $u->{$config{auth}{user_access_column}});
	}
	else {	  # if ($config{auth}{type} eq "cookie") {
	    my $cname = $config{site}{identity_cookie};
	    my $cval  = join(":",$u->{login},crypt($u->{password},"wd"));
	    # set identity for this request
	    &edit_cookie($cname,$cval);
	    $share{identity} = {
		name   => $u->{$config{auth}{user_login_column}},
		uid    => $u->{$config{auth}{user_uid_column}},
		gid    => $u->{$config{auth}{user_gid_column}},
		access => $u->{$config{auth}{user_access_column}},
	    };
	    $this->{authenticated} = 1;
	    # set identity for subsequent requests
	    $this->info("setting identity cookie: $cname=$cval");
	    $cookie = &set_cookie($cname,$cval);
	    my $page = $share{Page};
	    if ($page) {
		$page->header($cookie);
	    }
	    else {
		print $cookie;
	    }
	}

	# optional logging of log-ins
	$this->run_handler("log_login");
	return $u->{access};
    }
    else {
	$this->warn("login: no such user: $user");
	return 0;
    }
}	    

sub login_old {
    my ($this,$login,$passwd) = @_;
    # check for custom login algorithm
    my $stat = $this->run_handler("login");
    return $stat if (defined $stat);
    if ($config{auth}{type} eq "cookie") {

	# set a pseudo-cookie

	my $cookie = "$config{site}{identity_cookie}=".
	    join(":",$login,crypt($passwd,"wd"));
	$ENV{HTTP_COOKIE} = $cookie;

	# try out this id/pwd combo

	my $access;
	$this->clear_login;
	if ($access = $this->authorize()) {  # yes, that's =, not ==

	    # good login!

	    my $cookie = &set_cookie($config{site}{identity_cookie},
				     join(":",$login,crypt($passwd,"wd")) );
	    my $page = $share{Page};
	    if ($page) {
		$page->header($cookie);
	    }
	    else {
		print $cookie;
	    }
	    # optional logging of log-ins
	    $this->run_handler("log_login");
	    return $access;
	}
    }
    elsif ($config{auth}{type} eq "session") {

	# try out this id/pwd combo

	my $access;
	$this->clear_login;
	my $user = &lookup($this,$login);
	if (!$user) {
	    # unknown username
	    $this->warn("no such user \"$this->{identity}->{name}\"");
	    return 0;
	}
	elsif (! $user->{$config{auth}->{user_pwd_column}}) {
	    # password field is blank - deny access
	    $this->warn("blank password for user \"$user->{$config{auth}->{user_login_column}}\"");
	    return 0;
	}
	# compare entered and stored passwords
	my $pass;
	if ($config{auth}->{password_storage_method} eq "cleartext") {
	    # encrypt the db password before comparing with identity
	    my $salt = substr($share{identity}{password},0,2);
	    $pass = ($share{identity}{password} eq
		     crypt($user->{$config{auth}->{user_pwd_column}},$salt));
	}
	else {
	    # encrypted db passwd can be compared directly with identity
	    $pass = ($share{identity}{password} eq $user->{$config{auth}->{user_pwd_column}});
	}
	if ($pass) {
	    # passwords match
	    $this->{authenticated} = 1;
	    # store extra user info in identity
	    $share{identity}{gid} = $user->{$config{auth}->{user_gid_column}};
	    ####?
	    $share{identity}{access} = $user->{$config{auth}->{user_access_column}};
	    return 1;
	}
	else {
	    # bad password
	    $this->warn("incorrect password for user \"$user->{$config{auth}->{user_login_column}}\" - access denied");
	    $this->{authenticated} = 0;
	    return 0;
	}
	if ($access = $this->authorize()) {  # yes, that's =, not ==

	    # good login!

	    my $cookie = &set_cookie($config{site}{identity_cookie},
				     join(":",$login,crypt($passwd,"wd")) );
	    my $page = $share{Page};
	    if ($page) {
		$page->header($cookie);
	    }
	    else {
		print $cookie;
	    }
	    # optional logging of log-ins
	    $this->run_handler("log_login");
	    return $access;
	}
    }
    return 0;
}

# unset the identity cookie

### FIXME: session management

sub logout {
    my $this = shift;

    $this->clear_login;  # this is sufficient for session logins

    if ($config{auth}{type} eq "cookie") {

	# clear existing identity value for remainder of this request

	my $cookie = &edit_cookie($config{site}{identity_cookie}, undef );

	# clear cookie for future requests

	my $cookie_domain = $config{server}{host} ? 
	    ".$config{server}{domain}" : $config{server}{domain};
	$cookie = &set_cookie($config{site}{identity_cookie},
			      undef,
			      $config{server}{CGIpath},
			      $cookie_domain,
			      "Thu, 01-Jan-1970 00:00:00 GMT",
			      );
	my $page = $share{Page} || $share{PageBuf};
	if ($page) {
	    $page->header($cookie);
	}
	else {
	    print $cookie;
	}
    }
    return;
}

#----------------------------------------------------------------------
# validate password : compare a password with the stored version
#
# Returns true (1) if the passed password matches the user's

sub validate_password {
    my ($this,$password,$user) = @_;
    my $udata;
    if (! defined $user) {
        $udata = $this->my_user_record;
    }
    elsif (ref $user eq "HASH") {
	$udata = $user;
    }
    else {
	$udata = $share{DB}->lookup($user);
    }
    if (!$udata) {
	# unknown user
	$this->warn("validate_password: unknown user");
	return 0;
    }
    elsif (! $udata->{$config{auth}->{user_pwd_column}}) {
	# password field is blank - deny access
	$this->warn("blank password for user \"$udata->{$config{auth}->{user_login_column}}\"");
	return 0;
    }

    if (! $password) {
	# blank given password - deny
	return 0;
    }
    if (! $udata->{password}) {
	# blank stored password - deny
	return 0;
    }

    my $pass;
    my $passtype = $config{auth}{password_storage_method};
    if ($passtype eq "des") {
	# encrypt the password claim before comparing with stored version
	my $salt = substr($udata->{password},0,2);
	$pass = ($udata->{password} eq crypt($password,$salt));
    }
    elsif ($passtype eq "md5") {
	# encrypt the password claim before comparing with stored version
	$pass = ($udata->{password} eq &md5_hex($password));
    }
    elsif ($passtype eq "crypt") {
	# decrypt the real password and compare with claim
	require ExSite::Crypt;  # don't load this unless we need it
	my $c = new ExSite::Crypt(mode=>"strong");
	$pass = ($c->decrypt($udata->{password}) eq $password);
    }
    else {  # cleartext
	# by default we just compare the two values directly
	$pass = ($udata->{password} eq $password);
    }
    return $pass;
}

# change password

sub change_password {
    my ($this,$old,$new,$new2) = @_;
    if ($new) {
	if (! defined $new2 || ($new eq $new2)) {
	    my $re = $share{DB}{map}->regexp('password');
	    if ($new =~ /^$re$/) {
		if (! defined $old || $this->validate_password($old)) {
		    my $stat = $this->set_password($new);
		    return $stat if ($stat); # error
		    # update the login cookie
		    $this->clear_login;
		    $this->login($this->my_name,$new);
		    return undef;
		}
		else {
		    return "password update failed - incorrect password";
		}
	    }
	    else {
		return "insecure password - needs more characters";
	    }
	}
	else {
	    return "new passwords do not match";
	}
    }
    else {
	return "cannot set a blank password";
    }
}

sub set_password {
    my ($this,$password,$uid) = @_;
    if (! $uid) {
	$uid = $this->my_uid;
    }
    if ($uid) {
	my $pass_store;
	my $passtype = $config{auth}{password_storage_method};
	if ($passtype eq "des") {
	    # des-encrypt the password
	    $pass_store = crypt($password,&randtext(2,'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'));
	}
	elsif ($passtype eq "md5") {
	    # md5 hash the password
	    $pass_store = &md5_hex($password);
	}
	elsif ($passtype eq "crypt") {
	    # encrypt the password using ExSite encryption
	    require ExSite::Crypt;  # don't load this unless we need it
	    my $c = new ExSite::Crypt(mode=>"strong");
	    $pass_store = $c->encrypt($password);
	}
	else { 
	    # plaintext storage
	    $pass_store = $password;
	}

	return $share{DB}->update($config{auth}{user_table},
			  {$config{auth}{user_pwd_column}=>$pass_store},
			  {$config{auth}{user_uid_column}=>$uid});
    }
    return "no such user";
}

# Authentication Token handling:
# an authtoken is an encrypted string that can be included in the
# query string, which will automatically authenticate the user.

# make_authtoken : generate an authtoken

sub make_authtoken {
    my ($this,$username) = @_;
    my $udata = $share{DB}->lookup($username);
    my ($sec,$min,$hour,$mday,$mon,undef) = localtime(time);
    $mon++;
    my $authtoken_plaintext = "$mon/$mday/$hour/$min:".$udata->{login};
    my $crypt = new ExSite::Crypt;
    return $crypt->encrypt($authtoken_plaintext);
}

# validate_authtoken : return the username, if the authtoken is valid

sub validate_authtoken {
    my ($this,$authtoken) = @_;
    my $crypt = new ExSite::Crypt;
    my $authtoken_plaintext = $crypt->decrypt($authtoken);
    $authtoken_plaintext =~ /^(\d+)\/(\d+):(.+)$/;
    my ($m,$d,$login) = ($1,$2,$3);
    if ($config{auth}{authtokens_expire}) {
	# ensure this authtoken hasn't expired
	my ($sec,$min,$hour,$mday,$mon,undef) = localtime(time);
	$mon++;
	if (($m != $mon) || ($d != $mday)) {
	    $this->warn("attempt to use expired authtoken for $login");
	    return 0;
	}
    }
    if ($login) {
	my $udata = $share{DB}->lookup($login);
	if (ref $udata eq "HASH") {
	    return $login;
	}
	$this->warn("attempt to use invalid authtoken for $login");
	return 0;
    }
    $this->warn("attempt to use invalid authtoken (no user)");
    return 0;
}

# process_authtoken : login the user, if the authtoken is valid

sub process_authtoken {
    my ($this,$authtoken) = @_;
    if (my $login = $this->validate_authtoken($authtoken)) {
	$this->do_login($login);
    }
    return;
}

# fetch all admin tickets

sub my_sites {
    my ($this) = @_;
    return $share{DB}->fetch_m2m("section",$config{auth}{user_table},
				 $this->my_uid,"member_site_link");
}

# ditto, but do not count website templates

sub my_real_sites {
    my ($this) = @_;
    my @real_sites;
    foreach my $site ($this->my_sites) {
	if ($site->{type} !~ /template/) {
	    push @real_sites, $site;
	}
    }
    return @real_sites;
}

# return list of sites the user has member-level access to

sub my_access_keys {
    my $this;

    # start with all sites with admin access
    my @sites = $this->my_real_sites;
    
    # add sites with regular member access
    my $user = $share{DB}->my_user_record;
    if ($user->{section_id}) {
	push @sites, $share{DB}->fetch("section",$user->{section_id});
    }
    
    return @sites;
}

# return the current site data (may require some guesswork)

sub this_site {
    my ($this) = @_;
    if ($share{Page}) {
	# we're on a page, so return the page's section
	return $share{Page}->fetch("section",$share{Page}->get("page","section_id"));
    }
    elsif ($share{Section}) {
	# we're on a section, so return it
	return $share{Section}->get();
    }
    elsif ($share{DB}->my_user_record()->{section_id}) {
	# user has a primary site defined
	return $share{DB}->fetch("section",$share{DB}->my_user_record()->{section_id});
    }
    else {
	my @sites = $this->my_real_sites;
	if (@sites > 0) {
	    # ambiguous result if more than one - return first one
	    my @s = $share{DB}->sort_rows("section",\@sites,"section_id");
	    return $s[0];
	}
    }
    return undef;
}

# test for "ownership" of a site, ie. an admin key

sub owns {
    my ($this,$site) = @_;    # $site can be a section hash, or a section ID
    if (! $site) { $site = $this->this_site(); }
    my $sid = ref $site eq "HASH" ? $site->{section_id} : $site;
    return 0 if (! $sid);
    my @temp = $share{DB}->fetch_match("member_site_link",
				       { member_id => $this->my_uid,
					 section_id => $sid } );
    return (@temp > 0);
}

# test for license to use a service on a site

sub license {
    my ($this,$service,$site) = @_;
    if (eval "require Modules::$service") {
	my $s = eval "new Modules::$service";
	if ($s->ioctl("isRestricted")) {
	    # this service requires licenses
	    if (!$site && $share{Page}) {
		$site = $share{Page}->id("section");
	    }
	    if ($site) {
		my $licenses = $share{DB}->fetch_match("service",
						       {section_id => $site,
							name => $service }
						       );
		return (scalar @$licenses > 0);
	    }
	    # hmm, that's strange, we're not on a page
	    $this->warn("$service: unknown website, cannot confirm license");
	    return 0;
	}
	# this modules is not a service - anyone can use it
	return 1;
    }
    # module does not seem to exist
    $this->warn("$service: no such service");
    return 0;
}

# test for license to use a service's control panel
# ie. test that a web app is licensed on any of the user's sites

sub license_control_panel {
    my ($this,$service) = @_;
    if (! $share{licenses}) {
	$share{licenses} = {};
	my $db = $share{DB};
	if ($db->{dbtype} =~ /SQL/i) {
	    # use a join query to fetch all licensed services
	    my $uid = $db->my_uid;
	    my @services = $db->custom_query("select service.name from service,section,member_site_link where service.section_id=section.section_id and service.section_id=member_site_link.section_id and member_site_link.member_id=$uid");
	    $share{licenses} = &keywise("name",\@services);
	}
	else {
	    # brute-force loop over all services/sites
	    my @sites = $db->my_real_sites();
	    foreach my $l ($share{DB}->fetch_all("service")) {
		foreach my $s (@sites) {
		    if ($l->{section_id} == $s->{section_id}) {
			$share{licences}{$l->{name}} = 1;
			next;
		    }
		}
	    }
	}
    }
    return (exists $share{licenses}{$service});
}

#
# user management
#

# fetch_admins : fetch all users with unrestricted access

sub fetch_admins {
    my ($this,$site) = @_;
    return $this->fetch_users_by_level($site,$config{auth}{root_auth_level}-1);
}

# fetch all users with manager level access (does not include admins)

sub fetch_managers {
    my ($this,$site) = @_;
    return $this->fetch_users_by_level($site,$config{auth}{root_auth_level}-2);
}

# fetch all users with a certain level of access (does not include admins)

sub fetch_users_by_level {
    my ($this,$site,$level) = @_;
    if (! defined $level) { $level = 1; }
    my @users_by_level;
    if ($site) {
	my $sid;
	$sid = (ref $site eq "HASH") ? $site->{section_id} : $site;

	# fetch everyone with an admin key
	my @users = $share{DB}->fetch_m2m($config{auth}{user_table},"section",$sid,"member_site_link");
	# select only those at our level of access
	foreach my $u (@users) {
	    if ($u->{access} == $level) {
		push @users_by_level, $u;
	    }
	}

	# fetch everyone with this as their home site
	@users = $share{DB}->fetch_match($config{auth}{user_table}, 
					 { access=>$level, section_id=>$sid});
	my @extras;
	# weed out the dupes
      USER:
	foreach my $u (@users) {
	  ADMIN:
	    foreach my $a (@users_by_level) {
		next USER if ($u->{member_id} == $a->{member_id});
	    }
	    # no dupe detected
	    push @extras, $u;
	}
	# add the non-dupes to our admin list
	push @users_by_level, @extras;
    }
    else {
	# no site given, just take everyone
	@users_by_level = $share{DB}->fetch_match($config{auth}{user_table},{access=>$level});
    }
    return @users_by_level;
}

# find most authoritative user on a site 

sub fetch_owner {
    my ($this,$site) = @_;
    if (! $site) { $site = $this->this_site; }
    while ($site) {
	my @u = $this->fetch_admins($site);
	if (@u > 0) {
	    return $u[0];
	}
	@u = $this->fetch_managers($site);
	if (@u > 0) {
	    return $u[0];
	}
	# no luck;  try parent site
	if ($site->{parent_id}) { 
	    $site = $share{DB}->fetch("section",$site->{parent_id}); 
	}
    }
    return undef;
}

sub owner_email {
    my ($this,$site) = @_;
    my $u = $this->fetch_owner($site);
    if (ref $u eq "HASH" && $u->{email}) {
	return $u->{email};
    }
    # no owner email found; use config administrator email
    return $config{site}{administrator};
}

sub adduser {
    my $this = shift;
    my $udata = shift;  # member datahash
    my @keys = @_;      # admin keys
    my $db = $share{DB};
    my @err;

    # 1) validate the password

    my $stat = $db->validate_column($config{auth}{user_table},"password",$udata->{password});
    if ($stat) {
	push @err, "Invalid/insecure password - please choose another";
    }

    # 2) ensure the login ID is unique

    my @u = $db->fetch_match($config{auth}{user_table}, { login=>$udata->{login} });
    if (@u > 0) {
	push @err, "That login ID is already in use - please try another";
    }

    # 3) name

    if ($udata->{name} =~ /^\s*$/) {
	push @err, "Cannot insert member - no name provided";
    }

    # 4) validate the email

    $stat = $db->validate_column($config{auth}{user_table},"email",$udata->{email});
    if ($stat) {
	push @err, "Invalid email address";
    }

    # 5) home site

    if ($udata->{section_id}) {
	$stat = $db->validate_column($config{auth}{user_table},"section_id",$udata->{section_id});
	if ($stat) { 
	    push @err, $stat;
	}
    }

    # abort if there are errors
    if (@err > 0) {
	$this->error(@err);
	return 0;
    }
    
    # insert the user
    if ($udata->{member_id}) { 
	$this->warn("removing member ID of new user");
	delete $udata->{member_id}; 
    }
    my $uid = $db->insert("member",$udata);
    if (! $uid) {
	$this->error("failed to add user $udata->{login}");
	return 0;
    }

    # add admin keys

    foreach my $key (@_) {
	next if (ref $key ne "HASH");
	$key->{member_id} = $uid;
	my @err = $db->validate_record("member_site_link",$key);
	if (@err) {
	    $this->error(@err);
	}
	else {
	    $db->insert("member_site_link",$key);
	}
    }

    return $uid
}

sub deluser {
    my $this = shift;
    my $user = shift;
    my $uid = ref $user eq "HASH" ? $user->{member_id} : $user;
    $share{DB}->trash_r($config{auth}{user_table},$uid);
}

1;

