#-----------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
=pod

=head1 ExSite::Cache

ExSite::Cache is a generic cache tool for saving re-useable data.

The cache is an in-memory hash of arbitrary data values.
The data can be indexed based on an arbitrary amount of
indexing data, so that the cache can store completely unrelated
data with low risk of collision.

The primary use in ExSite is to remember results of database
queries, so that we don't need to make a DB server call to
re-fetch something we've already fetched (but forgotten).

For instance, if you've fetched data record #123 from table XYZ,
you could save it in cache this way:

C<my $cache = new ExSite::Cache;>

C<$cache-E<gt>save($data,"XYZ","123");>

The first value is the data to save;  the remaining values (there 
can be any number) are used to uniquely index the data.  In this case,
the table name and record number are sufficient to uniquely specify
this piece of data.

If later someone asks for the same data, eg:

C<$db-E<gt>fetch("XYZ",123);>

then you could first check the cache for a value before querying the
the database:

C<$data = $cache-E<gt>get("XYZ",123);>

C<if (! $data) { # go to database...>

If practice, all of the above is done automatically by the 
ExSite::DB class.

Caching is controlled by a couple of configuration parameters,

=over 4

=item C<cache.max_items> (= 100 by default)
the maximum number of items that will be cached at once.  If this number
is exceeded, the cache will try to prune itself by dumping the largest
unused item it holds.  (Unused means it has not yet been fetched from
cache.)

=item C<cache.max_size> (= 1000000 [1 MB] by default)
the maximum size of an item that will be cached.  If this number
is exceeded, the item will not be saved in the cache at all.  That makes
it safer to simply throw everything at the cache without risk of having 
the cache become a giant memory leak.

=back

=cut

package ExSite::Cache;
use strict;
use ExSite::Config;
use ExSite::Misc;  # sizeof()
use Digest::MD5 qw(md5_hex);

# Cache constructor - bare bones

sub new ($%) {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->{_stats} = {};
    return $obj;
}

# save some data to the cache;
# first arg is the data to save;
# remaining args are used to index the data.
#
# performs a somewhat expensive safety check to ensure that the data
# isn't too large to cache.

sub save ($$@) {
    my $this = shift;
    my $data = shift; # this is the value you are saving in the hash
    # abort if data is unreasonably large
    return if (! $this->safe($data));
    # prune if cache is full
    $this->prune() if (scalar keys %{$this} > $config{cache}{max_items} + 1);
    # remaining args are the values to hash
    my $key = md5_hex(@_);
    $this->{$key} = $data;
    $this->{_stats}{$key} = 0;  # number of times this item has been fetched
}

# ditto, but assumes the data is safe and saves it without a safe check
# faster, but the caller has to know whether or not the data is safe to cache

sub save_safe ($$@) {
    my $this = shift;
    my $data = shift; # this is the value you are saving in the hash
    # prune if cache is full
    $this->prune() if (scalar keys %{$this} > $config{cache}{max_items} + 1);
    # remaining args are the values to hash
    my $key = md5_hex(@_);
    $this->{$key} = $data;
    $this->{_stats}{$key} = 0;  # number of times this item has been fetched
}

# fetch data from the cache;
# args are used to determine the index to the data.

sub get ($@) {
    my $this = shift;
    my $key = md5_hex(@_);
    if (exists $this->{$key}) { 
	$this->{_stats}{$key}++;
	return $this->{$key}; 
    }
    return undef;
}

# delete data from the cache;
# args are used to determine the index to the data.

sub delete ($@) {
    my $this = shift;
    my $key = md5_hex(@_);
    if (exists $this->{$key}) { delete $this->{$key}; } 
}

# prune the cache - try to reduce the cache size
# Currently removes largest unused item.  Slow, unfortunately.

sub prune ($) {
    my $this = shift;
    my ($id,$size);
    foreach my $key (keys %$this) {
	next if ($key eq "_stats");
	if (! $this->{_stats}{$key}) {
	    my $len = sizeof($this->{$key});
	    if ($len > $size) {
		$size = $len;
		$id = $key;
	    }
	}
    }
    if ($id) {
	delete $this->{$id};
    }
}

# clear the entire cache.

sub clear ($) {
    my $this = shift;
    foreach my $key (keys %$this) {
	delete $this->{$key};
    }
}

# test data to see if it is safe to cache

sub safe ($$) {
    my ($this,$data) = @_;
    if (ref $data eq "ARRAY") {
	# guesstimate size of large arrays to try and save cycles
	my $array_size = scalar @$data;
	if ($array_size > 5 && 
	    sizeof($data->[0]) * $array_size > 2 * $config{cache}{max_size}) {
	    # guesstimated size of array is more than twice 
	    # the cacheable object size limit
	    return 0;
	}
    }
    if (sizeof($data,$config{cache}{max_size}) > 
	$config{cache}{max_size}) {
	# real size of data exceeds cacheable object size
	return 0;
    }
    return 1;
}

1;
