#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Config;
use strict;
use myConfig;

=pod

=head1 ExSite::Config - ExSite configuration file

C<Config.pm> contains general configuration parameters,
site-specific setup data, and global variables used throughout
the ExSite system.

This file is designed so that it never has to be modified.
To set configuration variables, use an "exsite.conf" file, which 
will be loaded automatically, overriding the presets below.
Handlers and custom setup routines can be defined in the module
myConfig.pm.

=for html <hr>

=head2 Usage:

C<use ExSite::Config;>

In addition, every program should also include the following first line 
of executable code:

C<&exsite_init;>

=head2 Configuration Management

All system parameters are stored in one of three configuration hashes,
which are exported automatically to every program that uses Config.pm:

=over 4

=item C<%config>

C<%config> is used for system constants that do not normally change.

The system constants in C<%config> are read first from the hard-coded
values in Config.pm, and then from the local configuration file 
"exsite.conf", which can override the former values.  
Modifications to the site configuration should be made 
in exsite.conf, not in Config.pm.

Specific configuration parameters are described in the ExSite
Configuration documentation, or in Config.pm.

Any module or other package can read its own configuration file(s) into
C<%config>, by calling: C<&ExSite::Config::read_exsite_conf()>.

=item C<%share>

C<%share> is used as a general-purpose shared-memory area.  
C<%share> is cleared on each page request.

The C<%share> hash is used for sharing data between widely separated
modules and routines.  The primary difference between C<%config> and
C<%share> is that the former is regarded as persistent (values stay
the same from request to request), while the latter is volatile
(values vary from request to request).

The differences are not significant for normal CGI setups, since C<%config>
must be reloaded on each request anyway.  But in a mod_perl setup,
C<%config> may only be configured once at server startup, whereas
C<%share> will be cleared and rebuilt on every request.  It is good 
practice to follow this convention, to ease the transition to a mod_perl
setup, when it comes.

C<%share> is empty when initialized by default, and is cleared every
time C<&exsite_init()> is called.  C<%share> will be used to store
pointers to the current database and page objects (C<$share{DB}> and
C<$share{Page}>, respectively), plus any other shareable data placed
there by localization routines and content modules.

=item C<%session> 

C<%session> is used for session constants that need to be remembered
for short times between page requests, but not for extended periods.

The C<%session> hash stores semi-persistent data.  That is, it
remembers data between different page requests, but clears the data
after a certain amount of time has passed.  It is a useful mechanism
for temporarily preserving state.  By default, session data is stored
for a maximum of 1 hour if no page requests renew the session, or 24
hours if the session is continually renewed.  After this time, the
session will be closed, and a new one will have to be opened.

If session management is enabled on a particular installation of ExSite,
then %session will automatically be filled with the current session data.
(See C<ExSite::Session> for technical details.)  These data include
C<_atime> and C<_ctime>, which are timestamp values (Unix system times)
of the last access, and the creation time of the session.  If session 
management is not enabled, then the contents of %session are undefined 
(but will probably be blank in a conventional CGI setup).

B<Warning>: session management is B<not> enabled by default in the 
standard ExSite distribution.  In this case, C<%session> will be cleared 
on each page request, like C<%share>.

=back

=cut

#=====================================================================
# Exported names:

BEGIN {
    use Exporter;
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

    $VERSION = "3.3.0";

    @ISA = qw(Exporter);

    @EXPORT      = qw(%config %share %session
		      &exsite_init &exsite_close 
		      &page_header &page_footer
		      &get_obj
		      );

    @EXPORT_OK   = qw(&read_exsite_conf &install_handlers);

}

use vars (@EXPORT, @EXPORT_OK);

#=====================================================================
# Non-Exported names:

use vars qw(
	    %server_config %site_config %auth_config 
	    %form_config %report_config %log_config
	    %webtop_config %crypt_config %session_config
	    $config_flag
	    );

##################################################################
################ static configuration            #################
##################################################################

# static configuration parameters are valid for every site setup.
# They may still be overridden in particular scripts, by setting
# the given parameters to a new value.

#======================================================================
# SITE CONFIGURATION:
# parameters that define the website as a whole

%site_config = (

    # site name - used as a default title
    name => "ExSite",

    # site ID (no spaces) - used as a cookie prefix
    id => "ExSite",

    # identity cookie name
    identity_cookie => "ExSiteID",

    # site logo <IMG ...> tag
    logo => "",

    # sysadm's email
    administrator => "morgan\@exware.com",

);

#======================================================================
# AUTHENTICATION CONFIGURATION
# parameters controlling, login, authorization, etc.

%auth_config = (

    # authentication type (fixed, none, cookie, session)
    type => "fixed",

    # define the password storage method: "cleartext", "des", "md5", or "crypt"
    password_storage_method => "cleartext",

    # define the default identity to use in case of fixed auth type, 
    # or for simple shell-based tests with no cookie handling.
    default_identity => {
	       name => "nobody",
	       uid => 1,
	       gid => 1,
	       access => 3,
	   },

    # define the tables and columns where login and authentication
    # information can be found
    user_table => "member",
    group_table => "section",
    user_login_column => "login",
    user_pwd_column => "password",
    user_access_column => "access",
    user_uid_column => "member_id",
    user_gid_column => "section_id",

    # what authorization level corresponds to unlimited access?
    root_auth_level => 4,

    # what user level is permitted to link to records they don't own?
    foreign_link_level => 1,

    # login targets
    login_destination => [ "/index.html", 
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   ],

    # CGI program access levels, to be filled in exsite.conf
    cgi => { },

    # check http referer on database peeks (eg. getdata.cgi, content.cgi)
    referer_check => 1,
);

#======================================================================
# FORM CONFIGURATION
# parameters controlling form layout

%form_config = (

    # input field separator character
    sepchar => '#',

    # multiple-input separator character
    multi_sep => '; ',

    #---------------------------------
    # form style
    #---------------------------------
    # form_style can be:
    # "list" (labels above inputs)
    # "table" (labels to the left of aligned inputs)
    # "none" (labels to the left of unaligned inputs)
    # additional styles can be defined by adding entries to %fieldtag, below

    style => "table",

    # field tags (1st goes before label, 2nd goes between label and input,
    #             3rd goes after input)

    fieldtag => {
		 list => ["<P>","<br>","</P>"],
		 table => ["<TR><TD>",":</TD><TD>","</TD></TR>"],
		 none => ["",":","<br>"]
		},

    # reqd_field_style can be:
    # "bold" (labels shown in bold type)
    # "star" (labels marked with an asterisk)
    # "css" (will be dealt with by css)
    # "none" (required fields are not distinct)

    reqd_field_style => "bold",

    # label tags (1st goes before field name, 2nd goes after)

    labeltag => {
		 bold => ["<B><SPAN CLASS=\"reqd_label\">","</SPAN></B>"],
		 star => ["<SPAN CLASS=\"reqd_label\">","<SUP>*</SUP></SPAN>"],
		 css => ["<SPAN CLASS=\"reqd_label\">","</SPAN>"],
		 none => ["<SPAN CLASS=\"label\">","</SPAN>"]
	        },

    # maximum input field width
    max_cols => 60,
    # maximum rows in a textarea input
    max_rows => 10,

    # form mode:
    # "cgi" to have forms processed by CGI programs (eg. doform.cgi)
    # "exsite" to have forms processed by WebDB DCD
    mode => "cgi",

    #---------------------------------
    # data handling options
    #---------------------------------

    # trim leading/trailing whitespace from form data before 
    # inserting/updating?

    trim_whitespace => 0,

    # default date settings for undefined dates: "today" or "null"

    default_date => "today",

    # default setting for the html editor

    htmleditor => {
    		cms => 0, # use text editor in cms mode?
		# cut, copy and paste commands:
		# Gecko browser: required to add additional configuration parameters to 
		# 	the file(user.js or prefs.js) in the profile directory.
		#
		# user_pref("capability.policy.default.Clipboard.cutcopy","allAccess");
		# user_pref("capability.policy.default.Clipboard.paste", "allAccess");
		#
		# All icons are turned on by default
		op_selection=>0, # cut, copy and paste
		op_change=>1, # undo and redo
		op_table=>1, # table wizard and table editor
		op_image=>1, # image dialogue
		op_image_upload => 1, # enable file upload form in the insert image/document popup
		op_link=>1, # link dialogue
		op_dyncontent=>1, # dynamic content dialoge
		op_style=>1, # bold,italic and underline
		op_color=>1, # text color and background color
		op_alignment=>1, # justify, align left, align right and align center
		op_list=>1, # ordered list and bulleted list
		op_indent=>1, # outdent, indent
		op_cleanup => 1, # html scrubber		
		complete_html_content=>0,
		editor_width=>700,
		editor_height=>512,
    },

    # file upload limits (in btyes)
    # "file" type is capped by field size (64K, which is reasonable)
    # "bigfile" type allows up to 16MB, which is probably unreasonable
    # for web uploads.  Furthermore DB engines like MySQL limit query
    # sizes so that you can't actually insert large files into a SQL
    # record (MySQL limit is typically 1 MB).  Set the actual bigfile 
    # size limit here:

    bigfile_maxsize => 1048576,  # 1 MB

    # display of foreign key selectors
    # set to 1 to include foreign key id in selector
    # set to 0 to use only general description of foreign record
    show_foreign_key_id => 1,

);

#======================================================================
# REPORT CONFIGURATION
# parameters controlling report formatting

%report_config = (

    # currency character
    currency_char => '$',

    # maximum number of records to report at a time
    page_length => 20,

    # for table reports, this flag determines whether each row is classified 
    # as A or B for the purposes of highlighting alternate lines.  Note that 
    # turning this flag on does not by itself highlight lines;  you must also 
    # setup your stylesheet to recognize the A/B classes of <TR>.  This 
    # setting will override $reportrecordtag->{table}->[0].
    highlight_alternate_lines => 0,

    # report link mode:
    # "cgi" to have db_ops linked to CGI programs (eg. form.cgi)
    # "exsite" to have db_ops linked back to the page (ie. to WebDB DCD)
    linkmode => "cgi",

    # foreign key linking permissions
    # when reporting foreign keys, cross-link to the foreign record if the
    # viewer has this permission on the foreign record
    link_to_foreign_key_perm => "read",

    # email format:
    # "plain" displays the email using a regular mailto: URI
    # "obfuscated" uses javascript to obfuscate the address from spambots

    email => "plain",

    #---------------------------------
    # reporting of text data
    #---------------------------------

    text => {
	
	# allow HTML tags found in normal text

	honor_html => 0,

	# treat newlines as meaningful in normal text

	honor_nl => 0,

    },

    #---------------------------------
    # report styles
    #---------------------------------

    # Report styles are:
    #
    # table: records listed in a tabular format; column names are shown as first 
    #        header row
    # full: records listed in a 2-column column/value format, one table per record
    # dump: records listed as "column=value", one column per line.
    # index: records listed in a paragraph, with values delimited by commas
    # list-numbered: like index, but enumerated
    # list-bulleted: like index, but bulleted
    # address: records listed in a paragraph, with values delimited by linebreaks
    # description: records listed using <DT> for first value, <DD> for remaining 
    #	 values

    # default style
    style => "table",

    # the following tags go before and after the report, respectively
    report_tag => {
	table => [ "<TABLE CLASS=\"report_table\">", "</TABLE>" ],
	full => [ "", "" ],
	dump => [ "<hr>", "" ],
	index => [ "", "" ],
	list_numbered => [ "<OL>", "</OL>" ],
	list_bulleted => [ "<UL>", "</UL>" ],
	address => [ "", ""],
	description => [ "<DL>", "</DL>"],
    },

    # the following tags go before and after each record, respectively
    row_tag => {
	table => [ "<TR>", "</TR>" ],
	full => [ "<TABLE CLASS=\"report_record\">", "</TABLE><P>" ],
	dump => [ "<PRE>", "</PRE><hr>" ],
	index => [ "<P>", "</P>" ],
	list_numbered => [ "<LI><P>", "</P></LI>" ],
	list_bulleted => [ "<LI><P>", "</P></LI>" ],
	address => [ "<P>", "</P>" ],
	description => [ "<DT>", "<P>" ],
    },

    # the following tags go before and after each column value, respectively
    column_tag => {
	table => [ "<TD>", "</TD>" ],
	full => [ "<TR><TD CLASS=\"label\">", "</TD></TR>" ],
	dump => [ "", "\n" ],
	index => [ "", ", " ],
	list_numbered => [ "", ", " ],
	list_bulleted => [ "", ", " ],
	address => [ "", "<br>" ],
	description => [ "", "<DD>" ],
    },

    # report_labeltag: if defined, the column label will be shown along with 
    # each data value, with the given string separating the two.  If not 
    # defined, column labels will not be shown with each value.
    label_tag => {
	table => undef,
	full => "</TD><TD CLASS=\"data\">",
	dump => "=",
	index => undef,
	list_numbered => undef,
	list_bulleted => undef,
	address => undef,
	description => undef,
    },

);

#======================================================================
# SERVER CONFIGURATION
# this stuff varies between sites, so the following defaults are
# mostly just placeholders for configuration data to follow.
# The auto flag determines whether autoconfiguration is done by
# default.  Set it to 1 in exsite.conf if you want autoconfiguration.

%server_config = (
    db => { name=>"exsite", server=>"localhost", 
	    user=>"exsite", type=>"text" },
    map => { name=>"dbmap", type=>"text" },
    HTMLroot => "/var/www/html",
    HTMLpath => "",
    CGIroot => "/var/www/cgi-bin",
    CGIpath => "/cgi-bin",
    CSSpath => "/_ExSite/css",  # deprecated
    CSSfile => "ExSite.css",    # deprecated
    protocol => "http",
    host => "localhost",
    domain => "localdomain",
    port => "",
    server => "http://localhost.localdomain",
    temppath => "/tmp",
    imgdir => "/images",
    auto => 0,
);

#======================================================================
# LOG CONFIGURATION
# parameters controlling event logging

%log_config = (

    # Each parameter refers to the logging level for a particular log file.
    # Log levels go from 0 (no logging) to 3 (log everything).
    # Log files must exist and be writable by the webserver.

    # db - database log file
    # 1: log database errors
    # 2: log database modifications
    # 3: log database connections & selects
	       
    db => 0,

    # cms - content management system audit trail
    # 1: log publicly-visible changes
    # 2: log all website, page, and content updates
    # 3: log all cms operations

    cms => 0,

    # diag - general ExSite diagnostics
    # 1: log errors
    # 2: log warnings
    # 3: log info messages

    diag => 0,

    # the log files themselves... (paths are relative to cgi directory)

    logfile => {
	db => "db.log",
	cms => "cms.log",
	diag => "diag.log",
    },

);

#======================================================================
# WEBTOP CONFIGURATION 

%webtop_config = (

    # number of icons across
    x => 6,

);

#======================================================================
# CRYPTO CONFIGURATION 

%crypt_config = (

    # CRYPTOGRAPHY:
    # ExSite can automatically encrypt and decrypt data going to and from
    # the database, if the column has "crypt" datatype.  Crypt fields are
    # treated same as text fields, but the data is encrypted according
    # to the following options:

    # encryption mode can be:
    # "strong" - best security, good for storing critical data
    # "fast" - trades security for speed; 2-3 times faster than strong
    #          mode; good for obfuscating data communicated over http, 
    #          but not for secure storage.

    default_mode => "strong",

    # keys for coding/decoding
    # It is good practice to define different keys for fast and strong
    # encryption.  Otherwise the regular key will be used.

    key_fast => "fast dummy pass-phrase",
    key_strong => "strong dummy pass-phrase",
    key => "dummy pass-phrase",

    # number of mixing rounds (higher is better)

    nround_fast => 4,
    nround_strong => 32,

    # decryption rules:
    # admin => only administrator can view plaintext
    # owner => owner of record can view plaintext

    access => "admin",

);

#======================================================================
# SESSION CONFIGURATION 

%session_config = (

    # session management is disabled by default.  To enable it, set
    # this parameter to a non-zero value in exsite.conf, and create the
    # session file (DBM) and the session lockfile, both with mode 666,
    # in the cgi directory.

    on => 0,

    # the DBM file in which session data is stored.  Note that DBM
    # may add its own suffix to this file, depending on the implementation
    # that is used.  For this reason, we cannot reliably test for session
    # support by looking for this filename;  instead we check for the lock
    # file.

    file => "exsite.session",

    # the session lockfile.  This contains no data, but is simply gives us a 
    # file descriptor to which we can apply locks to prevent session file
    # collisions.

    lockfile => "exsite.session.lock",

    # max_lifetime - the maximum time (in minutes) that a session will
    # last before the garbage collector junks it.  (Default 1 day.)

    max_lifetime => 1440,

    # max_idle - the maximum idle time (time between requests, in minutes) 
    # that a session will last before the garbage collector junks it.  
    # (Default 1 hour.)

    max_idle => 60,

);

#======================================================================
# GENERAL CONFIGURATION 

%config = (

    #---------------------------------
    # HTML output options:
    #---------------------------------

    # PAGE BUFFERING
    # set true if using the page buffering module, ExSite::Page.pm, by default.
    # set false if output goes straight to stdout by default.
    # NB: page buffering can be changed from the default for specific scripts 
    # by setting this flag at the top of the script.

    page_buffering => 0,

    # Thumbnail size (max edge length in pixels)

    thumbnail_size => 100,

    # handling of database images
    # set true if ExSite::Report::show_data should display images inline
    # set false if ExSite::Report::show_data should display a link to the image

    show_images => 1,

    # DBmap type

    dbmap_type => "external",

    # maximum upload file size
    # NB: form.bigfile_maxsize limits the size of a file that can be loaded
    # into a bigfile DB record.  This parameter, on the other hand, limits 
    # the size of a file that can be uploaded through a web form for other
    # reasons (eg. publish straight to disk).

    max_upload_size => 16000000,

    # handling of database operations
    # each operation has the following parameters:
    #   approve: the low-level operation that needs to be approved by DB::approve
    #   url: the script that executes the operation
    #   img: the image for the operator icon
    #   img_width, img_height: the dimensions of the image
    #   label: the text string to use if the image is not present
    #   table: 1 means operation can be performed on whole tables
    #          (otherwise, op is restricted to individual records)
    # Each url is passed ?tab=<table>[&id=<record>]

    db_ops => {
	edit => { approve=>"update", url=>"webdb.cgi?action=edit", 
		  img=>"edit12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"edit", table=>0 },
	copy => { approve=>"insert", url=>"webdb.cgi?action=copy", 
		  img=>"copy12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"copy", table=>0 },
	delete => { approve=>"delete", url=>"webdb.cgi?action=delete", 
		  img=>"rm12.gif", level=>3,
	          img_width=>12, img_height=>13, label=>"delete", table=>0 },
	new => { approve=>"insert", url=>"webdb.cgi?action=new", 
		  img=>"new12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"add new", table=>1 },
	view => { approve=>"read", url=>"webdb.cgi?action=view", 
		  img=>"view12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"view details", table=>0 },
	list => { approve=>"read", url=>"webdb.cgi?action=list", 
		  img=>"list12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	report => { approve=>"read", url=>"webdb.cgi?action=report", 
		  img=>"list12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	searchquery => { approve=>"read", url=>"webdb.cgi?action=query", 
		  img=>"search12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"search", table=>1 },
    },

    db_ops_grps => {
	#all => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
	all => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],
	all_edit => ["edit", "copy", "delete"],
	all_view => ["list", "view", "search"],
	link => ["view"],  # kludge to support old "link" args
    },

    # preferred order for displaying db operators, when more than one is shown
    # ops not in this list will not be displayed at all.

    #db_ops_order => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
    db_ops_order => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],

    # HTML snippets (user-defined)

    snippet => {},

    # include the other configuration hashes here

    site => \%site_config,
    auth => \%auth_config,
    form => \%form_config,
    report => \%report_config,
    log => \%log_config,
    webtop => \%webtop_config,
    crypt => \%crypt_config,
    session => \%session_config,
    server => \%server_config,  # conditional

    # CACHE - maximum size of the data cache

    cache => {
	max_items => 100,
	max_size => 10000,   # 10 K
    },

);

#### end of static configuration


##################################################################
################ setup and configuration routines ################
##################################################################

=pod

=head1 Setup and Configuration Routines

=head2 server_autoconf() - autoconfigure ExSite

This routine attempts to set the various server configuration
parameters (ie. file paths and machine names) automatically, by
inferring them from the server environment.

WARNING: auto-configuration will probably fail when testing CGI
programs from the command line for testing/debugging purposes.  If you
need to have correct configuration, then either hard-code your
configurations in exsite.conf, or set some environment variables on
your command line to resemble the server environment.  The following
environment variables are used for auto-configuration: DOCUMENT_ROOT,
SCRIPT_NAME, SCRIPT_FILENAME, HTTPS, HTTP_HOST.

=cut

sub server_autoconf {
    $server_config{HTMLroot} = $ENV{DOCUMENT_ROOT};
    $ENV{SCRIPT_NAME} =~ m:^.*/([^/]+)$:;
    my $script_name = $1;
    $ENV{SCRIPT_FILENAME} =~ /^(.*)\/$script_name$/;
    $server_config{CGIroot} = $1;
    $server_config{CGIpath} = ($ENV{SCRIPT_NAME} =~ /^(.*)\/.*\.(cgi|pl)$/)? $1 : "";
    $server_config{protocol} = exists $ENV{HTTPS} ? "https" : "http";
    $server_config{HTMLpath} = $server_config{CGIpath};
    $server_config{HTMLpath} =~ s/\/cgi(-bin)?\//\//;
    $server_config{HTMLpath} =~ s/\/cgi(-bin)?$/\//;
    if ($server_config{HTMLpath} eq "/") {
        $server_config{HTMLpath} = "";
    }

    # obsolete path parsing methods
    #$server_config{HTMLpath} =~ s/\/cgi(-bin)?\//\//;
    #$server_config{HTMLpath} = (($server_config{CGIpath} =~ /^\/.+?(\/.+)$/)? $1 : ""),

    # obsolete host parsing method - only does 3-part hostnames
    #$ENV{HTTP_HOST} =~ /^(.+?)\.(.*?):?(\d+)?$/;
    #$server_config{host} = $1;
    #$server_config{domain} = $2;
    #$server_config{port} = $3 ? ":$3" : "";

    # new method - handles 2- or 3-part names
    $ENV{HTTP_HOST} =~ /^((.+?)\.)?(.*?\..*?)(:\d+)?$/;
    $server_config{host} = $2;
    $server_config{domain} = $3;
    $server_config{port} = $4;

}

=pod

=head2 read_exsite_conf() - load config parameters from a file

This routine loads configuration values from files.  By default 
it looks for "exsite.conf" files in the conf, ExSite and . directories.
An alternate filename can be given.  By default configuration values 
are read straight into the root of C<%config>, but a starting key can 
be given, eg. C<&read_exsite_conf("MyModule.conf","MyModule")> will
read the values in MyModule.conf into C<$config{MyModule}{...}>.

Configuration files contain parameters and values, separated by "=" 
(with any amount of whitespace around the "=".  Values extend from the 
first non-whitespace character after the "=", to the end of line.
Simple parameters are placed directly into C<%config>'s keys.  
Parameters of the form "a.b.c" are stored as C<$config{a}{b}{c}>. 
Config array references can be created using "+=" to set the values
in the array.

Example exsite.conf settings:

C<trim_whitespace = 0>

C<form.style=list>

C<db_ops.new.url = append.cgi>

=cut

sub read_exsite_conf {
    my ($file,$key) = @_;
    if (! $file) { $file = "exsite.conf"; }
    if ($key) { $config{$key} = {}; }
    my @confpath = ("./conf", "ExSite", ".");
    foreach my $dir (@confpath) {
	if (-e "$dir/$file") {
	    if (open CONF,"<$dir/$file") {
		foreach (<CONF>) {
		    next if (/^\#/);
		    next if (/^\s*$/);
		    chomp;
		    my ($param,$value,$append);
		    if (/\+=/) { 
			($param,$value) = split /\s*\+=\s*/,$_,2;
			$append = 1;
		    }
		    else {
			($param,$value) = split /\s*=\s*/,$_,2;
			$append = 0;
		    }
		    my @param = split /\./,$param;
		    my $conf = $key ? $config{$key} : \%config;
		    foreach my $par (@param[0..$#param-1]) {
			if (! exists $conf->{$par}) {
			    $conf->{$par} = {};
			}
			# this will crash if not a hash ref
			$conf = $conf->{$par};
		    }
		    if ($append) {
			# append parameter to list
			if (! exists $conf->{$param[-1]}) {
			    $conf->{$param[-1]} = [$value];
			}
			else {
			    if (ref $conf->{$param[-1]} ne "ARRAY") {
				my $tmplist = [ $conf->{$param[-1]} ];
				$conf->{$param[-1]} = $tmplist;
			    }
			    push @{$conf->{$param[-1]}},$value;
			}
		    }
		    else {
			# simple parameter
			$conf->{$param[-1]} = $value;
		    }
		}
		close CONF;
	    }
	}
    }
}

=pod

=head2 exsite_init() - generic ExSite initialization routine

This routine configures ExSite, and performs any local initializations.
This routine is called automatically at the start of all
generic CGI scripts, and should be called by custom scripts
as well.

Local initialization can be done in C<&myConfig::my_exsite_init>,
which is called automatically from here.  Note that C<my_exsite_init()> 
is called on every initialization, so it should try to avoid 
reconfiguring static configurations (eg. handler registration) on 
every pass.

=cut

sub exsite_init {

    if (! $config_flag) {
	# Only setup the config hash if $config_flag is unset;
	# this prevents unnecessary reloading in mod-perl setups.

	# load any extra configuration from conf files
	&read_exsite_conf();

	# autoconfigure server?
	if ($config{server}->{auto}) {
	    &server_autoconf;
	}
	
	# build the server address from the various config params
	if ($config{server}{host}) {
	    $config{server}{server} = "$config{server}{protocol}://$config{server}{host}.$config{server}{domain}$config{server}{port}";
	}
	else {
	    $config{server}{server} = "$config{server}{protocol}://$config{server}{domain}$config{server}{port}";
	}
	
	$config_flag = 1;
    }

    # reset share hash
    %share = (
	      diagnostics => [],
	      );

    # setup session hash, if supported
    if ($config{session}{on}                # if sessions are enabled
	&& $config{session}{loaded}         # and the session code is loaded
	&& -w $config{session}{lockfile}    # and the session file is writeable
	) {
	# (we test the lockfile instead of the session file, because DBM
	# may append its own suffix to the end of the session file name)

	# we appear to have a writeable session lockfile;
	# assume that we have a writeable session file;
	# tie the %session hash to it

	### FIXME: we should not create a session for bots

	tie %session, 'ExSite::Session';

	### FIXME: probably not necessary to redo this step on each 
	### request in a mod-perl setup
    }

    # localized initialization
    &my_exsite_init(\%config,\%share,\%session,@_);

}

=pod

=head2 exsite_close() - clean up

=cut

sub exsite_close {
    # localized clean-up;
    &my_exsite_close(@_);
    # close default database
#    if ($share{server}->{DB}) {
#	$share{server}->{DB}->disconnect;
#    }
}

=pod

=head2 page_header($title) - standard top-of-page HTML

Outputs the standard HTML to begin an ExSite administration page
(eg. a control panel).  Optionally, a page title can be provided.

Invokes C<&myConfig::my_page_header>, which should contain any
localized code for generating locally customized page headers.

=cut

sub page_header {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_header",@_);
    }
    if (! $out) {
	$out = &my_page_header(@_);
	if (! $out) { 
	    # no custom header - use generic ExSite system wrapper
	    my ($title,$opt) = @_;
	    $out = "<HTML>
<HEAD>
<TITLE>$title</TITLE>
<meta http-equiv=\"Content-Type\" content=\"text/html; charset=ISO-8859-1\">
<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"$config{server}{HTMLpath}/_ExSite/css/ExSite.css\">
<script LANGUAGE=javascript>
function popup_help(url) {
  window.open(url,\"\",\"width=500,height=500,resizable,scrollbars=1,menubar=0\")
}
</SCRIPT>
</HEAD>
<BODY MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\" TOPMARGIN=\"0\" LEFTMARGIN=\"0\" RIGHTMARGIN=\"0\" BGCOLOR=\"#9999CC\">
<table cellspacing=0 cellpadding=0 border=0 bgcolor=#ccccdd width=100%>
<tr><td colspan=3 bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td></tr>
<tr><td bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td>
<td WIDTH=100%>

<!-- begin menubar -->

<table cellspacing=0 cellpadding=4 border=0 bgcolor=#ccccdd width=100%>

<td align=left WIDTH=100%>

<!-- logo -->
<a href=http://www.ExsiteWebware.com><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/exsite-button-round.gif height=30 width=30 border=0></a>

</td>
<td><A HREF=$config{server}{CGIpath}/home.cgi><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-home.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"ExSite Home\"></A></td>
<td><A HREF=\"javascript:popup_help('$config{server}{CGIpath}/exsite-help.cgi')\"><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-help.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Help\"></A></td>
<td><A HREF=$config{server}{CGIpath}/logout.cgi><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-exit.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Log Out\"></A></td>
</tr>
</table>

<!-- end menubar -->

</td>
<td bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td>
</tr>
<tr><td colspan=6 bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td></tr>
</table><br>\n\n";

	    if (! $opt || ref $opt) {
		# opt flags the webtop in the default presentation
		# if !opt, then draw a window
		$out .= "<div align=center>
    <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0 WIDTH=90% CLASS=framed>
      <TR>
        <TD CLASS=titlebar><h1>$title</h1></TD>
      </TR>
      <TR>
        <TD BGCOLOR=#FFFFFF><DIV CLASS=htmlbox>\n";
	    }
	    elsif ($opt != 1) {
		# opt is actual content, not a flag
		$out .= $opt;
	    }
	}
    }
	
    # do not modify anything below this point
    
    if ($config{page_buffering} || $share{PageBuf}) { 
	return $out; 
    }
    else { 
	print "Content-type: text/html\n\n";
	print $out; 
    }
}

=pod

=head2 page_footer() - standard end-of-page HTML

Outputs the standard HTML to end an ExSite administration page.
Invokes C<&myConfig::my_page_footer>, which should contain any
localized code for generating the page footer.

=cut

sub page_footer {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_footer",@_);
    }
    if (! $out) {
	$out = &my_page_footer(@_);
	if (! $out) { 
	    # no special page footer - use ExSite system wrapper
	    my ($ref) = @_;
	    if (! $ref || ref $ref) {
		# close the pseudo-window
		$out = "        </DIV></TD>
      </TR>
    </TABLE>
</div>\n";
	    }
	    elsif ($ref != 1) {
		# custom footer
		$out = $ref;
	    }
	    if ($share{DB}) {
		# there is a database object - check for any diagnostics to report
		if ($share{DB}->errorcheck()) {
		    $out .= "<h3>Errors:</h3>\n";
		    $out .= $share{DB}->show_diagnostics("error","html");
		}
		if ($share{DB}->errorcheck("warn")) {
		    $out .= "<h3>Warnings:</h3>\n";
		    $out .= $share{DB}->show_diagnostics("warn","html");
		}
	    }
	    $out .= "</BODY></HTML>\n";
	}
    }
    if ($config{page_buffering} || $share{PageBuf}) { return $out; }
    else { print $out; }
}

##################################################################
################ object management                ################
##################################################################

=pod 

=head2 get_obj() - fetch a standard object

There are a handful of global objects that ExSite commonly works with.
The most common are the primary database object, and the current page
being constructed.  These are normally stored in %share to make them 
available to all components.

You can use get_obj() to fetch a predefined object safely.  That is,
it will return the predefined object if it exists, and it will create
the object if it does not exist.

Example:

C<my $db = &get_obj('DB');>

=cut

sub get_obj {
    my $obj = shift;
    if ($share{$obj}) {
	if (ref $obj =~ /$obj/) {
	    return $obj;
	}
	# error: no such object, and the %share spot is occupied
	return undef;
    }
    
    # no object found in %share;  look for class in ExSite
    if (eval "require ExSite::$obj") {
	my $o = eval "new ExSite::$obj";
	if ($o) {
	    $share{$obj} = $o;
	    return $o;
	}
    }

    # hm, nothing in ExSite;  look for class in Modules
    if (eval "require Modules::$obj") {
	my $o = eval "new Modules::$obj";
	if ($o) {
	    $share{$obj} = $o;
	    return $o;
	}
    }
    
    # nothing found
    return undef;
}

##################################################################
################ handlers                         ################
##################################################################

=pod

=head2 install_handlers() - register handlers

Handlers allow plug ins or other subsystems to register special 
business logic that should take precedence over the default logic
in certain cases.

The ExSite base system has to do this in a few cases for its content
management tables, which have their own logic that differs from the
generic database handling.  Otherwise, it is expected that most 
handlers will be site-specific (registered in myConfig::my_handlers() )
or plug-in-specific (registered from the plug-in module in 
ExSite::Modules).

=cut

sub install_handlers {
    my $db = $share{DB};
    if ($db) {
	# standard handlers here
	$db->handler("select_foreign_key",\&CMS_select_foreign_key);
    }
    # additional handlers unique to this installation 
    &my_handlers(@_);
}

# handler for foreign key selection on CMS tables
#
# - for *.template_id, select only valid templates
# - for *.section_id, select only from my_sites()
# - for page.parent_id, select only from other pages in same section
# - for content.page_id, select only from pages in same section
# - for section.parent_id, select only from other non-template sections

sub CMS_select_foreign_key {
    my ($this, $table, $rec, $col, $ftable) = @_;
    # ignore, if not using the CMS database
    if ($this->{name} eq $config{server}{db}{name}) {
	if ($col eq "section_id") {
	    if (! $this->is_admin) {
		my @sites;
		@sites = $this->my_real_sites();
		return \@sites;
	    }
	    return undef;
	}
	if ($table eq "page") {
	    if ($col eq "template_id") {
		# select only from templates
		### more authorization needed here, eg. disallow 3rd-party private templates
		return $this->fetch_match("page",{type=>"template"});
	    }
	    elsif ($col eq "parent_id") {
		my (@possible_parents,$page,@parents);
		# select only from pages in same section
		$page = $this->fetch("page",$rec);
		@parents = $this->fetch_match("page",{type=>"page",section_id=>$page->{section_id}});
		# don't allow user to select a page as its own parent
		foreach my $parent (@parents) {
		    if ($parent->{page_id} ne $rec) {
			push @possible_parents, $parent;
		    }
		}
		return \@possible_parents;
	    }
	}
	if ($table eq "content") {
	    if ($col eq "page_id") {
		my ($content,$page);
		$content = $this->fetch("content",$rec);
		$page = $this->fetch("page",$content->{page_id});
		return $this->fetch_match("page",{section_id=>$page->{section_id}});
	    }
	}
	if ($table eq "section") {
	    if ($col eq "parent_id") {
		my $sql = "select * from section where type != 'website template'";
		$sql .= ($rec =~ /^\d+$/) ? " and section_id != $rec" : "";
		return $this->custom_query($sql);
	    }
	    elsif ($col eq "template_id") {
		return $this->fetch_match("page",{type=>"template"});
	    }
	}
    }
    return undef;
}

1;

