#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Content;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentData;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentData);

#------------------------------------------------------------------------

=pod

=head1 Content.pm : manage content objects

A content object is an element of a web page, such as text, html, 
stylesheets, scripts, images, or downloadable files.  This module
provides methods to act on generic content objects.

new: defines a new content object

Also inherits from the ContentData and ContentBase classes.

=cut

#------------------------------------------------------------------------

# setup info for Content class

sub setup {
    &setupContent(@_);
}

sub setupContent {
    my ($this,%option) = @_;

    # fetch the object in question
    ###### CHECK PERMISSIONS.
    my $content;
    if ($option{id}) {
	# direct lookup
	$content = $this->fetch("content",$option{id});
    }
    elsif ($option{content}) {
	# record already fetched
	$content = $option{content};
    }
    elsif ($option{name} && $option{page}) {
	########################
	### search templates for matching content for general display of page,
	### but don't search in CMS management mode

	my @content = $this->fetch_match("content",
					 {name=>$option{name},
					  page_id=>$option{page}},
					 "content_id");
	# should be only 1, but in case there are more, we'll take the
	# most recently added version.
	$content = $content[-1];
    }

    if (ref $content eq "HASH") {
	# found something
	$this->info("setting up Content $content->{content_id}");
	$this->{data} = $content;
	if ($option{version}) {
	    # this will load a version, allowing the content to be rendered
	    $this->set_context("child",$option{version});
	}
	else {
	    # this is faster for simple content queries
	    $this->set_context("none");
	}
	
	#$this->{name} = $content->{name};
#	my $context = "none";
#	if (! defined $option{context} || $option{context}) {
#	    $context = $option{context} ? $option{context} : "all";
#	}
#	$this->set_context($context,$content);
    }
    else {
	$this->info("setting up null Content");
    }
}

# validate 

# validate - check integrity of datahash

sub validate {
    my ($this,$data) = @_;
    my $content;
    if ($data) {
	$content = new ExSite::Content(content=>$data);
    }
    else {
	$content = $this;
	$this->set_context("child");
	$data = $this->get("content");
    }
    my @err = $share{DB}->validate_record("content",$data);
    if (! $data->{name}) {
	push @err, "content has no name";
	return undef;
    }
    if ($data->{page_id}) {
	if (! $content->owns()) {
	    push @err, "ownership mismatch";
	}
    }
    return @err;
}


# make_content : insert a new content record
#
# can optionally pass a content data record to seed it with

sub make {
    return &make_content(@_);
}

sub make_content {
    my ($this,$content,$cdata,$opt) = @_;
    if ($content->{content_id}) {
	# invalid content
	$this->error("invalid attempt to insert an already-defined content record");
	return undef;
    }
    if (! $this->allow_insert("content",$content)) {
	# permission denied
	$this->error("make_content: permission denied");
	return undef;
    }
    my @err = ExSite::Content::validate($this,$content);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }
    if (! $content->{type}) {
	$content->{type} = "editorial";
    }
    # remove any residual timestamps
    delete $content->{mtime};
    delete $content->{ctime};
    $share{DB}->run_handler("pre_make_content",$content);
    my $cid = $share{DB}->insert("content",$content);
    if (! $cid) {
	$this->error("failed to install new content $content->{name}");
    }
    else {
	$this->info("installed new content $content->{name}");
	if (ref $cdata eq "HASH") {
	    $cdata->{content_id} = $cid;
	    my $cdid = $this->make_content_data($cdata,$opt);
	    # ignore return code; make_content_data will report its own errs
	}
	$content->{content_id} = $cid;
	$share{DB}->run_handler("post_make_content",$content);
    }
    return $cid;
}

# update the content - synonym for make_content_data

sub update {
    return &ExSite::ContentData::make_content_data(@_);
}

# roll back to the previous version of the content

sub rollback {
    my $this = shift;
    $this->set_context("child");
    my $cdata = new ExSite::ContentData(content_data=>$this->{context}{content_data});
    $cdata->delete;
    if ($cdata->errorcheck()) {
	# copy the errors into this object, so we can see them later
	$this->error($cdata->fetch_diagnostics("error"));
    }

    ### this will have unintended consequences if the version in use
    ### is not the latest version!

    # reconfigured the context with the previous version
    $this->{context}{"content_data"} = $this->find_content_data();
}

# count the number of versions on file

sub count {
    my $this = shift;
    return $share{DB}->count("content_data",
			     {content_id=>$this->id("content")});
}

# authtype = editorial|design

sub authtype {
    my $this = shift;
    return $this->get("content","type");
}

# allow the current user to modify the content

sub authorize {
    my $this = shift;
    $this->set_context();
    my $section_id = $this->get("section","section_id");
    if (! $section_id) {
	my $p = $share{DB}->fetch("page",$this->get("content","page_id"));
	$section_id = $p->{section_id};
    }
    return 0 if (! $section_id);
    
    my @key = $share{DB}->fetch_match("member_site_link",
				      {section_id=>$section_id,
				       member_id=>$share{DB}->my_uid});
    my $type = $this->authtype;
    foreach my $key (@key) {
	return 1 if ($type eq "editorial" && $key->{priveleges} =~ /editor/);
	return 1 if ($type eq "design" && $key->{priveleges} =~ /designer/);
    }
    # no matching priveleges found
    return 0;
}

# everything else is inherited :-)

### FIXME:
# - is_published() - determine whether content needs to be published
# - publish_date() - determine when content was last published

# delete the content object

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("content",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("content",$id);
    }
}

1;
