#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentBase;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Base;
use ExSite::Input;
use ExSite::Cache;

use vars qw(@ISA @class_structure %class_index);
@ISA = qw(ExSite::Base);
@class_structure = ("section", "page", "content", "content_data");
%class_index = ("ExSite::Section"=>0, 
		"ExSite::Page"=>1, 
		"ExSite::Content"=>2, 
		"ExSite::ContentData"=>3, 
		"ExSite::ContentData2"=>3);

#------------------------------------------------------------------------

=pod

=head1 ContentBase.pm : generic content-management methods

This is the base class for the ExSite core content management
framework.  (Core content pertains to basic web page content
such as HTML, images, stylesheets, etc.)  The base class contains
tools that can be used/inherited by all CMS data types (ie. sites,
pages, content definitions, and content data), as well as tools
that deal with the entire content stack at once.

NOTE: CMS objects can exist in two forms:  (1) as data objects (a hash
of attribute keys/values, equivalent to a DB record), or (2) as code
objects.  The names of the data objects correspond to the names of 
the DB tables used to store their attributes.  The names of the code
objects correspond to the Perl modules/classes that define them. 
These names are similar, but do not match exactly:

    +------------------------------+-----------------------------+
    |  DATA OBJECT                 |  CODE OBJECT                |
    +------------------------------+-----------------------------+
    |  section                     |  Section                    |
    |  page                        |  Page                       |
    |  content                     |  Content                    |
    |  content_data                |  ContentData                |
    +------------------------------+-----------------------------+

It can be considered a bug that we use both conventions when referring to
object types, not always consistently.

=head2 C<new(%option)> - generic constructor

This is a general-purpose constructor that is used by all CMS
classes.  As such, it is generally called via one of the other
CMS classes (ie. C<ContentData>, C<Content>, C<Page>, or C<Section>).
Usage is typically:

    my $page = new ExSite::Page(id=>$page_id);      # numeric ID

    my $page = new ExSite::Page(page=>$page_data);  # page datahash

where "C<page>" can be replaced with C<ContentData>, C<Content>, 
or C<Section>).

=cut

#------------------------------------------------------------------------


# options
# name => content object name
# page => page this content is assigned to
# id => direct lookup of content version
# version => the version to select
# DB => database object (optional)

sub new ($%) {
    my($this, %option) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;

    # setup content cache - in share, so that multiple objects can share it

    if (! exists $share{ccache}) {
	$share{ccache} = new ExSite::Cache;
    }

    # setup Base class features

    $cobj->initialize_object;

    # register CMS-specific handlers

#    $cobj->{DB}->handler("select_foreign_key",&CMS_select_foreign_key);

    # setup context hash

    $cobj->{context} = {};

    # end generic constructor; now run local class setup

    $cobj->setup(%option);
    return $cobj;
}

######################### CONTEXT METHODS ############################

=pod

=head1 Context Handling

The "context" consists of the related elements in the content "stack".
For instance, an image object will belong to a certain page, which will
belong to a certain section;  it will also have a current version among 
the known versions.

=head2 C<set_context($mode,$version)> - find related content

The rules for identifying parents in the stack is simply to use the
foreign key reference in the child record.  The rules for identifying
children in the stack are more complicated:

=over 4

=item * the child of a site is the index page for the site

=item * the child of a page is the "page" content object for the page

=item * the child of a content object is the current version of the content

=back

Setting the context consists of finding and loading the other objects 
in this stack.  It is assumed that the current data object is already 
loaded.

C<$mode> is C<none|all|parent|child>, depending on whether one needs
no context, full context, parent context, or child context only.

C<$version> is not really used at this time, but can be used to identify
non-current versions that you wish to select instead of the current 
version.

Many of the context-testing functions below will not return useful
results if the context has not been set.  However, we don't want to 
automatically set the context every time, because it is costly and 
many basic operations do not require it.

=head2 C<clear_context()> - unset the context

Removes any context settings that have already been made.

=head2 C<reset_context($mode,$version)> - set the context again

C<set_context()> saves its results so that it will not execute again
unnecessarily.  C<reset_context()> forces the context to be set over
again. Parameters are the same as C<set_context()>

=cut

sub set_context {
    my ($this,$mode,$version) = @_;
    return if $this->{context_set};
    $mode = $mode || "all";
    if ($mode eq "all") { $this->{context_set} = 1; }
    # start with current object
    my $class = ref $this;
    my $index = $class_index{$class};
    $this->{context}{$class_structure[$index]} = $this->{data};
    return if ($mode eq "none");
    if ($mode =~ /all|parent/) {
	# move up thru the content stack by following foreign key refs
	for (my $i = $index; $i > 0; $i--) {
	    my $table = $class_structure[$i];
	    my $ptable = $class_structure[$i-1];
	    my $pkey = $ptable."_id";
	    my $id = $this->{context}{$table}{$pkey};
	    if ($id) {
		$this->{context}{$ptable} = $this->fetch($ptable,$id);
	    }
	    else {
		$this->warn("no $table found");
		last;
	    }
	}
    }
    if ($mode =~ /all|child/) {
	# move down thru the content stack
	my @cstack = @class_structure;
	while (scalar @cstack && $cstack[0] ne 
	       $class_structure[$class_index{$class}]) { 
	    shift @cstack; 
	};
	foreach my $ctype (@cstack) {
	    next if (ref $this->{context}{$ctype} eq "HASH"); # already loaded
	    if ($ctype eq "page") {
		$this->{context}{$ctype} = $this->find_page("index.%");
	    }
	    elsif ($ctype eq "content") {
		$this->{context}{$ctype} = $this->find("page","data");
	    }
	    elsif ($ctype eq "content_data") {
		$this->{context}{$ctype} = $this->find_content_data($version);
	    }
	}
    }
}

# clear the context

sub clear_context {
    my $this = shift;
    $this->{context_set} = 0;
    $this->{context} = {};
    # $this->{data} = {};   # not the context
    return;
}

# force new context, even if context was already set

sub reset_context {
    my $this = shift;
    $this->clear_context;
    $this->set_context(@_);
}

#------------------------------------------------------------------------

=pod

=head2 C<whatami()> - return the type of the current object

Returns one of "Page", "Section", "Content", or "ContentData".

=head2 C<is($type)> - test if the current object is of a certain type.

C<$type> can be one of "Page", "Section", "Content", or "ContentData".

Example - given a content object C<$x>:
    
    if ($x->is("Page")) { # we are a Page object

=head2 C<has($type)> - test if the context contains a C<$type>

C<$type> can be one of "Page", "Section", "Content", or "ContentData".
C<has()> returns true if the given type is found in the current object's
context.

Example - given a content object C<$x>:
    
    if ($x->has("Section")) { # we have our section data loaded into the context

=head2 C<has_content()> - test if the object has any content

Returns false if the object's content is empty/undefined, true otherwise.

Example - given a content object C<$x>:
    
    if ($x->has_content()) { # this content object contains real content

=head2 C<get($type,$attribute)> - fetch a content attribute

C<$type> can be one of "Page", "Section", "Content", or "ContentData".
C<$attribute> can be any attribute of the C<$type>.  If C<$attribute>
is not given, all attributes are returned in a datahash.

Example - given a content object called C<$img>, get the filename of 
the page it is located in:
    
    my $filename = $img->get("page","filename");

=head2 C<id($type)> - return numeric ID of C<$type>

C<$type> can be one of "Page", "Section", "Content", or "ContentData",
which must be loaded into the context.  If C<$type> is not given, 
returns the numeric ID of the current object.

Example - given a content object called C<$img>, get the ID of 
the page it is located in:
    
    my $id = $img->id();            # ID of the image
    my $page_id = $img->id("page"); # ID of the image's page

=head2 C<secure()> - are we running in secure mode?

C<$type> can be one of "Page", "Section", "Content", or "ContentData",
which must be loaded into the context.  If C<$type> is not given, 
returns the numeric ID of the current object.

Example - given a content object called C<$img>, get the ID of 
the page it is located in:
    
    my $id = $img->id();            # ID of the image
    my $page_id = $img->id("page"); # ID of the image's page

=cut

#------------------------------------------------------------------------

sub whatami {
    my ($this,$datafmt) = @_;
    my $iam = ref $this;
    $iam =~ s/^.+::(.+)$/$1/;
    if ($datafmt) {
	# underscore kludge
	$iam = lc($iam);
	return $iam eq "contentdata" ? "content_data" : $iam;
    }
    return $1;
}

sub is {
    my ($this,$type) = @_;
    # return true if this object is a $type
    return (ref $this =~ /^$type$/i);
}

sub has {
    my ($this,$type) = @_;
    # return true if this object has a $type loaded in its context
    return defined $this->{context}{$type};
}

sub has_content {
    my ($this) = shift;
    # return true if this has content data defined 
    return $this->has("content_data");
}

sub get { 
    my ($this,$type,$attribute) = @_;
    if ($type) {
	# default to ::Base behaviour if $type is not a context item
	return $this->{$type} if (ref $this->{context}{$type} ne "HASH");
	# if $type is a context item, return data for it
	return $attribute ? 
	    $this->{context}{$type}{$attribute} : 
	    $this->{context}{$type};
    }
    else {
	# no args - return this object's data
	return $this->{context}{$class_structure[$class_index{ref $this}]};
    }
}

sub id {
    my ($this,$type) = @_;
    my $type = $type ? $type : $class_structure[$class_index{ref $this}];
    return $this->{context}{$type}{$type."_id"};
}

#------------------------------------------------------------------------

=pod

=head1 CMS Access Control

Access control in the ExSite CMS is normally handled on a section-by-section
basis.  Each section has its own set of managers, who can each have the
following types of access to the section:

=over 4

=item editorial

The manager can create, update, and delete editorial objects.

=item design

The manager can create, update, and delete design objects and templates.

=item administrator

The manager can create/delete pages.

=back

Who has manager access to a section, and what types of access they have, 
is determined by their administrator keys to the section.  An 
administrator key is a hash with the following keys/values:

=over 4

=item member_id

The UID of the user whom this key belongs to.

=item section_id

The section ID the key is for.

=item priveleges

A plain text list of the types of access (above) that are granted by
this key.

=back 

If you have a group of items that require different access
priveleges from the rest of the section, consider placing these items
into their own section and issuing a new set of keys for them.

=head2 C<get_keys()>

Returns the key(s) that the current user has for the current section.
They are returned in an array (or arrayref in scalar context), even
if there is only one.

=head2 C<allow_update($type,$data)> - can the user modify the C<$type>?

C<$type> can be one of "page", "section", "content", or "content_data",
C<$data> is the attribute datahash for the C<$type>, if not using the
one in the context.  Returns true/false.

=head2 C<allow_insert($type,$data)> - can the user create a new C<$type>?

C<$type> can be one of "page", "section", "content", or "content_data",
C<$data> is the attribute datahash for the C<$type>.  Returns true/false.

=head2 C<allow_delete($type,$data)> - can the user delete the C<$type>?

C<$type> can be one of "page", "section", "content", or "content_data",
C<$data> is the attribute datahash for the C<$type>, if not using the
one in the context.  Returns true/false.

=head2 C<owns()> - does the user "own" the object?

The user owns the object if they have a key for the section that
the object belongs to.

=cut

#------------------------------------------------------------------------

sub get_keys {
    my $this = shift;
    $this->set_context();  # need section ID to do authorization
    my $db = $share{DB};
    return $db->fetch_match("member_site_link",
			    {member_id=>$db->my_uid(),
			     section_id=>$this->id("section")});
}

sub allow_update {
    my ($this,$type,$data) = @_;
    if (! $type) { $type = $this->whatami(1); }
    if (! $data) { $this->set_context(); $data = $this->{context}{$type}; }
    foreach my $key ($this->get_keys()) {
	if ($key->{priveleges} =~ /administrator/) { return 1; }
    }
    $this->error("Configure $type: permission denied - you need administrator priveleges on this site.");
    return 0;
}

sub allow_insert {
    my ($this,$type,$data) = @_;
    my $db = $share{DB};
    if (! $type) { $type = $this->whatami(1); }

    # with inserts, we often have no preexisting object to set a context
    # on, so we must manually figure out what section the insert is 
    # going into.
    my $section_id;
    if ($type eq "section") {
	return 1 if ($share{DB}->is_admin);
    }
    elsif ($type eq "page") {
	$section_id = $data->{section_id};
    }
    elsif ($type =~ /content/) {
	my $page_id;
	if ($type eq "content") {
	    $page_id = $data->{page_id};
	}
	else {
	    my $content = $db->fetch("content",$data->{content_id});
	    if ($content) { $page_id = $content->{page_id}; }
	}
	if ($page_id) {
	    my $page = $db->fetch("page",$page_id);
	    if ($page) { $section_id = $page->{section_id}; }
	}
    }
    my @keys;
    if ($section_id) {
	@keys = $db->fetch_match("member_site_link",
				 {member_id=>$db->my_uid(),
				  section_id=>$section_id});
    }

    foreach my $key (@keys) {
	$_ = $key->{priveleges};
	if ($type =~ /^section$/) {
	    return 1 if ($share{DB}->is_admin);
	}
	elsif ($type =~ /^page$/) {
	    return 1 if (/administrator/);
	    return 1 if (/designer/ && $data->{type} =~ /template|library/);
	    # editors cannot create pages
	}
	elsif ($type =~ /^content$/) {
	    return 1 if (/administrator|designer/);
	    return 1 if (/editor/ && $data->{type} eq "editorial");
	}
	elsif ($type =~ /^content_data$/) {
	    my $content = $db->fetch("content",$data->{content_id});
	    my $ctype = $content->{type};
	    return 1 if (/designer/ && $ctype eq "design");
	    return 1 if (/editor/ && $ctype eq "editorial");
	}
    }
    $this->error("Insert new $type: permission denied - your administrator key does not allow you to do this.");
    return 0;
}

sub allow_delete {
    my ($this,$type,$data) = @_;
    if (! $type) { $type = $this->whatami(1); }
    if (! $data) { $this->set_context(); $data = $this->{context}{$type}; }
    foreach my $key ($this->get_keys()) {
	$_ = $key->{priveleges};
	if ($type =~ /^section$/) {
	    return 1 if (/administrator/ && $share{DB}->is_admin);
	}
	elsif ($type =~ /^page$/) {
	    return 1 if (/administrator/);
	    return 1 if (/designer/ && $data->{type} =~ /template/);
	    # editors cannot remove pages
	}
	elsif ($type =~ /^content$/) {
	    return 1 if (/administrator|designer/);
	    return 1 if (/editor/ && $data->{type} eq "editorial");
	}
	elsif ($type =~ /^content_data$/) {
	    return 1 if (/administrator/);
	}
    }
    $this->error("Delete $type: permission denied - your administrator key does not allow you to do this.");
    return 0;
}

sub owns {
    my ($this) = @_;
    return ($this->get_keys() > 0);
}

### DEPECATED:  see ExSite::Auth for access control tools
sub access {
    my ($this,$method,$class,$id) = @_;
    my $db = $share{DB};
    my $uid = $db->my_uid;
    my @sites = $db->fetch_m2m("section","member",$uid,"member_site_link");
    $this->set_context;

    # crude access control for now - simply check for a section match

    foreach my $site (@sites) {
	if ($site->{section_id} == $this->id("section")) {
	    return 1;
	}
    }
    return 0;
}

#------------------------------------------------------------------------

=pod

=head1 Content Searches

When building a page, ExSite encounters various tags that reference
content objects such as images, HTML, etc.  ExSite must search for
matching objects to resolve these references, within the parameters of
the current context.  (For example, there could be many C<index.html>
pages in the system; which one is meant depends on where one begins
looking for it.)

=head2 C<find($name,$mode)> - locate related content objects

Locates a named content object, starting from the current context
(which will be set automatically if not already set).  The object is
sought in the following locations:

    - in current page's content objects
    - in templates' content objects, including inherited templates
    - in content libraries for this section
    - in content libraries for ancestral sections

The first eligible match satisfies the algorithm, and searching will stop.

Returns the found object's attribute hash if C<$mode> is "data".  If
C<$mode> is "object", C<find()> will return a new content object.

=head2 C<find_page($name,$section_id)> - find a page

Returns the attribute hash for a page with filename C<$name>.  The
C<$section_id> may optionally be specified, if looking outside the
current context.

=head2 C<find_content($name,$page_id)> - find a content object

Returns the attribute hash for a content object named C<$name>.  The
C<$page_id> may optionally be specified, if looking outside the
current context.  This is used by the more powerful C<find()> method
to search individual pages, templates, and libraries for matching
content.

=head2 C<find_content_data($type,$content_id)> - find content data

Returns the attribute hash for the content data in the content object 
indicated by C<$content_id>.  C<$content_id> is optional, and only needed
in the case where the content object is different from that in the 
context.  C<$type> is not currently meaningful, and can be left undefined.

=cut

#------------------------------------------------------------------------

sub find {
    my ($this,$cname,$mode) = @_;

    if (! $this->{context_set}) {
    #if (! $this->{context}{page}) {
	# missing our context - can't just check context_set flag, because
	# set_context calls us before the context is fully set.
	$this->set_context;
    }

    my $page = $this->{context}{page};
    return undef if (! $page);

    # search the current page
    my $c = $this->find_content($cname);

    # search the master page
    if (! $c && $page->{type} eq "alternate") {
	$c = $this->find_content($cname,$page->{parent_id});
    }

    # search the template stack
    while (! $c && $page->{template_id}) {
	$page = $this->fetch("page",$page->{template_id});
	$c = $this->find_content($cname,$page->{page_id});
    }

    # search the libraries, if this is not the body
    if (! $c && $cname ne "body") { 
	# get current section
	my $section = $this->{context}{section};
	######## set the parent context first, to be sure it is defined?
	return undef if (! $section);
	my $sid = $section->{section_id};
	while ($sid) {
	    # get all libraries in this section
	    my @libs = $this->fetch_match("page",{type=>"library",
						  section_id=>$sid});
	    if (@libs > 0) {
		# search each library for a matching content object
		my $lib = shift @libs;
		while ($lib && ! $c) {
		    $c = $this->find_content($cname,$lib->{page_id});
		    $lib = shift @libs;
		}
		last if ($c);
	    }
	    # goto previous section
	    $sid = $section->{parent_id};
	    if ($sid) {
		$section = $this->fetch("section",$sid);
	    }
	}
    }

    return $c if ($mode eq "data");
    require ExSite::Content;
    return new ExSite::Content(content=>$c,version=>"current");
}

# more find tools

sub find_page {
    my ($this,$name,$sid) = @_;
    return undef if (! $this->{context}{section});
    if (! $sid) { $sid = $this->{context}{section}{section_id}; }
    return undef if (! $sid);
    my @page = $this->fetch_match("page",{filename=>$name,section_id=>$sid},
				  "page_id");
    return $page[-1];
}

sub find_content {
    my ($this,$name,$pid) = @_;
    return undef if (! $this->{context}{page});
    if (! $pid) { $pid = $this->{context}{page}{page_id}; }
    return undef if (! $pid);
    my @content = $this->fetch_match("content",{name=>$name,page_id=>$pid},
				     "content_id");
    return $content[-1];
}

sub find_content_data {
    my ($this,$type,$cid) = @_;
    return undef if (! $this->{context}{content});
    if (! $cid) { $cid = $this->{context}{content}{content_id}; }
    return undef if (! $cid);
    my @content_data = $this->fetch_child("content_data",
					  "content",
					  $cid,
					  "content_data_id");
    if ($type =~ /current|active/) {
	# return most recent approved version
	for (my $ic = $#content_data; $ic >= 0; $ic--) {
	    # if status flag is set, this content is waiting for approval
	    if (! $content_data[$ic]->{status}) {
		return $content_data[$ic];
	    }
	}
	return undef;
    }
    else {
	# return most recent version
	return $content_data[-1];
    }
}


#------------------------------------------------------------------------

=pod

=head1 Page Building and Content Expansion tools

=head2 C<get_start_html($mode)> - initialize the HTML for an object

C<page> content objects are typically built starting from an HTML 
framework that comes out of a template.  This template might be precompiled,
which means that it resides on disk as a partially-processed HTML file.  
If not precompiled, then it resides in the database as raw HTML.  This
method identifies and returns the appropriate block of initial HTML to
begin building the page from.

For non-page content objects, this simply returns an HTML representation
of the object.

=head2 C<expand(%options)> - expand/convert CMS tags

Expands the content in the current context, by replacing CMS tags with
the URLs or the raw content that they represent.  Returns the
HTML for the content.

Options:

=over 4

=item content

a hash of content names and values to use preferentially over 
those in the database

=item method

a combination of B<data, url, content, page, module,> depending on 
which expansions you want to apply to this content.  All are 
selected, by default.  You can also specify B<dummy-content> to 
substitute placeholder graphics instead of full content.

=item expand

C<template|design> - 
this flag requests partial expansion of the content.  Only
template/design content is expanded;  remaining tags
are left in place.  DCD tags are left unexpanded in both cases.

=item cms

if true, include CMS tools in the output, for interactive page
editing.

=item mode

"dynamic" (fetch everything from DB source) or "static" (default;
fetches published and precompiled versions where possible)

=back

The C<substitute()> method is an old synonym for C<expand()>.

=cut

#------------------------------------------------------------------------

sub get_start_html {
    my ($this,$mode) = @_;

    # for "page" objects, check to see if the template is precompiled,
    # as that will save us a lot of content lookups later.

    if ($mode ne "dynamic" &&
	$this->{context}{page} && 
	$this->{context}{content}) {
	# If we have a full page loaded,
	if ($this->{context}{page}{type} eq "page") {
	    # and it really is a page (not a template),
	    if ($this->{context}{content}{name} eq "page") {
		# and the content object is the base HTML for the page,
		if ($this->{context}{content}{page_id} != 
		    $this->{context}{page}{page_id}) {
		    # and this content object comes from an external template,
		    # ...then we can use a precompiled template, if available.
		    my $template_id = $this->get("page","template_id");
		    my $t = $share{DB}->fetch("page",$template_id);
		    if ($t->{publish_method} eq "static") {
			my $template = new ExSite::Page(page=>$t);
			my $path = $template->diskpath;
			# check for a precompiled template
			if (-e "$path/index.html") {
			    return &get_file("$path/index.html");
			}
		    }
		}
	    }
	}
    }
    # no precompiled template found;  convert the content_data object to HTML
    return $this->get_html($mode);
}

sub substitute {
    return &expand(@_);
}

sub expand { 
    my ($this,%option) = @_;

    return $this->{html} if ($this->{expanded} || $this->{noexpand});
    $this->info("expanding html");

    # monitor the content objects that we have expanded, to avoid
    # infinite loops
    $this->{expanded} = {};

    # fetch raw template data
    #$this->{html} = $this->get_html($option{mode});
    # fetch prepublished template data, if possible; otherwise fetch raw data
    $this->{html} = $this->get_start_html($option{mode});

    if (! exists $option{content}) { $option{content} = {}; }
    my $method = $option{method} || "data,url,content,page,module"; 
    my $cmsmenu = 0;

    # build a hash of meta data - combine page and section datahashes
    my $meta = $this->{context}{page};
    foreach my $key (keys %{$this->{context}{section}}) {
	$meta->{"site_".$key} = $this->{context}{section}{$key};
    }

    do {

	# clear substitution counter
	$this->{htmlsub} = 0;

	# 1) meta data substitutions, eg. <!--$name-->

	if ($method =~ /data/ && !($option{expand} eq "template") ) { 
	    $this->{html} =~ s/<!--\$(\w+)-->/$meta->{$1}/gs;
	}

	# 2) content-URL substitutions, eg. SRC=[[content-object-name]]

	if ($method =~ /url/) {
	    $this->{html} =~ s/\[\[([\w\.-]+)\]\]/$this->get_content_url($1,\%option)/ges;
	}

	# 3) page-URL substitutions, eg. HREF={{page-name/id}}

	if ($method =~ /page/) { 
	    $this->{html} =~ s/\{\{([\w\.-]+)\}\}/$this->get_page_url($1,\%option)/ges;
	}

	# 4) HTML content substitutions, eg. <!--&content(content-obj-name)-->

	if ($method =~ /content/) { 
	    # repeat until all substitutions are done
	    while ($this->{html} =~ /<\!--\&?content\((.+?)\)-->/) {
		$this->{html} =~ s/<\!--\&?content\((.+?)\)-->/$this->get_content($1,\%option)/ges;
	    }
	}

	# NOTE:  <!--&content(foo)--> is now deprecated in favor of 
	# <!--content(foo)--> which isn't syntactically ambiguous with the
	# DCD tags, below.
	    
	# 5) Dynamic Content Driver substitutions, eg. <!--&module.method(args)-->
	if ($method =~ /module/) {
	    # add a CMS menu DCD tag in advanced CMS mode
	    # NB: editmode selects a non-default way of handling inline CMS 
	    #     funcs; only add the menu if this is not set
	    if ($option{cms} && (! $config{cms}{editmode}) && ! $cmsmenu ) {
		$this->{html} =~ s/(<body[^>]*>)/$1\n<!--\&CMS\(cmscmd=pagemenu\)-->/i;
		# ensure that cmsmenu is only inserted once
		$cmsmenu = 1;
	    }
	    # insert DCD content
	    do {
		$this->{expand_redo} = 0;
		my $html = $this->{html};
		# direct substitution of dynamic content (same server request)
		$html =~ s/<\!--\&(\w+)\((.*?)\)-->/$this->get_dynamic_content($1,$2,\%option)/ges;
		# indirect substitution of dynamic content (separate request)
		$html =~ s/<\!--\&\&(\&)?(\w+)\((.*?)\)-->/$this->get_dynamic_content_indirect($2,$3,\%option,$1)/ges;
		if (! $this->{expand_redo} && ! $this->{noexpand}) {
		    $this->{html} = $html;
		}
	    } until (! $this->{expand_redo} || $this->{noexpand});

	    # Normally there is only one pass through above do...until loop.
	    # However, some DCDs may set flags to change the logic:
	    #    expand_redo - start over
            #    noexpand - abort

	}
    } until ($this->{htmlsub} == 0);

    if ($option{expand}) {
	# partial expansion was requested, which means we should now have
	# one or more masked CMS tags in our HTML.  Now we have to unmask them.
	$this->{html} = &restore_cms_tags($this->{html});
    }

    # flag expansion as done; also frees the hash we used to keep track
    # of expansions
    $this->{expanded} = 1;  
    return $this->{html};
}

=pod

=head2 C<get_page_url($id)> - return replacement text for a {{...}} CMS tag

Tags such as C<{{123}}> or C<{{index.html}}> will be replaced with 
the correct URL to the indicated page (referenced by a numeric page ID,
or the page file name). 

=head2 C<get_content_url($name,$option)> - return replacement text for a [[...]] CMS tag

Tags such as C<[[image.gif]]> will be replaced with the correct URL to 
the indicated content object.  C<$name> is the name of the content object.
C<$option> is a reference to an option hash;  only one option is currently
recognized:

=over 4

=item mode

C<static|dynamic> - force the URL to fetch the published or database versions
of the content, respectively.  This is optional;  the system defaults to
static, if it is available, dynamic otherwise.

=back

=head2 get_content($name,$option) - return replacement HTML for a C<E<lt>!--content(...)--E<gt>> CMS tag

This is similar to C<get_content_url()>, except the full HTML is substituted
instead of just the URL.  Accepts the same options as C<expand()>, and 
acts on the following:

=over 4

=item content

This can point to a hash of predefined content definitions, which you can
use to override the content database.

=item method

Same as the method item passed to C<expand()>.  The "B<dummy-content>" 
method is the only one used in this method.

=item expand

"B<template>" if we are precompiling a template, and wish to leave
page-specific data unexpanded.

=item cms

Inserts interactive page editing tools, if true.

=back

=cut

sub get_page_url {
    my ($this,$id,$option) = @_;
    my $p;
    if ($id =~ /^\d+$/) {
	# page id
	$p = new ExSite::Page(id=>$id,noload=>1);
    }
    else {
	# page name
	$p = new ExSite::Page(filename=>$id,
			      section_id=>$this->id("section"));
    }
    return $p->get_url;
}

# get_content_url : return replacement text for a [[...]] CMS tag

sub get_content_url {
    my ($this,$id,$option) = @_;
    # content reference
    my ($subobj, $vfile);
    $subobj = $this->find($id);
    #$subobj->set_context("child");  # load versions
    if (! $subobj || ! $subobj->defined) {
	# no matching content - escape the tag chars in case this
	# is not a CMS tag
	$vfile = &html_escape("[[$id]]","[]");
    }
    else {
	$vfile = $subobj->get_url($option->{mode});
    }
    return $vfile;
}

# get_content : return replacement text for a <!--&content(...)--> CMS tag

sub get_content {
    my ($this,$id,$option) = @_;

    $this->{htmlsub}++;
    my ($start,$end,$cobj,$cobjname,$csub,$newtext,@cobjname);
    @cobjname = split(/,/,$id);
    # first content object in list is substituted directly
    $cobjname = shift @cobjname;
# FIXME: removed this, as it blocks multiple calls to a compound content object
#    if ($this->{expanded}{$cobjname}) {
#	return "<span class=\"error\">ExSite Error: aborting recursive substitution of $cobjname</span>";
#    }
    $this->{expanded}{$cobjname} = 1;

    if ($option->{method} =~ /dummy-content/) {
	# we're not really substituting content, 
	# so use a placeholder graphic instead
	$newtext = $this->placeholder_image($id);
    }
    elsif ($option->{content}->{$cobjname}) {
	# predefined content object
	$newtext = $option->{content}->{$cobjname};
    }
    else {
	# look for matching content object in DB
	$cobj = $this->find($cobjname);
	#$cobj->set_context();  # load full context
	if ($option->{expand} eq "template" && 
	    $cobj->get("content_data","page_id") == $cobj->get("page","page_id") ) {
	    # defer substitution of local page content
	    # (we are using the placeholder image to "escape" the CMS tag)
	    $newtext = $this->placeholder_image($id);
	}
	elsif ($option->{expand} eq "design" && 
	    $cobj->get("content","type") eq "editorial") {
	    # defer substitution of editorial content 
	    $newtext = $this->placeholder_image($id);
	}
	else {
	    # substitute now
	    my $in = new ExSite::Input;
	    my $q = $in->query;

	    # FIXME:
	    # first two clauses that follow are experiments in inline
	    # content editing for a simplified CMS interface.  This should
	    # be cleaned up when it becomes clear which method is preferred.

	    if ($id eq "body" && $config{cms}{editmode} eq "easy" && 
		$option->{cms} && $q->{cmscmd} =~ /update/) {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    elsif ($id eq "body" && $config{cms}{editmode} eq "MySite" && 
		$option->{cms} && $q->{ecmd} eq "update") {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    else {
		$newtext = $cobj->get_html;
	    }
	    if ($option->{cms} && 
		$cobj->{context}{content}{page_id} == 
		$this->{context}{page}{page_id}) {
		###### should allow hotlinks to undefined content
		###### objects that are being inherited from template
		# can only manage content tied directly to this page
		# to manage templates, view the template in cms mode
		
		# inline editing links
		if ($share{DB}->owns() && 
		    (($config{cms}{editmode} ne "MySite") || $id eq "body")) {
		    $newtext .= $this->edit_content_link($cobj);
		    # highlight editable content sections
		    $newtext = "<div style=\"border-width:1px; border-color:red; border-style:dashed; padding:0px;\">
$newtext";
		    if ($option->{cms} > 1 && $config{cms}{editmode} eq "MySite") {
			# for the noobs
			$newtext .= "\n<br><a href=\"javascript:window.close()\">Return to website control panel</a>.\n";
		    }
		    $newtext .= "</div>\n";
		}
	    }
	}
    }
    
    # remaining content objects in list are content objects to 
    # insert into the first content object
    
    my $icobj = 1;
    foreach $cobjname (@cobjname) {
	if ($cobjname =~ /^\".+\"$/) {
	    # quoted string - plaintext substitution
	    # simple strings only - (no commas!)
	    $cobjname =~ s/^\"//;
	    $cobjname =~ s/\"$//;
	    $newtext =~ s/<!--\&?content\($icobj\)-->/$cobjname/g;
	}
	else {
	    # unquoted string - treat as a content object reference
	    $newtext =~ s/<!--\&?content\($icobj\)-->/<!--content($cobjname)-->/g;
	}
	$icobj++;
    }
    return $newtext;
}

=pod

=head2 C<get_dynamic_content($module,$args,$option)> - return replacement text for a C<E<lt>!--&Module(...)--E<gt>> CMS tag

Loads up the C<$module> plug-in, and calls its C<write()> method, passing
the C<$args> to it.  Returns the output of the module.  If the module 
fails to load/compile, a Perl error string will be returned.

C<$option> is the same as is passed to C<expand()>.  If the "B<expand>" option
or "B<dummy-method>" methods are selected, a placeholder image will be
substutited instead of the module content.

=head2 C<get_dynamic_content($module,$args,$option)> - return replacement text for a C<E<lt>!--&Module(...)--E<gt>> CMS tag

Loads up the C<$module> plug-in, and calls its C<write()> method, passing
the C<$args> to it.  Returns the output of the module.  If the module 
fails to load/compile, a Perl error string will be returned.

C<$option> is the same as is passed to C<expand()>.  If the "B<expand>" option
or "B<dummy-method>" methods are selected, a placeholder image will be
substutited instead of the module content.

=head2 C<get_dynamic_content_indirect()> - AJAX version of C<get_dynamic_content()>

This method returns replacement text for CMS tags of the form:

    <!--&&Module(...)-->    indirect substitution, direct re-links
    <!--&&&Module(...)-->   indirect substitution, indirect re-links

This fetches the DCD content using a separate server request, instead 
of inlining it directly.

The main advantage is that you can publish the main page to a static 
HTML file, yet keep some page elements dynamic.  This is especially 
useful for index.html pages, which must be static, but may contain 
dynamic elements (eg. recent news, upcoming events, current specials,
etc.).  The solution is either continuous republishing of the index
page (which may still be a better solution for heavily loaded sites)
or using an indirect dynamic content fetch.

The disadvantage is that the full page is slower (although the base
page may be much faster), and that JavaScript must be enabled to 
perform the secondary content fetches.

The '&&' variant does direct re-links, ie. links from the dynamic content
point to full URLs that generate a new page.  The '&&&' variant does 
indirect re-links, ie. links back to the same module in the same page
only fetch the DCD content and inline it dynamically into the current
page without generating a whole new page.

=cut

sub get_dynamic_content {
    my ($this,$module,$modargs,$option) = @_;

    my ($endname,$modname,$modmethod,$newtext);
    # save dcd objects for re-use
    if (! $this->{dcd}) { $this->{dcd} = {}; }

    # keep track of called DCDs, to avoid infinite recursion
    # FIXME: removed this, as it blocks multiple calls, not just recursion
    #if ($this->{expanded}{"$module($modargs)"}) {
	#return "<span class=\"error\">ExSite DCD Error: aborting repeated call to $module($modargs)</span>";
    #}
    #$this->{expanded}{"$module($modargs)"} = 1;

    # flag a substitution as having occured, so we can expand these results
    $this->{htmlsub}++;

    ($modname,$modmethod) = split(/\./,$module,2);
    if (!$modmethod) { $modmethod = "write"; }
    
    # run the DCD
    
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$modname/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = "<img alt=\"$module($modargs)\" src=$icon title=\"$module($modargs)\" name=\"$module($modargs)\" border=\"3\" style=\"border-style:outset;\">";
    }
    elsif ($modmethod =~ /^(read|write|ioctl)$/) {
	# find the module
	my $stat = eval "require Modules::$modname";
	if ($stat) {
	    if (! $this->{dcd}{$modname}) {
		# create the DCD object
		$this->{dcd}{$modname} = eval "new Modules::$modname";
	    }
	    my $dcd = $this->{dcd}{$modname};
	    if ($dcd) { 
		# check licensing
		my $license_ok = 1;
		if ($dcd->ioctl("isRestricted")) {
		    # this service requires licenses
		    my @licenses = $share{DB}->fetch_match("service",
							   {section_id => $this->id("section"),
							    name => $modname }
							   );
		    $license_ok = (@licenses > 0);
		    if (! $license_ok) {
			# log the problem
			$this->warn("$modname: unlicensed access from page ".$this->id("page"));
		    }
		}
		
		if ($license_ok) {
		    # execute the DCD method
		    $newtext = $this->{dcd}{$modname}->$modmethod($modargs) || $@;
		    # disable substitution for read and ioctl methods
		    if ($modmethod ne "write") {
			$newtext = "<!-- $modname.$modmethod executed -->";
		    }
		}
		else {
		    $newtext = "$modname: permission denied (no license)";
		}
	    }
	    else {
		$newtext = "$modname: $@";
	    }
	}
	else {
	    # $@ is $EVAL_ERROR
	    $newtext = "<H4>Error in $modname dynamic content driver:</H4>\n<P>\n$@\n</P>";
	    $this->error("$modname: $@");
	}
    }
    else {
	$newtext = "<!-- Error: invalid dynamic content driver method: $modmethod -->";
	$this->error("invalid DCD method: $modmethod");
    }
    return $newtext;
}

# AJAX version
#
# FIXME: in principle, we don't have to display a warning message in
# <noscript>...</noscript>.  We could actually display the the real
# content, as in a direct substitution.  Unfortunately, that would involve
# running the DCD twice, once for the <noscript> direct sub, and again for
# the JS indirect sub.  If we cached this data (in the %session?) we could
# spare ourselves that extra hit, and give ourselves some robustness to
# non-JS clients (eg. GoogleBot).
#
# Actually, there's no need to put it in the <noscript> section;  you could
# put it in the regular DIV, in which case, all clients see it and use it.
# Javascript-enabled clients then overwrite it with the most current 
# dynamic version, if there is one.  The double-hit is a reasonable price
# to pay if we are publishing, since the 1st hit only occurs once for the
# lifetime of the file.

sub get_dynamic_content_indirect {
    my ($this,$module,$modargs,$option,$indlink) = @_;
    my $newtext;
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$module/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = "<img alt=\"&${indlink}$module($modargs)\" src=$icon title=\"&${indlink}$module($modargs)\" name=\"&${indlink}$module($modargs)\" border=\"3\" style=\"border-style:outset;\">";
    }
    else {
	my $ref;
	if ($this->is_publishing) {
	    # referrer isn't really publish.cgi; should probably be page.cgi
	    my $pid = $this->id("page");
	    # note escaped =
	    $ref = "$config{server}{CGIpath}/page.cgi%3f_id=$pid";
	}
	else {
	    $ref = &url_escape($ENV{REQUEST_URI});
	}
	my $il = $indlink ? "_ajax=$module&" : ""; 
	$newtext = "<script langauge=\"JavaScript\" src=\"$config{server}{HTMLpath}/_ExSite/js/httprequest.js\"></script>
<div id=\"$module\">
<script language=\"JavaScript\">
<!--//
document.write(getRequestData('$config{server}{CGIpath}/dcd.cgi/$module?$modargs&${il}_ref=$ref'));
//-->
</script>
<noscript>
(You must have javascript enabled to view $module.)
</noscript>
</div>\n";
    }
    return $newtext;
}

sub placeholder_image {
    my ($this,$id) = @_;
    $id =~ s/\"/%22/g;
    $id =~ s/</%3C/g;
    $id =~ s/>/%3E/g;
    return "<img alt=\"content($id)\" src=\"$config{server}{HTMLpath}/_ExSite/images/checker.png\" title=\"$id\" name=\"$id\" HEIGHT=50 WIDTH=100>";
}

=pod

=head1 CMS Data Fetching Methods

These are wrappers for the DB data fetching methods.  The only
difference over the generic DB data fetching methods is that we
use more aggressive caching, since we know the number of records 
returned will be relatively low.

The parameters are the same as the equivalent methods in C<ExSite::DB>.

=over 4

=item C<fetch()>

=item C<fetch_child()>

=item C<fetch_match()>

=back

=cut

### TODO: additonal fetching/caching performance improvements can be
### realized by prefetching all content in advance, rather than doing
### individual searches for each content object.


sub fetch {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $share{DB}->fetch(@_);
	$share{ccache}->save($dat,@_);
    }
    if (wantarray) { return %$dat; }
    return $dat;
}

sub fetch_child {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $share{DB}->fetch_child(@_);
	# cache the whole query
	$share{ccache}->save($dat,@_);
	# now cache the individual records
	#my $pkey = $share{DB}->get_key($_[0]);  # calls DB server
	my $pkey = $share{DB}{map}->get_table_attr($_[0],"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$_[0],$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

sub fetch_match {
    my ($this,$table,$match,$sort) = @_;
    # expand the match hash for best indexing
    # match can be a scalar or hash ref
    my @match = ref $match ? %$match : ($match);
    my $dat = $share{ccache}->get($table,@match,$sort);
    if (! $dat) {
	$dat = $share{DB}->fetch_match($table,$match,$sort);
	$share{ccache}->save($dat,$table,@match,$sort);
	# now cache the individual records
	#my $pkey = $share{DB}->get_key($table); # calls DB server
	my $pkey = $share{DB}{map}->get_table_attr($table,"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$table,$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

=pod

=head2 Plug-in Fetching Methods

=over 4

=item C<get_dcd_all()>

Returns an array of installed DCD names.

=item C<get_dcd()>

Returns an array of DCD names available to the current site.

=back

=cut

# get_dcd_all : return an array of installed DCD names 

sub get_dcd_all {
    my $this = shift;
    my @dcd;
    opendir MODDIR, "$config{server}{CGIroot}/Modules";
    @dcd = readdir MODDIR;
    closedir MODDIR;
    @dcd = grep /\.pm$/, @dcd; # only accept Perl modules
    foreach (@dcd) { s/\.pm$//; }
    return sort @dcd;
}

# get_dcd : return an array of DCD names available to the current site

sub get_dcd {
    my $this = shift;
    my $section = shift;
    if (!$section) {
	$section = $this->id("section");
    }
    my @dcd = $this->get_dcd_all;
    my @dcd_allow;
    foreach my $dcd (@dcd) {
	my ($mod,undef) = split /\./,$dcd;
	if (eval "require Modules::$mod") {
	    my $s = eval "new Modules::$mod";
	    if ($s) {
		if ($s->ioctl("isRestricted")) {
		    # this service requires licenses
		    my @licenses = $share{DB}->fetch_match("service",
							   {section_id => $section,
							    name => $mod }
							   );
		    if (@licenses > 0) {
			push @dcd_allow, $dcd;
		    }
		}
		else {
		    push @dcd_allow, $dcd;
		}
	    }
	    else {
		$this->error($@);
	    }
	}
    }
    return @dcd_allow;
}

#------------------------------------------------------------------------
#
# placeholder methods : these are dummies that will be overloaded in
# higher-level CMS classes.  They are provided here to allow ContentBase
# to run stand-alone for testing.

sub get_raw {
    return;
}

sub get_html {
    return;
}

sub get_url {
    return;
}

sub setup {
    return;
}

sub edit_content_link {
    return;
}
    

1;
