#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentData;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentBase;
use ExSite::Cache;
use ExSite::Image;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentBase);

#------------------------------------------------------------------------

=pod

=head1 ContentData.pm : tools for manipulating content data 

A ContentData object is a particular piece of raw web content that is
attached to a content object.  This modules provides methods to act on
ContentData objects.

new: define a new ContentData object

content_type: return the mime type of the content data

get: return the object in the specified mode

get_raw: return the object's raw data

get_html: return the HTML tags needed to inline the object into a page

get_url: return the URL needed to fetch the object in a separate request

=cut

#------------------------------------------------------------------------

# setup info for ContentData class

sub setup {
    &setupContentData(@_);
}

# fetch the content_data record

sub setupContentData {
    my ($this,%option) = @_;
    my $cdata;
    if ($option{id}) {
	$cdata = $this->fetch("content_data",$option{id});
    }
    elsif ($option{content_data}) {
	$cdata = $option{content_data};
    }
    if (ref $cdata eq "HASH") {
	$this->info("setting up ContentData $cdata->{content_data_id}");
	$this->{data} = $cdata;
	$this->set_context("none");
    }
    else {
	$this->warn("setting up null ContentData");
    }
}

# is this content defined?

sub defined {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	return ($data->{data} || $data->{fdata} || 
		$data->{file} || $data->{url});
    }
    return 0;
}
	
# content queries

# return the content-type

sub content_type {
    my $this = shift;
    return $this->get("content_data","mime_type");
}

# return the timestamp

sub timestamp {
    my $this = shift;
    return $this->get("content_data","ctime");
}

# determine whether content is in a virtual file

sub is_file {
    my $this = shift;
    my $dat = $this->get("content_data");
    return ( ! $dat->{data} && $dat->{fdata} );
}

# determine whether content is an image

sub is_image {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^image\//;
}

# determine whether content is text

sub is_text {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^text\//;
}

# determine whether content is publishable to a file
# -> essentially true for all non-HTML mime types

sub publishable {
    my $this = shift;
    my $mimetype = $this->get("content_data","mime_type");
    return ($mimetype && $mimetype !~ /html/);
}

# in the event that the content data is an upload, return the file info
# (filename, size (K), mime-type)

sub get_fileinfo {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{fdata}) {
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$data->{fdata};
	    my $mime = &MimeType($file);
	    my $size = int(length($fdata)*6/8/100)/10;
	    return ($file,$size,$mime);
	}
	else {
	    return undef;
	}
    }
    return undef;
}

# return the raw content

##### check for published file, file pointer, DB file, and raw content

sub get_raw {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    return $data->{data};
	}
	elsif ($data->{fdata}) {
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$data->{fdata};
	    return decode_base64($fdata);
	}
	elsif ($data->{file}) {
	    return &get_file($data->{file});
	}
    }
    $this->warn("ContentData: no raw content defined");
    return undef;
}

# return the HTML tags needed to inline the content into a page

sub get_html {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    my $attributes = " $data->{attributes}";
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	my $type = $this->content_type;
	if ($type =~ /^text\/html$/) {
	    # html can be returned straight-up
	    return $this->get_raw;
	    #return $this->substitute;  ##### weird error on page views here
	}
	elsif ($type =~ /^text\/css$/) {
	    # stylesheets are referenced using a link tag
	    my $url = $this->get_url;
	    return "<link rel=\"stylesheet\" type=\"text/css\" href=\"$url\">";
	}
	elsif ($type =~ /^text/) {
	    # plain text is rendered preformatted
	    return "<pre>\n".$this->get_raw."</pre>\n";
	    # alternately, return it as-is
	    #return $this->get_raw;
	}
	elsif ($type =~ /^image/) {
	    # images are rendered using IMG tags
	    my $url = $this->get_url(@_);
	    return "<img src=\"$url\"$attributes>";
	}
	else {
	    # miscellaneous types are shown as hotlinks
	    my $url = $this->get_url(@_);
	    my $filename = $this->filename || "download $type";
	    return "<a href=\"$url\"$attributes>$filename</a>";
	}
    }
    $this->warn("ContentData: no html content defined");
    return undef;
}

# return a URL that fetches the content as a separate HTTP request
# mode = static|dynamic (defaults to static, if found, dynamic otherwise)

sub get_url {
    my $this = shift;
    my $mode = shift;
    if ($this->{publish}) { $mode = undef; }
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	# if there is a published version, use it
	if ($mode ne "dynamic" && ($data->{data} || $data->{fdata})) {
	    my $file = $this->filename;
	    my $diskpath = $this->diskpath;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		my $url;
		if ($data->{type} eq "page") {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    require ExSite::Page;
		    my $p = new ExSite::Page(page=>$data);
		    $url = $p->get_url;
		}

		## make sure we know our section
		#$this->set_context; 
		#require ExSite::Section;
		#my $sec = new ExSite::Section(section=>$this->get("section"));
		## content is saved in the section, presumably
		#if (-e "$config{server}{HTMLroot}$config{server}{HTMLpath}$url/$file") {
		### fails if $url contains server info
		    return $url."/".$file;
		#}
	    }
	}
	if ($mode ne "static") {
	    # use dynamic version
	    my $fname = $this->filename;
	    my $url = "$config{server}{server}$config{server}{CGIpath}/content.cgi/${fname}?id=".$this->id("content_data");
	    # $this->set_context;  # needed to get name, but do we really need that?
	    my $cobj = $this->get("content");
	    my $name = &url_escape($cobj->{name});
	    $url .= "&name=$name"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_static_url {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $file = $this->filename;
	    my $diskpath = $this->diskpath;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		my $url;
		if ($data->{type} eq "page") {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    require ExSite::Page;
		    my $p = new ExSite::Page(page=>$data);
		    $url = $p->get_url;
		}
		return $url."/".$file;
	    }
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_dynamic_url {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $url = "$config{server}{server}$config{server}{CGIpath}/content.cgi?id=".$this->id("content_data");
	    my $cobj = $this->get("content");
	    $url .= "&name=$cobj->{name}"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}{server}$config{server}{HTMLpath}/_ExSite/images/checker.png";
}

# return the HTML tags needed to inline a thumbnail of the content into a page

sub get_thumb_html {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	if ($data->{thumb}) {
	    # yay, a real thumb!
	    my $url = $this->get_thumb_url;
	    return "<img src=\"$url\" $data->{thumb_attributes} border=0>";
	}
	elsif ($this->content_type =~ /^image/) {
	    # fake a thumb by shrinking the image
	    my $url = $this->get_url;
	    my $attr = $data->{attributes};
	    $attr =~ /width=\"?(\d+)\"?/i;
	    my $width = $1;
	    $attr =~ /height=\"?(\d+)\"?/i;
	    my $height = $1;
	    if ($width !~ /^\d+$/ || $height !~ /^\d+$/) {
		# missing height, width attributes - compute from scratch
		my $img = new ExSite::Image($data->{fdata});
		($width,$height) = $img->dim;
	    }
	    my $th_attr;
	    if ($width =~ /^\d+$/ && $height =~ /^\d+$/) {
		if ($width <= 100 && $height <= 100) {
		    $th_attr = "width=$width height=$height";
		}
		elsif ($width > $height) {
		    my $h = ($height / $width) * 100;
		    $th_attr = "width=100 height=$h";
		}
		else {
		    my $w = ($width / $height) * 100;
		    $th_attr = "width=$w height=100";
		}
	    }
	    return "<img src=\"$url\" $th_attr border=0>";
	}
	else {
	    return $this->get_html;
	}
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

# return the URL to a thumbnail

sub get_thumb_url {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	if ($data->{thumb}) {
	    my $diskpath = $this->diskpath;
	    $diskpath =~ s/\.([^\.]*)$/_T.$1/;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		my $url;
		if ($data->{type} eq "page") {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    require ExSite::Page;
		    my $p = new ExSite::Page(page=>$data);
		    $url = $p->get_url;
		}
		my $file = $this->filename;
		$file =~ s/\.([^\.]*)$/_T.$1/;
		return $url."/".$file;
	    }
	    # not published - return dynamic url
	    return "$config{server}{CGIpath}/getdata.cgi?tab=content_data&col=thumb&id=$data->{content_data_id}";
	}
	else {
	    return $this->get_url;
	}
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

sub get_thumb_raw {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{thumb}) {
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$data->{thumb};
	    return decode_base64($fdata);
	}
	else {
	    return undef;
	}
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

### FIXME: do we need a get method that returns an administrative link to 
### the content data (or content object)?

# determine the filename of a content datum "virtual file"

sub filename {
    my $this = shift;
    my $file;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    my $suffix = &MimeToFile($data->{mime_type});
	    $file = "$data->{content_id}.$suffix";
	}
	elsif ($data->{fdata}) {
	    ($file, undef) = split /$config{form}{sepchar}/,$data->{fdata};
	    # strip inconvenient characters from filenames
	    $file = &clean_filename($file);
	}
	elsif ($data->{file}) {
	    # strip path info
	    my @path = split /[:\\\/]/, $data->{file};
	    $file = $path[-1];
	}
	else {
	    # strip path info
	    my @path = split /\//, $data->{url};
	    $file = $path[-1];
	}
	return $file;
    }
    return undef;
}

# determine the path to the file, if published

sub httppath {
    my $this = shift;
    $this->set_context;
    require ExSite::Page;
    my $page = new ExSite::Page(page=>$this->get("page"));
    # the path to the content data is the same as the path to its page
    return $page->httppath;
}

# determine the path to the file, if published

sub diskpath {
    my $this = shift;
    my $dir = shift;  # return directory only if true
    $this->set_context;
    #return undef if (! $this->{cdata});
    #return $config{server}->{HTMLroot}.$this->httppath;
    my $cdata = $this->get("content_data");
    if ($cdata->{data} || $cdata->{fdata}) {
	my $data = $this->get("page");
	my $path;
	if ($data->{type} eq "page") {
	    # page files are written directly to the section directory
	    require ExSite::Section;
	    my $sec = new ExSite::Section(section=>$this->get("section"));
	    $path = $sec->diskpath;
	}
	else {
	    # page files are written to the library/template subdirectory
	    require ExSite::Page;
	    my $p = new ExSite::Page(page=>$data);
	    $path = $p->diskpath;
	}
	return $path if ($dir);
	$path .= "/".$this->filename;
	return $path;
    }
    elsif ($cdata->{file}) {
	if ($dir) {
	    my $path = $cdata->{file};
	    my @path = split /[:\\\/]/, $cdata->{file};
	    pop @path;
	    return join "/",@path;   # *nix only
	}
	return $cdata->{file};
    }
    elsif ($cdata->{url}) {
	return undef;
    }
}

# preview: display the content for an administrator

sub preview {
    my $this = shift;
    my $mode = shift;
    my $type = $this->content_type;
    if ($type eq "text/html" && $mode ne "raw") {
	# expand the content
	# do not use published elements (ie. always show latest version)
	# do not expand dynamic modules
	return $this->expand(mode=>"dynamic",method=>"data,url,content,page");
    }
    elsif ($type =~ /^text/) {
	# display raw content for inspection, rather than convert it to HTML
	my $raw = $this->get_raw;
	$raw =~ s/&/&\#38;/g;
	$raw =~ s/</&\#60;/g;
	$raw =~ s/>/&\#62;/g;
	return "<pre>\n$raw\n</pre>\n";
    }
    else {
	# convert to HTML - force dynamic preview mode for images
	return $this->get_html("dynamic");
    }
}

# write the content data out to a file

sub publish {
    my ($this,$path,$filename) = @_;
    my $out;
    if (! $path) {
	$path = $this->diskpath(1);
	if ($path) { 
	    $out .= $this->mkdir($path); 
	    
	}
    }
    if (! $filename) {
	$filename = $this->filename();
    }
    $out .= "<p>Publishing $filename...</p>\n";
    if ($path) {
	my $data = $this->get("content_data");
	my $cobjname = $this->get("content","name");
	if (ref $data eq "HASH") {
	    my ($file,$fdata);
	    $file = $filename || $this->filename;  # redundant
	    if ($data->{data} || $data->{fdata}) {
		# raw data is stored in database
		$fdata = $this->get_raw;
	    }
	    elsif ($data->{file}) {
		# raw data is stored in a disk file...
		# but that means one of 2 things:
		# - it is alreay published
		# - user has uploaded a file too large for version control
		# In either case, check POST data for a newly uploaded
		# version of the file, and republish it in that case.
		my $in = new ExSite::Input;
		my $updata = $in->post;
		if ($updata->{from}) {
		    # raw data is stored in a temp file, which needs to be
		    # copied to the destination
		    my $safe_tmpfile = $updata->{from};
		    $safe_tmpfile =~ /^(\/tmp\/ExSite[0-9]+)$/;
		    $safe_tmpfile = $1;
		    if ($safe_tmpfile) {
			$fdata = &get_file($updata->{from});
			if ($fdata =~ /^[^\#]{1,40}\.[\w]{2,4}\#[A-Za-z0-9+\/=]+$/) {
			    # exsite-encoded file
			    my (undef,$mimedata) = split /\#/,$fdata,2;
			    $fdata = &decode_base64($mimedata);
			}
			# remove the temp file
			unlink $safe_tmpfile or 
			    $out .= "<p>Remove temp file: $!</p>\n";
		    }
		    else {
			$this->error("permission denied: $updata->{from}");
			return undef;
		    }
		}
		elsif ($updata->{fdata}) {
		    if ($updata->{fdata} =~ /^[^\#]{1,40}\.[\w]{2,4}\#[A-Za-z0-9+\/=]+$/) {
		    #if ($updata->{fdata} =~ /^[^\#]{1,40}\.[\w]{2,4}\#/) {
			# looks like an exsite-encoded file
			my (undef,$mimedata) = split /\#/,$updata->{fdata},2;
			$fdata = &decode_base64($mimedata);
		    }
		    else {
			# could be a multipart-encoded file
			$fdata = $in->fetch_file("fdata","raw");
		    }
		    $out .= (length $fdata)." bytes in file<p>\n";
		    if (! $fdata) {
			$this->warn("ContentData::publish: No file data found");
			return undef;
		    }
		}
		else {
		    # nothing to publish
		    if (! -e $data->{file}) {
			# the file pointer goes nowhere...
			return $this->warn("ContentData::publish: $data->{file} does not exist");
		    }
		    return undef;
		}
	    }
	    else {
		# nothing to publish
		return undef;
	    }
	    my $outputpath = $path ? "$path/$file" : $file;
	    $outputpath =~ /^(((\/[\w-]+)+)\/[\w][\w\-\.]+)$/;
	    my $secure_outputpath = $1;
	    $this->info("publish content data $data->{content_data_id} to $secure_outputpath");
	    if (open OUTFILE, ">$secure_outputpath") {
		print OUTFILE $fdata;
		close OUTFILE;
		$out .= "Successfully published $cobjname to $file.<br>\n";
		if ($data->{thumb}) {
		    # add _T to filename for thumnail
		    $secure_outputpath =~ s/\.([^\.]*)$/_T.$1/;
		    $file =~ s/\.([^\.]*)$/_T.$1/;
		    $fdata = $this->get_thumb_raw;
		    if (open OUTFILE, ">$secure_outputpath") {
			print OUTFILE $fdata;
			close OUTFILE;
			$out .= "Successfully published thumbnail of $cobjname to $file.<br>\n";
		    }
		    else {
			$out .= "Failed to publish thumbnail of $cobjname to $file: $!<br>\n";
		    }
		}
		return $out;
	    }
	    else {
		$this->error("Failed to publish $cobjname to $file: $!");
		my $x = $this->get("content_data");
		return "<p class=error>Failed to publish $cobjname to $file: $!</p>\n";
	    }
	}
    }
    $this->warn("ContentData::publish: No publication directory specified");
    return undef;
}

# remove the file associated with the content data

sub unpublish {
    my ($this,$path) = @_;
    my $out;
    if (! $path) { $path = $this->diskpath(1); }
    if ($path) {
	my $data = $this->get("content_data");
	my $cobjname = $this->get("content","name");
	if (ref $data eq "HASH") {
	    my ($file,$fdata);
	    $file = $this->filename;
	    my $outputpath = $path ? "$path/$file" : $file;
	    $outputpath =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}((\/[\w-]+)*)\/[\w][\w\-\.]+)$/;
	    my $secure_outputpath = $1;
	    if (-e $secure_outputpath) {
		$this->info("removing file $secure_outputpath (content data $data->{content_data_id})");
		if (unlink $secure_outputpath) {
		    $out .= "Removed $file.<br>\n";
		}
		else {
		    $out .= $this->error("Failed to remove $file: $!");
		}
	    }
	    else {
		return $this->error("ContentData::unpublish: $outputpath: permission denied");
	    }
	    
	    if ($data->{thumb}) {
		# add _T to filename for thumnail
		$secure_outputpath =~ s/\.([^\.]*)$/_T.$1/;
		$file =~ s/\.([^\.]*)$/_T.$1/;
		if (-e $secure_outputpath) {
		    $this->info("removing file $secure_outputpath (thumbnail for content data $data->{content_data_id})");
		    if (unlink $secure_outputpath) {
			$out .= "Removed $file.<br>\n";
		    }
		    else {
			$out .= $this->error("Failed to remove $file: $!");
		    }
		}
	    }

	    return $out;
	}
    }
    $this->warn("ContentData::unpublish: cannot determine content location");
    return undef;
}

# utility for creating publication directories

sub mkdir {
    my ($this,$dir) = @_;
    my $out = " ";  # dummy return value
    # can only create directories in the HTMLpath
    #$dir =~ /(\/[\w_\-\/]+)/;
    $dir =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}(\/[\w-\/]+)?)$/;
    my $secure_dir = $1;
    if (! $secure_dir) { 
	$this->error("mkdir: invalid directory: $dir");
	return undef;
    }
    my @path = split(/\//,$secure_dir);
    shift @path; # first one is a dummy
    my $path;
    foreach my $dir (@path) {
	$path .= "/$dir";
	next if (-e "$path");
	$out .= "<P>Creating site directory $dir... ";
	if (! mkdir "$path") {
	    $this->error("mkdir $dir: failed: $!");
	    return undef;
	}
	else {
	    $this->info("created site directory $dir");
	    $out .= "succeeded.</P>\n";
	}
    }
    return $out;
}

# utility for removing publication directories

sub rmdir {
    my ($this,$dir) = @_;
    my $out = " ";  # dummy return value
    # can only remove directories in the HTMLpath
    $dir =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}\/[\w-\/]+)$/;
    my $secure_dir = $1;
    if (! $secure_dir) { 
	$this->error("rmdir: invalid directory: $dir");
	return undef;
    }
    elsif (-e $secure_dir) {
	opendir D,$secure_dir;
	my @files = readdir D;
	closedir D;
	foreach my $f (@files) {
	    next if ($f =~ /^\.\.?$/);
	    $f =~ /^(.+)$/;
	    my $secure_file = $secure_dir."/".$1;
	    if (-d $f) {
		# subdir; better remove it, too
		$out .= $this->rmdir("$dir/$f");
	    }
	    else {
		# must move unlink to its own if statement, or it causes
		# an insecure dependency error 
		if (unlink $secure_file) {
		    $out .= "Removed $f.<br>\n";
		}
		else {
		    $this->error("rmdir: failed to remove $f: $!");
		    return undef;
		}
	    }
	}
	if (rmdir $secure_dir) {
	    $out .= "Removed $secure_dir.<br>\n";
	}
	else {
	    $this->error("rmdir: failed to remove $secure_dir: $!");
	    return undef;
	}
    }
    return $out;
}

# utility for removing published files

sub rm {
    my ($this,$path) = @_;
    my $out = " ";  # dummy return value
    # can only remove files in the HTMLpath
    if ($path =~ /\.\./) {
	$this->error("rm: not removing suspicious path: $path");
	return undef;
    }
    $path =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}\/.+)$/;
    my $secure_path = $1;
    if (! $secure_path) { 
	$this->error("rm: invalid path: $path");
	return undef;
    }
    elsif (-e $secure_path && ! -d $secure_path) {
	if (unlink $secure_path) {
	    $out .= "Removed $secure_path.<br>\n";
	}
	else {
	    $this->error("rm: failed to remove $secure_path: $!");
	    return undef;
	}
    }
    return $out;
}

# validate - check integrity of datahash

sub validate {
    my ($this,$data) = @_;
    my $cdata;
    if ($data) {
	$cdata = new ExSite::ContentData(content_data=>$data);
    }
    else {
	$cdata = $this;
	$data = $this->get("content_data");
    }
    my @err = $share{DB}->validate_record("content_data",$data);
    if ($cdata->{content_id}) {
	if (! $cdata->owns()) {
	    push @err, "permission denied - ownership mismatch";
	}
    }
    if (! $data->{data} && 
	! $data->{fdata} && 
	! $data->{file} && 
	! $data->{url}) {
	push @err, "content version is empty";
    }
    return @err;
}

# make_content_data : insert a new content_data record
#
# this provides some generic algorithms for:
# - guessing the mime type (automatic)
# - determining image attributes (on request)
# - thumbnailing the image (on request)

sub make {
    return &make_content_data(@_);
}

sub make_content_data {
    my ($this,$cdata,$option) = @_;
    if ($cdata->{content_data_id} || ! $cdata->{content_id}) {
	# invalid content
	$this->error("failed to make improperly defined content_data");
	return undef;
    }
    if (! $this->allow_insert("content_data",$cdata)) {
	# permission denied
	$this->error("make_content_data: permission denied");
	return undef;
    }

    # remove any residual timestamps
    delete $cdata->{ctime};

    my ($file,$filedata,$length,$filepath);

    # data size

    if ($cdata->{data}) {
	delete $cdata->{fdata};   # clear ambiguous definition
	$length = length $cdata->{data};
	if ($length > $config{form}{bigfile_maxsize}) {
	    $this->error("data too large");
	    return undef;
	}
    }
    elsif ($cdata->{fdata}) {
	($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	$length = (length $filedata) / 4 * 3;
	if ($length > $config{max_upload_size}) {
	    $this->error("file too large");
	    return undef;
	}
	elsif ($length > $config{form}{bigfile_maxsize}) {
	    # too big for database, but not to big to write out to disk
	    delete $cdata->{fdata};
	    $file = &clean_filename($file);
	    # determine the future path to the published file
	    my $cid = $cdata->{content_id};
	    my $content = $share{DB}->fetch("content",$cid);
	    if ($content) {
		require ExSite::Page;
		my $page = new ExSite::Page(id=>$content->{page_id});
		$filepath = $page->diskpath(1);
		if ($filepath) {
		    $cdata->{file} = "$filepath/$file";
		}
		else {
		    $this->error("could not determine path to $file");
		}
	    }
	    else {
		$this->error("could not determine context for publishing $file");
	    }
	}
    }

    # Now that file size issues are sorted out, we can run normal validation.
    # Otherwise, we risk getting oversize file errors on files that aren't
    # really going into the database anyway.

    # brute force the validate call into this class
    # $this->validate(...) fails for content updates, because $this is a 
    # Content object and we end up using Content::validate() instead.
    my @err = &ExSite::ContentData::validate($this,$cdata);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    # set the mime-type

    if (! $cdata->{mime_type}) {
	if ($cdata->{data}) {
	    # look for suggestive patterns
	    if ($cdata->{data} =~ /<(!--|[\/a-zA-Z])/) {
		$cdata->{mime_type} = "text/html";
	    }
	    elsif ($cdata->{data} =~ /\/\*|[\w-]+:\s*[\w\#]+\;|^\w*\.\w+\s*\{/) {
		$cdata->{mime_type} = "text/css";
	    }
	    else {
		$cdata->{mime_type} = "text/plain";
	    }
	}
        else {
	    # look at file type
	    my $file;
	    if ($cdata->{fdata}) {
		($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	    }
	    else {
		$file = $cdata->{"file"} || $cdata->{"url"};
	    }
	    $cdata->{mime_type} = &MimeType($file);
	}
    }

    # image handling - optional

    if ($cdata->{fdata} && $cdata->{mime_type} =~ /^image\/(gif|jpeg|png)$/) {
	my $img = new ExSite::Image($cdata->{fdata});
	if ($option =~ /attribute/ && ! $cdata->{attributes}) {
	    my ($x,$y) = $img->dim;
	    if ($x && $y) {
		$cdata->{attributes} = "width=\"$x\" height=\"$y\"";
	    }
	}
	if ($option =~ /thumb/ && ! $cdata->{thumb}) {
	    if ($img->thumb) {
		$cdata->{thumb} = "th_".$img->encode;
		my ($x,$y) = $img->dim;
		if ($x && $y) {
		    $cdata->{thumb_attributes} = "width=\"$x\" height=\"$y\" border=\"0\"";
		}
	    }
	}
    }
    elsif ($cdata->{fdata}) {
	($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	$length = $filedata / 4 * 3;
	if ($length > $config{max_upload_size}) {
	    $this->error("file too large");
	    return undef;
	}
	elsif ($length > $config{form}{max_upload_size}) {
	    # too big for database, but not to big to write out to disk
	    $file = &clean_filename($file);
	}
    }

    # comment

    if (! $cdata->{note}) {
	$cdata->{note} = "inserted by ExSite::ContentData::make()";
    }

    my $cdid = $share{DB}->insert("content_data",$cdata);
    if (! $cdid) {
	$this->error("failed to install new version $file");
    }
    else {
	$this->info("installed new version $cdid $file as $cdata->{mime_type}");
	if ($filepath && ! $cdata->{fdata}) {
	    # looks like fdata was removed from the record, which means
	    # we're expected to publish the file directly
	    $this->write_to_file($cdata->{content_id},"$file#$filedata");
	}
    }
    return $cdid;
}

# write a content file straight to disk
# -> outsource job to publish.cgi program to get necessary permissions
# cobj = content ID
# path = file to publish to
# data = exsite-encoded file data 

sub write_to_file {
    my ($this,$cobj,$data) = @_;

    # make a temporary copy of the file

#    my $tmpfile = "$config{server}{temppath}/ExSite$$";
#    open TMP,">$tmpfile";
#    print TMP $data;
#    close TMP;
#    my $postdata = "from=$tmpfile";

    my $postdata = "fdata=$data";
    $postdata =~ s/\+/%2B/g;          # mask MIME chars from http

    # save current environment
    my $qstr = $ENV{QUERY_STRING};
    my $clen = $ENV{CONTENT_LENGTH};
    my $reqm = $ENV{REQUEST_METHOD};
    my $ctyp = $ENV{CONTENT_TYPE};
    # overload environment
    $ENV{QUERY_STRING} = "content=$cobj&silent=1";
    $ENV{CONTENT_LENGTH} = length $postdata;
    $ENV{REQUEST_METHOD} = "POST";
    $ENV{CONTENT_TYPE} = "application/x-www-form-urlencoded";

    # fire up the publishing program, and post data to it
    open F, "|$config{server}{CGIroot}/publish.cgi" or print "$?\n";
    print F $postdata,"\n";
    close F;

    # restore the environment of the current request, just to be safe
    $ENV{QUERY_STRING} = $qstr;
    $ENV{CONTENT_LENGTH} = $clen;
    $ENV{REQUEST_METHOD} = $reqm;
    $ENV{CONTENT_TYPE} = $ctyp;

    return;
}

# restore a previous version

sub restore {
    my $this = shift;
    my $cdata = shift || $this->get("content_data");
    return 0 if (! $cdata);
    my %data;
    if (ref $cdata eq "HASH") {
	%data = %$cdata;
    }
    else {
	%data = $share{DB}->fetch("content_data",$cdata);
    }
    return 0 if (scalar keys %data == 0);
    delete $data{content_data_id};
    if ($this->allow_insert("content_data",\%data)) {
	delete $data{ctime};
	$data{note} .= " (RESTORED)";
	my $cdid = $share{DB}->insert("content_data",\%data);
	if (! $cdid) {
	    $this->error("content_data::restore: failed to restore old version");
	}
	else {
	    $this->info("content_data::restore: restored old version $cdid");
	}
	return $cdid;
    }
    return 0;
}

# delete the version

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("content_data",$this->{data})) {
	$share{DB}->trash_key("content_data",$id);
    }
}

# purge older versions

sub purge {
    my $this = shift;
    my $data = $this->get("content_data");
    if ($data && $this->allow_delete("content_data",$data)) {
	my $db = $share{DB};
	my @cdata = $db->fetch_child("content_data","content",$data->{content_id},"content_data_id");
	foreach my $cdat (@cdata) {
	    last if ($cdat->{content_data_id} >= $data->{content_data_id});
	    $db->trash_key("content_data",$cdat->{content_data_id});
	}
    }
}

1;
