#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
# ExSite::DBM - DBM database driver for ExSite web development package
#
# This driver will perform a set of SQL-like operations on a DBM
# database.  The DBM driver supports two data formats, described
# below.  The following assumptions are made about the structure of the
# database and its files:
#
# - the database is contained in a single directory
# - the tables are optionally listed, on per line, in the text file .tables
# - each table has its own DBM file
# - each record consists of one entry in the file
# - the key of each DBM entry is the primary key of the record
# - the value of each DBM entry is an URL-encoded set of column=value pairs
#
# All ExSite DB drivers must contain the following methods:
#
# new
# get_tables
# get_columns
# get_key
# select
# insert
# update
# delete
#
# All remaining methods are unique to the DBM driver, and should not be
# used if portability to other drivers is desired.  They are intended
# to be used internally only.
#-------------------------------------------------------------------------

package ExSite::DBM;
use strict;
use ExSite::Misc;

sub new {
    my ($this, %opt) = @_;
    my %newdb;
    # build a list of known tables, columns, and primary keys
    # uses .config file if present, or best guess based on existing files
    # pre-opens all DBM tables
    $newdb->{tables} = [];
    if (-e ".tables") {
	# table and column structure is explicitly described
	open CONF,".tables" or die "failed to open database";
	foreach my $line (<CONF>) {
	    next if ($line =~ /^\#/);  # comment
	    chomp;
	    my ($table,$columns) = split(/\t/,$line,2);
	    push @{$newdb->{tables}}, $table;
	    if ($columns) {
		$newdb->{columns} = {};
		my @cols = split /,/,$columns;
		$newdb->{columns}->{$table} = \@cols;
	    }
	}
    }
    else {
	# table and column structure must be guessed
	opendir DB,".";
	my @allfiles = readdir DB;
	closedir DB;
	$newdb->{tables} = [];
	for (@allfiles) {
	    next if (/^\./);
	    my ($base,undef) = split /\./;
	    next if (scalar grep /$base/,@files);  # already got it
	    push @{$newdb->{tables}},$base;
	}
    }
    # pre-open all DBM files
    foreach my $table (@{$newdb->{tables}}) {
	$newdb->{$table} = {};
	##### FIXME:  use Any_DBM instead
	dbmopen(%{$newdb->{$table}},$table,00666) or delete $newdb->{$table};
	# delete bad tables from tables array?
    }
    my $class = ref($this) || $this;
    bless $newdb, $class;
    return 
}

sub config {
    # reads a config file
    my $this = shift;
    open CONF,".config" or die "failed to open database";
    $this->{tables} = [];
    foreach my $line (<CONF>) {
	next if ($line =~ /^\#/);  # comment
	chomp;
	my ($table,undef) = split(/\t/,$line,2);
	push @{$this->{tables}}, $table;
    }
}

#sub open_table {
#    my ($this,$table) = @_;

#sub close_table {
#    my ($this) = @_;

sub get_tables {
    # returns list of known tables
    my $this = shift;
    return @{$this->{tables}};
}

sub get_columns {
    # returns list of columns for a table
    my ($this,$table) = @_;
    # grab an arbitrary record from the table
    my ($key, $val) = each %{$this->{$table}};
    my %data = &DecodeString($val);
    return keys %data;
}

sub get_key {
    my ($this,$table) = @_;
    # DBM files have no primary key name; use "."
    return ".";
}

# QUERY methods


sub fetch {
    # look up a single record in a table by its key
    my ($this,$table,$key) = @_;
    return &DecodeString($this->{$table}->{$key});
}

# fetch_all : get all records

sub fetch_all ($$$) {
    my ($this,$table) = @_;
    
    ################################# FIX ME!

}

sub fetch_match {
    # fetch a list of records matching a match hash
    my ($this,$table,$rmatch) = @_;
    my @records;
    # scan the entire table
  record:
    foreach my $key (keys %{$this->{$table}}) {
	my %data = &DecodeString($this->{$table}->{$key});
	$data{"."} = $key;
	foreach my $mkey (keys %{$rmatch}) {
	    my $mval = $rmatch->{$mkey};
	    $mval =~ s/%/.*/g;  # allow SQL wildcards - slow?
	    next record if ($data{$mkey} !~ /$mval/);
	}
	# this record is a match
	push @records,\%data;
    }
    return @records;
}

sub fetch_child {
    # fetch records that reference a given primary key from another table
    my ($this,$table,$ptable,$pkey) = @_;
    my @records;
    # reset hash iterator
    keys %{$this->{$table}};
    # scan the entire table
  record:
    foreach my $key (keys %{$this->{$table}}) {
	my %data = &DecodeString($this->{$table}->{$key});
	next record if ($data{"${ptable}_id"} eq $pkey);
	# this record is a match
	push @records,\%data;
    }
    return @records;
}

# EDIT methods

# create_db : initialize a new flat DB
############## FIXME!

sub create_table {
    # creates a new table
    my ($this,$table) = @_;
    $this->{$table} = {};
    if (dbmopen(%{$this->{$table}},$table,00666)) {
	return 1;
    }
    else {
	delete $this->{$table};
	return 0;
    }
}

# delete_table : remove a table from the DB
############## FIXME!

# rename_table : change the name of a table
############## FIXME!

sub insert {
    # add a new record 
    my ($this,$table,$rdata);
    return 0 if (!exists $rdata->{"."});  # no key
    return 0 if (exists $this->{$table}->{$rdata->{"."}});  # key exists
    my $key = $rdata->{"."};
    delete $rdata->{"."};
    $this->{$table}->{$key} = &EncodeHash(%$rdata);
    return 1;
}

sub update {
    # alter an existing record
    my ($this,$table,$rdata);
    return 0 if (!exists $rdata->{"."});  # no key in record
    return 0 if (!exists $this->{$table}->{$rdata->{"."}});  # no such key
    my $key = $rdata->{"."};
    delete $rdata->{"."};
    $this->{$table}->{$key} = &EncodeHash(%$rdata);
    return 1;
}

sub update_match {
    # alter all matching records
    my ($this,$table,$rmatch,$rdata);
    return 0 if (!exists $rdata->{"."});  # no key in record
    return 0 if (!exists $this->{$table}->{$rdata->{"."}});  # no such key
    my $key = $rdata->{"."};
    delete $rdata->{"."};
    my $nupdate = 0;
  record:
    foreach my $key (keys %{$this->{$table}}) {
	my %data = &DecodeString($this->{$table}->{$key});
	$data{"."} = $key;
	foreach my $mkey (keys %$rmatch) {
	    $mval = $rmatch->{$mkey};
	    $mval =~ s/%/.*/g;  # allow SQL wildcards - slow?
	    next record if ($data{$mkey} !~ /$mval/);
	}
	# this record is a match - update matching columns
	foreach my $col (keys %data) {
	    if (exists $rdata->{$col}) {
		$data{$col} = $rdata->{$col};
	    }
	}
	delete $data{"."};
	$this->{$table}->{$key} = &EncodeHash(%data);
	$nupdate++;
    }
    return $nupdate;
}

sub delete {
    # deletes a record - no integrity checks
    my ($this,$table,$key);
    delete $this->{$table}->{$key};
}

sub delete_match {
    # deletes all matching records - no integrity checks
    my ($this,$table,$key);
    delete $this->{$table}->{$key};

    $key = $rdata->{"."};
    delete $rdata->{"."};
    my $nupdate = 0;
  record:
    foreach my $key (keys %{$this->{$table}}) {
	my %data = &DecodeString($this->{$table}->{$key});
	$data{"."} = $key;
	foreach my $mkey (keys %$rmatch) {
	    $mval = $rmatch->{$mkey};
	    $mval =~ s/%/.*/g;  # allow SQL wildcards - slow?
	    next record if ($data{$mkey} !~ /$mval/);
	}
	# this record is a match - update matching columns
	foreach my $col (keys %data) {
	    if (exists $rdata->{$col}) {
		$data{$col} = $rdata->{$col};
	    }
	}
	delete $data{"."};
	$this->{$table}->{$key} = &EncodeHash(%data);
	$nupdate++;
    }
    return $nupdate;
}

sub remove {
    # deletes a record and its children, and clears all references to 
    # those deleted records
}

1;
