#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
# dbmap : database map object
#
# tools for querying the database map
#
# methods:
#   new : create dbmap object
#   get_tablemap : returns map for one table
#   get_dbmap : loads map for all tables
#   get : returns a parameter for a column (same as get_column_attr)
#   get_column : returns all parameters for a column
#   get_column_attr : returns one parameter for a column
#   get_table : returns all parameters for a table (from the .table table)
#   get_table_attr : returns one parameter for a table
#   get_datatype : returns all parameters for a datatype (from the .datatype table)
#   get_datatype_attr : returns one parameter for a datatype
#   get_parents : returns all parents of a table
#   get_children : returns all children of a table
#   make_sql_create_script : prints a SQL script to create a database
#   make_sql_drop_script : prints a SQL script to destroy a database


# The dbmap associates the following parameters with every column
# in the database:
#
# column    the column name in the database
# datatype  the datatype of the column (DBform datatype; this is
#           converted to a SQL datatype using %datatype2sql)
# size      the data size of the column (used in form generation and
#           SQL script generation)
# label     the full column description (used in forms and reports)
# display   display logic:
#           no - do not display this field publicly
#           full - display this field only in full reports
#           brief - also display this field in brief (summary) reports
#           key - primary key for display purposes
# validate  validation logic:
#           no - do not validate data in this field
#           yes - validate any data given for this field
#                 (null data will not be validated)
#           soft - required: validate this field even if null
#                  (administrator may override)
#           hard - required: validate this field even if null
#                  (administrator may not override)
# write     write level (0-4)
# read      read level (0-4)
#
# read/write levels refer to the access level of the current user.
# 0 = public (no authentication performed)
# 1 = user (authenticated)
# 2 = supervisor (user with enhanced privledges)
# 3 = administrator (user with privacy violation priveleges)
# 4 = god (user with unlimited override capabilities)

package ExSite::DBmap;

use strict;
use ExSite::Config;
use ExSite::Text;
use ExSite::Misc;    # columnwise(), keywise()

# dbmap inherits from Text

#use vars (@ISA);
our @ISA = qw(ExSite::Text);

use vars qw(%datatype_default %map_default %tablemap_default %datatype2sql);

##################################################################
# default map info

# default datatype definitions, if none are provided
# these are sufficient to manage the dbmap itself, with a few
# extra common features.
%datatype_default = (
  text => { tag=>"text", regexp=>'.+' },
  string => { tag=>"text", regexp=>'.+' },
  date => { tag=>"text", regexp=>'[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} ?' },
  time => { tag=>"text", regexp=>'[012]?[0-9]:[0-5][0-9](:[0-5][0-9])?' },
  email => { tag=>"text", regexp=>'^[\w](\.?[\w\-]+)*@([\w\-]+\.)+[A-Za-z]{2,4}$' },
  url => { tag=>"text", regexp=>'(http://)?.+\..+' },
  key => { tag=>"text", regexp=>'[0-9]+' },
  'key#' => { tag=>"text", regexp=>'[0-9]+' },
  int => { tag=>"text", regexp=>'[0-9]+' },
  decimal => { tag=>"text", regexp=>'[0-9]+\.?[0-9]+' },
  password => { tag=>"password", regexp=>'.{4,}' },
  money => { tag=>"text", regexp=>'[0-9]*(.[0-9]{2})?' },
  "list:yn" => { tag=>"radio", regexp=>'Y|N' },
  "list:display" => { tag=>"select", regexp=>'no|full|brief|key' },
  "list:validate" => { tag=>"select", regexp=>'no|yes|soft|hard' },
  "list:datatype" => { tag=>"select", regexp=>'text|date|time|email|url|key|int|decimal|password|money|list:yn|list:display|list:validate|list:datatype' },
);

# default table map
%tablemap_default = (
		     "*" => {
			 table => "*",
			 key => "id",
			 title => "",
			 report => $config{report}->{style},
			 read => 0,
			 insert => 3,
			 update => 3,
			 delete => 3,
			 search => 1,
		     }
);

# default map, sufficient for dbmap
%map_default = (
  column => ["column","datatype","size","label","display","read",
        "write","validate"],
  datatype => ["string","string","int","string","list:display","int",
          "int","list:validate"],
  size => [20,20,0,40,10,0,0,10],
  label => ["column","datatype","size","label","display","read",
       "write","validate"],
  display => ["key","brief","brief","brief","brief","brief","brief","brief"],
  read => [1,1,1,1,1,1,1,1],
  write => [1,1,1,1,1,1,1,1],
  validate => ["soft","soft","soft","soft","soft","soft","soft","soft"],
);

# DBmap -> SQL conversion
%datatype2sql = (
  key => "int",
  'key#' => "int",
  text => "text",
  html => "text",
  string => "varchar",
  int => "int",
  url => "varchar",
  email => "varchar",
  email_list => "varchar",
  date => "date",
  time => "time",
  decimal => "decimal",
  money => "decimal",
  datetime => "datetime",    # MySQL only
  list => "varchar",
  file => "text",            # 64K max, after base64 encoding
  bigfile => "mediumtext",   # 16M max, after base64 encoding
  set => "varchar",
  password => "varchar",
  timestamp => "timestamp",
  crypt => "text",
);


sub new ($%) {
    ### extended version of ExSite::Text constructor
    my ($this,%opt) = @_;

    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    # optional config parameters
    if (!exists $opt{name}) {
        # no map named, use the default
	$opt{dbmap} = 1;
	$opt{name} = undef;
    }
    $obj->{name} = $opt{name};
    $obj->{dbmap} = $opt{dbmap};

    if (exists $opt{delim}) {
	$obj->{delim} = $opt{delim};
    }
    else {
	$obj->{delim} = "\t";
    }

    # table cache
    $obj->{columns} = {};
    $obj->{key} = {};

    #### map-specific attributes ####

    if (exists $obj->{preload}) { &get_dbmap($obj); }

    # get datatypes
    $obj->{datatype} = {};

    if (! defined $obj->{name}) {
      # no map defined - use default datatypes
      # this will be adequate for mapping dbmaps
      $obj->{datatype} = \%datatype_default;
      $obj->{tables} = \%tablemap_default;
    }
    else {
      # mapping a real db
      if (-e "$obj->{name}/.datatype") {
	# load custom datatypes
	$obj->{datatype} = keywise("type",$obj->fetch_all(".datatype"));
      }
      else {
	# use default datatypes
	$obj->{datatype} = \%datatype_default;
      }
      # table descriptions (optional)
      if (-e "$obj->{name}/.table") {
	# load table descriptions
	$obj->{tables} = keywise("table",$obj->fetch_all(".table"));
      }
      else {
        $obj->{tables} = \%tablemap_default;
      }
    }

    # table maps
    $obj->{map} = {};

    return $obj;
}

#-----------------------------------------------------------------------
# fetch raw maps...  
# use with caution: no guarantee these data structures will be stable.

# get map for a single table - sorted columnwise because column order is significant

sub get_map {
  my ($this,$table) = @_;
  if (! $this->{map}->{$table}) {
    if (! defined $this->{name}) {
      # undefined map
      $this->{map}->{$table} = \%map_default;
    }
    else {
	my @all = $this->fetch_all($table);
	$this->{map}->{$table} = columnwise(\@all);
    }
  }
  return $this->{map}->{$table};
}

# get maps for all tables

sub get_dbmap {
  my ($this) = shift;
  foreach my $table ($this->get_mapped_tables) {
    if (! $this->{map}->{$table}) {
      if (! defined $this->{name}) {
	$this->{map}->{$table} = \%map_default;
      }
      else {
	my @all = $this->fetch_all($table);
	$this->{map}->{$table} = columnwise(\@all);
      }
    }
  }
  return $this->{map};
}

#-----------------------------------------------------------------------
# fetch map parameters...

#-----------------------------------------------------------------------
# fetch column parameters
# columns may be specified by name or number

sub get_column {
  my ($this,$table,$id) = @_;
  if (! $this->{map}->{$table}) {
    # load map for this table
    return undef if (! get_map($this,$table));
  }
  my $i;
  if ($id =~ /^[0-9]*$/) { $i = $id; }
  else {
    # Appears to be a column name - see above.
    for ($i = 0; $i <= $#{$this->{map}->{$table}->{column}}; $i++) {
      if ($this->{map}->{$table}->{column}->[$i] eq $id) {
	last;
      }
    }
  }
  if ($i <= $#{$this->{map}->{$table}->{column}}) {
    my %data;
    foreach my $key (keys %{$this->{map}->{$table}}) {
      $data{$key} = $this->{map}->{$table}->{$key}->[$i];
    }
    return %data;
  }
  else {
    # could not find any matching column
    return undef;
  }
}

# return one attribute for a given column

sub get_column_attr {
  my ($this,$table,$id,$param) = @_;
  if (! $this->{map}->{$table}) {
    # load map for this table
    return undef if (! get_map($this,$table));
  }
  if ($id =~ /^[0-9]*$/) { #################### does this work for numbers?
    # appears to be a column number (fast)
    return $this->{map}->{$table}->{$param}->[$id];
  }
  else {
    # Appears to be a column name - must search for the column (slow).
    # Could speed this up by also indexing the map by column name, but
    # the time savings would be offset by the time to actually build the
    # map.  Since these sorts of lookups are generally infrequent,
    # it is probably more efficient to use this low-speed method.
    for (my $i = 0; $i <= $#{$this->{map}->{$table}->{$param}}; $i++) {
      if ($this->{map}->{$table}->{column}->[$i] eq $id) {
	return $this->{map}->{$table}->{$param}->[$i];
      }
    }
  }
  # could not find any matching column
  return undef;
}

# return number of columns in a table

sub get_ncol {
  my ($this,$table) = @_;
  &get_map($this,$table);
  return $#{$this->{map}->{$table}->{column}} + 1;
}

#### deprecated methods:

# get is synonymous with get_column_attr

sub get {
    return &get_column_attr(@_);
}


#-----------------------------------------------------------------------
# datatype queries...

sub get_datatype {
  my ($this,$datatype) = @_;
  if (exists $this->{datatype}->{$datatype}) {
    return $this->{datatype}->{$datatype};
  }
  return {};
}

sub get_datatype_attr {
  my ($this,$datatype,$par) = @_;
  if (exists $this->{datatype}->{$datatype}) {
    return $this->{datatype}->{$datatype}->{$par};
  }
  return undef;
}


#### deprecated methods:

# get datatype map - sorted keywise

sub get_typemap {
  my ($this) = shift;
  return $this->{datatype};
}

sub get_datatypes {
  my ($this) = shift;
  return $this->{datatype};
}

# return tag for a particular datatype

sub tag {
    my ($this,$datatype) = @_;
    return $this->get_datatype_attr($datatype,"tag");
}

# return regexp for a particular datatype

sub regexp {
    my ($this,$datatype) = @_;
    return $this->get_datatype_attr($datatype,"regexp");
}

#-----------------------------------------------------------------------
# tablemap queries...

# return all attributes of a table
sub get_table {
  my ($this,$table) = @_;
  if (exists $this->{tables}) {
      if (exists $this->{tables}->{$table}) {
	  return $this->{tables}->{$table};
      }
      elsif (exists $this->{tables}->{"*"}) {
	  return $this->{tables}->{"*"};
      }
  }
  return {};
}

# return a particular attribute of a table
sub get_table_attr {
  my ($this,$table,$attr) = @_;
  if (exists $this->{tables}) {
    if (exists $this->{tables}->{$table}) {
      if (exists $this->{tables}->{$table}->{$attr}) {
	return $this->{tables}->{$table}->{$attr};
      }
    }
    elsif (exists $this->{tables}->{"*"}) {
	if (exists $this->{tables}->{"*"}->{$attr}) {
	    return $this->{tables}->{"*"}->{$attr};
	}
    }
  }
  return undef;
}

# set list of mapped tables
sub set_mapped_tables {
    my ($this) = shift;
    $this->{mapped_tables} = \@_;
}

# get list of mapped tables
sub get_mapped_tables {
    my ($this) = shift;
    if ($this->{mapped_tables}) { 
	return @{$this->{mapped_tables}};
    }
    elsif ($this->{tables}->{"*"}) {
	# wildcard applies to all tables known to the underlying database
	return $this->get_tables;  ## this doesn't work
    }
    else {
	# only recognize the mapped tables
	return keys %{$this->{tables}};
    }
}

# table_title : returns a title for a given table
#
# looks for "title" and "plural" attributes, otherwise, makes up
# a title based on the table name.
#
# INPUT:
#    - table name
#    - plural flag (try to pluralize title, if true)
#
# OUTPUT:
#    - returns a title string

sub get_table_title {
    my ($this,$table,$plural) = @_;
    # title the table
    my $table_title;
    my $tmap = $this->{tables};
    if ($tmap) {
      if ($tmap->{$table}) {
	if ($plural && $tmap->{$table}->{plural}) {
	  return $tmap->{$table}->{plural};
	}
	elsif ($tmap->{$table}->{title}) {
	  $table_title = $tmap->{$table}->{title};
	}
      }
    }
    # no table title defined
    if (! $table_title) { $table_title = $table; } #### non-capitalized

    # pluralize, if required
    if ($plural) {
      # no plural defined; make something up
      $table_title =~ s/_/ /g;
      $table_title =~ s/y$/ie/;
      return $table_title."s";
    }
    else {
      return $table_title;
    }
}

# return the name of the table in the low-level DB

sub get_table_name {
    my ($this,$table) = @_;
    my $tmap = $this->{tables}->{$table};
    return $tmap->{alias} ? $tmap->{alias} : $table;
}

#### deprecated methods:

sub table_title {
    return &get_table_title(@_);
}

sub table {
    return &get_table_attr(@_);
}

# get entire table attribute map - sorted keywise
sub get_tablemap {
  my ($this) = shift;
  return $this->{tables};
}

#-----------------------------------------------------------------------
# table relationships...

# return parents of a table

sub get_parents {
  my ($this,$table) = @_;
  if (! $this->{map}->{$table}) {
    return undef if (! get_map($this,$table));
  }
  my @fkeys = grep /^key:/, @{$this->{map}->{$table}->{datatype}};
  foreach my $key (@fkeys) {
    $key =~ s/^key://;
  }
  return @fkeys;
}

# return children of a table

sub get_children {
  my ($this,$table) = @_;
  &get_dbmap($this);
  my @fkeys;
  my @children;
  foreach my $ctable (keys %{$this->{map}}) {
    next if ($ctable =~ /^\.|^db$/);
    @fkeys = grep /^key:/, @{$this->{map}->{$ctable}->{datatype}};
    if (grep(/^key:${table}$/,@fkeys) > 0) {
      push @children, $ctable;
    }
  }
  return @children;
}

# return the column in table $child that references table $parent

sub get_reference_column {
  my ($this,$child,$parent) = @_;
  my $map = &get_map($this,$child);
  my $ncol = &get_ncol($this,$child);
  for (my $i; $i < $ncol; $i++) {
    if ($map->{datatype}->[$i] =~ "key[:=]$parent") {
      return $map->{column}->[$i];
    }
  }
  return undef;
}

#-----------------------------------------------------------------------
# some useful functions from DB

#------------------------------------------------------------------------
# fetch_all : get all records
# arguments:
#   $table : table to read from
# returns an array of datahashes
#------------------------------------------------------------------------

sub fetch_all ($$$) {
    my ($this,$table) = @_;
    return $this->select($table);
}

#-----------------------------------------------------------------------
# SQL functions...

# make sql script to create a database

sub make_sql_create_script {
  my ($this,$flavour) = shift;
  if (!$flavour) { $flavour = "mysql"; }
  if ($flavour !~ /^(mysql|postgres)$/) {
    print STDERR "unknown SQL flavour: $flavour;  using MySQL\n";
    $flavour = "mysql";
  }
  &get_dbmap($this);
  if ($flavour eq "mysql") {
    print "# MySQL create script - autogenerated by ExSite::DBmap\n\n";
    print "# create all tables\n\n";
  }
  elsif ($flavour eq "postgres") {
    print "-- PostgreSQL create script - autogenerated by ExSite::DBmap\n\n";
    print "-- create all tables\n\n";
  }
  foreach my $table (keys %{$this->{map}}) {
    next if ($table =~ /^\.|^db$/);
    next if ($table eq "CVS");
    next if ($table ne $this->{tables}->{$table}->{alias});
    print "CREATE TABLE $table (\n";
    my $ncol = $#{$this->{map}->{$table}->{column}};
    for (my $i=0; $i <= $ncol; $i++) {
      my $column = $this->{map}->{$table}->{column}->[$i];
      print "        $column";
      if ($this->{map}->{$table}->{datatype}->[$i] =~ /key\#/ &&
	  $flavour eq "postgres") {
	print " serial";
      }
      else {
	my ($datatype,$datatype_option) =
	  split /[:=]/, $this->{map}->{$table}->{datatype}->[$i];
	print " $datatype2sql{$datatype}";
	if ($datatype2sql{$datatype} =~ /^(varchar|decimal|char)$/) {
	  ################ make configurable
	  my $size = $this->{map}->{$table}->{size}->[$i];
	  if (!$size) {
	    print STDERR "WARNING: no size given for $datatype2sql{$datatype}\n";
	  }
	  print "($size)";
	}
	if ($this->{map}->{$table}->{validate}->[$i] eq "hard" ||
	    $this->{map}->{$table}->{datatype}->[$i] =~ /key[:=]/) {
	  print " NOT NULL";
	}
	if ($this->{map}->{$table}->{datatype}->[$i] =~ /key\#?$/) {
	  print " PRIMARY KEY";
	}
	if ($this->{map}->{$table}->{datatype}->[$i] =~ /key\#/ &&
	    $flavour eq "mysql") {
	  print " AUTO_INCREMENT";
	}
	if ($i < $ncol) { print ",\n"; }
      }
    }
    print "\n        );\n";
    #print ");\n";
  }
}

# make sql script to destroy a database

sub make_sql_drop_script {
  my ($this,$flavour) = shift;
  if (!$flavour) { $flavour = "mysql"; }
  if ($flavour !~ /^(mysql|postgres)$/) {
    print STDERR "unknown SQL flavour: $flavour;  using MySQL\n";
    $flavour = "mysql";
  }
  &get_dbmap($this);
  if ($flavour eq "mysql") { print "# drop all tables\n"; }
  elsif ($flavour eq "postgres") { print "-- drop all tables\n"; }
  foreach my $table (keys %{$this->{map}}) {
    if ($table =~ /^%/) { next; }
    next if ($table =~ /^\.|^db$/);
    print "drop table $table;\n";
  }
}

1;


