package ExSite::Image;

=pod

=head1 ExSite::Image - image manager

Provides a few basic image management tools for the ExSite CMS.
This is just a simple wrapper/API to whatever underlying image processing
tools are going to do the heavy lifting, so it is easy to replace or
extend with more powerful tools if needed.

This version tries to use Image::Info for fetching image attributes, 
and the convert shell command for modifying the image itself.

To start working with a new image:

    my $img = new ExSite::Image($name,$rawdata);   # OR
    my $img = new ExSite::Image($encoded_data);

To get the image attributes:

    my $info = $img->info;  # $info is a hash ref

To get the image size:

    my ($x,$y) = $img->dim;

To change the image size:

    my $stat = $img->scale(500,500,$flag);  

    # $flag > 0   scale up to that size, if smaller
    # $flag == 0  scale to that size
    # $flag < 0   scale down to that size, if larger
    # retain aspect ratio in all cases

To scale the image to exactly the passed dimensions (the edges may
get cropped to achieve the new aspect ratio without stretching the
original image):

    my $imgdata = $img->scale_crop($x,$y);
   
    # $x: image width
    # $y: image height
   
To thumbnail an image (preserve aspect ratio):

    my $stat = $img->thumb;  # equivalent to $img->scale(100,100,-1);  

To get a square thumbnail (some cropping may be involved):

    my $imgdata = $img->square_thumb($x);
    
    # $x: image width
    
To get image raw data:

    my $imgdata = $img->get;

To get image data in exsite binary file encoding format:

    my $imgdata = $img->encode;

To get the size of the image data in bytes:

    my $nbyte = $img->size;

To shrink the size of the image:

    $img->shrink(%opt);

You can pass the following options in the hash: C<width>, C<height>
(maximum width and height of the image), C<dim> (maximum width OR
height), and C<size> (maximum file size in kilobytes).  Without a 
C<size> option, C<shrink()> works analagously to C<scale()>, and merely
changes the image dimensions.  With a C<size> option, shrink attempts to
reduce the file size as well, by scaling the image size smaller iteratively
until it is below the maximum file size.

=cut

use strict;
use MIME::Base64;
use IO::Handle;
use FileHandle;
use IPC::Open3;
use ExSite::Base;
use ExSite::Config;
use ExSite::Misc;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

sub new {
    my $this = shift;
    # objectify object
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    return $obj->setup(@_);
}

sub setup {
    my $this = shift;
    if (@_ == 1) {
	($this->{filename},my $mimedata) = split /\#/,shift,2;
	$this->{filedata} = &decode_base64($mimedata);
    }
    elsif (@_ == 2) {
	my @filepath = split /[:\\\/]/, shift;
	$this->{filename} = @filepath[-1];
	$this->{filedata} = shift;
    }
    else {
	$this->error("invalid image data");
	return undef;
    }
    return $this;
}

sub info {
    my $this = shift;
    if (require Image::Info && require IO::String ) {
	my $fdata = $this->{filedata};
	return &Image::Info::image_info(\$fdata);
    }
    # no info tool, so let's kludge something up using convert
    my $pid = open3(*W, *R, *E, "convert -verbose - -");
    print W $this->{filedata}; close W;
    # we have to read (and discard) stdout data before the stderr output
    # will be given to us
    while (<R>) { 1; } 
    close R;
    # now read stderr, which contains the -verbose data we want
    # expected output is something like:
    # /tmp/ABC123=>- GIF 239x239+0+0 PseudoClass 256c 8-bit 0.0u 0:01
    my $imgdata = <E>; close E; 
    waitpid $pid, 0; # perlipc recommends this to avoid zombies
    my @imgdata = split / /,$imgdata;
    $imgdata[2] =~ /([\d]+)x([\d]+)/;
    my $ext = lc($imgdata[1]);
    return {
	file_ext => $ext,
	file_media_type => &MimeType("dummy.$ext"),
	width => $1,
	height => $2,
    }
}

sub dim {
    my $this = shift;
    my $info = $this->info;
    return ($info->{width},$info->{height});
}

sub crop_parameter{
	my ($this,$type,$old1,$old2,$new1,$new2) = @_;
	require POSIX;
	my $out;
	my $resize_ratio = $old1 / $new1;
	my $resize1 = $old1;
	my $resize2 = POSIX::ceil($new2 * $resize_ratio);
	my $offset = 0;
	if ($old2 >= $resize2){
		$offset = POSIX::ceil(($old2 - $resize2) / 2);
		$out = " -crop";
		if ($type =~ /x/i){
			$out .= " ${resize1}x${resize2}";
			$out .= "+0+${offset}";
		}else{
			$out .= " ${resize2}x${resize1}";
			$out .= "+${offset}+0";
		}
		
	}else{
		$resize_ratio = $old2 / $new2;
		$resize2 = $old2;
		$resize1 = POSIX::ceil($new1 * $resize_ratio);
		$offset = POSIX::ceil(($old1 - $resize1) / 2);
		$out = " -crop";
		if ($type =~ /x/i){			
			$out .= " ${resize1}x${resize2}";
			$out .= "+${offset}+0";
		}else{
			$out .= " ${resize2}x${resize1}";
			$out .= "+0+${offset}";	
		}
	}
	return $out;
}
sub scale_crop{
	my ($this,$x,$y) = @_;
	$x = $x || 0;
	$y = $y || 0;
	# You can't create a image with 0 width or height
	return if !$x || !$y;
	# Original image size
	my ($width,$height) = $this->dim();	
	my $crop_param;
	my $scale_param;
	if ($x >= $y){
		$crop_param = $this->crop_parameter("x",$width,$height,$x,$y);
	}else{
		$crop_param = $this->crop_parameter("y",$height,$width,$y,$x);
	}
	$scale_param = " -geometry ${x}x${y}";
	my $convert_command = "convert${crop_param}${scale_param} - -";	
	my $pid = open3(*W,*R,*E,$convert_command);
	print W $this->{filedata};
	close W;	
    	my $newdata;
    	my $size = 0;
    	my $continue = 1;
    	while ($continue) {
		$continue = read(R,$newdata,1024,$size);
		$size += $continue;
    	}
    	$this->{filedata} = $newdata;
    	close R;
    	close E;  # don't need stderr in this case
    	waitpid $pid, 0;  # anti-zombie spell
	return $size;
}
sub square_thumb{
	my ($this,$size,$flag) = @_;	
	my $thumb_size = $size || $config{thumbnail_size};
	return $this->scale_crop($thumb_size,$thumb_size);
}

sub scale {
    my ($this,$x,$y,$flag) = @_;
    #
    # A better solution here might be to install Perl::Magick and
    # use its interface to the ImageMagick libraries.  However, that
    # package is HUGE and technically unnecessary, since if the 
    # ImageMagick libraries are already in place then we can just
    # use the "convert" shell command to do the job, without the 
    # hassle of installing and supporting Perl::Magick.
    #
    my $geom = "${x}x${y}";
    if ($flag > 0) { $geom .= '\<'; }
    elsif ($flag < 0) { $geom .= '\>'; }
    my $pid = open3(*W, *R, *E, "convert -geometry $geom - -");
    print W $this->{filedata}; close W;
    my $newdata;
    my $size = 0;
    my $continue = 1;
    while ($continue) {
	$continue = read(R,$newdata,1024,$size);
	$size += $continue;
    }
    $this->{filedata} = $newdata;
    close R;
    #my $imgdata = <E>;
    close E;  # don't need stderr in this case
    waitpid $pid, 0;  # anti-zombie magic spell
    return $size;
}

# shortcut for thumbnailing

sub thumb {
    my $this = shift;
    my $thumbdim = $config{thumbnail_size};
    my ($x,$y) = $this->dim;
    # don't thumbnail if the image is already small
    return 0 if ($x < $thumbdim && $y < $thumbdim);
    return $this->scale($thumbdim,$thumbdim,-1);
}

sub size {
    my $this = shift;
    return length $this->{filedata};
}

# reduce image size
#
# options:
# dim => max x or y dimension  (default 500)
# height => max y dimension  (default 500)
# width => max x dimension  (default 500)
# size => max file size (K)  (default 48)

sub shrink {
    my ($this,%opt) = @_;

    my $shrunk = 0;
    my $info = $this->info;
    my $max_height = $opt{height} || $opt{dim} || 500;
    my $max_width = $opt{width} || $opt{dim} || 500;
    my $max_size = $opt{size};
    my $npass = 0;  # abort after 5 attempts to shrink

    while (! $shrunk && $npass < 5) {
	if ($info->{width} > $max_width || $info->{height} > $max_height) {
	    $this->scale($max_width,$max_height,-1);
	}
	my $size = $this->size;
	if ($max_size && $size > $max_size * 1024) {
	    # file size too large; scale dimensions
	    my $proportional_difference = ($max_size * 1024) / $size;
	    my $scale_proportion = $proportional_difference < 0.8 ?
		$proportional_difference : 0.8;
	    $max_width *= $scale_proportion;
	    $max_height *= $scale_proportion;
	}
	else {
	    $shrunk = 1;
	}
	$npass++;
    }
    return;
}

# return raw image data to user

sub get () {
    my $this = shift;
    return $this->{filedata};
}

# encode image data in exsite format (filename#mime_data)

sub encode {
    my $this = shift;
    my $fname = &url_escape($this->{filename});
    return $fname.$config{form}{sepchar}.&encode_base64($this->{filedata},"");
}

1;
