package ExSite::Input;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

use strict;
use ExSite::Base;
use ExSite::Misc;
use ExSite::Config;
use MIME::Base64;

# inherit from Base
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

#========================================================================

=pod

=head1 ExSite::Input  --  tools for reading http inputs

Usage: C<my $in = new ExSite::Input;>

=head2 Reading Input

Input is returned as a hash reference of parameters names =E<gt> values.

=head3 C<query()>

Usage: C<my $querydata = $in-E<gt>query()>

Reads input passed via GET method (ie. QUERY_STRING)

=head3 C<post()>

Usage: C<my $postdata = $in-E<gt>post()>

Reads input passed via POST method. 

=head3 C<query_or_post()>

Usage: C<my $data = $in-E<gt>query_or_post()>

Reads from query preferentially, otherwise from post data.

=head3 C<post_or_query()>

Usage: C<my $data = $in-E<gt>post_or_query()>

Reads from post data preferentially, otherwise from query string.

=head3 C<combine()>

Usage: C<my $data = $in-E<gt>combine()>

Combines the post data and query data into a single hash.  Post data
is taken preferentially.

=head2 Raw Input Data

The raw input is also saved, and can be retrieved using:

    my $raw = $in->fetch("post","raw");   # or, use "get" instead of "post"

=head2 Input Order

If the order of the inputs is important, you can fetch the raw key order
using:

    my $keylist_ref = $in->fetch("post","keylist");  # or "get"...

=head1 Notes

The Input class is a replacement for the generic input parsing routines in 
ExSite::Misc::, which are deprecated for processing inputs to page.cgi.

The main purpose in having an Input class is so that POST input can 
be shared by multiple modules working independently.  Without an Input
class, the first module to get its mitts on STDIN gobbles all the input,
and the remaining modules see nothing.  

=head2 Multipart-encoded Forms

Multipart-encoding is handled automatically using the CGI:: class.
Multipart-encoding is normally only enabled by ExSite when the form
accepts file uploads.

To fetch the decoded file data directly, use:

    my $raw_file = $in->fetch_file("input_name","raw"); # binary file data

    $my $encoded_file = $in->fetch_file("input_name");  # encoded file data

ExSite encoded file data looks like this:

    filename.jpg#MIMEDATA....

To convert this back to a regular file, split on the separator character
('#' by default) to get the file name, and the mime-encoded data.  Then use

    use Mime::Base64;
    decode_base64($mimedata);

If you request the raw input data from a multipart-encoded form, you
will receive a CGI:: object.

=head2 Input Storage

ExSite::Input stores all input data in C<%share>, where other modules
will find it if they also play nice and use ExSite::Input.  The input
is parsed and cached automatically, and ExSite::Input is smart enough not 
to redo this work once it has been done.

Input is saved in the following structure:

C<$share{input}{get|post|combine}{raw|data|keylist}>

where get/post/combine selects the input stream, and raw/data/keylist
selects the format you would like to view the data in.

=head2 Hacking Input

To modify the input cache after input has already been read in, use the
following methods:

    $in->set("parameter","newvalue","post"); # or use "get", "combine" instead of "post"

To remove a form parameter from the input cache, use:

    $in->delete("parameter");

=cut

#========================================================================

# my $preset_input = new ExSite::Input({post=>"mypostdata",get=>"myquery"});

sub new ($$) {
    my($this, $preset) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;
    $cobj->info("ExSite::Input object created");

    # input data is stored in share area, not in object, so that it 
    # can serve multiple components simultaneously.

    if (!$share{input}) {
	$share{input} = {};
	foreach my $action ("post","get","combine") {
	    $share{input}{$action} = {};
	    $share{input}{$action}{keylist} = [];
	    $share{input}{$action}{data} = {};
	}
    }

    # check for preset data: allows programs to manually preset the input
    # (useful for automated testing)

    if (keys %$preset > 0) {
	$share{input}{preset} = $preset;
    }
    elsif (! $share{input}{preset}) {
	$share{input}{preset} = {};
    }

    return $cobj;
}

# query : process QUERY_STRING input, ie. "GET" method

sub query {
    my ($this,$fetch_as) = @_;
    if (! $share{input}{get}{raw}) {
	$share{input}{get}{raw} = 
	    $share{input}{preset}{get} || $ENV{QUERY_STRING};
	$this->decode("get",$share{input}{get}{raw});
    }
    return $this->fetch("get",$fetch_as);
}

# get is synonym for query
sub get { return &query(@_); }

# post : process "POST" method input

sub post {
    my ($this,$fetch_as) = @_;
    if (! $share{input}{post}{raw}) {
	if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
	    $this->info("decoding multipart form");
	    #$share{input}{post}{data}{debug} = $ENV{CONTENT_TYPE};
#	    if (! require CGI) : or { 
#		$this->error("CGI:: module failed to load: $@");
#		return; 
#	    }
	    require CGI;
	    # CONTENT_TYPE=multipart/form-data
	    # (this form contains file uploads; use CGI module to decode it;
	    # that means that the "raw" data is really a CGI:: handle.)
	    $share{input}{post}{raw} = 
		$share{input}{preset}{post} || CGI::new();
	    # must use CGI:: to decode
	    if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
		# CONTENT_TYPE=multipart/form-data
		my $in = $share{input}{post}{raw};
		foreach my $par ($in->param) {
		    push @{$share{input}{post}{keylist}}, $par;
		    # fetch multiple values, in case of repeated inputs
		    my @pars = $in->param($par);
		    if (@pars > 1) { 
			# there are multiple values; concatenate them
			$share{input}{post}{data}{$par} = 
			    join($config{form}{multi_sep},@pars);
		    }
		    else {
			# single value; do not concatenate, or filehandles
			# will be wrecked
			$share{input}{post}{data}{$par} = $in->param($par);
		    }
		}
	    }
#	    $this->info("Input processing done.");
	}
	else {
	    # CONTENT_TYPE=application/x-www-form-urlencoded
	    $this->info("decoding regular form");
	    my $clen = $ENV{'CONTENT_LENGTH'};
	    if ($share{input}{preset}{post}) {
		$share{input}{post}{raw} = $share{input}{preset}{post};
	    }
	    elsif ($clen) {
		my $in;
		read(STDIN,$in,$ENV{'CONTENT_LENGTH'});
		$share{input}{post}{raw} = $in;
	    }
	    $this->decode("post",$share{input}{post}{raw});
	}
    }
    return $this->fetch("post",$fetch_as);
}

# post_or_query, query_or_post : return one or the other, but not both

sub post_or_query {
    my ($this,$fetch_as) = @_;
    my $data = $this->post($fetch_as);
    return $data if (ref $data eq "HASH" && keys %$data > 0);
    return $data if (ref $data eq "ARRAY" && @$data > 0);
    $data = $this->query($fetch_as);
    return $data;
}

sub query_or_post {
    my ($this,$fetch_as) = @_;
    my $data = $this->query($fetch_as);
    return $data if (ref $data eq "HASH" && keys %$data > 0);
    return $data if (ref $data eq "ARRAY" && @$data > 0);
    $data = $this->post($fetch_as);
    return $data;
}

# combine : combine post and query data together into a single input stream

sub combine {
    my ($this,$fetch_as) = @_;
    # no raw input for combined input
    $share{input}{combine}{raw} = undef;
    # load both input streams
    $this->query;
    $this->post;
    # combine the data hashes and key lists
    my @tmplist = 
	(@{$share{input}{get}{keylist}},@{$share{input}{post}{keylist}});
    $share{input}{combine}{keylist} = \@tmplist;

    my %tmphash = %{$share{input}{post}{data}};
    foreach my $key (keys %{$share{input}{get}{data}}) {
	if (! exists $tmphash{$key}) { 
	    $tmphash{$key} = $share{input}{get}{data}{$key};
	}
    }
    $share{input}{combine}{data} = \%tmphash;
    return $this->fetch("combine",$fetch_as);
}

# decode : break input string down into key, value pairs
# arguments: input method, input string, separator character (optional)

sub decode {
    my $this = shift;
    my $method = shift;
    my $sep = $_[1] || "&";
    my @in = split(/$sep/,$_[0]);
    my (%out, @okeys);
    my $out = $share{input}{$method};
    foreach my $i (0 .. $#in) {
        # Convert pluses to spaces
        $in[$i] =~ s/\+/ /g;

        # Split into key and value.
        (my $key, my $val) = split(/=/,$in[$i],2); # splits on the first =

        # Convert %XX from hex numbers to alphanumeric
        $key =~ s/%(..)/pack("c",hex($1))/ge;
        $val =~ s/%(..)/pack("c",hex($1))/ge;

	# save key order
	if (! exists $out->{data}{$key}) { push @{$out->{keylist}}, $key; }
        # Associate key and value. Multiply-defined inputs are concatenated.
	$out->{data}{$key} .= 
	    $config{form}{multi_sep} if (defined($out->{data}{$key}));
	$out->{data}{$key} .= $val;
    }
    return;
}

# fetch : retrieve data in various formats

sub fetch {
    my $this = shift;
    my $method = shift;
    $_ = shift;
    if (/raw/)     { return $share{input}{$method}{raw}; }
    elsif (/keys/) { return $share{input}{$method}{keylist}; }
    else           { return $share{input}{$method}{data}; }
}

sub fetch_file {
    my ($this,$name,$format) = @_;
    my $f = $share{input}{post}{data}{$name};
    if ($f) {
	if ($format eq "name") {
	    return $f;
	}
	my $fdata = undef;
	my $size = 0;
	my $mimedata = "";
	# read file contents
	my $continue = 1;
	while ($continue) {
	    $continue = read($f,$fdata,1024,$size);
	    $size += $continue;
	}
	if ($size > $config{max_upload_size}) {
	    $this->error("file too large - $config{max_upload_size} bytes max");
	    return undef;
	}

	# return raw file contents, if requested

	if ($format eq "raw") {
	    return $fdata;
	}

	# otherwise return MIME-encoded file contents
	$mimedata = encode_base64($fdata,"");
	my @path = split /[:\\\/]/, $f;
	# strip filesystem cruft
	# convert all quirky characters to _
	my $filename = &clean_filename($path[-1]);
	my $mime = &MimeType($filename);
	return "$filename#$mimedata";
    }
    return undef;
}

# modify the input buffers

sub set {
    my ($this,$key,$val,$action) = @_;
    foreach my $act ("post","get","combine") {
	if (! $action || $action eq $act) {
	    $share{input}{$act}{data}{$key} = $val;
	}
    }
}

sub delete {
    my ($this,$key) = @_;
    # WARNING: only deletes value from data hash, not raw data or keylist
    foreach my $action ("post","get","combine") {
	delete $share{input}{$action}{data}{$key};
    }
}


1;

