#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Page;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Section;
use ExSite::Content;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Content);

# new() constructor is inherited from ContentData class.

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupPage(@_);
}

# options:
# id = id of page to look up
# page = prefetched page record
# noload = skip content loading phase (much faster if not rendering page)

sub setupPage { 
    my ($this,%option) = @_;

    my $page;
    if ($option{id}) {
	# lookup page record
	$page = $this->fetch("page",$option{id});
	if (! $page) {
	    $this->error("page $option{id} not found");
	}
    }
    elsif ($option{page}) {
	# page record was passed
	if (ref $option{page} eq "HASH") {
	    $page = $option{page};
	}
	else {
	    $this->error("invalid page parameters passed to setupPage");
	}
    }
    elsif ($option{filename} && $option{section_id}) {
	# lookup page by name & section
	my @pages = $this->fetch_match("page",{filename=>$option{filename},
					   section_id=>$option{section_id}});
	$page = @pages[-1];
	if (! $page) {
	    $this->error("\"$option{filename}\" not found");
	}
    }

    if ($page) {
	$this->info("setting up Page $page->{page_id}");
	$this->{data} = $page;
	$this->set_context("none");
    }
    else {
	$this->info("setting up null Page");
	$this->{data} = {};
    }

    # see if an alternate version has been requested
    if ($option{version}) {
	$this->set_version($option{version});
    }

    # publish flag
    if (exists $option{publish}) { 
	$this->{publish} = $option{publish}; 
    }
}

sub set_version {
    my ($this,$version) = @_;
    if ($version ne $this->{page}{version}) {
	my $master = $this->{page}{version} ? 
	    $this->{data}{parent_id} : $this->{data}{page_id};
	
	my @pages = $this->fetch_match("page",{
	    section_id=>$this->{data}{section_id},
	    parent_id=>$master,
	    version=>$version});
	if (@pages > 0) {
	    $this->{data} = $pages[-1];
	    $this->{context_set} = 0;
	    $this->set_context("none");
	    $this->info("setting up $version version of Page $this->{page}{page_id}");
	}
	else {
	    $this->warn("$version version of Page $this->{page}{page_id} not found.");
	}
    }
}

### basic queries

sub title { 
    my $this = shift;
    return $this->{data}{title};
}

sub description { 
    my $this = shift;
    return $this->{data}{description};
}

sub keywords { 
    my $this = shift;
    return $this->{data}{keywords};
}

sub is_template { 
    my $this = shift;
    return ($this->{data}{type} eq "template");
}

sub is_library { 
    my $this = shift;
    return ($this->{data}{type} eq "library");
}

sub is_page { 
    my $this = shift;
    return ($this->{data}{type} =~ /page|alternate/);
}

sub is_public { 
    my $this = shift;
    return ($this->{data}{access} eq "public");
}

# does this page ever get published?

sub is_publishable { 
    my $this = shift;
    return ($this->{data}{access} eq "public" && 
	    $this->{data}{publish_method} eq "static");
}

# has the latest version of this page's content been published?

### warning: this only gives accurate results for normal pages
### warning: as a check to see if we need to publish, this is probably 
###          costlier than just publishing

sub is_published { 
    my $this = shift;

    if ($this->is_publishable) {
    
	# the page timestamp is a MySQL timestamp
	my $time = $this->timestamp;
	$time =~ /^(....)(..)(..)(..)(..)(..)$/;
	my ($y1,$M1,$d1,$h1,$m1,$s1) = ($1,$2,$3,$4,$5,$6);

	# the published file has a Unix timestamp
	my @stat = stat($this->diskpath);
	my ($s2,$m2,$h2,$d2,$M2,$y2,undef,undef,undef) = 
	    localtime $stat[9]; # ie. mtime
	$M2 += 1;
	$y2 += 1900;
    
	# compare the two timestamps;  return TRUE if the published
	# file is older than the page data

	if ($y2 < $y1) { return 0; }
	elsif ($y2 > $y1) { return 1; }

	if ($M2 < $M1) { return 0; }
	elsif ($M2 > $M1) { return 1; }

	if ($d2 < $d1) { return 0; }
	elsif ($d2 > $d1) { return 1; }

	if ($h2 < $h1) { return 0; }
	elsif ($h2 > $h1) { return 1; }

	if ($m2 < $m1) { return 0; }
	elsif ($m2 > $m1) { return 1; }

	if ($s2 < $s1) { return 0; }
	elsif ($s2 > $s1) { return 1; }

    }

    return 0;

}

# is the page currently being published?

sub is_publishing { 
    my $this = shift;
    return $this->{publish};
}

sub publishing { return &is_publishing(@_); } # deprecated

# valid pages are those that belong to the site they are invoked from.
# Ie. if site x.com owns page 98 and site y.org owns page 99, then the
# url http://x.com/cgi/page.cgi?_id=99 is not valid.
# However, we waive validity requirements for administrators.

sub is_valid {
    my $this = shift;
    return 1 if ($share{DB}->is_admin);
    my $domain;
    if ($config{server}{auto}) {
	# config's domain info should be accurate
	$domain = $config{server}{domain};
    }
    else {
	# parse the domain right out of the HTTP_HOST info
	$ENV{HTTP_HOST} =~ /^((.+?)\.)?(.*?\..*?)(:\d+)?$/;
	$domain = $3;
    }
    my $s = new ExSite::Section(id=>$this->id("section"));
    my $url = $s->get_host;
    if ($url =~ /^(\w+\.)?$domain$/) {
	# the section URL matches the calling host
	return 1;
    }
    return 0;
}

# is the current user permitted to view this page dynamically?

sub authorize {
    my $this = shift;

    # skip authcheck if we are publishing, since the page is public by nature

    return 1 if ($this->publishing);

    # sysadmins can view anything from anywhere

    return 1 if ($share{DB}->is_admin);

    # regular admins can view pages from their own sites from any URL

    my $this_site = $this->get("section");
    my @sites = $share{DB}->my_sites;
    foreach my $site (@sites) { 
	return 1 if ($site->{section_id} == $this_site->{section_id});
    }

    # in all other cases, the page URL must be valid

    if (! $config{cms}{valid_page_check} || $this->is_valid) {
	# restricted-access pages
	my $pdata = $this->get;
	return 0 if ($pdata->{access} eq "members" && ! $share{DB}->authorize);
	return 1;
    }

    return 0;
}

## deprecated?
sub flag { 
    my ($this,$flag) = @_;
    if ($this->{data}{flags} =~ /$flag/) {   ## not a very robust regexp
	return 1;
    }
    return 0;
}

# return the timestamp of the page
# defined as the most recent of the timestamps of:
# - the template
# - the page
# - the content objects in the page

sub timestamp {
    my $this = shift;
    my @times;
    # timestamp of current page
    push @times, $this->get("page","mtime");
    # timestamp of all content objects in this page
    my @cobjs = $this->fetch_all_content_objs(version=>"newest");
    foreach my $cobj (@cobjs) {
	push @times, $cobj->timestamp;
    }
    # plus the same, for this page's template
    my $template_id = $this->get("page","template_id");
    if ($template_id) {
	my $template = new ExSite::Page(id=>$template_id);
	push @times, $template->timestamp;
    }
    my @sorted_times = sort @times;
    return pop @sorted_times;
}

### some page display tools

sub show {
    my $this = shift;
    
    # dump headers

    my $content_type = $this->content_type || "text/html";
    print $this->{http_header};
    if ($this->{http_header} !~ /content-type:/i) {
	print "Content-type: $content_type\n";
    }
    print "\n";

    # dump page

    if (! $this->{redirect}) {
	print $this->expand;
    }
}

# some HTTP header functions

sub header {
    my $this = shift;
    my $hdr = shift;
    if ($hdr !~ /\n$/) { $hdr .= "\n"; }
    $this->{http_header} .= $hdr;
    return;
}

# redirect using a location: header (address bar does not change)

sub redirect {
    return &location(@_);
}

sub location {
    my ($this,$url) = @_;
    $this->header("Location: $url");
    $this->{redirect} = 1;
    return;
}

sub cookie {
    my ($this,$name,$value,$path,$domain,$expiry) = @_;
    if (! $path) { $path = $config{server}{CGIpath}; }
    if (! $domain) { $domain = $config{server}{domain}; }
    my $cookie = "Set-cookie: $name=$value; path=$path; domain=$domain";
    if ($expiry) { $cookie .= "; Expires=$expiry"; }
    $this->header($cookie);
    return;
}

sub errorpage {
    my ($this,$html) = @_;
    

    # see if the section has defined a special template for exceptions

    ### FIXME: this only finds error templates in the current site;
    ### Would be nice to inherit error templates from parent sites.
    ### Also: does this fail to identify the section, if the URL has our
    ### domain, but no page identifier info?
    my @t = $share{DB}->fetch_match("page",{type=>"template",
					    section_id=>$this->id("section"),
					    filename=>"_ERROR"});

    if (@t > 0) {
	# reset the page template
	$this->{data}{template_id} = $t[-1]{page_id};
	$this->reset_context;
	$this->{expanded} = 0;
	$this->expand(content=>{body=>$html});
	return;
    }
    else {
	# no error-handler page;
	# just output the bare html
	$this->{html} = $html;
	$this->{noexpand} = 1;
	return;
    }
}

# exception_page : change the template to handle irregular occurences
#
# Special templates can be defined to handle exceptional conditions
# (errors, permission denied, login screens, etc.).  When the code calls
# exception_page($html,"ABC"), then the system will attempt to replace
# the correct template for the page with a template named '_ABC' (always
# upper case), and then substitute $html into the body of this template.
#
# ExSite knows about and will try to use the following exception templates:
#
# _LOGIN : authentication failures, requiring login to proceed
# _ERROR : errors in building or displaying the page
#
# Note that '_ERROR' is the default exception template, and will be used
# if no explicit exception template is named, or in cases where another 
# exception template was requested but not found.  For example, if the
# "LOGIN" template is requested, but not defined, but an "ERROR" template
# _is_ defined, the _ERROR template will be used instead.
#
# Any other exception templates may be defined, and invoked by plug-in
# modules or other custom code.

sub exception_page {
    my ($this,$html,$exception) = @_;

    $exception = $exception ? uc($exception) : "ERROR";

    # see if the section has defined a special template for exceptions

    ### FIXME: this only finds error templates in the current site;
    ### Would be nice to inherit error templates from parent sites.
    ### Also: does this fail to identify the section, if the URL has our
    ### domain, but no page identifier info?
    my @t = $share{DB}->fetch_match("page",{type=>"template",
					    section_id=>$this->id("section"),
					    filename=>"_$exception"});
    if (@t == 0 && $exception ne "ERROR") { 
	# requested exception page not found;
	# try _ERROR instead.
	@t = $share{DB}->fetch_match("page",{type=>"template",
					     section_id=>$this->id("section"),
					     filename=>"_ERROR"});
    }

    if (@t > 0) {
	# reset the page template
	$this->{data}{template_id} = $t[-1]{page_id};
	$this->reset_context;
	$this->{expanded} = 0;
	$this->expand(content=>{body=>$html});
	return;
    }
    else {
	# no error-handler page;
	# just output the bare html
	$this->{html} = $html;
	$this->{noexpand} = 1;
	return;
    }
}

# link - like get_url, but always returns a dynamic link, and you can add 
# your own parameters to it.

sub link {
    my ($this,$args) = @_;
    my $id = $this->id("page");
    my $url = "$config{server}{CGIpath}/page.cgi?_id=$id";
    if ($args) {
	if (ref $args eq "HASH") {
	    $args = &EncodeHash($args);
	}
	$url .= "&$args";
    }
    return $url;
}

# name of page file - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $filename = $page->{filename};
	if ($this->is_page()) {
	    $filename = &clean_filename($filename);
	}
	else {
	    # "." not allowed in dir names
	    $filename =~ s/[^\w]/_/g;
	    $filename =~ s/_{2,}/_/g;
	}

	if ($page->{type} eq "page") {
	    return $filename || "page_".$page->{page_id}.".html";
	}
	else {
	    # directory name for templates and libraries
	    return $filename || $page->{page_id};
	}
    }
    return undef;
}

# httppath: the URL path to the page, minus the server
######### DEPRECATED ############

sub httppath {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $s_httppath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_httppath = $s->httppath;
	}
	else {
	    $s_httppath = $config{server}{HTMLpath};
	}
	$_ = $page->{type};
	if (/template/) {
	    return "$s_httppath/_Templates/$page->{filename}";
	}
	elsif (/library/) {
	    return "$s_httppath/_Libraries/$page->{filename}";
	}
	else {
	    return $s_httppath;
	}
    }
    return undef;
}

# disk path to file

sub diskpath {
    my $this = shift;
    my $dir = shift;  # return directory only if true
    my $page = $this->get("page");
    if ($page) {
	my $s_diskpath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_diskpath = $s->diskpath;
	}
	else {
	    $s_diskpath = $config{server}{HTMLroot};
	}
	if ($s_diskpath) {
	    if (! $dir || $page->{type} ne "page") { 
		if ($page->{type} eq "template") { 
		    $s_diskpath .= "/_Template";
		}
		elsif ($page->{type} eq "library") { 
		    $s_diskpath .= "/_Library";
		}
	    }
	    return $s_diskpath if ($dir && $page->{type} eq "page");
	    return "$s_diskpath/".$this->filename;
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the page
#### check for published page
#### else, return a link for dynamic page

sub get_url {
    my $this = shift;
    my $static_req = shift; # static url requested?
    my $static = defined $static_req ? $static_req : 1;  # default to static
    #if ($this->{publish}) { $static = 1; }
    my $page = $this->get;
    return undef if (! $page);
    #return undef if ($page->{type} eq "library");
    if ($page->{section_id}) {
	my $file = $this->filename;
	my $s = new ExSite::Section(id=>$page->{section_id});
	my $s_url = $s->get_url;
	my $s_diskpath = $s->diskpath;
	my ($path_to_page,$url_to_page);
	if ($page->{type} eq "template") { 
	    $path_to_page = $s_diskpath."/_Template/".$file;
	    $url_to_page = $s_url."/_Template/".$file;
	}
	elsif ($page->{type} eq "library") { 
	    $path_to_page = $s_diskpath."/_Library/".$file;
	    $url_to_page = $s_url."/_Library/".$file;
	}
	else {
	    $path_to_page = $s_diskpath."/".$file;
	    $url_to_page = $s_url."/".$file;
	}
	########### fixme:  remote sites need new logic here

	# if CMScookie is set, use dynamic pages only, so that we can stay
	# in edit mode while we navigate
	my $CMScookie = &get_cookie("ExSiteCMS");

	# return the page URL
	if ($page->{type} ne "page") {
	    # always return path for libraries and templates
	    return $url_to_page;
	}
	elsif ($this->is_publishable()) {
	    if ($static || 
		(-e $path_to_page && ! $CMScookie)) {
		return $url_to_page;
	    }
#	elsif ($this->is_publishable() && -e $path_to_page) {
#	    # there is a published version to show
#	    # use it, unless we are in a CMS preview mode
#	    if (! $CMScookie || $static) {
#		return $url_to_page;
#	    }
	}
	# cannot show a published version - fall back on dynamic
	return "$config{server}{CGIpath}/page.cgi?_id=$page->{page_id}";
	# don't use full path - admin server may not be the same as the
	# public server
	#return "$config{server}{server}$config{server}{CGIpath}/page.cgi?_id=$page->{page_id}";
    }
    return undef;
}

# fetch_all_content_objs - returns array of content objects for this page
#### only returns ones belonging to this page.
#### Might be useful to have a function to return all content objects used
#### by this page, including templated objects, etc.

sub fetch_all_content_objs {
    my ($this,%options) = @_;
    my @contents = $this->fetch_child("content","page",$this->id(),"content_id");
    my @cobjs;
    foreach my $content (@contents) {
	push @cobjs, new ExSite::Content(content=>$content,%options);
    }
    return @cobjs;
}

# create a hotlink to edit content data

sub edit_content_link {
    my ($this,$cobj) = @_;
    return if (! $cobj);

    #return "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$this->{page}->{section_id}/$this->{page_id}/$cobj->{content_id}?action=update\" target=\"update\"><img src=\"$config{server}->{HTMLpath}$config{server}->{imgdir}/edit20.gif\" height=20 width=20 border=0 title=\"update $cobj->{name}\"></a>";

    my $path = "$this->{context}{page}{section_id}/$this->{context}{page}{page_id}/$cobj->{context}{content}{content_id}";

    my $name = "$cobj->{context}{content}{name}";

    my $out;

    #my $url = "$config{server}->{CGIpath}/CMS.cgi/$path?";
    #my $url = "$config{server}->{CGIpath}/ctrl-panel.cgi/CMS?path=/$path";

    if ($config{cms}{editmode} eq "easy") {
	$out .= "<!--&CMS(mode=inline&cmscmd=edit_content_link&name=${name}&path=/$path)-->";
    }
    elsif ($config{cms}{editmode} eq "MySite") {
	$out .= "<!--&MySite(ecmd=edit_content_link)-->";
    }
    else {

	my $url = "$config{server}->{CGIpath}/ctrl-panel.cgi/CMS?path=/$path";

	# update content, if user has editorial/design priviledges

	$out .= "<a href=\"${url}&cmscmd=update\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/plus16.gif\" height=16 width=16 border=0 title=\"update $name\"></a>";

	# manage content, if user has editorial/design priviledges

	$out .= "<a href=\"$url\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/box16.gif\" height=16 width=16 border=0 title=\"manage $name\"></a>";
    }

=pod
    # approve content, if user has suprvisory priviledges, and content is new

    $out .= "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$path?action=approve\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/check16.gif\" height=16 width=16 border=0 title=\"approve $cobj->{name}\"></a>";

    # remove/rollback/unapprove content, if user has sufficient priviledges

    $out .= "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$path?action=rollback\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/x16.gif\" height=16 width=16 border=0 title=\"restore previous $cobj->{name}\"></a>";
=cut
    return $out;
}

# write the page to a file

sub publish {
    my $this = shift;
    my $out;
    #$share{Page} = $this;  # save page pointer for DCDs
    $this->{publish} = 1;  # flag this page as going to disk
    my $url = $this->get_url(1); # get static url
    my $diskpath = $this->diskpath(1); # directory only
    my $page = $this->get;
    if (! $page) { 
	$this->error("Page::publish: no page to publish");
	return "No page to publish!";
    }
    else {
	return "" if ($page->{status} eq "disabled");
	$this->info("Page::publish: publishing page $page->{page_id}");
	$this->set_context;
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    return $this->show_diagnostics("error","html");
	}
	$out .= $mkdir;
	# First, publish virtual files and external content objects.
	# This allows the page to find the published files and link to them
	# instead of the dynamic versions.
	#### Might be better to save all content objects actually 
	#### inserted into the page during substitute(), and use those.
	my @contents = $this->fetch_all_content_objs(version=>"current");
	foreach my $content (@contents) {
	    if ($content->publishable) {
		$out .= $content->publish($diskpath);
	    }
	}
        if ($page->{type} eq "page" && 
            $page->{access} eq "public" &&
            $page->{publish_method} eq "static" ) {
	    # publish page
	    my $filename = $this->filename;
	    $filename =~ /^([\w-]+.(html?|php))$/;
	    my $secure_filename = $1;
	    ##### do we want to allow publication anywhere on the server?
	    $diskpath =~ /^((\/[\w-]+)+)$/;
	    my $secure_diskpath = $1;
	    if ($secure_filename) {
		#$this->{publish} = 1;  # flag this page as going to disk
		$out .= "<P>Saving page $this->{page_id} to <a href=$url>$secure_filename</a>... ";
		if (open OUTFILE, ">$secure_diskpath/$secure_filename") {
		    # WARNING: tampering with environment to spoof normal 
		    # page generation context
		    $ENV{SCRIPT_NAME} = "$config{server}{CGIpath}/page.cgi";
		    $ENV{QUERY_STRING} = "_id=$page->{page_id}";
		    print OUTFILE $this->expand;
		    close OUTFILE;
		    $out .= "succeeded.</P>\n";
		}
		else {
		    $this->error("Page::publish: publish $secure_filename failed: $!");
		    $out .= "<SPAN CLASS=error>FAILED: $!</SPAN></P>\n";
		    return $out;
		}
	    }
	    else {
		$this->error("Page::publish: invalid filename to publish to: $filename");
		$out .= "<P class=error>Invalid filename to publish to: $filename.</p>\n";
	    }
	}
	elsif ($page->{type} eq "template" && $page->{publish_method} eq "static" ) {
	    # publish precompiled template
	    my $secure_filename = "index.html";	
	    $diskpath =~ /^((\/[\w-]+)+)$/;
	    my $secure_diskpath = $1;
	    #$this->{publish} = 1;  # flag this page as going to disk
	    $out .= "<P>Precompiling template $this->{page_id}... ";
	    if (open OUTFILE, ">$secure_diskpath/$secure_filename") {
		# WARNING: tampering with environment to spoof normal 
		# page generation context
		$ENV{SCRIPT_NAME} = "$config{server}{CGIpath}/page.cgi";
		$ENV{QUERY_STRING} = "_id=$page->{page_id}";
		print OUTFILE $this->expand(expand=>"template");
		close OUTFILE;
		$out .= "succeeded.</P>\n";
	    }
	    else {
		$this->error("Page::publish: publish $secure_filename failed: $!");
		$out .= "<SPAN CLASS=error>FAILED: $!</SPAN></P>\n";
		return $out;
	    }
	}
    }
    return $out;
}

# remove the page file(s)

sub unpublish {
    my $this = shift;
    my $out;
    my $diskpath = $this->diskpath();
    my $page = $this->get;
    if (! $page) { 
	return $this->error("Page::unpublish: no page to remove");
    }
    else {
	$this->info("Page::unpublish: removing $page->{filename}");
	$this->set_context;
        if ($page->{type} eq "library" || $page->{type} eq "template") {
	    # remove the library/template directory
	    $this->rmdir($diskpath);
	}
        else {
	    # remove the page file
	    $this->rm($diskpath);
	}
    }
    return $out;
}

# get child pages, in appropriate order

sub get_child_pages {
    my ($this,$version) = @_;

    # use the current version by default
    if (! $version) { $version = $this->{page}{version}; }

    # determine the parent page to find children of
    my $parent = $this->{page}{type} eq "alternate" ?
	$this->{page}{parent_id} : 
	$this->id;
    my @pages = $this->fetch_match("page",{parent_id=>$parent},"rank");

    # create an array of Page objects
    my @Page;
    foreach my $page (@pages) {
	if ($version) {
	    # alternate version requested
	    my @alt = $this->fetch_match("page",{parent_id=>$page->{page_id},
						 type=>"alternate",
						 version=>$version,
						 },"page_id");
	    if (@alt > 0) {
		# alternate version of this page was found
		# swap it in place
		$page = $alt[-1];
	    }
	}
	push @Page, new ExSite::Page(page=>$page);
    }
    return @Page;
}

# get master page, as a Page object

sub master {
    my ($this) = @_;
    if ($this->{data}{type} eq "alternate") {
	return new ExSite::Page(id=>$this->{data}{parent_id});
    }
    return $this;
}

# get parent page, as a Page object

sub parent {
    my ($this) = @_;
    my $page = $this->get();
    my $version = $page->{version};
    # get master
    if ($page->{type} eq "alternate") {
	$page = $this->fetch("page",$page->{parent_id});
    }
    # take the parent of the master, in the appropriate version
    return new ExSite::Page(id=>$page->{parent_id},version=>$version);
}

# change template for this page

sub set_template {
    my ($this,$template) = @_;
    my $tid = ref $template eq "HASH" ? $template->{page_id} : $template;
    $this->set_context();

    # sanity checks

    if (! $tid) {
	return $this->error("set_template: no template given");
    }
    if (! $this->owns()) {
	return $this->error("set_template: you do not have sufficient priveleges to modify this page");
    }
    my $s = new ExSite::Section(section=>$this->get("section"));
    my @allowed_templates = $s->get_templates(0);
    my $t;
    foreach $t (@allowed_templates) {
	if ($t->{page_id} == $tid) {
	    $share{DB}->update("page",{template_id=>$tid},{page_id=>$this->id()});
	    return undef;
	}
    }
    return $this->error("set_template: you do not have permission to use this template");
}

# change body for this page

sub set_body {
    my ($this,$body) = @_;
    my ($bid,$content);
    if (ref $body eq "HASH") {
	$bid = $body->{content_id};
	$content = new ExSite::Content(content=>$body);
    }
    else {
	$bid = $body;
	$content = new ExSite::Content(id=>$body);
    }
    $content->set_context();

    # sanity checks

    if (! $bid) {
	return $this->error("set_body: no page body");
    }
    if ($content->get("content","name") ne "body") {
	return $this->error("set_body: not a page body");
    }
    if (! $content->owns()) {
	return $this->error("set_body: you do not have sufficient priveleges to work with this page body");
    }

    my $s = new ExSite::Section(section=>$content->get("section"));
    my @allowed_bodies = $s->get_bodies;
    my $b;
    foreach $b (@allowed_bodies) {
	if ($b->{content_id} == $bid) {
	    # get actual body
	    my $db = $share{DB};
	    my @pbody = $db->fetch_match("content",{page_id=>$this->id(),
						    name=>"body"});
	    my $bodydata = $content->get("content_data");
	    delete $bodydata->{content_data_id};
	    $bodydata->{content_id} = $pbody[0]{content_id};
	    if (@pbody == 0) {
		# no body - make one!
		$this->make_content({name=>"body",
				     page_id=>$this->id(),
				     type=>"editorial"},
				    $bodydata);
	    }
	    # update the existing body
	    my $c = new ExSite::Content(content=>$pbody[0]);
	    $c->update($bodydata);
	    return undef;
	}
    }
    return $this->error("set_body: you do not have sufficient priveleges to work with this page body");
}

# validate a page record
#
# FIXME: validate list: datatypes

sub validate {
    my ($this,$data) = @_;
    my ($page,$type,@err);
    if (! $data) { $data = $this->get("page"); }

    # generic validation (using DBmap validation rules)
    push @err, $share{DB}->validate_record("page",$data);

    # special validation
    $type = $data->{type};
    if ($type eq "page" || $type eq "alternate") {
	if (! $data->{template_id}) {
	    push @err, "page is missing its template";
	}
	if (! $data->{filename} || $data->{filename} !~ /\.(html?|php|asp)$/) {
	    push @err, "page has an invalid filename";
	}
	if (! $data->{label}) {
	    push @err, "page is missing a label";
	}
	if (! $data->{access}) {
	    push @err, "page has no access control defined";
	}
	if (! $data->{visibility}) {
	    push @err, "page has no publish method defined";
	}
    } 
    elsif ($type eq "template") {
	if (! $data->{filename} || $data->{filename} !~ /^[\w_\-]$/) {
	    push @err, "template has an invalid publication name";
	}
	if (! $data->{access}) {
	    push @err, "template has no access control defined";
	}
    }
    elsif ($type eq "library") {
	if (! $data->{filename} || $data->{filename} !~ /^[\w_\-]$/) {
	    push @err, "library has an invalid publication name";
	}
	if (! $data->{access}) {
	    push @err, "library has no access control defined";
	}
    }
    if (@err == 0) {
	if ($this->is_dupe($data)) {
	    push @err, "a $type with that publish name already exists";
	}
    }
    return @err;
}

# is_dupe : returns TRUE if a page with matching publication characteristics
# exists in the system

sub is_dupe {
    my ($this,$data) = @_;
    return 0 if (! $data->{section_id} ||
		 ! $data->{type} ||
		 ! $data->{filename});

    # check for dupe pages in this site

    my @p = $this->fetch_match("page",{section_id=>$data->{section_id},
				       type=>$data->{type},
				       filename=>$data->{filename}});
    if (@p > 1) {
	return 1;
    }
    elsif (@p == 1) {
	if ($data->{page_id}) {
	    return ($data->{page_id} != $p[0]{page_id});
	}
	else {
	    return 1;
	}
    }
    return 0;
}

# make_page : insert a new page record
#
# can optionally pass a body (content_data) record to seed it with
#
# FIXME:  should be able to pass a body template ID, too

sub make {
    return &make_page(@_);
}

sub make_page {
    my ($this,$page,$body,$opt) = @_;
    if ($page->{page_id}) {
	# invalid page
	$this->error("invalid attempt to insert an already-defined page record");
	return undef;
    }
    if (! $this->allow_insert("page",$page)) {
	# permission denied
	$this->error("make_page: permission denied");
	return undef;
    }
    my @err = $this->validate($page);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    # remove any residual timestamps
    delete $page->{mtime};
    delete $page->{ctime};
    $share{DB}->run_handler("pre_make_page",$page);
    my $pid = $share{DB}->insert("page",$page);
    if (! $pid) {
	$this->error("failed to install new content $page->{filename}");
    }
    else {
	$this->info("installed new page $page->{filename}");
	my $cid;
	if (ref $body eq "HASH") {
	    $cid = $this->make_content({name=>"body",
					type=>"editorial",
					page_id=>$pid},
				       $body,$opt);
	    # ignore return code; make_content_data will report its own errs
	}
	$body->{content_id} = $cid;
	$share{DB}->run_handler("post_make_page",$page);
    }
    return $pid;
}

# delete the page

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("page",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("page",$id);
	return 1;
    }
    return 0;
}

1;


