#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::PageBuf;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Base;

our @ISA = qw(ExSite::Base);

#------------------------------------------------------------------------

=pod

=head1 PageBuf.pm : ExSite Web Page Builder

This is a legacy page-building framework for ExSite, which has largely been
superceded by ExSite V3's page.cgi.  However, it still has some practical
uses for non-page.cgi CGI programs, due to its ability to reset output, 
or redirect to other pages.

Features:

=over 4

=item * run in buffered or stream mode

=item * add content to different page sections, in any order

=item * clear or replace whole sections with new content

=item * cleanly abort pages that are already fully or partly constructed

=item * redirect to other URLs even if page is partly built

=item * dump to stdout with HTTP headers, or publish to a file

=item * benchmarking and diagnostic tools for sending extra info to client

=back

In buffered mode, all page sections are stored until the end, and
dumped in the proper order at that time.  This allows pages to be
assembled in non-linear fashion, and for page sections (or entire
pages) to be cleared and re-composed after already being laid out.  In
addition, it allows multiple pages to be built at once, which can be
useful for publishing applications.  In stream mode, page output goes
straight to stdout; diagnostic buffers are dumped after the whole page
is output.

Most of the advanced features are only available in buffered mode.
Stream mode is supported mostly for compatibility with scripts that
write directly to stdout.

A standard ExSite page has up to 5 sections:

=over 4

=item 1. body - the section containing the main page content

=item 2. head - standard top-of-page masthead

=item 3. foot - standard end-of-page footer

=item 4. left sidebar - auxiliary column of content on the left side of page

=item 5. right sidebar - auxiliary column of content on the right side of page

=back

In addition, there is a section 0, which contains the HTTP headers.
Additional sections, if they become defined, are used as diagnostic
buffers, and output following the page, if the diagnostics are on.

Not all of these sections are enabled; the parameter $DEFAULT_NSEC
determines which ones are used by default (every section from 1 to the
number of sections is enabled).  This can be overridden in the Page
constructor.  The diagnostic buffers are off by default, but can be
turned on by calling $page->debug, or by setting the debug option in
the Page constructor.

To abort or abandon a page in progress, do one of:

=over 4

=item 1 C<$page-E<gt>clear;  # erases page so you can start over>

=item 2 C<$page-E<gt>error("error html"); # outputs an error page instead>

=item 3 C<$page-E<gt>location("new_url"); # redirect to another page>

=back

=cut

#------------------------------------------------------------------------
# Methods:
#    new : initialize a page
#    httphead : append to HTTP headers
#    header : build page header
#    footer : build page footer
#    print : append HTML to page body
#    printTo : append HTML to a page section
#    build : assemble the whole page
#    show : output page to stdout, with HTTP headers
#    publish : output page HTML to a file
#    clear : erase current page
#    error : cancel all page output so far, display error message
#    location : cancel page, redirect to the given page
#    DESTROY : automatically show the page if not already done
#    debug : enable the diagnostic buffers
#    benchmark : display benchmarking info in the diagnostic buffers
#    note : add an arbitrary note to the diagnostic buffers
#    bug : same as note
#    dhead : add a title to a diagnostic buffer
#    printenv : display the server environment in the diagnostic buffers
#    showdb : display the database object in the diagnostic buffers


#package ExSite::Page;

#use strict;
#use ExSite::Config;
#use ExSite::Misc;

BEGIN {
  use Exporter   ();
  use vars       qw(@EXPORT_OK);
  @EXPORT_OK =   qw($DEFAULT_NSEC $DEFAULT_MODE
		    $HTTP $BODY $HEAD $FOOT $LEFT $RIGHT);
}

use vars (@EXPORT_OK);

$DEFAULT_MODE = "buffered";
$DEFAULT_NSEC = 3;

$HTTP = 0;
$BODY = 1;
$HEAD = 2;
$FOOT = 3;
$LEFT = 4;
$RIGHT = 5;

#------------------------------------------------------------------------

=pod

=head2 new() - Page constructor

Creates a new page.

options (all are optional)

=over 4

=item * title = page title

=item * benchmark = initialize benchmark time counters.  Benchmarking 
results are written to the diagnostic headers, and are only displayed 
if diagnostics are turned on.

=item * debug = enable the diagnostic buffers, if true.

=item * stylesheet = the stylesheet to use for this page 
(only needed to override the default).

=item * httphead = HTTP header text.  (Optional, if Content-type is 
the only header.)

=item * headtype = optional argument to pass to C<&page_header>, to
provide extra flexibility in laying out the top of page.

=item * foottype = optional argument to pass to C<&page_footer>, to
provide extra flexibility in laying out the bottom of page.

=item * mode = buffered | stream.  In stream mode, page content is dumped 
direct to stdout as soon as received.  In buffered mode, content is saved
and assembled/dumped only when the page is complete.

=item * publish = set true if this page is intended to be published (serves as
             a flag to alert other programs of different disposition for
             the page).

=item * nsec = number of page sections

=over 4

=item 1. body only

=item 2. body and header

=item 3. body, header, and footer

=item 4. body, header, footer, and left sidebar

=item 5. body, header, footer, left, and right sidebar

=back

=back

B<Examples:>

C<my $page = new ExSite::Page( { title=E<gt>"Cool Stuff" } );>

=cut

#------------------------------------------------------------------------

sub new ($%) {
    my($this, %option) = @_;

    my ($class,$title,%page);
    %page = %option;

    # general benchmarking feature
    if (exists $page{benchmark} || $ENV{QUERY_STRING} =~ /benchmark/) {
      ($page{user},$page{system}) = times;
    }

    # store page components
    $page{section} = [];

    # set the mode
    if (!exists $page{mode}) { 
	$page{mode} = $config{page_mode} ? $config{page_mode} : $DEFAULT_MODE;
    }

    # look for header & footer flags
    if (!exists $page{header}) { $page{header} = 1; }
    if (!exists $page{footer}) { $page{footer} = 1; }

    # set page title
    if (!exists $page{title}) { $page{title} = "$config{site}->{name} $0"; }

    # set stylesheet
    if (!exists $page{stylesheet}) { $page{stylesheet} = "$config{server}{HTMLpath}/_ExSite/css/ExSite.css"; }

    # how many sections for this page?
    if ($page{nsec} < 1 || $page{nsec} > 5) { $page{nsec} = $DEFAULT_NSEC; }

    # HTTP headers
    $page{section}->[0] = $page{httphead};

    # set page buffering flag (in Config.pm) on
    #$config{page_buffering} = 1;

    # objectify object
    my $rpage = \%page;
    $class = ref($this) || $this;
    bless $rpage, $class;

    # save PageBuf object in share
    $share{PageBuf} = $rpage;

    # automatically invoke header & footer methods
    if ($page{header}) { &header($rpage,%page); }
    if ($page{footer}) { &footer($rpage,%page); }

    $rpage->info("created buffered page \"$page{title}\"");
    return $rpage;
}

#------------------------------------------------------------------------

=pod

=head2 httphead($string) - append to HTTP headers

Appends the passed string to the HTTP headers

B<Example:>

C<$page-E<gt>httphead("Set-Cookie: arf-arf!\n");>

=cut

#------------------------------------------------------------------------

sub httphead ($$) {
  my($this, $hdr) = @_;
  $this->{section}->[$HTTP] .= $hdr;
}

#------------------------------------------------------------------------

=pod

=head1 Automatic Page Header and Footer Generation

=head2 header() - initialize the top of page

Builds the HTML for the top of the page.  The standard header for
the site is taken from C<&ExSite::Config::page_header>, which in 
turns gets it from C<&myConfig::my_page_header>.

C<header()> is automatically invoked by C<new()> in the default case.
It will be skipped if the C<header> argument is set to 0.

The C<headtype> (set in the call to C<new()>) will be passed to
C<&page_header()> and then to C<&my_page_header()>, to pass extra
information for building the top of the page.

If you need to use a custom header, use C<$page-E<gt>printTo[2]>.

=head2 footer() - initialize the bottom of page

Builds the HTML for the bottom of the page.  The standard footer for
the site is taken from C<&ExSite::Config::page_footer>, which in 
turns gets it from C<&myConfig::my_page_footer>.

C<footer()> is automatically invoked by C<new()> in the default case.
It will be skipped if the C<footer> argument is set to 0.

The C<foottype> (set in the call to C<new()>) will be passed to
C<&page_footer()> and then to C<&my_page_footer()>, to pass extra
information for building the bottom of the page.

If you need to use a custom footer, use C<$page-E<gt>printTo[3]>.

=cut

#------------------------------------------------------------------------

sub header ($%) {
  my($this, %option) = @_;
  # begin standard page
  $this->{section}->[$HEAD] = &page_header($this->{title},$this->{headtype});
}

sub footer($) {
    my $this = shift @_;
    $this->{section}->[$FOOT] = &page_footer($this->{headtype});
}

#------------------------------------------------------------------------
# general printing functions

=pod

=head1 General Printing and Output Functions

=head2 print() - add content to the page body

Appends all of the passed arguments to the body of the page.
C<printBody()> is a synonym for this.

B<Example:>

C<$page-E<gt>print("E<lt>H1E<gt>NewsE<lt>/H1E<gt>\n", "E<lt>PE<gt>No news is good news.E<lt>/PE<gt>\n");>

=head2 printTo() - add content to a particular page section

The first argument is the page section to print to;  the remaining
arguments are the content to print.  In stream mode, the content only
prints if the section matches the current section.

B<Example:>

C<# print to left sidebar>

C<$page-E<gt>printTo(4,"E<lt>H1E<gt>NewsE<lt>/H1E<gt>\n", "E<lt>PE<gt>No news is good news.E<lt>/PE<gt>\n");>

=head2 printLeft() - add content to the left sidebar

Adds its arguments to the left sidebar (section 4).

B<Example:>

C<# print to left sidebar>

C<$page-E<gt>printLeft("E<lt>H1E<gt>NewsE<lt>/H1E<gt>\n", "E<lt>PE<gt>No news is good news.E<lt>/PE<gt>\n");>

=head2 printRight() - add content to the right sidebar

Similar to printLeft, above.

=cut

#------------------------------------------------------------------------

sub print {
  &printBody(@_);
}

sub printTo($$@) {
    my $this = shift @_;
    my $section = shift @_;
    if ($this->{mode} eq "buffered") {
	$this->{section}->[$section] .= join($, , @_);
    }
    elsif ($this->{mode} eq "stream") {
	if ($section == $this->{current_section}) {
	    print join($, , @_);
	}
    }
}

sub printBody($@) {
    my $this = shift @_;
    $this->printTo($BODY,@_);
    #$this->{section}->[$BODY] .= join($, , @_);
}

sub printLeft($@) {
    my $this = shift @_;
    $this->printTo($LEFT,@_);
    #$this->{section}->[$LEFT] .= join($, , @_);
}

sub printRight($@) {
    my $this = shift @_;
    $this->printTo($RIGHT,@_);
    #$this->{section}->[$RIGHT] .= join($, , @_);
}

#------------------------------------------------------------------------
# page construction functions

=pod

=head1 Page Construction Features

=head2 build() - assemble a page

Assembles all page sections into a complete HTML page, and returns
the complete HTML in a string.

=head2 show() - view a page

Dumps a page to stdout, for view by the client browser.  This must be
called for the page to become visible; if the calling agent neglects
to call it, however, the DESTROY method should detect that condition
and make the call automatically.  C<show()> can only be called once
per page.

=head2 publish($file) - write the page out to a file

file = the filename to write to (including path, optionally)

Security Note: this will not generally work from a CGI script. To make
it work, create a C wrapper that is setuid to the owner of the directory
where the file will be written.  You may also need to fix the Path at
the top of your publish script, eg:

C<BEGIN { push @INC,"."; }>

=cut

#------------------------------------------------------------------------

sub build($) {
    my $this = shift @_;
    my $out;

    # page header

    if ($this->{section}->[$HEAD] && $this->{nsec} >= 2) {
	$out .= $this->{section}->[$HEAD];
    }

    # left sidebar

    if ($this->{nsec} >= 4) {
	$out .= "<TABLE CELLSPACING=0 CELLPADDING=2 BORDER=0><TR><TD>\n";
	if ($this->{section}->[$LEFT]) {
	    $out .= $this->{section}->[$LEFT];
	}
	$out .= $this->{section}->[$HEAD];
	$out .= "</TD><TD>\n";
    }
    
    # body

    if ($this->{section}->[$BODY]) {
	$out .= $this->{section}->[$BODY];
    }

    # right sidebar

    if ($this->{nsec} >= 4) {
	$out .= "</TD></TD>\n";
	if ($this->{section}->[$LEFT]) {
	    $out .= $this->{section}->[$LEFT];
	}
	$out .= $this->{section}->[$HEAD];
	$out .= "</TD></TR></TABLE>\n";
    }

    # footer

    if ($this->{nsec} >= 3) {
	if ($this->{section}->[$FOOT]) {
	    $out .= $this->{section}->[$FOOT];
	}
    }

    return $out;
}

sub show($) {
    my $this = shift @_;
    return if ($this->{done});

    # HTTP headers

    if (! exists $this->{to_file}) {
	print $this->{section}->[$HTTP];
	if ($this->{section}->[$HTTP] !~ /Content-type:/s) {
	    # add a content-type header, if missing
	    print "Content-type: text/html\n\n";
	}
	elsif ($this->{section}->[$HTTP] !~ /\n\n$/s) {
	    # add a blank link to indicate end of headers
	    print "\n";
	}
    }

    # output the page itself

    print $this->build;

    # diagnostics

    if (exists $this->{benchmark} || $ENV{QUERY_STRING} =~ /benchmark/) {
	# benchmark, if requested
	$this->benchmark;
    }
    if ($this->{debug}) {
	print "<HR><H2>Page Diagnostics:</H2>\n<OL>\n";
	for (my $idbg = 6; $idbg < scalar @{$this->{section}}; $idbg++) {
	    print "<LI>".$this->{section}->[$idbg]."</LI>\n";
	}
	print "</OL>\n";
    }

    $this->{done} = 1;
}

sub publish($) {
    my ($this,$file) = @_;
    return if ($this->{done});
    open OUTFILE, ">$file" or return 0;
    print OUTFILE $this->build;
    close OUTFILE;
    return 1;
}

#------------------------------------------------------------------------
# page abort features

=pod

=head1 Page Abort Features

=head2 error($message) - substitute an error page for the current page

This method aborts the current page in progress, and replaces it 
with an error page, the primary content of which is in the string
C<$message>.  The error content only replaces the body of the page;
other sections remain as they were.

=head2 location($url) - substitute another URL for the current page

This method aborts the current page in progress, and redirects the
client to another URL.

=head2 clear($section) - erase contents of a section

Clears a particular section, if specified, or all sections, if not.

=head2 cancel() - cancel the page and output nothing

Cancels the page and outputs nothing at all.

=cut

#------------------------------------------------------------------------

sub error($$) {
    my ($this,$message) = @_;

    # replace any existing body with an error message
    $this->{section}->[$BODY] = $message;
    $this->warn("page aborted: $message");

    # FIXME:
    # add bug report feature
    # add error logging
    
    $this->show;
}

sub location($$) {
    my ($this,$url) = @_;
    print "Location: $url\n\n";
    $this->{done} = 1;
}

sub redirect {
    return &location(@_);
}

sub clear($$) {
    my ($this,$section) = @_;
    if ($section) {
	$this->{section}->[$section] = "";
    }
    else {
	foreach my $section (@{$this->{section}}) {
	    $section = "";
	}
    }
}

sub cancel($) {
    my ($this) = @_;
    $this->{done} = 1;
}

#------------------------------------------------------------------------
# destructor

=pod

=head2 DESTROY() - Page destructor

DESTROY() is the destructor method for the Page class;  if it senses 
that the page has not been disposed of yet, it will invoke the 
C<show()> method before destroying the object.

=cut

#------------------------------------------------------------------------

sub DESTROY($) {
    my $this = shift @_;
    if (!$this->{done}) { &show($this); }
}

#========================================================================
# DIAGNOSTIC TOOLS

=pod

=head1 Diagnostic Tools:

=head2 debug($flag) - turn diagnostics on/off

With no arguments, this toggles the setting of the diagnostic buffers.
Otherwise, the flag is taken to be the setting for diagnostics
(should be 1/0).

When diagnostics are on, additional information in the diagnostic
buffers can be displayed at the bottom of the page.  By turning
diagnostics off, the page is restored to its normal display.

Diagnostics are stored in buffers numbered from 1 on up.  If a
particular buffer is specified, a diagnostic will be stored there;
otherwise buffer 1 is used by default.  When the diagnostics are
dumped, they are dumped in order, allowing you to group related
diagnostics together.

=head2 note($note, $section) - adds a note to the diagnostic buffers.  

C<$note> is an arbitrary string of HTML or text, and are automatically
terminated with a "E<lt>brE<gt>".  C<$section> defaults to 1.

=head2 bug($string, $section) - synonym for note()

=head2 dhead($title, $section) - add a diagnostic header

Diagnostic headers are titles placed at the front of the diagnostic
buffer.  C<$section> defaults to 1.

=head2 benchmark($section) - benchmarking diagnostics

When called the first time, this will initialize the benchmarking
counters.  (This can be done automatically in the C<new()> options.)

If called again, this will add the current report total u/s time for 
process and u/s time since page creation to the diagnostics.

C<$section> is optional.

=head2 printenv($section) - display the server environment

C<printenv()> displays all environment variables in the server
environment.  C<$section> is optional.

=head2 showdb() - display the ExSite::DB object, if defined.

The DB object (or the Report/Form object that inherit from it) is
displayed in the diagnostics.

DANGEROUS: the db object may contain passwords, so be careful about 
using this method on any pages that might be publicly viewable.

=cut

#------------------------------------------------------------------------

sub debug {
    my ($this) = shift;
    if ($#_ == -1) { $this->{debug} = 1; }
    else { $this->{debug} = shift; }
}

sub note {
    my ($this,$note,$dsection) = @_;
    my $section;
    if ($dsection > 0) { $section = $dsection + 5; }
    else { $section = 6; }
    $this->{section}->[$section] .= "$note<br>\n";
}

sub bug { 
    &note(@_);
}

sub dhead {
    my ($this,$title,$dsection) = @_;
    my $section;
    if ($dsection > 0) { $section = $dsection + 5; }
    else { $section = 6; }
    $this->{section}->[$section] .= 
	"<H3>$title</H3>\n".$this->{section}->[$section]; 
}

sub benchmark {
    my ($this,$dsection) = @_;
    my $section;
    if ($dsection > 0) { $section = $dsection + 5; }
    else { $section = 6; }
    if (exists $this->{user}) {
	# already initialized the times, report current usage
	my ($user,$system) = times;
	my $utime = $user - $this->{user};
	my $stime = $system - $this->{system};
	$this->{section}->[$section] .= "Benchmark: total time=$user/$system;  since page creation=$utime/$stime<br>\n";
    }
    else {
	# initialize the benchmarks
	($this->{user},$this->{system}) = times;
    }
}

sub printenv {
    my ($this,$dsection) = @_;
    my $section;
    if ($dsection > 0) { $section = $dsection + 5; }
    else { $section = 6; }
    $this->{section}->[$section] .= "<DL><DT><b>Environment:</b></DT><DD>\n";
    foreach my $evar (keys %ENV) {
	$this->{section}->[$section] .= "$evar=$ENV{$evar}<br>\n";
    }
    $this->{section}->[$section] .= "</DD></DL>\n";
}

sub showdb {
    my ($this,$dsection) = @_;
    return if (! $this->{db});
    my $section;
    if ($dsection > 0) { $section = $dsection + 5; }
    else { $section = 6; }
    $this->{section}->[$section] .= &ShowHash($this->{db});
}

1;


