#------------------------------------------------------------------------
# ExSite::ReportBuilder
#----------------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
# ReportBuilder : view component for reporting system
#
# This is a utility for building tabulated reports.  It is
# content-agnostic, and simply handles the formatting of data into a
# table, depending on the input provided.  If the report is to be
# interactive (links, JS, etc.), those features must be included in
# the content of the report; ie. you must code the links, divs,
# etc. yourself in the values that you pass.
#
# INPUT:
# all input is passed in a control hash
#    title => content for titlebar; <th class="ReportTitle">
#    headers => array of header content; <th>
#    data => array of arrays of data content; <td>
#    tools => content for toolbar; <th class="ReportTools">
#    foot => content for footer; <td class="ReportFooter">
#    nodata => no data warning message
#    show_tools_if_nodata => flag controlling tool display if no data
#    show_data_if_nodata => flag controlling display of data area if no data
#
# The show_* flags are true/false values to control the display of
# those report fields/rows if there is no content for the respective
# rows.  In other cases, the row is left out completely if there is no
# content.
#
# OUTPUT:
#    HTML for report
#
# Reports have 3 structures:
#
# 1) if data is an array of arrays:
#
# +----------------------------------------------+
# | title                                        |
# +------------+---------------------+-----------+
# | header1    | header 2            | header 3  |
# +------------+---------------------+-----------+
# | data1A     | data2A              | data3A    |
# +------------+---------------------+-----------+
# | data1B     | data2B              | data3B    |
# +------------+---------------------+-----------+
# | tools                                        |
# +----------------------------------------------+
# | foot                                         |
# +----------------------------------------------+
#
# 2) if data is a simple array:
#
# +----------------------------------------------+
# | title                                        |
# +------------+---------------------------------+
# | header1    | data1                           |
# +------------+---------------------------------+
# | header2    | data2                           |
# +------------+---------------------------------+
# | header3    | data3                           |
# +------------+---------------------------------+
# | tools                                        |
# +----------------------------------------------+
# | foot                                         |
# +----------------------------------------------+
#
# 3) if data is a scalar
# +----------------------------------------------+
# | title                                        |
# +----------------------------------------------+
# | data                                         |
# +----------------------------------------------+
# | tools                                        |
# +----------------------------------------------+
# | foot                                         |
# +----------------------------------------------+
#--------------------------------------------------------------------

package ExSite::ReportBuilder;
use strict;
use ExSite::Base;
use ExSite::Config;

use vars qw(@ISA);

@ISA = qw(ExSite::Base);

sub new {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    # copy options into object
    while (my ($key,$val) = each %opt) {
	$obj->{$key} = $val;
    }
    $obj->initialize_object;
    if ($obj->{data}) { $obj->format; }
    return $obj;
}

sub init {
    my ($this,%opt) = @_;
    # clear previous report parameters
    foreach my $param (qw(title headers data tools foot nodata show_tools_if_nodata format)) {
	delete $this->{$param};
    }
    # copy new parameters into object
    while (my ($key,$val) = each %opt) {
	$this->{$key} = $val;
    }
    if ($this->{data}) { $this->format; }
    return;
}

# content for titlebar
sub title {
    my $this = shift;
    $this->{title} = shift;
}

# array of column headings
sub headers {
    my ($this,$head) = @_;
    if (ref $head eq "ARRAY") {
	$this->{headers} = $head;
    }
}

# content for toolbar
sub tools {
    my $this = shift;
    $this->{tools} = shift;
}

# content for footer region
sub foot {
    my $this = shift;
    $this->{foot} = shift;
}

# message if no data exists
sub nodata {
    my $this = shift;
    $this->{nodata} = shift;
}

# flag to show toolbar if there is no data
sub show_tools_if_nodata {
    my $this = shift;
    $this->{show_tools_if_nodata} = shift;
}

# flag to show toolbar if there is no data
sub show_data_if_nodata {
    my $this = shift;
    $this->{show_data_if_nodata} = shift;
}

# the data itself, as an array, or an array of arrays
sub data {
    my ($this,@data) = @_;
    $this->{data} = \@data;
    $this->format;
}

# decide which format (brief/full) to use
sub format {
    my ($this,$fmt) = @_;
    if ($fmt) { 
	$this->{format} = $fmt; 
	return $fmt;
    }
    return $this->{format} if ($this->{format});
    if (ref $this->{data} eq "ARRAY") {
	if (ref $this->{data}[0] eq "ARRAY" || 
	    scalar @{$this->{data}} == 0) {
	    $this->{format} = "brief";
	    $this->{show_tools_if_nodata} = 1;
	}
	else {
	    $this->{format} = "full";
	    $this->{show_tools_if_nodata} = 0;
	}
    }
    elsif (!ref($this->{data}) && defined($this->{data}) && $this->{data}){
    	$this->{format} = "simple";
	$this->{show_tools_if_nodata} = 1;
    }
    else {
	$this->{format} = "brief";
	$this->{show_tools_if_nodata} = 1;
    }
    return $this->{format};
}

# how many columns?
sub ncol {
    my $this = shift;
    if ($this->format eq "brief") {
	if ($this->{headers}) {
	    if (ref $this->{headers} eq "ARRAY") {
		return scalar @{$this->{headers}};
	    }
	    return 1;
	}
	elsif ($this->{data}) {
	    if (ref $this->{data} eq "ARRAY") {
		if (ref $this->{data}[0] eq "ARRAY") {
		    return scalar @{$this->{data}[0]};
		}
		else {
		    return scalar @{$this->{data}};
		}
	    }
	    return 1;
	}
	return 1;
    }
    else {
	return 2;
    }
}

# generate the report
sub make {
    my $this = shift;
    $this->info("ExSite::Report::report_generic called");

    # use handler to build the report, if one is defined
    my $out = $this->run_handler("report_generic",$this);
    return $out if (defined $out);

    my $ncol = $this->ncol;

    # proceed with standard generic report
    $out = "<table class=\"Report\"";
    if ($this->{width}) { $out .= " width=\"$this->{width}\""; }
    $out .= ">\n<tbody>\n";
    if ($this->{title}) {
	$out .= "<tr><th class=\"ReportTitle\" colspan=$ncol>$this->{title}</th></tr>\n";
    }

    if ($this->format eq "brief") {
	my $ndat = ref $this->{data} eq "ARRAY" ? scalar @{$this->{data}} : 0;
	if ($ndat) {
	    if ($this->{headers}) {
		$out .= "<tr>";
		foreach my $head (@{$this->{headers}}) {
		    $out .= "<th>$head</th>";
		}
		$out .= "</tr>\n";
	    }
	    if ($this->{data}) {
		my $nrow = 0;
		foreach my $row (@{$this->{data}}) {
		    # alternate row classes A/B for CSS highlighting purposes
		    $out .= "<tr class=\"".($nrow % 2 ? "B" : "A")."\">";
		    foreach my $col (@$row) {
			$out .= "<td>$col</td>";
		    }
		    $out .= "</tr>\n";
		    $nrow++;
		}
	    }
	}
	else {
	    if (! $this->{show_data_if_nodata}) {
		my $nodata = $this->{nodata} || "No data";
		$out .= "<tr><td><p class=\"error\">$nodata</p></td></tr>\n";
	    }
	    if (! $this->{show_tools_if_nodata}) {
		# don't show the tools
		delete $this->{tools};
	    }
	}
    }
    elsif ($this->format eq "simple") {
    	$out .= "<tr><td class=\"data\">$this->{data}</td></tr>\n";
    }
    else {
	if ($ncol == 2) {
	    while (scalar @{$this->{headers}}) {
		my $head = shift @{$this->{headers}};
		my $data = shift @{$this->{data}};
		$out .= "<tr><td class=\"label\">$head</td><td class=\"data\">$data</td></tr>\n";
	    }
	}
	else {
	    if ($this->{show_data_if_nodata}) {
		$out .= "<tr><th><p class=\"error\">$this->{nodata}</p></th></tr>\n";
	    }
	    if (! $this->{show_tools_if_nodata}) {
		# don't show the tools
		delete $this->{tools};
	    }
	}
    }

    if ($this->{tools}) {
	$out .= "<tr><th class=\"ReportTools\" colspan=$ncol>$this->{tools}</th></tr>\n";
    }
    if ($this->{foot}) {
	$out .= "<tr><td class=\"ReportFooter\" colspan=$ncol>$this->{foot}</td></tr>\n";
    }
    $out .= "</tbody>\n</table>\n";
    return $out;
}

1;
