package ExSite::SQL;

#------------------------------------------------------------------------
# ExSite::SQL
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# Generic SQL Driver for ExSite web development package.
# (Designed to work with MySQL, but should be easily adaptable to any
# SQL engine, since it goes through Perl DBI.)
#
# ExSite DB drivers have the following methods:
#
# new : make a new database handle
# dbget_tables : retrieve all table names
# dbget_columns : retrieve all column names from a table
# dbget_key : retrieve the primary key for a table
# dbselect : retrieve data from tables
# dbinsert : add new data to tables
# dbupdate : modify old data in tables
# dbdelete : remove data from tables
#
# See the ExSite::DB package for additional wrapper and convenience
# methods to fetch data from and modify the database.
#
# SECURITY NOTE: since the DB package is unaware of the underlying
# engine, it cannot perform proper security checks on the data being
# passed to the DB server.  Therefore all SQL parameters must be 
# validated within each driver, ie. in this file.
#------------------------------------------------------------------------

=pod

=head1 ExSite::SQL - generic SQL database driver for ExSite

The SQL class serves as a glue layer between ExSite::DB and the Perl
DBI classes.  Perl DBI (DataBase-Indepedent interface) supports many
different database engines, so this class can serve as a model for 
connecting to a variety of back-ends.  As supplied with ExSite, it is 
configured to operate with MySQL.

As with all ExSite Database drivers, ExSite::SQL implements the following
standard calls:

=over 4

=item new()

make a new database handle

=item dbget_tables()

retrieve all table names

=item dbget_columns()

retrieve all column names from a table

=item dbget_key()

    retrieve the primary key for a table

=item dbselect()

retrieve data from tables

=item dbinsert()

add new data to tables

=item dbupdate()

modify old data in tables

=item dbdelete()

remove data from tables

=back

There are two useful concepts for understanding how ExSite represents
data records:

=over 4

=item Data hash 

this is a record stored in a hash, or associative
array.  The hash keys are the column names, and the values are the 
field values for that record.  Groups of records are represented 
as arrays of hash references, so that (for example) the value of the
"price" field in the 3rd record is C<$data[2]{price}>.

=item Match hash

this is similar to a data hash, except that
the hash values represent data to match to.  For instance,
$match{price} = "4.95" will select all records whose price field
contains the value 4.95.  If there are multiple keys/values in the
match hash, all will be matched (ie. they imply a logical AND).
The ExSite::SQL driver accepts SQL wildcards ("%") in match values.

=back

Many ExSite::SQL methods will also accept an explicit SQL WHERE clause
in place of a match hash.  However, that mode of use is not portable,
and is recommended only when the simple AND logic of the standard match
hashes is insufficient.

=cut

use strict;
use Carp;
use DBI;
use ExSite::Config;
use ExSite::Base;
our @ISA = qw(ExSite::Base);

#------------------------------------------------------------------------

=pod

=head2 C<new(%options)> : SQL driver constructor

Usage: C<my $db = new ExSite::SQL(%options);>

Connects to the SQL server with the appropriate authentication 
information, and prepares to accept database queries.

C<%options> containts the database connection info.  Up to 6 parameters
are expected:

=over 4

=item driver

the DBI driver to use ("mysql" by default)

=item server

the server to connect to

=item port

the port number to connect to (optional)

=item name

the name of the database

=item user

the user name to connect under

=item pass

the password to authenticate with

=back

Example: C<my $db = new ExSite::SQL(name=E<gt>"my_db",server=E<gt>"my.server.com",user=E<gt>"my_name",pass=E<gt>"my_pwd");>

=cut

#------------------------------------------------------------------------

sub new($%) {
  my($this, %db) = @_;
  my($obj,$class);

  $obj = {};
  $class = ref($this) || $this;
  bless $obj, $class;

  if ($db{debug}) {
    $obj->{debug} = $db{debug};
    delete $db{debug};
  }

  # store most-recent operation status
  $obj->{query_error} = undef;
  # table info cache
  $obj->{sql_table_info} = {};

  # Connect!
  # typical $connect values
  # dbi:DriverName:database_name
  # dbi:DriverName:database_name@hostname:port
  # dbi:DriverName:database=database_name;host=hostname;port=port
  # BUT . . .
  # I'm following MySQL format by default and inserting
  #     modifications for other drivers
  # dbi:DriverName:database_name;host=hostname;port=port

  # default to MySQL
  $db{'driver'} = $db{'driver'} ? $db{'driver'} : "mysql";

  my $connect = join(":", "dbi", $db{'driver'}, $db{'name'});
  if($db{'server'}) {
    $connect .= ";host=".$db{'server'};
    if($db{'port'}) {
      $connect .= ";".$db{'port'};
    }
  }
  $obj->{dbh} = DBI->connect($connect, $db{'user'}, $db{'pass'},
			     { AutoCommit=>1, RaiseError=>0, PrintError=>0 });
  if (! $obj->{dbh}) {
    $obj->error("ExSite::SQL - failed to connect: $DBI::errstr; $!");
  }
  $obj->info("database connection established");

  # store db access parameters for later lookup
  $obj->{db} = {%db};

  return $obj;
}

#------------------------------------------------------------------------

=pod

=head2 C<get_tables()> : return a list of tables in the database

Example: C<my @tables = $db-E<gt>get_tables();>

=cut

#------------------------------------------------------------------------

sub get_tables($) {
  my ($this) = shift;
  my $sth = $this->query("SHOW TABLES");
  return () if ($this->{query_error});
  my @tables = ();
  # return value has ridiculous formatting; extract important info
  foreach my $tab (@{$sth->fetchall_arrayref}) {
    push @tables, $tab->[0];
  }
  return @tables;
}

#------------------------------------------------------------------------

=pod

=head2 C<dbget_columns($table)> : return a list of columns in a table

Example: C<my @columns = $db-E<gt>get_columns("sales_data");>

=cut

#------------------------------------------------------------------------

sub dbget_columns($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::get_columns: invalid table: $table");
      return;
  }
  $this->get_table_info($table);
  return if ($this->{query_error});
  my @columns = ();
  foreach my $col (@{$this->{sql_table_info}->{$table}}) {
    push @columns, $col->{Field};
  }
  return @columns;
}

#------------------------------------------------------------------------

=pod

=head2 C<dbget_key($table)> : return the primary key of a table

Example: C<my $key = $db-E<gt>dbget_key("sales_data");>

=cut

#------------------------------------------------------------------------

sub dbget_key($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::dbget_key: invalid table: $table");
      return;
  }
  $this->get_table_info($table);
  return undef if ($this->{query_error});
  foreach my $col (@{$this->{sql_table_info}->{$table}}) {
    # assume key info is stored in column "Key"
    if ($col->{Key} =~ /^pri/i) {
      # this appears to be a primary key
      return $col->{Field};
    }
  }
  return undef;
}

#------------------------------------------------------------------------

=pod

=head2 C<dbselect($rtable,$rcolumn,$cond,$sort,$limit)> : fetches records from the database

Returns an array of datahashes matching the input criteria:

=over 4

=item $rtable

a table name, or reference to an array of table names

=item $rcolumn

a column name, or reference to an array of column names, 
or references to a hash of tablename => array of column names, or NULL to
select all columns.

=item $cond

a WHERE clause string, or a reference to a match hash

=item $sort

a sort string, or a reference to a list of columns to sort 
the results by

=item $limit

the maximum number of records to fetch

=back

Most of these parameters are optional.

Examples:

C<@data = $db-E<gt>dbselect("telephone");> - fetch all telephone records

C<@data = $db-E<gt>dbselect("telephone",["number","type"]);> - fetch only
the number and type fields of all telephone records

C<@data = $db-E<gt>dbselect("telephone",undef,{type=E<gt>"fax"});> - fetch all 
telephone records whose type field is set to "fax".

C<@data = $db-E<gt>dbselect("telephone",undef,undef,"number"});> - fetch all 
telephone records, sorted by the number field.

C<@data = $db-E<gt>dbselect("telephone",undef,undef,undef,100});> - fetch
the first 100 telephone records

=cut

#------------------------------------------------------------------------

sub dbselect($$;$$$$) {
    my ($this,$rtable,$rcolumn,$cond,$sort,$limit) = @_;

    my ($tlist,$clist,$where,$order,$sql,@result);

    $tlist = $this->sql_table($rtable);
    return undef if (! defined $tlist);

    $clist = $this->sql_column($rcolumn);
    return undef if (! defined $clist);

    $where = $this->sql_condition($cond);
    return undef if (! defined $where);

    if ($sort) {
	$order = $this->sql_sort($sort);
	return undef if (! defined $order);
    }
    if ($limit) {
	$limit = $this->sql_limit($limit);
	return undef if (! defined $limit);
    }

    # execute the query
    my $hstat = $this->run_handler("pre_select",@_);
    $sql = "SELECT $clist FROM $tlist $where $order $limit";
    my $sth = $this->query($sql);
    @result = ();
    my $row;
    while ($row = $sth->fetchrow_hashref) {
      push @result, $row;
    }
    $hstat = $this->run_handler("post_select",@_);
    if (wantarray) { return @result; }
    return \@result;
}

#-------------------------------------------------------------------------

=pod

=head2 C<dbcount($rtable,$condition)> : returns number of matching records

C<$rtable> = table, or reference to array of tables

C<$cond> = explicit conditional, or a reference to a match hash, which
can optionally contain wildcards

Only C<$rtable> is required for the function to work.

=cut

#-------------------------------------------------------------------------

sub dbcount($$;$) {
    my ($this,$rtable,$cond) = @_;

    my $tlist = $this->sql_table($rtable);
    return 0 if (! $tlist);

    my $where = $this->sql_condition($cond);
    #return 0 if (! $where);  # unconditional counting ok

    # execute the query
    my $sql = "SELECT COUNT(*) FROM $tlist $where";
    my $sth = $this->query($sql);
    my @result = ();
    my $row = $sth->fetchrow_arrayref;
    return $row->[0];
}

#------------------------------------------------------------------------

=pod

=head2 C<dbinsert($table,$data)> : inserts a new record into a table.

C<$data> is a reference to a datahash.  This is inserted as a new record
into C<$table>.

Returns the primary key of the new record.

=cut

#------------------------------------------------------------------------

sub dbinsert($$$) {
  my ($this,$table,$data) = @_;
  if (ref $data eq "HASH") {
    # pre-handler?
    my $hstat = $this->run_handler("pre_insert",$table,$data);
    my $clist = join(",", keys %$data);
    if ($clist !~ /^(\w+,)*\w+$/) {
	$this->error("SQL::insert: invalid column list: $clist");
	return 0;
    }
    # values
    my @values = ();
    for my $val (values %$data) {
	if (ref $val) { push @values, $$val; }  # SQL literal - careful!
	else { push @values, $this->{dbh}->quote($val); } # normal data
    }
    my $vlist = join(",", @values);

    # if primary key is defined, use "REPLACE" instead of "INSERT"
    my $pkey = $this->dbget_key($table);
    my $operation;
    if ($data->{$pkey}) {
      $operation = "REPLACE";
    }
    else {
      $operation = "INSERT";
    }

    my $sql = "$operation INTO $table ($clist) VALUES ($vlist)";
    
    # Lock
    $this->query("LOCK TABLES $table WRITE");
    # Insert
    #print "Content-type: text/plain\n\n$sql\n";
    #my $new_id = 667;
    $this->query($sql);
    my $new_id = $this->{dbh}->{mysql_insertid};
    $this->{last_insert_id} = $new_id;
    # Unlock
    $this->query("UNLOCK TABLES");
    # post-handler?
    $hstat = $this->run_handler("post_insert",$table,$data);
    return $new_id;
  }
  else {
    $this->error("ExSite::SQL::insert: bad data reference");
    return 0;
  }
}


#------------------------------------------------------------------------

=pod

=head2 C<dbupdate($table,$data,$cond)> : modifies records in a table.

C<$table> is the table that the update operation will be performed on.

C<$data> is a reference to a datahash, containing the fields and the values
they should be changed to.

C<$cond> is an optional match hash.  Only records matching this will
be updated.  

If C<$cond> is undefined, then the primary key of the record to be modified
is taken from C<$data>.  If the primary key is not defined in C<$cond>,
this will throw an exception.

Returns an error messge if there were problems.

Examples:

C<$db-E<gt>dbupdate("member",{password=E<gt>"",member_id=E<gt>100});> -
sets an empty password for member #100 (assuming that member_id is
the primary key of the member table).

C<$db-E<gt>dbupdate("member",{password=E<gt>""},{status=E<gt>"expired"});> -
clears the password field for all members whose status field is set to
"expired".

=cut

#------------------------------------------------------------------------

sub dbupdate($$$;$) {
  my ($this,$table,$data,$cond) = @_;
  # pre-handler?
  my $hstat = $this->run_handler("pre_update",$table,$data,$cond);

  my ($set,$where);
  # data
  if (ref $data eq "HASH") {
    my @set = ();
    while (my ($key,$val) = each %$data) {
      if ($key !~ /^\w+$/) {
	my $error = "SQL::update: invalid column: $key";
	$this->error($error);
	return $error;
      }
      my $qval;
      if (ref $val) { $qval = $$val; }  # SQL literal - careful!
      else { $qval = $this->{dbh}->quote($val); } # normal data
      push @set, "$key=$qval";
    }
    $set = join(",",@set);
  }
  else {
    my $error  = "ExSite::SQL::update: bad data reference";
    $this->error($error);
    return $error;
  }

  # condition
  $where = $this->sql_condition($cond);
  if (! $where) {
      if (! defined $where) {
	  return "ExSite::SQL::update: invalid where clause";
      }
      else {
	  # no condition!  Select rows matching the primary key.
	  my $pkey = $this->dbget_key($table);
	  #if (exists $data->{$pkey}) {
	  if ($data->{$pkey}) {
	      $where = "WHERE $pkey=".$this->{dbh}->quote($data->{$pkey});
	  }
	  else {
	      # no primary key!  Select all rows (!!!)
	      #$where = "";
	      #$this->warn("ExSite::SQL::update: warning! unconditional update!");
	      # if you'd rather this be an error, replace with this:
	      my $error = "ExSite::SQL::update: trapped unconditional update";
	      $this->error($error);
	      return $error;
	  }
      }
  }

  my $sql = "UPDATE $table SET $set $where";

  # Lock
  $this->query("LOCK TABLES $table WRITE");
  # update
  $this->query($sql);
  my $stat = $this->{query_error};
  # Unlock
  $this->query("UNLOCK TABLES");
  # post-handler?
  $hstat = $this->run_handler("post_update",$table,$data,$cond);
  return $stat;
}

#------------------------------------------------------------------------

=pod

=head2 C<dbdelete($table,$cond)> : deletes records from a table.

Deletes records in C<$table> matching the match hash referenced by C<$cond>.
C<$cond> can optionally name an explicit SQL WHERE condition, but this 
is not portable to other drivers.

Examples:

C<$db-E<gt>dbdelete("member",{status=E<gt>"expired"});> -
removes member records whose status field is set to "expired".

=cut

#------------------------------------------------------------------------

sub dbdelete($$$) {
  my ($this,$table,$cond) = @_;
  # pre-handler?
  my $stat = $this->run_handler("pre_delete",$table,$cond);

  my $where;
  # condition
  if (ref $cond eq "HASH") {
    # condition is a match hash
    $where = $this->where_match($cond);
    if (! $where) {
	my $error = "SQL::delete: invalid delete condition";
	$this->error($error);
	return $error;
    }
  }
  elsif ($cond) {
    # condition is an explicit WHERE clause
    ##### no sanity check!
    $where = "WHERE $cond";
  }
  else {
    #$this->warn("ExSite::SQL::update: warning! unconditional delete!");
    my $error = "ExSite::SQL::delete: trapped unconditional delete";
    $this->error($error);
    return $error;
  }

  my $sql = "DELETE FROM $table $where";

  # Lock
  $this->query("LOCK TABLES $table WRITE");
  # Insert
  $this->query($sql);
  $stat = $this->{query_error};
  # Unlock
  $this->query("UNLOCK TABLES");
  # post-handler?
  $stat = $this->run_handler("post_delete",$table,$cond);
  return $stat;
}

#------------------------------------------------------------------------

=pod

=head1 Internal and non-standard methods

The section describes ExSite::SQL methods that do not conform to
the ExSite database driver API.  These methods are for internal use 
only, or provide non-portable extensions to the API.

=head2 C<query($sql_statement)> : generic query method

This method prepares and executes an arbitrary SQL query, and
returns the statement handle.  It does not fetch the results.

Status and errors are stored in the object for later reporting.  

Example: C<$db-E<gt>query("SELECT * FROM SALESDATA");>

=cut

#------------------------------------------------------------------------

sub query($$) {
  my ($this, $sql) = @_;
  my ($sth);
  # clear the error status
  $this->{query_error} = undef;
  # prepare the statement
  $sth = $this->{dbh}->prepare($sql);
  $this->info($sql);
  if (! $sth) {
    # failure
    $this->{query_error} = "ExSite::SQL::query - failed to prepare \"$sql\"; $DBI::errstr; $!";
    $this->error($this->{query_error});
  }
  # execute the statement
  if (! $sth->execute) {
    # failure
    $this->{query_error} = "ExSite::SQL::query - failed to execute \"$sql\"; $DBI::errstr; $!";
    $this->error($this->{query_error});
  }
  return $sth;
  # calling routine is responsible for getting the results, if any
}

#------------------------------------------------------------------------

=pod

=head2 C<where_match($match_hash)> : constructs a WHERE clause

Returns an SQL WHERE clause built from the passed match hash.
Is sensitive to wildcards in the match hash, and modifies the
WHERE clause (to use "LIKE") appropriately.
Attempts to perform some basic sanity/security checks on the
column names.

This is an internal routine, not normally called from outside.

Example: C<my $where = $db-E<gt>where_match(quantity=E<gt>9, size=E<gt>"L");>

=cut

#------------------------------------------------------------------------

sub where_match ($$) {
  my ($this,$match) = @_;
  my @wlist = ();
  my $where;
  while (my ($key, $val) = each(%$match)) {
    # sanity check on keys
    if ($key !~ /^[\w\.]+$/) {
	$this->error("SQL::where_match: invalid column: $key");
	return undef;
    }	
    # default operator
    my $op = "=";
    # if wildcards are present, use "LIKE"
    if($val =~ /%/) {
      if($val =~ /^%|[^\\]%/) {
	# unescaped wildcard
	$op = "LIKE";
	# quote while protecting "\%"
	# (quote() escapes the backslash otherwise)
	my $ph = "~_jwosja_~"; # a highly improbable string unaffected by quote()
	$val =~ s/\\%/$ph/g;
	$val = $this->{dbh}->quote($val);
	$val =~ s/$ph/\\%/g;
      }
      else {
	# strip backslashes before quoting; "%" is not a meta-char for "="
	$val =~ s/\\%/%/g;
	$val = $this->{dbh}->quote($val);
      }
    }
    elsif (ref $val) {
      # if val is a reference, it points to a SQL literal
      # WARNING! no sanity check
      $val = $$val;
    }
    else {
      $val = $this->{dbh}->quote($val);
    }
    # build the where
    push(@wlist, "$key $op $val");
  }
  if (scalar @wlist) {
    $where = "WHERE ".join(" && ", @wlist);
  }
  return $where;
}

#------------------------------------------------------------------------

=pod

=head2 C<get_table_info($table)> : fetches and stores table attributes

This is an internal utility routine that is used by C<get_key> and C<get_columns>.

=cut

#------------------------------------------------------------------------

sub get_table_info($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::get_columns: invalid table: $table");
      return;
  }
  if (! $this->{sql_table_info}->{$table}) {
    my $sth = $this->query("SHOW COLUMNS FROM $table");
    return if ($this->{query_error});
    $this->{sql_table_info}->{$table} = [];
    while (my $row = $sth->fetchrow_hashref) {
      push @{$this->{sql_table_info}->{$table}},$row;
    }
  }
}

#------------------------------------------------------------------------

=pod

=head2 C<custom_query($sql_statement)> : perform an explicit SQL query.

This method prepares and executes an arbitrary SQL query.  Unlike the
lower-level C<query()> method, this method fetches the results and
returns them in an array of datahashes.  This makes it a more user-friendly
way to execute SQL statements that are more complex than the standard
API allows for.

WARNING: no sanity checking or taint checking is performed.  The
caller is responsible for validating their own SQL statements.

This method is not portable to other ExSite database drivers.

=cut

#------------------------------------------------------------------------

sub custom_query($$){
  my ($this,$string)=@_;
  my ($sth, $rowvP, @rowvPs);
  $sth = $this->query($string);
  # log the query
  $this->info($string);
  ### return the data as a reference to an array of hash references
  while($rowvP = $sth->fetchrow_hashref) { push @rowvPs, {%$rowvP};}
  $sth->finish;
  if (wantarray) { return @rowvPs; }
  return \@rowvPs;
}

#-------------------------------------------------------------------------
# convert perl args to SQL snippets

sub sql_table {
    my ($this,$rtab) = @_;
    my $tlist;
    # table(s) to select from
    if (ref $rtab eq "ARRAY") {
	# array of tables was passed
	$tlist = join(",",@$rtab);
	if ($tlist !~ /^(\w+,)*\w+$/) {
	    # validate 
	    $this->error("SQL::tlist: invalid table list: $tlist");
	    return undef;
	}
    }
    else {
	# single table was passed as a scalar
	$tlist = $rtab;
	if ($tlist !~ /^\w+$/) {
	    # validate 
	    $this->error("SQL::tlist: invalid table: $tlist");
	    return undef;
	}
    }
    return $tlist;
}

sub sql_column {
    my ($this,$rcolumn) = @_;
    my $clist;
    # column(s) to select
    if (ref $rcolumn eq "ARRAY") {
	# array of columns was passed
	$clist = join(",",@$rcolumn);
	if ($clist !~ /^([\w\.]+,)*[\w\.]+$/) {
	    $this->error("SQL::select: invalid column list: $clist");
	    return undef;
	}
    }
    elsif (ref $rcolumn eq "HASH") {
	# hash of table => reference_to_column_list was passed
	my @joinlist = ();
	while (my ($table,$list) = each %$rcolumn) {
	    if ($table !~ /^\w+$/) {
		$this->error("SQL::select: invalid table: $table");
		return undef;
	    }
	    foreach my $col (@$list) {
		if ($col !~ /^\w+$/) {
		    $this->error("SQL::select: invalid column: $col");
		    return undef;
		}	
		push @joinlist,"$table.$col";
	    }
	}
	$clist = join(",",@joinlist);
    }
    elsif ($rcolumn) {
	# single column or explicit function was passed as a scalar
	if ($rcolumn !~ /^\w+$/) {
	    ####### this regexp will not accept functions!
	    $this->error("SQL::select: invalid column: $rcolumn");
	    return undef;
	}
	$clist = $rcolumn;
    }
    else {
	# no columns passed, select all columns
	$clist = "*";
    }
    return $clist;
}

sub sql_condition {
    my ($this,$cond) = @_;
    my $where = "";
    # condition to select on
    if (ref $cond eq "HASH") {
	# condition was passed as a match hash -
	# the selected records must match all elements of the hash.
	$where = $this->where_match($cond);
	if (! $where) {
	    $this->error("SQL::select: invalid where clause");
	    return undef;
	}	
    }
    elsif ($cond) {
	# assume condition is a scalar explicitly stating the condition.
	# WARNING: use of this feature is potentially non-portable to other
	# ExSite DB drivers.
	# Also, no sanity check on $cond, so be careful about using tainted data
	$where = "WHERE $cond";
    }
    return $where;
}

sub sql_sort {
    my ($this,$sort) = @_;
    my $order = "";
    # sort, if asked
    if ($sort) {
	my @ords = ();
	if (ref $sort eq "ARRAY") {
	    # sort is a list of columns to sort in default order
	    foreach my $col (@$sort) {
		if ($col =~ /^\w+$/) {
		    push @ords, $col;
		}
		elsif (ref($col) eq "HASH") {
		    while (my($key,$val) = each(%$col)) {
			if ($key !~ /^\w+$/ || $val !~ /^(asc|desc)$/i) {
			    $this->error("SQL::select: invalid sort column: $col");
			    return undef;
			}
			push @ords, "$key $val";
		    }
		}
		else{
		    $this->error("SQL::select: invalid sort column: $col");
		    return undef;
		}
	    }
	    $order = join(",", @ords);
	}
	elsif (ref $sort eq "HASH") {
	    # sort is a hash of column => direction

	    # WARNING - hash keys have an undefined order, so this may not
	    # sort well if more than one key is given

	    while (my ($col, $dir) = each(%$sort)) {
		if ($col !~ /^\w+$/ || $dir !~ /^(asc|desc)$/i) {
		    $this->error("SQL::select: invalid sort directive: $col $dir");
		    return undef;
		}
		push @ords, "$col $dir";
	    }
	    $order = join(",", @ords);
	}
	else {
	    # explicit sort list
	    # WARNING: may not be portable to other ExSite DB drivers.
	    if ($sort !~ /^(\w+( (asc|desc))?,\s*)*(\w+( (asc|desc))?\s*)$/i) {
		$this->error("SQL::select: invalid sort directive: $sort");
		return undef;
	    }
	    $order = $sort;
	}
	$order = "ORDER BY $order";
    }
    return $order;
}

sub sql_limit {	
    my ($this,$limit) = @_;
    if (! $limit) { $limit = ""; }
    # limit, if asked
    if ($limit) {
	if ($limit !~ /^(\d+,\s*)?\d+$/) {
	    $this->error("SQL::select: invalid limit directive: $limit");
	    return undef;
	}
	$limit = "LIMIT $limit";
    }
    return $limit;
}



1;
