#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Section;

# new, setup, setupSection (preload index page)
# filename - name of directory
# httppath - htdocs path to section
# diskpath - full disk path
# get_url - full web path
# publish - write out all files

########### needed:
# get_webmasters
# is_webmaster
# section_path
# get_pages ! this section only
# get_pages_from_path ! all sections in path

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Page;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Page);

# new() constructor is inherited from ContentData class.

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupSection(@_);
}

sub setupSection { 
    my ($this,%option) = @_;

    my $section;
    if ($option{id}) {
	# lookup section record
	$section = $this->fetch("section",$option{id});
    }
    elsif (ref $option{section} eq "HASH") {
	# section record was passed
	if (ref $option{section} eq "HASH") {
	    $section = $option{section};
	}
	else {
	    $this->error("invalid section parameters passed to setupSection");
	}
    }
    if ($section) {
	$this->info("setting up section $section->{section_id}");
	$this->{data} = $section;
	$this->set_context("none");
    }
    elsif ($option{id}) {
	$this->error("section $option{id} not found");
    }
    else {
	$this->warn("setting up null section");
    }
}

# previous - change the section object to next one up in path

sub previous {
    my $this = shift;
    my $section = $this->get;
    if ($section->{parent_id}) {
	$this->setupSection(id=>$section->{parent_id});
	return $section->{parent_id};
    }
    else {
	return undef;
    }
}

# name of section directory - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	if ($section->{type} =~ /native/) {
	    return $section->{publish_to};
	}
    }
    # filename is not defined for other website types
    return undef;
}

# disk path to section

sub diskpath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	# no disk path for templates and remote sites
	return undef if ($section->{type} =~ /template|remote/);
	if ($section->{type} =~ /stand-alone/) {
	    # stand-alone sites hold full path
	    return $section->{publish_to};
	}
	else {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id});
		my $parent_path = $psec->diskpath;
		# if parent is undefined, that's because we are nested 
		# under a non-local website
		return undef if ! defined $parent_path;
		return "$parent_path/$section->{publish_to}";
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}/$section->{publish_to}";
	    }
	    else {
		# root section of the native site
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}";
	    }
	}
    }
    return undef;
}

# web path to section   ######### DEPRECATED ###########

sub httppath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	# can only give a partially-qualified URL for native sites
	if ($section->{type} =~ /native/) {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id},
					       context=>"none");
		my $parent_path = $psec->httppath;
		# if parent is undefined, that's because we are nested 
		# under a non-native website
		return undef if ! defined $parent_path;
		return "$parent_path/$section->{publish_to}";
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLpath}/$section->{publish_to}";
	    }
	    else {
		# root section of the native site - return base HTML path
		return $config{server}{HTMLpath};
	    }
	}
	elsif ($section->{type} !~ /template/) {
	    return $section->{url};
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the section

sub get_url {
    my $this = shift;
    my $url;
    my $section = $this->get;
    return undef if ($section->{type} =~ /template/);
    while ($section) {
	if ($section->{type} =~ /native/) {
	    if ($section->{publish_to}) {
		$url = "/".$section->{publish_to}.$url;
	    }
	}
	else {
	    $url = $section->{url}.$url;
	    last;
	}
	if ($section->{parent_id}) {
	    $section = $this->fetch("section",$section->{parent_id});
	}
	else { 
	    $section = undef;
	}
    }
    if (! $url || $url =~ /^\/\w/) {
	#$url = $config{server}{HTMLpath}.$url;
	# DANGER: admin server may not be the same as public server
	$url = $config{server}{server}.$config{server}{HTMLpath}.$url;
    }
    elsif ($url =~ /^\w+\./) {
	$url = "http://$url";
    }
    return $url;
}

# get_host : return only the machine name part of the section URL

sub get_host {
    my $this = shift;
    my $url = $this->get_url;
    $url =~ /^https?:\/\/([^\/]*)/;
    return $1;
}

# sectionpath : return an array of the sites/sections traversed to get to 
#               this site

sub sectionpath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	my @parent_path;
	if ($section->{parent_id}) {
	    my $psec = new ExSite::Section(id=>$section->{parent_id});
	    @parent_path = $psec->sectionpath;
	}
	return (@parent_path,$section);
    }
}

# fetch_libraries : fetches all libraries in the site path.

sub fetch_libraries {
    my $this = shift;
    my @sections = $this->sectionpath;
    my @Library;
    my $first = 1;
    while (my $s = pop @sections) {
	my @libs = $this->fetch_match("page",{type=>"library",section_id=>$s->{section_id}});
	foreach my $lib (@libs) {
	    # only the last one in the path is accessible by administrators
	    next if ($lib->{access} eq "administrators" && ! $first);
	    push @Library,$lib;
	    $first = 0;
	}
    }
    return @Library;
}

# fetch_library_content : grabs all content objects in all libraries in the
#                         site path.  This could be useful for inserting 
#                         shared content into pages, or it could be a pig.
#
# mime_re is a regexp to filter fetched content with, eg. to select 
# all pdf files, or all jpgs, or all non-text items.

sub fetch_library_content {
    my ($this,$mime_re) = @_;
    my @sections = $this->sectionpath;
    my @libs = $this->fetch_libraries;
    my @Content;
    foreach my $l (@libs) {
	my @cont = $this->fetch_match("content",{page_id=>$l->{page_id}});
	foreach my $c (@cont) {
	    my $cobj = new ExSite::Content(content=>$c);
	    if (! $mime_re || $cobj->content_type() =~ /$mime_re/) {
		push @Content, $cobj;
	    }
	}
    }
    return @Content;
}

# get all pages for the section - returns array of Page objects
#
# %opt is a match hash for pages

sub get_all_pages {
    my ($this,%opt) = @_;
    if (! $opt{section_id}) {
	$opt{section_id} = $this->id;
    }
    my @pages = $this->fetch_match("page",\%opt);
    my @Page;
    foreach my $page (@pages) {
	push @Page, new ExSite::Page(page=>$page);
    }
    return @Page;
}

# get all pages for the section, organized into a map (ExSite::Tree)
#
# options are:
# version => alternate version/language
# hidden => true if the map should include hidden pages

sub get_mapped_pages {
    my ($this,$parent_id,$opt) = @_;
    my $match = {section_id=>$this->id(),
		 status=>"active",
		 type=>"page",
		 };
    if (! $opt->{hidden}) {
	$match->{visibility} = "visible";
    }
    my @pages = $this->fetch_match("page",$match,["rank","page_id"]);

    # make a basic Tree/map from the regular pages
    require ExSite::Tree;
    my $map = new ExSite::Tree("page_id","parent_id",@pages);

    # do we need to consider alterate pages?
    my $version = $opt->{version};
    if (! $version && $parent_id) {
	# no version explicitly specified, but we should check the version
	# of the parent page, just to be sure

	# see if the parent is already loaded in the map
	my $p = $map->getnode_data($parent_id);
	# if not, look it up in the DB
	if (!$p) { $p = $this->fetch("page",$parent_id); }
	if ($p->{version}) { $version = $p->{version}; }
    }
    
    # overlay the alternate pages
    if ($version) {
	$match->{type} = "alternate";
	@pages = $this->fetch_match("page",$match,["rank","page_id"]);
	foreach my $alt (@pages) {
	    # the alternate's parent points to the page it replaces
	    $map->replacenode($alt->{parent_id},$alt);
	}
    }
    return $map;
}

# old map builder - slow (too many DB calls)

sub get_mapped_pages_old {
    my ($this,$parent_id,$opt) = @_;
    if (! $parent_id) { $parent_id = 0; }
    my @map;
    my $match = {section_id=>$this->id(),
		 parent_id=>$parent_id,
		 status=>"active",
		 type=>"page",
		 };
    if (! $opt->{hidden}) {
	$match->{visibility} = "visible";
    }
    if (! $opt->{version} && $parent_id) {
	# check if we're already on an alternate version
	my $p = $this->fetch("page",$parent_id);
	if ($p->{version}) { $opt->{version} = $p->{version}; }
    }
    my @pages = $this->fetch_match("page",$match,"rank");
    foreach my $page (@pages) {
	my $pid = $page->{page_id};
	if ($opt->{version}) {
	    # alternate version requested
	    my @alt = $this->fetch_match("page",{parent_id=>$pid,
						 type=>"alternate",
						 status=>"active",
						 version=>$opt->{version},
						 },"page_id");
	    if (@alt > 0) {
		# alternate version of this page was found
		# swap it in place
		$page = $alt[-1];
	    }
	}
	my $p = new ExSite::Page(page=>$page);
	my @submap = $this->get_mapped_pages($pid,$opt);
	push @map, $p, \@submap;
    }
    return @map;
}

# get all templates available to the section
# - includes public and semi-private templates from other sections
#
# opt:
#   section_id - the section to search under, if not the current section
#   mode - the template selection method:
#      -1 => select all templates on the system
#           (administrators only, otherwise treated as 0)
#      0 => select templates useable by this site (default)
#           (use when changing template of a site)            
#      1 => select templates useable by children of this site
#           (use when creating a child site)
#      2 => select public templates only
#           (use when creating a new independent site)

sub get_templates {
    my ($this,%opt) = @_;
    my ($sid,$db,@templates);
    $sid = $opt{section_id} || $this->id;
    $db = $share{DB};

    if ($opt{mode} < 0 && $db->is_admin()) {
	# fetch all templates in the system
	@templates = $db->fetch_match("page",{type=>"template",
					      status=>"active"});
    }
    else {
	# fetch public templates
	@templates = $db->fetch_match("page",{type=>"template",
					      access=>"public",
					      status=>"active"});
	
	if ($opt{mode} < 2) {
	    # ... plus the member-only templates from the current section
	    push @templates, $db->fetch_match("page",{type=>"template",
						      access=>"members",
						      status=>"active",
						      section_id=>$sid});
	    if (! $opt{mode}) {
		my $sec = $db->fetch("section",$sid);
		if ($sec && $sec->{parent_id}) {
		    # ... plus the member-only templates from the parent section
		    push @templates, $db->fetch_match("page",{type=>"template",
							      access=>"members",
							      status=>"active",
							      section_id=>$sec->{parent_id}});
		}
		# ...plus the private templates from this section
		push @templates, $db->fetch_match("page",{type=>"template",
							  access=>"administrators",
							  status=>"active",
							  section_id=>$sid});
	    }
	}
    }
    return @templates;
}

# change template for all pages

sub set_template {
    my ($this,$template) = @_;
    my $tid = ref $template eq "HASH" ? $template->{page_id} : $template;
    my $sid = $this->id();

    # sanity checks

    if (! $tid) {
	return $this->error("set_template: no template given");
    }
    if (! $sid) {
	return $this->error("set_template: no section given");
    }
    if (! $this->owns()) {
	return $this->error("set_template: you do not have sufficient priveleges to modify this website");
    }
    my @allowed_templates = $this->get_templates(0);
    my $t;
    foreach $t (@allowed_templates) {
	if ($t->{page_id} == $tid) {
	    my $db = $share{DB};
	    # update site
	    $db->update("section",{template_id=>$tid},{section_id=>$sid});
	    # update pages
	    $db->update("page",{template_id=>$tid},{section_id=>$sid,type=>"page"});
	    $db->update("page",{template_id=>$tid},{section_id=>$sid,type=>"alternate"});
	    return;
	}
    }
    return $this->error("set_template: you do not have permission to use this template");

}

# get all bodies available to the section (and its parent sections)

sub get_bodies {
    my ($this) = @_;
    my ($sid,$db,@liblist,@libs,@body_templates,$sql);
    $sid = $this->id;
    $db = $share{DB};

    # find all the libraries that we can take bodies from

    @libs = $db->fetch_match("page",{type=>"library",section_id=>$sid});
    my $section = $this->get("section");
    while ($section->{parent_id}) {
	$sid = $section->{section_id};
	push @libs, $db->fetch_match("page",{type=>"library",section_id=>$sid});
	$section = $this->fetch("section",$sid);
    }

    # get all the bodies in our list of libraries
    foreach my $l (@libs) { 
	push @liblist, $l->{page_id};
    }
    # warning: not portable to non-SQL DB drivers
    $sql = "select * from content where name='body' and page_id in (".join(',',@liblist).")";
    @body_templates = $db->custom_query($sql);

    return wantarray ? @body_templates : \@body_templates;
}

# write the section to disk

sub publish {
    my ($this,$opt) = @_;
    my $out;
    my $section = $this->get;
    if (! $section) { 
	$this->error("Section::publish: no section to publish");
	print "No section to publish!";
	return;
    }
    elsif (! $section->{type} =~ /template/) { 
	$this->error("Section::publish: cannot publish website templates");
	print "Cannot publish website templates!";
	return;
    }
    else { # if native website
	my $diskpath = $this->diskpath;
	if (! $diskpath) { 
	    $this->error("Section::publish: non-local publishing not supported");
	    print "Non-local publishing not supported yet.";
	    return;
	}
	$this->info("Section::publish: publishing section $section->{section_id}");
	# 1) create the section directories
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    print $this->show_diagnostics("error","html");
	    return;
	}
	print $mkdir;
	return if ($opt eq "nofiles");
	# 2) publish all pages 
	# autoflush stdout so we can monitor the progress...
	my $old_flush = $|;
	$| = 1;
	# first, libraries
	my @Page = $this->get_all_pages(type=>"library",
					status=>"active"); ### access=>"public");
	foreach my $p (@Page) {
	    print $p->publish;
	}
	# next, templates
	my @Page = $this->get_all_pages(type=>"template",
					status=>"active"); ### access=>"public");
	foreach my $p (@Page) {
	    print $p->publish;
	}
	# last, pages
	my @Page = $this->get_all_pages(type=>"page",
					status=>"active",
					access=>"public",
					publish_method=>"static");
	my $found_index;
	foreach my $p (@Page) {
	    $share{Page} = $p;
	    print $p->publish;
	    my $page = $p->get;
	    if ($page->{filename} =~ /^(default|index).(php|html?)$/) {
		$found_index = $p;
	    }
	}
	if (! $found_index) {
	    $this->warn("Section::publish: no index page for section $section->{section_id}");
	    print "<p class=error>WARNING: no index page was found for this website.  
In most cases, it is recommended that you define an index page to serve 
as the front page of your site.  To set a particular page as the index page, 
edit that page's details, and set its filename to \"index.html\" 
(or \"index.php\" if using PHP pages).</p>\n";
	}
	# reset flush settings
	$| = $old_flush;
    }
    return;
}

# remove the published section files from disk

sub unpublish {
    my ($this) = @_;
    my $out;
    my $section = $this->get;
    if (! $section) { 
	print $this->error("Section::unpublish: no section to remove");
	return;
    }
    elsif (! $section->{type} =~ /template/) { 
	print $this->error("Section::unpublish: cannot unpublish website templates");
	return;
    }
    else { # if native or standalone website
	my $diskpath = $this->diskpath;
	if (! $diskpath) { 
	    print $this->error("Section::unpublish: non-local unpublishing not supported");
	    return;
	}
	elsif (-e $diskpath) {
	    $this->info("Section::unpublish: removing files for section $section->{section_id}");

	    # autoflush stdout so we can monitor the progress...
	    my $old_flush = $|;
	    $| = 1;

	    # unpublish any child sections
	    ########### FIXME: should we really do this?  
	    ########### Could we unpublish a section's files, but want
	    ########### to leave child sections intact?
	    my @csections = $share{DB}->fetch_match("section",{parent_id=>$this->id,
							       type=>"native website"});
	    foreach my $cs (@csections) {
		my $s = new ExSite::Section(section=>$cs);
		$s->unpublish;
		if ($s->errorcheck()) {
		    $this->error($s->fetch_diagnostics("error"));
		}
	    }
	
	    # remove libraries, templates, pages
	    my @Page = $this->get_all_pages();
	    foreach my $p (@Page) {
		print $p->unpublish;
		if ($p->errorcheck()) {
		    $this->error($p->fetch_diagnostics("error"));
		}
	    }

	    # remove library, template container dirs
	    my $rmdir =  $this->rmdir("$diskpath/_Template");
	    $rmdir .= $this->rmdir("$diskpath/_Library");

	    # remove the section directory if not a top-level site
	    if ($section->{type} eq "native website" &&
		$section->{publish_to}) {
		$rmdir .= $this->rmdir($diskpath);
		if (! $rmdir) {
		    print $this->show_diagnostics("error","html");
		    return;
		}
	    }
	    print $rmdir;

	    # reset flush settings
	    $| = $old_flush;
	}
    }
    return;
}

# sitemap tool
#
#

sub map {
    my ($this,$options,$pageoptions) = @_;
    my $out;
    my $db = $share{DB};
    my %match = (section_id=>$this->id,
		 type=>"page",
		 status=>"active",
		 parent_id=>0,);
    if (! $options->{show_hidden}) {
	$match{visibility} = "visible";
    }
    my @pages = $db->fetch_match("page",\%match,"rank");
    $out .= "<P><UL CLASS=\"sitemap\">\n";
    foreach my $page (@pages) {
	$out .= $this->map_r($page,$options,$pageoptions);
    }
    $out .= "</UL></P>\n";
    if (! $options->{show_hidden} && $options->{footnote}) {
	$out .= "<P>NOTE: hidden pages do not appear in the site map.</P>\n";
    }
    return $out;
}

sub map_r {
    my ($this,$page,$options,$pageoptions) = @_;
    # recursive mapping routine
    my $db = $share{DB};
    my $pid = $page->{page_id};
    my $url;

    if ($options->{link_to} eq "cms") {
	$url = &relink(path=>"/$page->{section_id}/$pid",cmscmd=>undef);
    }
    elsif ($options->{link_to} eq "dynamic") {
	my $query;
	if (ref $pageoptions eq "HASH") {
	    $pageoptions->{_id} = $pid;
	    $query = &EncodeHash(%$pageoptions);
	}
	else {
	    $query = "_id=$pid";
	}
	$url = "$config{server}{CGIpath}/page.cgi?$query";
    }
    else {
	my $p = new ExSite::Page(page=>$page);
	$url = $p->get_url;
    }
    
    my $out = "<LI><A HREF=$url";
    if ($options->{target}) {
	$out .= " target=\"$options->{target}\"";
    }
    my $label = $page->{label} || "(no label)";
    $out .= ">$label</A>";
    my @attr;
    if ($options->{show_filename}) { 
	$out .= " ($page->{filename})"; 
    }
    if ($page->{access} ne "public") { 
	push @attr, "RESTRICTED ACCESS"; 
    }
    if ($page->{visibility} eq "hidden") { 
	push @attr, "HIDDEN";
    }
    if (@attr > 0) {
	$out .= " [".join(", ",@attr)."]";
    }
    $out .= "</LI>\n";
    my $subpage = $db->fetch_match("page",{parent_id=>$pid,
					   type=>"page",
					   visibility=>"visible"},
				   "rank");
    if (@$subpage > 0) {
	$out .= "<UL CLASS=\"sitemap\">\n";
	foreach my $page (@$subpage) { 
	    $out .= $this->map_r($page,$options,$pageoptions); 
	}
	$out .= "</UL>\n";
    }
    return $out;
}

# validate - validate a section record

sub validate {
    my ($this,$data) = @_;
    my ($section,$type,@err,@dupe);
    if (! $data) { $data = $this->get("section"); }

    # generic validation (using DBmap validation rules)
    $share{DB}->{action} = "insert";  # ignore missing primary keys
    push @err, $share{DB}->validate_record("section",$data);

    # special validation
    $type = $data->{type};
    if ($type =~ /^native website$/) {
	@dupe = $share{DB}->fetch_match("section",{type=>$type,
						   parent_id=>$data->{parent_id},
						   publish_to=>$data->{publish_to}});
	if ($data->{publish_to} =~ /\//) {
	    push @err, "invalid publication folder for native website";
	}
    }
    elsif ($type =~ /^stand-alone website$/) {
	@dupe = $share{DB}->fetch_match("section",{type=>$type,
						   publish_to=>$data->{publish_to}});
	if (! $data->{publish_to}) {
	    push @err, "stand-alone website has no publication directory";
	}
	elsif ($data->{publish_to} !~ /^\/.+/) {
	    push @err, "stand-alone website requires a complete publication path";
	}
	if (! $data->{url}) {
	    push @err, "stand-alone website is missing its URL";
	}
    }

    if (@dupe > 0) {
	# check for dupes
	if (@dupe > 1) {
	    push @err, "a section with that publication location already exists";
	}
	elsif (@dupe == 1) {
	    if ($data->{section_id} != $dupe[0]{section_id}) {
		push @err, "a section with that publication location already exists";
	    }
	}
    }

    return @err;
}

# make_section : insert a new section record
#
# can optionally pass website template ID to configure it with

sub make {
    return &make_section(@_);
}

sub make_section {
    my ($this,$section,$template) = @_;
    my @err;
    if ($section->{section_id}) {
	push @err, "invalid attempt to insert an already-defined section record";
    }
    if (! $this->allow_insert("section",$section)) {
	# permission denied
	push @err, "make_section: permission denied";
    }
    push @err, $this->validate($section);
    
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    my $type = $section->{type};
    my $db = $share{DB};

    # remove any residual timestamps
    delete $section->{mtime};
    delete $section->{ctime};
    $db->run_handler("pre_make_section",$section,$template);
    my $sid = $db->insert("section",$section);
    if (! $sid) {
	$this->error("failed to install new site \'$section->{filename}\'");
    }
    else {
	$this->info("installed new page $section->{filename}");
	# add current user as an administrator
	$db->insert("member_site_link",{member_id=>$db->my_uid,
					section_id=>$sid,
					priveleges=>"editor; designer; administrator"});
	my $pid;
	if ($template) {
	    my @pages = $db->fetch_match("page",{section_id=>$template});
	    foreach my $page (@pages) {
		my %page = %$page;
		my %new = %page;
		delete $new{page_id};  # force an insert
		delete $new{ctime};
		delete $new{mtime};
		$new{template_id} = $section->{template_id};
		$new{section_id} = $sid;
		my $pid = $db->insert("page",\%new);
		if (! $pid) {
		    push @err, "failed to add page \"$page->{title}\"";
		}
		else {
		    # make copies of all editorial content objects
		    # my @content = $db->fetch_match("content",{page_id=>$page->{page_id},type=>"editorial"});
		    # make copies of all content objects
		    my @content = $db->fetch_match("content",{page_id=>$page->{page_id}});
		    foreach my $cont (@content) {
			my $oldcid = $cont->{content_id};
			delete $cont->{content_id};  # force an insert
			$cont->{page_id} = $pid;
			my $cid = $db->insert("content",$cont);
			if (! $cid) {
			    push @err, "failed to add \"$cont->{name}\" content to page \"$page->{title}\"";
			}
			else {
			    # copy latest version
			    my @cdata = $db->fetch_match("content_data",
							 {content_id=>$oldcid},
							 "content_data_id");
			    if (@cdata > 0) {
				delete $cdata[-1]->{content_data_id};
				$cdata[-1]->{content_id} = $cid;
				my $cdid = $db->insert("content_data",$cdata[-1]);
				if (! $cdid) {
				    push @err, "failed to initialize \"$cont->{name}\" in page \"$page->{title}\"";
				}
			    }
			}
		    }
		}
	    }
	}
    }
    $db->run_handler("post_make_section",$section,$template);
    return $sid;
}

# delete the section

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("section",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("section",$id);
	return 1;
    }
    return 0;
}

1;


