#------------------------------------------------------------------------
# ExSite::Session
#----------------------------------------------------------------------------
#
#   Copyright 2001-2005 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# An ExSite Session Manager
#
# A session is a hash of keys/values that is persistent between web
# site visits.  The session key is the address of the session, and is a
# hard-to-guess MD5 hash of originating IP, and the current 
# time.  Old session keys are cleaned up by a garbage collector.
#
# Sessions are stored in a DBM table for performance.  Because session 
# IDs are hard to guess, they obviate the necessity of revalidating 
# identities on each visit.
#
# We don't really care what version of DBM we use, since the entries
# are transient, anyway.
#
# Session information persists for a limited amount of time, assuming
# you run the garbage collector at regular intervals.  There are two
# different ways to expire a session:
#
# $config{session}->{max_lifetime} = max number of minutes since a
#                                    session was created.
# $config{session}->{max_delay}    = max number of minutes since a
#                                    session was last accessed.
#
# If either of these is exceeded, the session will be removed by the
# garbage collector.  max_delay defaults to 1 hour; max_lifetime defaults
# to 1 day.
#
# Sessions are useful for storing:
#   - user identity information
#   - account or website information
#   - current position in multi-page forms or reports
#   - modes and other state variables
#
# Pros of using sessions:
#   - helps to maintain state
#   - faster than re-authenticating on every transaction
# Cons of using sessions:
#   - extra infrastructure requirements (DBM session table)
#   - no persistent logins
#
# Methods:
#   new - constructor
#   status - report session table stats
#   garbage_collector - clean out old sessions
#   
#   get - get session hash
#   set - set session hash
#   getval - get a session value
#   setval - set a session value
#   save - write session data back to the session table
#   close - drop the session hash
#   DESTROY - save and drop fallback
#   lock - lock the session table for saves
#   lock_nb - non-blocking version of lock
#   unlock - unlock the session table
#------------------------------------------------------------------------
#
# In the last section of this file, we provide methods to 'tie' this
# session class to a hash.  That means that in practice, you should
# never have to work with this class directly.  All you need to do is:
#
# tie %session, 'ExSite::Session';
#
# and your session values should automatically appear in %session, by
# magic ;-)
#
# PS. this is done automatically by ExSite::Config::exsite_init.  
# If you find that it doesn't work, it's most likely due to one of the
# following problems:
#
# - your session files do not exist where ExSite is looking for them
# - your session files do not have the right permissions (666)
# - your program has not loaded the ExSite::Session package
#
#------------------------------------------------------------------------
#
# TODO:
#
# - do not issue the session ID and save the session until real
#   session data (non-timestamp) is stored to the session
#
#------------------------------------------------------------------------

package ExSite::Session;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use Digest::MD5 qw(md5_hex);
use Fcntl qw(:flock);
use ExSite::Base;
our @ISA = qw(ExSite::Base);

# flag the session code as loaded
# this lets us ignore sessions (and the resulting errors) on
# scripts that do not use them.
$config{session}{loaded} = 1;

# new: class constructor
#
# - gets session ID from HTTP_COOKIE
# - looks up session in session table, or creates a new one

sub new {
    my ($this) = @_;

    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    my $session_file = $config{session}{file};
    my %session_table;

    if (dbmopen(%session_table,$session_file,00666)) {
	$obj->{sessionDB} = \%session_table;
	$obj->garbage_collector;   # clean up any stale sessions
    }
    else {
	# no session table
	$obj->error("error opening session table");
	dbmclose(%session_table);
	return undef;
    }

    return $obj;
}

# begin working in a session
# (validates or creates a session, sets a cookie)

sub begin {
    my $this = shift;
    if (! $this->validate) {
	my $sid = $this->create;
	my $p = $share{Page} || $share{PageBuf};
	my $cookie = &set_cookie("$config{site}{id}_sessionID",$sid);
	if ($p) { $p->header($cookie); }
	else { print $cookie; }
    }
}

# stop working in a session
# (saves session data for next visit, closes the DBM file)

sub end {
    my $this = shift;
#    if (! $this->{done}) {
#	$this->save;
#    }
    dbmclose(%{$this->{sessionDB}});
#    $this->{done} = 1;
}

# validate a session

sub validate {
    my ($this,$id) = @_;
    if (! $id) {
	my $session_cookie = $config{site}->{id}."_sessionID";
	$id = &get_cookie($session_cookie);
    }
    if ($id) {
	$this->info("re-opening session $id");
	#if (exists $this->{sessionDB}{$id}) {
	if ($this->{sessionDB}{$id}) {
	    # this is a good session
	    # copy the saved session data into our object
	    my %session_data = &DecodeString($this->{sessionDB}{$id});
	    $session_data{'_atime'} = time();  # update access time
	    $this->{data} = \%session_data;
	    $this->{sessionID} = $id;
	    return 1;
	}
	else {
	    # no such session
	    $this->warn("session $id expired");
	}
    }
    return 0;
}

# start a new session

sub create {
    my ($this,@data) = @_;

    # get access time

    my $time = time();

    # make a session ID from a set of values unique to this visitor
	    
    if (@data == 0) {
	@data = ($ENV{REMOTE_ADDR}, $ENV{HTTP_USER_AGENT}, $time, int(rand(100000)));
    }
    $this->{sessionID} = md5_hex(@data);
    $this->info("creating new session $this->{sessionID}");

    $this->{data}{_ctime} = $this->{data}{_atime} = $time;

    # save the session
    $this->save;

    return $this->{sessionID};
}

# status report - display summary of active sessions

sub status {
    my ($this) = @_;
    my $time = time;
    my @sessions = keys %{$this->{sessionDB}};
    print scalar @sessions, " active sessions...\n";
    foreach my $sessionID (@sessions) {
	my %session = &DecodeString($this->{sessionDB}->{$sessionID});
	print "$sessionID: ",
	    $time - $session{_ctime}, " seconds old; ",
	    $time - $session{_atime}, " seconds since last access.\n";
    }
}

# garbage_collector - removes expired sessions
# sessions expire if more than 
# $config{session}->{max_lifetime} minutes elapsed since session creation
# $config{session}->{max_idle} minutes elapsed since last session activity
# - returns number of removed sessions

sub garbage_collector {
    my ($this) = @_;
    my $time = time();
    my @sessions = keys %{$this->{sessionDB}};
    my $lifetime = $config{session}{max_lifetime};
    my $latency = $config{session}{max_idle};
    my $gcount;
    foreach my $sessionID (@sessions) {
	my %session = &DecodeString($this->{sessionDB}->{$sessionID});
	if (($time > $session{_ctime} + ($lifetime * 60)) ||
	    ($time > $session{_atime} + ($latency * 60))) {
	    delete $this->{sessionDB}->{$sessionID};
	    $gcount++;
	}
    }
    if ($gcount) {
	$this->info("Session::garbage_collector: removed $gcount sessions");
    }
    return $gcount;
}

# get

sub get {
    my $this = shift;
    if (wantarray) { return %{$this->{data}}; }
    else { return $this->{data}; }
}

# set

sub set {
    my ($this,$rhash) = @_;
    # preserve session timestamps
    my $ctime = $this->{data}->{'_ctime'};
    my $atime = $this->{data}->{'_atime'};
    $this->{data} = $rhash;
    $this->{data}->{'_ctime'} = $ctime;
    $this->{data}->{'_atime'} = $atime;
}

# getval

sub getval {
    my ($this,$key) = @_;
    if (exists $this->{data}->{$key}) {
	return $this->{data}->{$key};
    }
    return undef;
}

# setval

sub setval {
    my ($this,$key,$val) = @_;
    if (defined $val) {
	$this->{data}->{$key} = $val;
    }
    else {
	delete $this->{data}->{$key};
    }
    $this->save;
}

# save - NB: can only save once per session

sub save {
    my ($this) = @_;
    #print "SAVE\n";
    if ($this->lock) {
	#print "lock good $this->{sessionID}\n";
	my $dat = &EncodeHash(%{$this->{data}});
	#print "$dat\n";
	#print "$this->{sessionID} => $this->{sessionDB}\n";
	# next line fails when called from DESTROY()
	$this->{sessionDB}->{$this->{sessionID}} = $dat; 
	#print "SAVE DONE\n";
	$this->unlock;
	#dbmclose(%{$this->{sessionDB}});
	#print "save complete $!\n";
	#$this->{done} = 1;
	$this->info("session $this->{sessionID} updated");
	return 1;
    }
    else {
	$this->warn("could not lock session file: session data NOT saved");
	return 0;
    }
}

# close - permanently ends the session

sub close {
    my $this = shift;
    delete $this->{sessionDB}->{$this->{sessionID}};
    $this->end;
}

#------------------------------------------------------------------------
# Locking tools, to prevent file collisions.
#------------------------------------------------------------------------
#
# Lock/unlock a DB for read-write access
# -> doesn't lock DB itself;  uses a lockfile for simplicity
#    will fail if lockfile is missing, or not writeable.
# -> returns TRUE (1) if lock was successful
# -> calling script must police itself - a misbehaving script
#    could ignore the lock and proceed anyway.
#
#------------------------------------------------------------------------

sub lock {
    my $this = shift;
    my $lockfile = $config{session}{lockfile};
    # first, test the lock file for write access
    if (open LOCK,$lockfile) {
	if (! flock(LOCK,LOCK_EX)) {
	    $this->warn("failed to lock session lockfile");
	    return 0;
	}
    }
    else {
	$this->warn("failed to open session lockfile");
	return 0;
    }
    return 1;
}

# non-blocking version - aborts after 10sec timeout
sub lock_nb {
    my $this = shift;
    my $lockfile = $config{session}{lockfile};
    # first, test the lock file for write access
    if (open LOCK,$lockfile) {
	my $count = 0;
	my $gotit = flock LOCK,(LOCK_EX|LOCK_NB);
	while (! $gotit && $count < 10) {
	    $count++;
	    sleep(1);
	    $gotit = flock LOCK,(LOCK_EX|LOCK_NB);
	}
        if (! $gotit) { 
	    $this->warn("session lockfile timed out");
	    return 0; 
	}
    }
    else { 
	$this->warn("failed to open session lockfile");
	return 0; 
    }
    return 1;
}

sub unlock {
    flock LOCK,LOCK_UN;
    close LOCK;
}

#------------------------------------------------------------------------
# Using perl's tie mechanism, we can implement this class implicitly
# using the %session variable.  To do that, we need to support the
# following class methods.
#------------------------------------------------------------------------

sub TIEHASH {
    my ($this) = @_;
    my $s = &new($this);
    $s->begin;
    return $s;
}

sub FETCH {
    my ($this,$key) = @_;
    return $this->getval($key);
}

sub STORE {
    my ($this,$key,$val) = @_;
    return $this->setval($key,$val);
}

sub DELETE {
    my ($this,$key) = @_;
    return $this->setval($key);
}

sub CLEAR {
    my ($this) = @_;
    foreach my $key (keys %{$this->{data}}) {
	next if ($key =~ /^_/);  # don't clear control variables
	delete $this->{data}{$key};
    }
    return;
}

sub EXISTS {
    my ($this,$key) = @_;
    return exists $this->{data}{$key};
}

sub FIRSTKEY {
    my ($this) = @_;
    my @keys = keys %{$this->{data}};
    return shift @keys;
}

sub NEXTKEY {
    # ugh, kludge...
    my ($this,$lastkey) = @_;
    my @keys = keys %{$this->{data}};
    my $k = 0;
    do { $k = shift @keys; } until ($k eq $lastkey || !defined $k);
    return shift @keys;
}

sub UNTIE {
    my ($this) = @_;
#    if (! $this->{done}) {
#	$this->save;
#    }
    return;
}

sub DESTROY {
    my $this = shift;
#    if (! $this->{done}) {
#	$this->save;
#    }
#    dbmclose(%{$this->{sessionDB}});
    return;
}

1;

