package ExSite::Time;

#-----------------------------------------------------------------------
#
#   Copyright 2001-2005 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# ExSite::Time - utilities for handling dates and times in various formats
#
# We support the following time formats
#
# sql_date		yyyy-mm-dd
# iso_date		yyyy-mm-dd
# sql_datetime		yyyy-mm-dd hh:mm
# time			hh:mm am/pm
# time24		hh:mm
# sql_timestamp		yyyymmddhhmmss
# unix_timestamp	nnnnnnnnnnn
# date			mmm dd, yyyy
# date_long		mmm dd, yyyy
# datetime		wday Mon dd hh:mm yyyy
# datetime_long	        Weekday Month dd hh:mm yyyy
# datetime2             hh:mm, Month dd, yyyy
# dd-mmm-yy             dd-Mon-yy
# weekday		wday
# weekday_long		Weekday
# year			yyyy
# month			Mon
# month_long		Month
#
# Internally, we convert all dates/times to a 6-element time array: 
# y, m, d, h, m, s.  The time array is initialized to the current datetime, 
# so if you only set a partial datetime (eg. only the time), the remaining 
# values are unchanged, and the full array still represents a complete 
# datetime.
#
#----------------------------------------------------------------------------
#
# Usage:
#
# Make a time object (initialized to current time):
# my $t = new ExSite::Time;
#
# Make a time object (initialized to some other time/format):
# my $t = new ExSite::Time("2001-12-25 09:00:00","sql_datetime");
#
# Set the time to the current time:
# $t->set;
#
# Set the time to some time/format:
# $t->set('Feb 1, 1968','date');
# $t->set('6:30 pm','time');
# $t->set('Fri Aug 26 12:14:43 2005','datetime');
#
# adjust the time by some amount:
# $t->add(30,'days');
# $t->add(-3,'months');
#
# Output the time in some format:
# print $t->write('date');
# print $t->write('unix_timestamp');
# print $t->write('weekday');
#
# Compare another time to the currently-set time:
# ... returns 1 if the compared time is in the future, -1 if it is in 
#     the past, and 0 if it is the same as the current time.
# $t->compare('20050826091500','sql_timestamp');
#
# Convenience call to compare two times in same format:
# $t->compare('Aug 26, 2005','Aug 28, 2005','datetime');
#
# There are also a number of legacy calls to emulate an older Time package.
#
#----------------------------------------------------------------------------


use strict;
use ExSite::Config;
use ExSite::Base;
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

use vars qw(@month_short @month @dayinmonth @weekday_short @weekday $weekday);

@month_short = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
@month = qw(January February March April May June July August September October November December);
@dayinmonth = qw(31 28 31 30 31 30 31 31 30 31 30 31);
@weekday_short = qw(Sun Mon Tue Wed Thu Fri Sat);
@weekday = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);

# new : create a Time object

sub new ($%) {
    my($this, $time, $format) = @_;

    # objectify object
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    # setup Base class features
    $obj->initialize_object;

    # load time
    $obj->set($time,$format);

    return $obj;
}

# set : set the current time, using a formatted time string and a format

sub set {
    my ($this,$value,$format) = @_;
    $weekday = undef;
    # eval the call to time() in case of a badly formatted value
    my $t = eval { $this->time($value,$format); };
    if ($@) {
	$this->error("ExSite::Time: bad time value ($format): $value [$@]");
	return $this->{time} = undef;
    }
    $this->{time} = $t;
    my @t = @$t;
    if (! $weekday) {
	use Time::Local;
	my $t = timelocal($t[5],$t[4],$t[3],$t[2],$t[1]-1,$t[0]-1900);
	(undef,undef,undef,undef,undef,undef,$weekday,undef,undef) =
	    localtime($t);
    }
    $this->{wday} = $weekday;
    return $this->{time};
}

# time : convert a formatted time string to a time array

sub time {
    my ($this,$value,$format) = @_;
    my $t = $this->now;
    my @time = @$t;

    if ($value) {
	$_ = $format;
	if (/^(sql|iso)_date$/) {
	    ($time[0],$time[1],$time[2]) = split /-/,$value;
	    $weekday = undef;
	}
	elsif (/^sql_datetime$/) {
	    my ($d,$t) = split / /,$value;
	    ($time[0],$time[1],$time[2]) = split /-/,$d;
	    ($time[3],$time[4],$time[5]) = split /:/,$t;
	    $weekday = undef;
	}
	elsif (/^sql_timestamp$/) {
	    ($time[0],$time[1],$time[2],
	     $time[3],$time[4],$time[6]) = 
		 ($value =~ /(....)(..)(..)(..)(..)(..)/);
	    $weekday = undef;
	}
	elsif (/^time$/) {
	    $value =~ /([012]?\d):(\d{2})(:(\d{2}))?/;
	    $time[3] = $1;
	    $time[4] = $2;
	    $time[5] = $4;
	    if ($value =~ /pm/i && $time[3] < 12) {
		$time[3] += 12;
	    }
	    elsif ($value =~ /am/i && $time[3] == 12) {
		$time[3] = 0;
	    }
	}
	elsif (/^time24$/) {
	    $value =~ /([012]?\d):(\d{2})(:(\d{2}))?/;
	    $time[3] = $1;
	    $time[4] = $2;
	    $time[5] = $4;
	}
	elsif (/^unix_timestamp$/) {
	    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
		localtime($value);
	    @time = ($year+1900,$mon+1,$mday,$hour,$min,$sec);
	    $weekday = $wday;
	}
	elsif (/^date$/) {
	    my ($m,$d,$y) = ($value =~ /(\w+) (\d+), (\d{4})/);
	    $time[0] = $y;
	    $time[1] = $this->which($m,@month_short);
	    $time[2] = $d;
	    $weekday = undef;
	}
	elsif (/^date_long$/) {
	    my ($m,$d,$y) = ($value =~ /(\w+) (\d+), (\d{4})/);
	    $time[0] = $y;
	    $time[1] = $this->which($m,@month);
	    $time[2] = $d;
	    $weekday = undef;
	}
	elsif (/^datetime$/) {
	    $value =~ /(\w+) (\w+) (\d+), ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $time[0] = $8;
	    $time[1] = $this->which($2,@month_short);
	    $time[2] = $3;
	    $time[3] = $4;
	    $time[4] = $5;
	    $time[5] = $6;
	    $weekday = $this->which($1,@weekday_short) - 1;
	}
	elsif (/^datetime_long$/) {
	    $value =~ /(\w+) (\w+) (\d+), ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $time[0] = $8;
	    $time[1] = $this->which($1,@month);
	    $time[2] = $3;
	    $time[3] = $4;
	    $time[4] = $5;
	    $time[5] = $6;
	    $weekday = $this->which($1,@weekday) - 1;
	}
	elsif (/^datetime2$/) {
	    $value =~ /(\d+):(\d+).*?, (\w+) (\d+), (\d+)/;
	    $time[0] = $5;
	    $time[1] = $this->which($3,@month);
	    $time[2] = $4;
	    $time[3] = $1;
	    $time[4] = $2;
	    if ($value =~ /pm/i && $time[3] < 12) {
		$time[3] += 12;
	    }
	    elsif ($value =~ /am/i && $time[3] == 12) {
		$time[3] = 0;
	    }
	}
	elsif (/^dd-mmm-yy$/) {
	    my ($d,$m,$y) =~ /(\d+)-(\w+)-(\d+)/i;
	    $time[0] = $y > 50 ? $y+1900 : $y+2000;
	    $time[1] = $this->which($m,@month_short);
	    $time[2] = $d;
	}
	elsif (/^weekday$/) {
	    $weekday = $this->which($value,@weekday_short) - 1;
	}
	elsif (/^weekday_long$/) {
	    $weekday = $this->which($value,@weekday) - 1;
	}
	elsif (/^year$/) {
	    $time[0] = $value;
	    $weekday = undef;
	}
	elsif (/^month$/) {
	    $time[1] = $this->which($value,@month_short);
	    $weekday = undef;
	}
	elsif (/^month_long$/) {
	    $time[1] = $this->which($value,@month);
	    $weekday = undef;
	}
	elsif (/^raw$/) {
	    @time = @$value;
	}
    }
    return \@time;
}

# write : output the time in an arbitrary format

sub write {
    my $this = shift;
    $_ = shift;     # format
    my $imonth = $this->{time}[1] - 1;
    if (/^(sql|iso)_date$/) {
	return sprintf "%04d-%02d-%02d", @{$this->{time}};
    }
    elsif (/^sql_datetime$/) {
	return sprintf "%04d-%02d-%02d %02d:%02d:%02d", @{$this->{time}};
    }
    elsif (/^sql_timestamp$/) {
	return sprintf "%04d%02d%02d%02d%02d%02d", @{$this->{time}};
    }
    elsif (/^time$/) {
	my $h = $this->{time}[3];
	my $m = $this->{time}[4];
	my $ampm = " AM";
	if ($h >= 24) {
	    $h -= 24;
	}
	elsif ($h > 11) {
	    if ($h > 12) { $h -= 12; }
	    $ampm = " PM";
	}
	return sprintf("%d:%02d", $h, $m).$ampm;
    }
    elsif (/^time24$/) {
	return sprintf "%02d:%02d", $this->{time}[3], $this->{time}[4];
    }
    elsif (/^unix_timestamp$/) {
	use Time::Local;
	return timelocal($this->{time}[5],$this->{time}[4],$this->{time}[3],
			 $this->{time}[2],$this->{time}[1]-1,$this->{time}[0]-1900);
    }
    elsif (/^date$/) {
	return "$month_short[$imonth] $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^date_long$/) {
	return "$month[$imonth] $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^datetime$/) {
	return "$weekday_short[$this->{wday}] $month_short[$imonth] $this->{time}[2] ".$this->write('time24')." $this->{time}[0]";
    }
    elsif (/^datetime_long$/) {
	return "$weekday[$this->{wday}] $month[$imonth] $this->{time}[2] ".$this->write('time24')." $this->{time}[0]";
    }
    elsif (/^datetime2$/) {
	return $this->write('time').", $month[$imonth] $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^dd-mmm-yy$/) {
	return sprintf("%d-%s-%02d",$this->{time}[2],$month_short[$imonth],substr($this->{time}[0],2,2));
    }
    elsif (/^weekday$/) {
	return $weekday_short[$this->{wday}];
    }
    elsif (/^weekday_long$/) {
	return $weekday[$this->{wday}];
    }
    elsif (/^year$/) {
	return $this->{time}[0];
    }
    elsif (/^month$/) {
	return $month_short[$imonth];
    }
    elsif (/^month_long$/) {
	return $month[$imonth];
    }
    elsif (/^raw$/) {
	return $this->{time};
    }
}

# now : convert the current time to a time array

sub now {
    my $this = shift;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = 
	localtime(time);
    $weekday = $wday;
    return [$year+1900,$mon+1,$mday,$hour,$min,$sec];
}

# add: add a number of time units (years, months, days, hours, minutes, 
# seconds) to the current time.

sub add {
    my ($this,$number,$what) = @_;
    my $t = $this->write("unix_timestamp");
    my $seconds = $number;
    if ($what =~ /year/i) {
	$seconds *= (365*24*60*60);
    }
    elsif ($what =~ /month/i) {    
	$seconds *= int(30.4375*24*60*60);  ### 1 month == 1/12 of a year
	#$seconds *= (30*24*60*60);          ### 1 month == 30 days
    }
    elsif ($what =~ /day/i) {
	$seconds *= (24*60*60);
    }
    elsif ($what =~ /hour/i) {
	$seconds *= (60*60);
    }
    elsif ($what =~ /min/i) {
	$seconds *= 60;
    }
    $t += $seconds;
    return $this->set($t,"unix_timestamp");
}

# compare : return -1, 0, +1 if passed date is past, present, or future 
# with respect to the current date

sub compare {
    my ($this, $date, $fmt) = @_;
    my $now = $date ? $this->time($date,$fmt) : $this->now;
    my $then = $this->{time};
   
    return 1 if ($now->[0] > $then->[0]);
    return -1 if ($now->[0] < $then->[0]);
	       
    return 1 if ($now->[1] > $then->[1]);
    return -1 if ($now->[1] < $then->[1]);
	       
    return 1 if ($now->[2] > $then->[2]);
    return -1 if ($now->[2] < $then->[2]);

    return 0;
}

sub compare2 {
    my ($this, $date1, $date2, $fmt) = @_;
    $this->set($date2,$fmt);
    return $this->compare($date1,$fmt);
}

# which : finds a named time value (eg. month, weekday) in an array
# and returns the array location plus 1.

sub which {
    my ($this,$val,@list) = @_;
    my $i = 0;
    foreach my $item (@list) {
	$i++;
	return $i if ($val =~ /^$item$/i);
    }
    return 0;
}

###================= LEGACY FUNCTIONS ====================

#sub CompareToNow
#pre	sql date
#post	if date is in future, returns 1
#post	if date is in past, returns -1
#post	if date is now, returns 0

sub CompareToNow {
    my ($this,$t1,$t2) = @_;
    if ($t2) { 
	return $this->compare2($t1,$t2,"sql_date");
    }
    else {
	$this->set;
	return $this->compare($t1,"sql_date");
    }
}

#
# makePreviousDate
# accepts a number of days
# returns the sql date for the date in
# the past equalt to today minus the number
# of days specified
# ie, if $days = 7, we get the date for one week ago

sub MakePreviousDate {
    my ($this, $days) = @_;
    $this->add(-$days,"days");
    return $this->write("sql_date");
}
	
sub GetCurrentSqlDate{
    my ($this) = @_;
    $this->set;
    return $this->write('sql_date');
}

sub ConvertToDate{
    my ($this,$value) = @_;
    if ($value) { $this->set($value,"sql_date"); }
    return $this->write("date");
}

sub ConvertSqlToTime{
    my ($this,$value) = @_;
    if ($value) { $this->set($value,"time24"); }
    return $this->write("time");
}

sub ConvertFromTimeStamp{
    my ($this,$data) = @_;
    if (length $data == 14){
	$this->set($data,"sql_timestamp");
	return $this->write("sql_date");
    }
    else{
	$this->set($data,"unix_timestamp");
	return $this->write("datetime");
    }
}

sub GetWeekDay{
    my ($this,$date) = @_;
    $this->set($date,"sql_date");
    return $weekday[$this->{wday}];
}

sub ConvertToShortDate{
    my ($this,$value) = @_;
    my ($year,$month,$day);
    if ($value) { $this->set($value,'sql_date'); }
    return $this->write('dd-mmm-yy');
}

sub ConvertSqlToDateTime{
    my ($this,$data)= @_;	
    if ($data) { $this->set($data,'sql_datetime'); }
    return $this->write('datetime2');
}

1;
