#-----------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
=pod

=head1 ExSite::Tree

ExSite::Tree is a generic tool for managing heirarchical tree
structures such as are found in ExSite.  It assumes that each tree
node is essentially an ExSite datahash, and each node has a
single parent node.

Each tree node consists of a datahash and some tree indexing parameters.  
The hash should include keys/values for the id (an arbitrary value used 
to reference the node) and the parent (the id value of the parent node).  
It can contain any other keys/values as well.

Nodes retain the order in which they were inserted into the tree.

=head2 Tree Construction

Each tree node is a datahash.  One of the hash keys is used as a node
ID, and another is used as a parent ID (collectively, these are the 
index keys).  The index keys can be specifed uniquely for each node,
for a bulk insert of multiple nodes, or they can be defined as 
defaults for all nodes.

ExSite Trees are designed to be self-constructing with typical ExSite
data structures.  See the example at the end for more info.

=head2 Tree Contruction

=over 4

=item Create a new tree object

C<my $tree = new ExSite::Tree("id_key", "parent_key", @data);>

"id_key" and "parent_key" are the index keys for the datahashes in C<@data>.
They are also the default index keys for all other nodes added to the tree.
C<@data> is optional.

=item Add nodes to a tree

C<$tree-E<gt>addnode($data);>

C<$tree-E<gt>addnode($data,$id_key,$parent_key);>

Adds one node whose data is in C<%$data>.  If the index keys are different
from the default for the tree, they can be specified as extra parameters.

C<$tree-E<gt>addnodes($data);>

C<$tree-E<gt>addnodes($data,$id_key,$parent_key);>

Same as above, except C<$data> is taken to be a reference to 
an array of datahashes.

C<$tree-E<gt>replacenode($id,$data);>

C<$tree-E<gt>replacenode($id,$data,$id_key,$parent_key);>

This replaces the data in the node indexed under C<$id>, with the contents
of C<%$data>.  The replaced node will be indexed under the original 
key as well as under a new key determined from the replacement data.

=back

=head2 Querying the tree

Tree nodes can be fetched as nodes (including the tree indexing parameters)
or as data (ie. the original datahash).  Fetching as nodes is typical for
internal use;  fetching as data is typical for calls from other
packages.  If you fetch the node, the data can be accessed as 
C<$node-E<gt>{data}{...}>.

=over 4

=item Fetch a node

C<$tree-E<gt>getnode($id,$create_flag);>

C<$tree-E<gt>getnode_data($id,$create_flag);>

Fetches the node indexed under C<$id>.  If C<$create_flag> is true, 
a dummy version of the node will be created if it is not found.

=item Fetch the parent of a node

C<$tree-E<gt>get_parent_node($id);>

C<$tree-E<gt>get_parent_data($id);>

Fetches the parent of the node indexed under C<$id>.

=item Fetch the children of a node

C<$tree-E<gt>get_child_nodes($id);>

C<$tree-E<gt>get_child_data($id);>

C<$tree-E<gt>get_child($id);>  # synonym for get_child_data()

Fetches the children nodes under node C<$id>.  Returns an array, or
array ref, depending on the list context.

=item Fetch the top-level of the tree

C<$tree-E<gt>get_topnodes();>

C<$tree-E<gt>get_topnodes_data();>

Fetches the nodes that have no parents.  Returns an array, or
array ref, depending on the list context.

=item Fetch the top-level ancestor of a node

C<$tree-E<gt>get_ancestor_node($id);>

C<$tree-E<gt>get_ancestor_data($id);>

Search up through the parent links for a node without a parent.

=item Dump the tree

C<print $tree-E<gt>dump();>    # dump IDs

C<print $tree-E<gt>dump($id);> # dump IDs, starting from $id

C<print $tree-E<gt>dump($id,"key1","key2");> # dump data keys, $id

Returns a string that illustrates the map as a nested list of text strings.
Each map entry is shown as its ID, by default, but if alternate keys are 
given, the values under those keys in the node's data will be listed instead.

=back

=head2 Example

This example creates a site map as a tree.

    # fetch a (somewhat) ordered list of pages
    my @pages = $db->fetch_match("page", {section_id=>$my_section, type=>"page"}, ["rank","page_id"]);

    # index the pages as a tree, using page.page_id and page.parent_id as the index keys
    my $map = new ExSite::Tree("page_id","parent_id",@data);

This tree consists of one node for each page, with the nodes nested 
exactly as they do in site menus and site maps.  At each level of the tree,
the pages are ordered correctly according to their page rank.

To fetch the top-level pages from the site, use:

    my @toppages = $map->get_topnodes_data();

This retrieves an ordered array of datahashes, representing the top-level
pages of the site.  For each of these pages, you can fetch the child pages
(submenus) beneath them with:

    my @subpages = $map->get_child_data($page->{page_id});

...and so on, to any depth.

To get a quick view of the site map, use:

    print $map->dump(0,"filename");

=cut

package ExSite::Tree;
use strict;

use vars qw($depth);

# Tree constructor

sub new ($$$@) {
    my ($this,$id_key,$parent_key,@data) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->{top} = [];   # top-level node order
    $obj->{nodes} = {}; # node storage
    $obj->{id_key} = $id_key;          # default keys
    $obj->{parent_key} = $parent_key;
    if (@data > 0) { $obj->addnodes(\@data); }
    return $obj;
}

# add a bunch of nodes

sub addnodes {
    my ($this,$data,$id_key,$parent_key) = @_;
    foreach my $d (@$data) {
	$this->addnode($d,$id_key,$parent_key);
    }
}

# add a single node

sub addnode {
    my ($this,$d,$id_key,$parent_key) = @_;
    my $id = $id_key || $this->{id_key};
    my $pid = $parent_key || $this->{parent_key};

    my $key = $d->{$id};
    # add node to the tree
    if (exists $this->{node}{$key}) {
	# overwrite existing node
	$this->{node}{$key}{id_key} = $id,
	$this->{node}{$key}{parent_key} = $pid,
	$this->{node}{$key}{data} = $d,
	$this->{node}{$key}{parent} => $d->{$pid},
    }
    else {
	# insert new node
	$this->{node}{$key} = {
	    id_key => $id,
	    parent_key => $pid,
	    data => $d,
	    parent => $d->{$pid},
	    child => [],
	};
    }
    my $node = $this->{node}{$key};
    
    # add node to the child list of its parent
    my $parent_id = $node->{data}{$node->{parent_key}};
    if ($parent_id) {
	my $parent = $this->getnode($parent_id,1);
	push @{$parent->{child}}, $node;
    }
    else {
	# no parent - must be a top-level node
	push @{$this->{top}}, $node;
    }
}

# replace node data, but maintain existing node relationships

sub replacenode {
    my ($this,$id,$d,$id_key,$parent_key) = @_;
    if (exists $this->{node}{$id}) {
	# overwrite existing node data
	$this->{node}{$id}{data} = $d;
	# add a pointer from new node key
	$this->{node}{$d->{$id_key || $this->{id_key}}} = $this->{node}{$id};
    }
    else {
	$this->addnode($d,$id_key,$parent_key);
    }
}

# fetch a node, optionally creating it if it does not exist

sub getnode ($$$) {
    my ($this,$id,$create_flag) = @_;
    if (! exists $this->{node}{$id} && $create_flag) {
	$this->{node}{$id} = {
	    id_key => $this->{id_key},
	    parent_key => $this->{parent_key},
	    data => {},
	    child => [],
	};
    }
    return exists $this->{node}{$id} ?
	$this->{node}{$id} :
	undef;
}

sub getnode_data ($$$) {
    my ($this,$id,$create_flag) = @_;
    my $node = $this->getnode($id,$create_flag);
    return $node ? $node->{data} : undef;
}

# fetch the parent of a node

sub get_parent_node ($$) {
    my ($this,$id) = @_;
    my $node = $this->getnode($id);
    if ($node && $node->{data}{$node->{parent_key}}) {
	return $this->getnode($node->{parent});
    }
    return undef;
}

sub get_parent_data ($$) {
    my ($this,$id) = @_;
    my $p = $this->get_parent_node($id);
    return $p ? $p->{data} : undef;
}

sub get_parent {
    return &get_parent_data(@_);
}

# fetch the children of a node

sub get_child_nodes ($$) {
    my ($this,$id) = @_;
    my @children;
    if (exists $this->{node}{$id}) {
	@children = @{$this->{node}{$id}{child}}
    }
    else {
	@children = $this->get_topnodes();
    }
    return wantarray ? @children : \@children;
}

sub get_child_data ($$) {
    my ($this,$id) = @_;
    my @children = $this->get_child_nodes($id);
    foreach my $c (@children) {
	$c = $c->{data};
    }
    return wantarray ? @children : \@children;
}

sub get_child {
    return &get_child_data(@_);
}

# return the top-level nodes

sub get_topnodes {
    my ($this) = @_;
    return wantarray ? @{$this->{top}} : $this->{top};
}

sub get_topnodes_data {
    my ($this) = @_;
    my @topnode = $this->get_topnodes();
    foreach my $n (@topnode) {
	$n = $n->{data};
    }
    return wantarray ? @topnode : \@topnode;
}

# return the top-level node relative to a starting node

sub get_ancestor_data {
    my ($this,$id) = @_;
    my $node; 
    while ($id) {
	if (exists $this->{node}{$id}) {
	    $node = $this->{node}{$id};
	    $id = $node->{data}{$node->{parent_id}};
	}
	else {
	    # non-existent id
	    $id = undef;
	    $node = undef;
	}
    }
    return $node;
}

sub get_ancestor_data {
    my ($this,$id) = @_;
    my $node = $this->get_ancestor($id);
    if (ref $node eq "HASH") {
	return $node->{data};
    }
    return undef;
}

# display a summary of the tree contents

sub dump {
    my ($this,$id,@keys) = @_;
    my ($out,@node);
    if (exists ($this->{node}{$id})) {
	@node = @{$this->{node}{$id}{child}};
    }
    else {
	@node = $this->get_topnodes();
    }
    $depth = 0;
    foreach my $n (@node) {
	if (@keys == 0) { @keys = ($n->{id_key}); }
	foreach my $k (@keys) {
	    $out .= $n->{data}{$k}." ";
	}
	$out .= "\n";
	$depth++;
	$out .= $this->dump_r($n,@keys);
	$depth--;
    }
    return $out;
}

sub dump_r {
    my ($this,$node,@keys) = @_;
    my ($out,@node);
    foreach my $n ($this->get_child_nodes($node->{data}{$node->{id_key}})) {
	if (@keys == 0) { @keys = ($n->{id_key}); }
	foreach (1..$depth) { $out .= "  "; }
	foreach my $k (@keys) {
	    $out .= $n->{data}{$k}." ";
	}
	$out .= "\n";
	$depth++;
	$out .= $this->dump_r($n,@keys);
	$depth--;
    }
    return $out;
}

1;
