#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Util;

use strict;
use ExSite::Misc;
use ExSite::Config;

#========================================================================

=pod

=head1 ExSite::Util --  miscellaneous utility functions needed by ExSite::

Miscellaneous routines and functions that depend on ExSite::Config only

=cut

#========================================================================

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT);
    $VERSION     = 1.00;
    @ISA         = qw(Exporter);
    @EXPORT = qw(   &replace_tag &restore_cms_tags 
		    &restore_cms_tag_img &restore_cms_tag_dcd
		    &set_cookie &get_cookie &edit_cookie &redirect
		    &snippet
		    );
}

=pod

=head2 set_cookie() - generates a Set-cookie header.

This routine does not actually send the cookie, it is only a
convenience routine for composing the cookie header.

Arguments:

=over 4

=item * name = name of cookie

=item * value = value of cookie

=item * path = path to return this cookie to (optional)

=item * domain = domain to return this cookie to (optional)

=item * expiry = expiry date of this cookie (optional)

=back

=head2 get_cookie($name,$sepchar) - retrieve a cookie value

Returns the value of cookie "$name".
If sepchar is defined, the cookie value is split on that character,
and an array of values is returned.

=head2 edit_cookie($name,$value) - changes current cookie value

This changes the value of cookie $name to $value.  It ONLY changes
the value for the current request - it does not re-set the coookie for 
subsequent requests.  To modify the cookie for future requests, you
must also call &ExSite::Util::set_cookie(). &edit_cookie() returns 
the new cookie value, which can be passed to set_cookie, if needed.

=cut

sub set_cookie {
    my ($name,$value,$path,$domain,$expiry) = @_;
    if (! $path) { $path = $config{server}->{CGIpath}; }
    if (! $domain) { 
	$domain = $config{server}{host} ? 
	    ".$config{server}{domain}" : $config{server}{domain}; 
    }
    my $out = "Set-cookie: $name=$value; path=$path; domain=$domain";
    if ($expiry) { $out .= "; Expires=$expiry"; }
    return "$out\n";
}

sub get_cookie { 
    my ($name,$sepchar) = @_;
    if (exists $ENV{'HTTP_COOKIE'}) {
        my @cookies = split /; /, $ENV{'HTTP_COOKIE'};
        foreach my $cookie (@cookies) {
            if ($cookie =~ /^$name=/) {
		my (undef,$value) = split /=/,$cookie,2;
		if ($sepchar) { 
		    return split /$sepchar/,$value;
		}
		else {
		    return $value;
		}
            }
        }
    }
    return undef;
}

sub edit_cookie { 
    my ($name,$value) = @_;
    my @cookies;
    my $done;
    if (exists $ENV{'HTTP_COOKIE'}) {
        @cookies = split /; /, $ENV{'HTTP_COOKIE'};
        foreach my $cookie (@cookies) {
            if ($cookie =~ /^$name=/) {
		$cookie = "$name=$value";
		$done = 1;
		last;
	    }
        }
    }
    if (! $done) {
	# add the cookie
	push @cookies, "$name=$value";
    }
    # redefine the cookie
    my $new_cookie = join "; ",@cookies;
    $ENV{HTTP_COOKIE} = $new_cookie;
    return $new_cookie;
}

=pod

=head2 redirect($goto) - generates a Location header to switch pages

This routine tries to use any instantiated page management objects to
redirect.  If none can be found, it prints a Location header directly.
C<$goto> is the URL to redirect to.

=cut

sub redirect {
    my $goto = shift;

    # look for page management objects that will handle redirection 
    # gracefully, even if we have already generated output.

    my $p = $share{Page} || $share{PageBuf};
    if ($p) { 
	$p->location($goto);
    }
    else {
	# no page management objects found
	# try printing the location header directly
	print "Location: $goto\n";
    }
    return;
}

=pod

=head2 snippet() - generates preformatted HTML snippets with data insertions

Snippets are stored in C<$config{snippet}-E<GT>{snippet-name}>.  Each
snippet is an array of values, that are simply copied to the output
string in order, subject to the following transformations:

=over 4

=item * strings: sent to output as-is

=item * numbers: sends the calling argument at that index to the output.

=item * code reference: executes that code, and sends the return value
to the ouput.

=back

=cut

sub snippet {
    my $snippet = $_[0];
    my $out;
    foreach my $dat (@{$config{snippet}->{$snippet}}) { 
	if (ref $dat eq "CODE") { $out .= &$dat; }
	elsif ($dat =~ /^\d$/) { $out .= $_[$dat]; }
	else { $out .= $dat; }
    }
    return $out;
}


=pod

=head1 CMS tag utilities

=head2 replace_tag($html) - replaces the CMS tags in $html with 
IMG tags that serve as placeholders, and allow the HTML to be displayed
out of context.  DEPRECATED: ExSite::ContentBase::expand() dummy modes
now do this task.

=head2 restore_cms_tags($html) - convert masked CMS tags (see replace_tag)
back to the original CMS tags.

This routine contains a lot of ugly kludges to clean up after some 
browsers that alter our HTML for reasons that probably seemed like a good
idea to the browser authors, but which have obligated us to call down a 
plague of locusts upon them for messing with our stuff.

For example:

=over 4

=item IE randomly rearranges the attributes in our IMG tags, making it hard 
to find the ones we just inserted into the code.

=item IE rewrites our HREF URLs to add hostname info, thereby trashing our
page pointer tags.

=item several browsers like to convert '&' in URLs to '&amp;' and then 
convert it back again so that scripts etc. don't notice.  But *we* notice...

=back

=cut


sub replace_tag
{
    my ($html) = @_;

    # replace <img src=[[...]]> tags

    $html =~ s/\<img src=(\'|\")?\[\[(\w+\.\w+)\]\](\'|\")?[^>]*\>/\<img alt=\"\[\[$2\]\]\" name=\"$2\" src=\"$config{server}->{HTMLpath}\/_ExSite\/images\/checker.png\" title=\"$2\" width=\"100\" height=\"50\"\>/gi;

    # replace <!--&content(...)--> tags

    $html =~ s/\<!--\&?content\((\w+)\)--\>/\<img alt=\"content($1)\" name=\"$1\" src=\"$config{server}->{HTMLpath}\/_ExSite\/images\/checker.png\" title=\"$1\" width=\"100\" height=\"50\"\>/gi;

    # replace DCD tags <!--&XYZ(...)-->

    $html =~ s/\<!--\&(\w+)\((\w*)\)--\>/\<img alt=\"$1($2)\" name=\"$1\" src=\"$config{server}->{HTMLpath}\/_ExSite\/images\/checker.png\" title=\"$1\" width=\"100\" height=\"50\"\>/gi;

    return $html;
}

sub restore_cms_tags {
    my ($data) = @_;
    # restore <!--&content(...)--> tags

    my $start = index($data,"<img alt=\"content(");
    my ($endc, $endi);
    while ($start != -1) {
        $endc = index($data,")",$start);
        $endi = index($data,">",$start);
        my $subname = substr($data,$start+18,$endc-$start-18);
        $subname =~ s/%22/\"/g;
        $subname =~ s/%3C/</g;
        $subname =~ s/%3E/>/g;
        
        $data = substr($data,0,$start)."<!--content(".$subname.")-->".substr($data,$endi+1);
        # check if there are any more HTML content substitutions to do
        $start = index($data,"<img alt=\"content(");
    }

    # restore dummy URL-substituted graphics
    $data =~ s/(<img ([^>]*)alt=\"?\[\[([-\w]+\.?\w+)\]\]\"?[^>]*\>)/&restore_cms_tag_img($1)/gie;

    # restore real URL-substituted graphcs
    $data =~ s/($config{server}{server})?$config{server}{CGIpath}\/content\.cgi(\/)?(.+?)\?id=(\d+)&(amp;)?name=([\w-\.]+)/[[$6]]/gi;
    #$data =~ s/src=\"?($config{server}{server})?$config{server}{CGIpath}\/content\.cgi\?id=(\d+)&(amp;)?name=([\w-\.]+)\"?/src=\"[[$4]]\"/gi;
    # restore dummy DCD objects
    #$data =~ s/\<img ([^>]*)alt=\"?(\w+)\(([^\"]*)\)\"?[^>]*\>/\<!--\&$2($3)--\>/gi;
    $data =~ s/\<img ([^>]*)alt=\"?(\w+)\(([^\"]*)\)\"?[^>]*\>/&restore_cms_tag_dcd($2,$3)/gie;

    # KLUDGE: unescape URL-substitution tags that are mysteriously escaped 
    # by some browsers
    $data =~ s/%7B%7B/{{/g;
    $data =~ s/%7D%7D/}}/g;
    $data =~ s/%5B%5B/\[\[/g;
    $data =~ s/%5D%5D/\]\]/g;

    # KLUDGE: fix page-URL substitutions that get trashed by IE
    # IE wants to convert '<a href={{x}}>' 
    #                  to '<a href="http://whatever.com/cgi/{{x}}>'
    # so we have to strip out the gak that it has so helpfully inserted
    # for us.
    $data =~ s/href=\"[^{>]+({{[^}]+}})\"/href=\"$1\"/g;
    $data =~ s/href=\"[^\[>]+(\[\[[^}]+\]\])\"/href=\"$1\"/g;

    return $data;
}

# restoring images is complicated, because IE scrambles the attributes, 
# which means we cannot match to reliable patterns.

sub restore_cms_tag_img {
    my $tag = shift;
    my ($name,$align);
    # 041116 - allow hyphens in content names
    # 050109 - allow escape chars (%) in content names
    if ($tag =~ /alt=\"?\[\[([-%\w]+\.?\w+)\]\]/i) { $name = $1; }
    my $newtag = "<img src=\"[[$name]]\"";
    if ($tag =~ /align=\"?(\w+)\"?[ >]/i) { 
	$align = $1; 
	$newtag .= " align=\"$align\"";
    }
    $newtag .= ">";
    return $newtag;
}

# DCD tags get the parameter separator '&' munged by browsers. We should
# switch to using ';' instead, but in the meantime, this kludge un-munges 
# the '&' characters in the tags.

sub restore_cms_tag_dcd {
    my ($dcd,$param) = @_;
    $param =~ s/\&amp\;/\&/g;
    return "<!--\&$dcd($param)-->";
}

1;


