#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Wizard;

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Base;
use ExSite::Misc;

our @ISA = qw(ExSite::Base);


#========================================================================

=pod

=head1 ExSite::Wizard -- create, manage, and process multi-screen forms

Wizards take the user through a multi-screen sequence of forms, collecting 
data at each step.  Collected data is stored in hidden fields in each 
screen.  The submission of the aggregate form data does not occur until
the final screen is submitted.  This allows wizard forms to have branching 
logic, or inputs that vary in character, depending on the inputs provided 
on previous screens.

To create a wizard:

    my $wiz = new ExSite::Wizard(title=>"My Wizard",action=>"mywiz.cgi");

    $wiz->add_step("First",\&create_step_1,\&do_step_1);

    $wiz->add_step("Second",\&create_step_2,\&do_step_2);

C<mywiz.cgi> is your wizard CGI program that processes the input of
each screen; 

C<&create_step_X> is the routine that composes and returns a form for step X.
It should return the HTML for the form in a string.  The Wizard automatically
outputs the form tags, submit buttons, and data from previous steps (as 
hidden fields), so you only need to output the new input fields for this
screen.

C<&do_step_X> is the routine that processes the form from each step.  It 
should return an array of error messages (such as input validation failures,
or other problems).  If this array has no entries, the form input is 
accepted, and the wizard automatically proceeds to the next step.

The last C<&do_step_X> routine in the wizard is responsible for doing
final processing of the collected data, ie. handling the final submit
of the aggregate wizard inputs.  It also returns an array of error strings 
if the final collection of data cannot be processed;  if the error array
is empty, the wizard is assumed to have completed.

If the final stage of the wizard has a C<&create_step_N> routine, but
no C<&do_step_N> routine, then ExSite will run the C<&create_step_N>
routine, but not display any wizard buttons to "do" this step.  In this
case you should do your final form processing in C<&do_step_[N-1]>.  
This configuration is useful if you require a specialized final screen 
for final instructions.

=cut

#========================================================================

# create the wizard object

sub new ($%) {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->{title} = $opt{title};
    $obj->{reply} = $opt{reply};
    $obj->{action} = $opt{action};
    $obj->{id} = $opt{id};
    $obj->{step_name} = [];
    $obj->{show_step} = [];
    $obj->{do_step} = [];
    $obj->{db} = $share{DB};
    $obj->{index} = $opt{index};
    $obj->info("creating new Wizard: $opt{title}");
    return $obj;
}

# configure the wizard steps/screens

sub add_step {
    my ($this,$name,$show,$do) = @_;
    if (ref $show ne "CODE") { return 0; }
#    if (ref $show ne "CODE" || ref $do ne "CODE") { return 0; }
    push @{$this->{step_name}}, $name;
    push @{$this->{show_step}}, $show;
    if (ref $do eq "CODE") { push @{$this->{do_step}}, $do; }
    return 1;
}

# execute the wizard

sub go {
    my ($this,$init) = @_;
    $this->{db}->get_data;
    my $input = $this->{input} = $this->{db}->{data};
    my $out; # = &ShowHash($input);

    $this->{current} = $input->{screen};
    
    # javascript resubmit kludge: if the wizard tries to reconfigure screen 1
    # and reload the form, the inputs are lost because the form re-initializes.
    # If we detect this condition, flag it, so that we can pass the form
    # data instead of the initialization data to the start screen.
    my $resubmit = ($this->{current} == -1);

    if ($this->{db}->{submit_action} eq "Cancel") { 
	return $this->abort; 
    }
    elsif ($this->{current} =~ /^\d+$/) {
	my @do_step = @{$this->{do_step}};
	# validate input
	my @error = ();
	if (defined $this->{do_step}[$this->{current}]) {
	    @error = &{$this->{do_step}[$this->{current}]}($this,$input);
	    if (@error > 0) {
		my $message = "<H3>Problem:</H3>\n<P CLASS=error>\n";
		foreach my $err (@error) {
		    $message .= "$err<br>";
		}
		$message .= "</P><P><FORM><INPUT TYPE=button VALUE=\"&lt; Back\" ONCLICK=\"history.back()\"></FORM></P>\n";
		return $message;
	    }
	}
	if (@error == 0) {
	    # goto next screen
	    #if ($this->{current} == scalar @{$this->{show_step}} - 1) {
	    if ($this->{current} >= scalar @{$this->{show_step}} - 1) {
		# no more screens!
		my $reply = $this->{reply} || $input->{reply};
		if ($reply && (exists $share{PageBuf} || exists $share{Page})) {
		    # we can safely do a redirect
		    my $p = $share{PageBuf} || $share{Page};
		    $p->location($this->{reply});
		}
		else {
		    # safe redirect uncertain, use a manual link
		    $out = "<H2>$this->{title} Complete!</H2>";
		    if ($this->{reply}) {
			$out .= "<p><a href=$this->{reply}>Continue</a> &gt;</p>\n";
		    }
		}
		return $out;
	    }
	    else {
		$this->{current} ++;
		$out .= &{$this->{show_step}[$this->{current}]}($this,$input);
		if (! $this->{do_step}[$this->{current}]) { 
		    $this->{terminate} = 1;
		}
		$out = $this->start_wizard_form.$out.$this->end_wizard_form;
	    }
	}
    }
    else {
	# start the wizard sequence
	$this->{current} = 0;
	if ($resubmit) {
	    # rebuild screen 1 based on screen 1 inputs
	    $out .= &{$this->{show_step}[$this->{current}]}($this,$input)
	}
	else {
	    $out .= &{$this->{show_step}[$this->{current}]}($this,$init);
	}
	$out = $this->start_wizard_form.$out;
	$out .= $this->end_wizard_form;
    }
    return $out;
}

# build standard HTML for start of wizard form

sub start_wizard_form {
    my $this = shift;
    my $step = $this->{current} + 1;
    my $out;
    if ($this->{title}) { $out .= "<H2>$this->{title}</H2>\n"; }
    my $method = "POST";
    $out .= $this->index;
    $out .= "<H3>Step $step: $this->{step_name}[$this->{current}]</H3>\n";
    if (! $this->{terminate}) {
	$out .= "<FORM";
	$out .= $this->{action} ? " ACTION=$this->{action}" : "";
	$out .= " METHOD=$method";
	if ($this->{db}->{htmleditor}) {
	    $out .= " onSubmit=\"validate()\"";
	    $out = $this->{db}->editor_javascript.$out;
	}
	if ($this->{db}->{enctype}) {
	    $out .= " enctype=\"$this->{db}->{enctype}\"";
	}
	$out .= ">\n<div class=\"form\">\n";
	$out .= $this->{db}->{htmleditor} ? "<input type=\"hidden\" name=\"$this->{db}->{htmleditor}\">\n" : "";
	$out .= $this->include_input;
    }
    return $out;
}

# build standard HTML for close of wizard form

sub end_wizard_form {
    my $this = shift;
    my $out;
    if (! $this->{terminate}) {
	$out = "</div>\n<INPUT TYPE=HIDDEN NAME=screen VALUE=$this->{current}>\n";
	if ($this->{current} > 0) {
	    $out .= "<INPUT TYPE=button VALUE=\"&lt; Back\" ONCLICK=\"history.back()\">&nbsp;&nbsp;";
	}
	my $next = ($this->{current} == scalar @{$this->{show_step}} - 1 ? 
		    "FINISH" : "NEXT &gt;");
	$out .= "<INPUT TYPE=submit NAME=submit_action VALUE=\"$next\">&nbsp;&nbsp;<INPUT TYPE=submit NAME=submit_action VALUE=\"Cancel\">&nbsp;&nbsp;\n<INPUT TYPE=reset>&nbsp;&nbsp;\n</FORM>\n";
    }
    return $out;
}

# abort wizard 
# (does nothing, really, except print a message instead of wizard form)

sub abort {
    my $this = shift;
    my $url = &relink();
    $this->{terminate} = 1;
    return "<H3>$this->{title} Aborted</H3>\n<p><a href=\"$url\">Back to start of wizard</a>.</p>\n";
}

# terminate wizard 
# blocks the wizard buttons, effectively terminating progression of the
# wizard.  The wizard control routines determine what the screen text
# will say.

sub terminate {
    my $this = shift;
    $this->{terminate} = 1;
    return;
}

# show a little index of where we are in the wizard sequence

sub index {
    my $this = shift;
    my $out = $this->{index} ? "<table align=\"$this->{index}\" class=\"wizard\"><tr><td>\n" : "";
    $out .= "<OL>\n";
    for (my $istep = 0; $istep < @{$this->{step_name}}; $istep++) {
	if ($istep == $this->{current}) { 
	    $out .= "<li><span style=\"background-color:yellow; font-weight:bold;\">$this->{step_name}[$istep]</span></li>\n"; 
	}
	else {
	    $out .= "<li>$this->{step_name}[$istep]</li>\n"; 
	}
    }
    $out .= "</OL>\n";
    $out .= $this->{index} ? "</td></tr></table>\n" : "";
    return $out;
}

# include_input : grab input from previous step, and insert it as hidden

sub include_input {
    my $this = shift;
    my $input = $this->{input};
    delete $input->{screen};
    delete $input->{submit_action};
    my $out;
    while (my ($key, $val) = each %$input) {
	$val = &html_escape($val,'&"');
	#$val = &html_escape($val,'&"',1);  # borks img & dcd tags
	#$val = &url_escape($val,'&<>%');
	$out .= "<INPUT TYPE=HIDDEN NAME=\"$key\" VALUE=\"$val\">\n";
    }
    return $out;
}

# should probably call this before final form processing...
# will act on a passed input hash, or the share input hash

sub unescape_input {
    my $this = shift;
    my $input = shift;
    if (! $input) { $input = $share{input}{post}; }
    foreach my $key (keys %$input) {
       $input->{$key} = &html_unescape($input->{$key},'&"',1);
       #$input->{$key} = &url_unescape($input->{$key});
    }
    return;
}

1;

