package Modules::BaseDCD;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------

=pod

BaseDCD : a base Dynamic Content Driver

BaseDCD gives a simple framework on which all Dynamic Content Drivers
should be based.  It may be inherited by any Dynamic Content Driver that
wishes to pick up its generic methods as a starting point for
development.

The driver framework is adapted from the Unix device driver API, which
is a simple I/O framework that is easy to understand, and which will
already be familiar to many programmers.

There are 3 standard methods used by all drivers:
    read: acquire data from an external source (such as the URL, stdin 
          [form input], a session manager, or database)
    write: output data for insertion into a web page
    ioctl: all operations that do not constitute either reads or writes
          (eg. user authentication, updating a log file, sending email)

Although DCDs may define other methods for internal use, only the above
three methods can be invoked directly by means of CMS tags.  The above
three methods may be thought of as the public driver methods.

The BaseDCD defines a set of generic DCD methods that can be
inherited by more advanced DCDs.  Generally speaking, the new()
constructor method is generic and may be reused by all DCDs.  The
read() method is fairly versatile, and should be adequate for parsing
most URLs and form input.  The write() method will almost certainly
need to be replaced with something appropriate to the DCD, as it is
the workhorse method in any CMS environment.  The ioctl() method,
while nominally useful here, serves more as an example of ioctl()
programming than as a useful piece of code.

In addition, the BaseDCD class defines a small set of utility functions 
that have proven to be generally useful for Plug-in development.

=cut

use strict;
use ExSite::Config qw(%config %share &read_exsite_conf);
use ExSite::Misc;
use ExSite::Input;

# inherit from Base class
use ExSite::Base;
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

# new : called implicitly once per page when a DCM is invoked

sub new {
    my $this = shift;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    # read the module configuration file
    $obj->read_conf;

    # automatically read the default input source
    $obj->read;

    return $obj;
}

# read : called automatically by default new() to fetch input data
#        can optionally be invoked using <!--&Module.read(options)-->

sub read {
    my ($this,$opt) = @_;
    # fetch overlayed query and form input
    my $in = new ExSite::Input;
    $this->{input} = $in->combine;
}

# write : the default DCD call, invoked using <!--&BaseDCD(options)-->
#         or <!--&BaseDCD.write(options)-->  (these are synonymous).

sub write {
    my ($this,$options) = @_;

    # BaseDCD::write will almost certainly be overloaded by another
    # DCD, but we'll make it do something useful nonetheless:
    # report on the system configuration.

    my $out = "<h3>Exsite System Configuration</h3>
<ul>
<li> ExSite WebWare Version: ${ExSite::Config::VERSION}</li>
<li> Installation name: $config{site}{name}</li>
<li> Server: $config{server}{server}</li>
<li> Base HTdocs Directory: $config{server}{HTMLpath}</li>
<li> Base CGI Directory: $config{server}{CGIpath}</li>
<li> Administrator: $config{site}{administrator}</li>
<li> Configuration: ".($config{server}{auto} ? "automatic" : "manual")."</li>
</ul>

<h4>Plug-in Web Applications:</h4>

<ul>\n";
    require ExSite::ContentBase;
    my $c = $share{Page} || new ExSite::ContentBase;
    foreach my $dcd ($c->get_dcd_all) {
	$out .= "<li>$dcd</li>\n";
    }
    $out .= "</ul>\n";
    return $out;
}

# ioctl : all operations that don't qualify as reads or writes
#         invoked as <!--BaseDCD.ioctl(request)-->

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    # dummy routine - do nothing
    return 0;
}

# check for a config file, and load it if one exists and hasn't already
# been loaded.  Module name can optionally be forced;  this is useful 
# for modules that run as special cases of other modules, and share the
# other module's configuration.

sub read_conf {
    my ($this,$module) = @_;
    if (!$module) { $module = $this->module_name; }
    my $conffile = "./conf/${module}.conf";
    if (! exists $config{$module} && -e $conffile) {
	&read_exsite_conf($conffile,$module);
    }
}

# link - relink to DCD, with different args
#
# In simple case, modifies the query string args of the current URL
#
# For services (DCDs that are served by particular pages), it may also 
# change the page ID.
#
# For AJAX calls into a DCD, it may generate AJAX javascript calls 
# instead of a direct link.
#
# Whatever method is in effect, the output can be used as an anchor HREF.

sub link {
    my ($this,%opt) = @_;
    if ($this->ioctl("isService")) {
	my @mod = split /::/, ref $this;
	my $page = $share{Page};
	my $site;
	if ($page) {
	    $site = $page->id("section");
	}
	if ($site) {
	    # call fetch_match via page to get benefits of CMS caching
	    my @service = $page->fetch_match("service",{name=>$mod[-1],
							section_id=>$site,
						    });
	    foreach my $s (@service) {
		if ($s->{page_id}) {
		    $opt{_id} = $s->{page_id};
		    last;
		}
	    }
	}
	return ExSite::Misc::relink(%opt);
    }
    elsif ($share{ajax_target}) {
	# NB: must load _ExSite/js/httprequest.js for this to work
	my $url = ExSite::Misc::relink(%opt);
	my $id = $share{ajax_target};
	return "javascript:subRequestData('$id','$url')";
    }
    return ExSite::Misc::relink(%opt);
}

# relink - relink to page, with different args
# may have a namespace clash with Misc::relink - use link instead
#sub relink {
#    my $this = shift;
#    return ExSite::Misc::relink(@_);
#}

# icon - return a URL to the icon for this DCD

sub icon {
    my $this = shift;
    my $name = (split /::/, ref $this)[-1];
    my $icon = "$config{server}{HTMLpath}/_Modules/$name/icon.gif";
    if (! -e "$config{server}{HTMLroot}$icon") {
	$icon =~ s/\.gif$/\.png/;
    }
    if (! -e "$config{server}{HTMLroot}$icon") {
	# no icon; use default ExSite app icon
	$icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
    }
    return $icon;
}

# return this module's name

sub module_name {
    my $this = shift;
    my $class = ref($this) || $this;
    my (undef,$name) = split /::/, $class;
    return $name;
}

# allow this DCD to run

sub allow {
    my $this = shift;
    if ($this->ioctl('isRestricted')) {
	my (@sites,$nallow,$db);
	$db = $share{DB};
	if ($share{Page}) {
	    @sites = ($share{Page}->get("section"));
	}
	else {
	    @sites = $db->my_real_sites;
	}
	foreach my $site (@sites) {
	    $nallow += $db->count("service",
				  { section_id => $site->{section_id},
				    name => $this->module_name,
				});
	}
	return ($nallow > 0);
    }
    return 1;
}

# return the section/site ID that we belong to

sub get_section_id {
    my $this = shift;
    if ($share{Page}) {
	return $share{Page}->get("page","section_id");
    }
    else {
	# No page!  We're probably in a control panel...
	my $db = $share{DB};
	my @sites = $db->my_real_sites;
	my $in = $this->{input};
	if ($in->{section_id}) {
	    # a section ID has been explicitly passed, but we need to validate
	    foreach my $site (@sites) {
		if ($in->{section_id} == $site->{section_id}) {
		    # looks good
		    return $in->{section_id};
		}
	    }
	}
	if (@sites == 1) {
	    return $sites[0]{section_id};
	}
    }
    return undef;
}

# Set the section/site ID that we belong to.  This is useful in some
# control panels when it isn't obvious which site we are attached to, and
# the web application requires site-level data privacy.

sub set_section_id {
    my $this = shift;
    my $out;
    if (! $share{Page}) {
	# only output a site selector on control panels
	my @sites = $share{DB}->my_real_sites;
	$out .= "<p><form>Select website: <select name=\"section_id\" onchange=\"document.forms[0].submit()\">\n";
	foreach my $s (@sites) {
	    $out .= "<option value=\"$s->{section_id}\">$s->{title}</option>\n";
	}
	$out .= "</select><input type=\"submit\" value=\"Submit\"></form></p>\n";
    }
    return $out;
}

1;

