package Modules::DMenu;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------
# Modules::DMenu
# Author: Morgan Burke
# (c) 2006 Exware Solutions
#-----------------------------------------------------------------------

=pod

=head1 Modules::DMenu - generic DHTML menu-builder for ExSite

DMenu is a multi-purpose DHTML menu-builder for ExSite websites.  It
will create context-specific menus on any web page in a site.
Context-specific means that the site navigation links will reflect the
current location in the site.

DMenu builds horizontal menubar-style menus.  The menubar items
correspond to top-level pages on a website.  If a page has a submenu
(ie. child pages), then DMenu will pop-up a drop-down menu of the
child pages to select from.

=head2 Usage

Usage is very simple.  Place the DMenu plug-in at the point on the page
at which the menubar should be inserted.  No special options/parameters
are required.  The plain CMS tag will look like this:

    <!--&DMenu()-->

DMenu will create a table with 100% width for the menubar, so the menubar
will span the region into which it was placed.  DMenu will also insert
the necessary Javascript to work the DHTML menu effects.  Javascript must 
be enabled to view the pop-up DHTML menus.  Even without Javascript, the
main menubar links will still work.

=head2 Customizing Menu Appearance

The menubar is constructed as a table, with width of 100%.  The table, 
cells (TD elements) and menubar links are all of class "menubar".

The drop-down menus are encapsulatd in DIV elements of class "DMenu",
and the links are of class "menu".

If any of the links in the menubar or drop-down menus corresponds to the
page being viewed, that link is given an ID of "thispage".

A template that uses DMenu to generate its menus must include style rules
for these elements in order to get the full appearance of a menubar and
drop-downs.  Without any special style rules, only the text menu labels
will be visible, whereas you typically want a distinct menubar, framed
pop-up menus, and rollover effects.  

=head2 Example Stylesheet

Here is a simple CSS example showing how you can obtain basic highlighting
and rollover effects.  The example uses a simple grey color
scheme, but can in principle be extended with borders, image backgrounds, 
and other effects.

    /* give the menubar a solid color */
    td.menubar {
        background-color:#cccccc;
        padding:0px;
    }
    /* style the menubar links */
    a.menubar { 
        font-family:arial;
        font-size:10pt;
        text-decoration:none;
        padding:3px;
        color:#666666;
    }
    /* style the drop-down menu links */
    a.menu { 
        font-family:arial;
        font-size:10pt;
        text-decoration:none;
        padding:3px;
        color:#666666;
    }
    /* rollover effects for menubar and menu */
    a.menu:hover { 
	background-color:#eeeeee;
        color:#000000;
    }
    a.menubar:hover { 
        background-color:#eeeeee;
        color:#000000;
    }
    /* the drop-down menu box */
    .DMenu {
        background-color:#dddddd;
    }


=cut

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Page;
use ExSite::Input;
use ExSite::Tree;
# inherit from BaseDCD class
use Modules::BaseDCD;
use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

sub write {
    my ($this, $opt) = @_;

    # setup 

    my $Page = $share{Page};
    return if (!$Page); 
    my $sid = $Page->id("section");
    if ($Page->is_page()) {
	my $version = $Page->get("page","version");
	my $s = new ExSite::Section(id=>$sid);
	$this->{map} = $s->get_mapped_pages(0,{version=>$version});
    }
    else {
	# probably a template; make a fake map
	$this->{map} = new ExSite::Tree("page_id","parent_id",
	      ({page_id=>1,parent_id=>0,label=>"Home",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>2,parent_id=>0,label=>"About",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>3,parent_id=>0,label=>"Contact",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>4,parent_id=>0,label=>"Links",access=>0,url=>"#",section_id=>$sid},
	       ));
    }

    # generate a menu

    return $this->menu($opt);
}

sub ioctl {
    my $this = shift;
    $_ = shift;
    if (/ModuleInfo/) {
	return "DMenu is a multi-purpose menu-building system.
It automatically assembles site menus and site maps by inspecting the
relationships between your webpages.  It displays the top-level pages
in a horizontal \"menu bar\".  Secondary pages are displayed in drop-down
menus that appear when you mouse over the menu bar.";
    }
    return undef;
}

sub menu {
    my $this = shift;
    my $opt = shift;

    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $pdata = $Page->get;

    my ($menubar,$menus);

    my @items;

    my $access = $db->authorize;
    my $parent = $Page->parent;
    my $ppid = $parent->id();

    # check if we need member-only links
    my $member_links;
    if ($db) {
	if ($Page) {
	    if (! $Page->{publish}) {
		my $level = $db->authorize;
		if ($level) {
		    # set this to the number of special member links
		    $member_links = 1;
		}
	    }
	}
    }

    # get the menubar links
    my @map = $this->{map}->get_topnodes_data();

    # how wide is a given menubar item?
    my $menubar_item_width = 100 / ($#map + $member_links + 1);

    $menubar = "

<!-- begin DMenu plug-in -->

<script type='text/javascript' src='$config{server}{HTMLpath}/_ExSite/js/menu_exsite.js'></script>
<script type='text/javascript' src='$config{server}{HTMLpath}/_ExSite/js/cbe_core.js'></script>
<script type='text/javascript' src='$config{server}{HTMLpath}/_ExSite/js/cbe_event.js'></script>
<script>
function windowOnload(){
    document.cbe.addEventListener('mousemove', menuHideListener, false);
}
</script>

<!-- 
This is the bare essential CSS for DMenu; designers can add their
own styles in the template stylesheets.
-->

<style>
a.menubar { 
    display:block; 
    width:100%; 
}
a.menu { 
    display:block; 
}
.DMenu {
    visibility:hidden;
    position:absolute;
    display:block;
    color:black;
}
</style>

<table width='100%' cellspacing='0' cellpadding='0' class='menubar'><tr>\n";
    my $n_bar_item = 0;
    while (@map > 0) {
	my $p = shift @map;
	my $menuPage = new ExSite::Page(page=>$p);
	my $url = $menuPage->get_url;
	my $id = $menuPage->id();
	my @submap = $this->{map}->get_child_data($id);

	if (@submap == 0) {
	    # no child pages
	    $menubar .= "<td class='menubar' width='$menubar_item_width\%'><a class='menubar' href='$url'";
	    if ($id == $this_page) {
		$menubar .= " id='thispage'";
	    }
	    $menubar .= ">$p->{label}</a></td>\n";
	}
	else {
	    # need submenu pop-up tags
	    $menubar .= "<td class='menubar' width='$menubar_item_width\%'><div id='menu$n_bar_item' onmouseover=Menu($n_bar_item)><a class='menubar' href='$url'";
	    if ($id == $this_page) {
		$menubar .= " id='thispage'";
	    }
	    $menubar .= ">$p->{label}</a></div></td>\n";
	    
	    # need a submenu

	    $menus .= "\n<div id='box$n_bar_item' class='DMenu'>
<table width='100%' cellspacing='0' cellpadding='0' border='0'><tr><td>\n";
	    foreach my $sub (@submap) {
		my $submenuPage = new ExSite::Page(page=>$sub);
		my $url = $submenuPage->get_url;
		my $id = $submenuPage->id();
		$menus .= "<a class='menu' href='$url'";
		if ($id == $this_page) {
		    $menus .= " id='thispage'";
		}
		$menus .= ">$sub->{label}</a>\n";
	    }
	    $menus .= "</td></tr></table>\n</div>\n";
	}
        $n_bar_item++;
    }

    # optional links for members only
    if ($member_links) {
        $menubar .= "<td class='menubar' width='$menubar_item_width\%'><a class='menubar' href='$config{server}{CGIpath}/logout.cgi'>Log out</td>\n";
    }

    $menubar .= "</div></td></tr></tr></table>
<div id='areaTag' class='clsCBE' style='display:none'></div>\n";

    return "$menubar\n$menus\n<!-- end DMenu plug-in -->\n\n";
}

1;
