package Modules::Document;

# Document : a simple document manager

=pod

=head1 Modules::Document - document management web applet

The Document Tool provides a simple document manager that easily integrates
into ExSite-driven websites.  It has the following features:

=over 4

=item * paginated reports of documents that are available for download

=item * title, description, filename, type, date, size are reported 

=item * sorted by date

=item * add, update, delete functions for administrators

=item * versioning (using ExSite's CMS tools)

=item * will create the document library if it doesn't exist

=item * can use alternate documemt libraries

=item * webservice-ready

=back

=cut

###  TODO:
###  
###  - search
###  - built-in publish function
###  - shortcut to CMS
###  - paginated control panel

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Content;
use Modules::BaseDCD;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

sub write {
    my ($this,$options) = @_;
    my $in = $this->{input};  # set by BaseDCD::read
    my %opt = &DecodeString($options);
    $this->{libname} = $opt{lib} || "docs";
    my $action = $opt{doccmd} || $in->{doccmd} || "list";
    #my $id = $in->{doc};
    if ($action eq "add") {
	return $this->add;
    }
    elsif ($action eq "update") {
	return $this->update($in->{docid});
    }
    elsif ($action eq "upload") {
	return $this->upload;
    }
    return $this->list;
}

# ioctl : all operations that don't qualify as reads or writes
#         invoked as <!--BaseDCD.ioctl(request)-->

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    if (/ModuleName/) {
	return "Documents";
    }
    elsif (/ControlPanel/) {
	return \&ctrl_panel;
    }
}

sub ctrl_panel {
    my $this = shift;
    my $db = $share{DB};
    my $in = $this->{input};
    $this->{admin} = 1;
    my $out;

    $this->{section_id} = $this->get_section_id;
    if (! $this->{section_id}) {
	$out .= "<h3>Select website:</h3>\n";
	$out .= $this->set_section_id;
	return $out;
    }

    $this->{libname} = $in->{lib} || "docs";
    my $lib = $this->getlib;

    if (! $lib) {
	# hmm, no library
	if ($in->{doccmd} eq "makelib") {
	    # create one
	    my $rlib = {section_id=>$this->{section_id},
			title=>"Documents",
			filename=>$this->{libname},
			type=>"library",
			status=>"active",
			access=>"public",};
	    my $lid = $db->insert("page",$rlib);
	    if ($lid) {
		$lib = $this->{lib} = $rlib;
	    }
	    else {
		return $db->error("failed to create document library");
	    }
	}
	else {
	    # ask if we should create one
	    my $makeurl = &relink(doccmd=>"makelib");
	    my $section = $db->fetch("section",$this->{section_id});
	    $out .= "<h3 class=\"error\">Document library \"$this->{libname}\" not found in $section->{title}.</h3>\n";
	    $out .= "<p><a href=\"$makeurl\">Click here to make document library</a>.</p>\n";
	    return $out;
	}
    }

    # options

    if ($in->{doccmd} eq "update") {
	return $this->update($in->{docid});
    }
    elsif ($in->{doccmd} eq "add") {
	return $this->add();
    }
    elsif ($in->{doccmd} eq "del") {
	return $this->delete($in->{docid});
    }
    elsif ($in->{doccmd} eq "upload") {
	return $this->upload();
    }

    # show the library contents, and administrator options

    return $this->list;
}

# get document library

sub getlib {
    my ($this) = @_;
    return $this->{lib} if ($this->{lib} && $this->{lib}{filename} eq $this->{input}{lib}); 
    if (! exists $this->{section_id}) {
	$this->{section_id} = $this->get_section_id;
    }
    return undef if (! $this->{section_id});

    my @doclib = $share{DB}->fetch_match("page",{filename=>$this->{libname},
						 type=>"library",
						 section_id=>$this->{section_id}});
    if (@doclib > 0) {
	return ($this->{lib} = $doclib[0]);
    }
    return undef;
}

# list documents by date, in groups of 10

sub list {
    my $this = shift;
    my $db = $share{DB};
    my $in = $this->{input};
    my $lib = $this->getlib;
    return "<p class=\"error\">No document library found.</p>\n" if (!$lib);

    my $out;
    my @content_records = reverse $db->fetch_child("content","page",
						   $lib->{page_id},"content_id");

    my @content;
    my $nrec = scalar @content_records;
    my $start = (defined $in->{st} && $in->{st} < $nrec) ? $in->{st} : 0;
    my $step = $in->{step} || 10;
    my $end = $start + ($step - 1);
    if ($end > $nrec) { $end = $nrec - 1; }

    $out .= "<dl>\n";
    foreach my $crec (@content_records[$start..$end]) {
	my $c = new ExSite::Content(content=>$crec,version=>"newest");
	$c->set_context();
	my $url = $c->get_url;
	my ($file,$size,$mime) = $c->get_fileinfo;
	my $time = $db->show_data("content_data","ctime",$c->timestamp);
	my $descr = $db->show_data("content","description",$c->get("content","description"));
	if ($descr) { $descr .= "<br>\n"; }
	my $info;
	if ($file) {
	    $info = "\n<i>$file ($mime, $size K) posted on $time</i>";
	}
	if ($c->has_content()) {
	    $out .= "<dt><a href=\"$url\">".
		$db->show_data("content","name",$c->get("content","name")).
		"</a></dt>\n<dd>$descr$info\n";
	}
	else {
	    $out .= "<dt>".
		$db->show_data("content","name",$c->get("content","name")).
		"</dt>\n<dd>$descr (no file available for download)\n";
	}
	if ($this->{admin} && $db->owns($this->{section_id})) {
	    my $update_url = &relink(doccmd=>"update",docid=>$crec->{content_id});
	    $out .= "<a href=\"$update_url\"><img src=\"$config{server}{HTMLpath}/_Modules/Document/update.gif\" height=\"12\" width=\"12\" border=\"0\" alt=\"update\" title=\"update\"></a>\n";
	    my $del_url = &relink(doccmd=>"del",docid=>$crec->{content_id});
	    $out .= "<a href=\"$del_url\"><img src=\"$config{server}{HTMLpath}/_Modules/Document/del.gif\" height=\"12\" width=\"12\" border=\"0\" alt=\"delete\" title=\"delete\"></a>\n";
	}
	$out .= "</dd>\n";
    }
    $out .= "</dl>\n";

    if ($end - $start < 0) {
	$out .= "<p>No documents found.</p>\n";
    }
    else {
	if ($end < $nrec - 1) {
	    my $url = &relink(st=>$end+1);
	    $out .= "<p><a href=\"$url\">More documents...</a></p>\n";
	}
	if ($end < $nrec - 1 || $start > 0) {
	    my $url = &relink(st=>0,step=>$nrec);
	    $out .= "<p><a href=\"$url\">All $nrec documents...</a></p>\n";
	}
	if ($start > 0) {
	    $out .= "<p><a href=\"javascript:history.back()\"><< Back</a></p>\n";
	}
    }

    # admin links

    if ($this->{admin} && $db->owns($this->{section_id})) {
	# current user owns this site, and is viewing the page dynamically;
	# give them some admin links
	my $url = &relink(doccmd=>"add");
	my $puburl = "$config{server}{CGIpath}/publish.cgi?page=$this->{lib}{page_id}";
	$out .= "<a href=\"$url\"><img src=\"$config{server}{HTMLpath}/_Modules/Document/new.gif\" height=\"12\" width=\"12\" border=\"0\" alt=\"new doc\" title=\"new doc\"> add new document</a><br>
<a href=\"$puburl\" target=\"publish\"><img src=\"$config{server}{HTMLpath}/_Modules/Document/publish.gif\" height=\"12\" width=\"12\" border=\"0\" alt=\"publish docs\" title=\"publish docs\"> publish all documents</a><br>\n";
    }
    
    return $out;
}

# add a new document

sub add {
    my $this = shift;
    my $url = &relink(doccmd=>"upload");
    my $out = "<h2>Upload new document</h2>
<form method=\"POST\" action=\"$url\" enctype=\"multipart/form-data\">
<p>Choose document:<br>
<input type=\"file\" name=\"upload\"></p>
<p>Title:<br>
<input type=\"text\" name=\"title\" size=\"40\"></p>
<p>Description:<br>
<textarea name=\"descr\" rows=\"4\" cols=\"60\"></textarea></p>
<p><input type=\"submit\" value=\"Upload\"></p>
</form>\n";
    return $out;
}

# update an existing document

sub update {
    my ($this,$id) = @_;
    my $doc;
    my $db = $share{DB};
    if ($id) {
	$doc = $db->fetch("content",$id);
	if ($doc) {
	    my $url = &relink(doccmd=>"upload");
	    my $name = $db->show_data("content","name",$doc->{name});
	    my $out = "<h2>Upload new version of $name</h2>
<form method=\"POST\" action=\"$url\" enctype=\"multipart/form-data\">
<p>Choose document:<br>
<input type=\"file\" name=\"upload\"></p>
<input type=\"hidden\" name=\"docid\" value=\"$id\">
<p><input type=\"submit\" value=\"Upload\"></p>
</form>\n";
	    return $out;
	}
	else {
	    return $this->error("Invalid document ID");
	}
    }
    else {
	return $this->error("No document to update");
    }
}

# update an existing document

sub upload {
    my $this = shift;
    my $in = $this->{input};
    my $cid;
    my $lib = $this->getlib;
    if (! $lib) {
	return $this->error("No document library found.");
    }
    
    if ($in->{upload}) {
	if (exists $in->{title}) {
	    # make content object
	    if (! $in->{title}) {
		my ($file,undef) = split/$config{form}{sepchar}/,$in->{upload};
		$in->{title} = $file;
	    }
	    $cid = $share{DB}->insert("content",{page_id=>$lib->{page_id},
						 name=>$in->{title},
						 description=>$in->{descr},
						 type=>"editorial",
					     }
				      );
	}
	else {
	    if ($in->{docid}) { 
		my %c = $share{DB}->fetch("content",$in->{docid});
		if ($c{page_id} == $lib->{page_id}) {
		    $cid = $in->{docid};
		}
		else {
		    # somebody's trying to fool us
		    return $this->error("Invalid document ID.");
		}
	    }
	    else {
		# should never happen
		return $this->error("No document to update!");
	    }
	}
	
	if ($cid) {
	    my $comment = "uploaded on ".(scalar localtime);
	    my $file = $in->{upload};
	    my $fdata = undef;
	    my $size = 0;
	    my $mimedata = "";
	    if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
		# read file contents
		my $continue = 1;
		while ($continue) {
		    $continue = read($in->{upload},$fdata,1024,$size);
		    $size += $continue;
		}
		# MIME-encode the file contents
		$mimedata = encode_base64($fdata,"");
	    }
	    # strip path from filename
	    my @path = split /[:\\\/]/, $file;
	    my $filename;
	    if ($path[-1]) { 
		$filename = $path[-1];
		$filename =~ s/ /_/g;                # conv space to _
		$filename = &url_escape($filename);  # escape other chars
	    }
	    if ($share{DB}->insert("content_data",{content_id=>$cid,
						   mime_type=>&MimeType($filename),
						   note=>$comment,
						   fdata=>"$filename$config{form}{sepchar}$mimedata",
					       })) {
		my $home = &relink(doccmd=>undef,docid=>undef);
		return "<h3>$file uploaded successfully</h3>
<p><a href=\"$home\">Back to document library</a>.</p>\n";
	    }
	    else {
		return $this->error("Error saving uploaded file.  It could be larger than the maximum allowed size.");
	    }
	}
	else {
	    return $this->error("Error configuring uploaded file.");
	}
    }
    else {
	return $this->error("Nothing to upload!");
    }
}

# update an existing document

sub delete {
    my $this = shift;
    my $db = $share{DB};
    my $in = $this->{input};
    my $cid;
    my $lib = $this->getlib;
    if (! $lib) {
	return $this->error("No document library found.");
    }

    if ($in->{confirm}) {
	if ($in->{docid}) {
	    # validate
	    my $doc = $db->fetch("content",$in->{docid});
	    if ($doc->{page_id} eq $lib->{page_id}) {
		# okay to proceed
		# use recursive delete to remove all versions
		
		### FIXME: published version remains on disk
		
		$db->delete_r("content",$in->{docid},1);
		my $home = &relink(doccmd=>undef,docid=>undef,confirm=>undef);
		return "<h3>\"$doc->{name}\" deleted</h3>
<p><a href=\"$home\">Back to documents control panel</a>.</p>\n";
	    }
	    else {
		return $this->error("Invalid document ID.");
	    }
	}
	else {
	    return $this->error("No document was specified.");
	}
    }
    else {
	my $url = &relink(confirm=>1);
	my $doc = $db->fetch("content",$in->{docid});
	return "<h3>Confirm deletion of \"$doc->{name}\"</h3>
<ul>
<li><a href=\"$url\">Yes, delete this document and all its versions</a>.</li>
<li><a href=\"javascript:history.back()\">NO, I changed my mind.</li>
</ul>\n";
    }
}

1;

