package Modules::Search;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------
#
#  The search tool builds a search index using 2 tables:
#  - searchurl: every URL that is searched is represented once here
#  - searchterm: every indexed word at each url is represented once here,
#                with a weight (based on location and number of ocurrences)
#
#  The search index is constructed by stripping out all extraneous tags and
#  text, and counting the incidence of each remaining word in the body of 
#  the html at each URL.  Words in filenames, titles, descriptions, and 
#  keywords are counted and weighted separately.
#
#  This version searches only the default representation of each page, 
#  and does not follow links to recursive instances of the same page.
#
#  The search logic is somewhat crude, and does not allow for searching
#  for partial words, quoted phrases, absense of words, etc.
#
#-----------------------------------------------------------------------
#
#  TODO:
#  - the query logic for looking up search terms could be more efficient
#  - do we need to do a full page build to get the searchable text?
#  - would be nice to exclude menus from the searchable text
#  - follow recursive page links at least one level
#

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Auth;
use ExSite::Page;
use Modules::BaseDCD;
use Digest::MD5;

use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

sub new {
    my $this = shift;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    $obj->read_conf;

    $obj->{weight} = $config{Search}{weight};
    $obj->{title} = {};
    $obj->{descr} = {};
    $obj->{site} = {};

    my @skipwords = split / /,$config{Search}{skipwords};
    $obj->{skipwords} = \@skipwords;

    $obj->read;
    return $obj;
}

sub write {
    my ($this, $opt) = @_;
    my $a = new ExSite::Auth;
    if ($a->license("Search",$this->{input}{section})) {
	my $in = new ExSite::Input;
	my $sdata = $in->post_or_query;
	my $term = $sdata->{searchterm} || $opt;
	if ($term) {
	    return $this->do_search($term);
	}
	else {
	    return $this->searchform($opt);
	}
    }
    return "Search: no license for this web application";
}

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    if (/ControlPanel/) {
	return \&ctrl_panel;
    }
    elsif (/isRestricted/) {
	return 1;
    }
    elsif (/isService/) {
	return 1;
    }
    elsif (/ModuleInfo/) {
	return "Search is a standard website search tool.  It allows you
to build search indexes for your website, and easily search for any set of
terms.  Search results are ranked and reported with the page title and
description.  Search indexes can be manually edited by experts, if you 
want to boost or lower the relevance of certain terms on certain pages.";
    }
    return;
}

sub ctrl_panel {
    my $this = shift;
    my $in = $this->{input};
    my $out;
    my $donewrite;
    if ($share{DB}->authorize < $config{auth}{root_auth_level} - 1) {
	return "Permission denied.";
    }
    if ($in->{action} eq "search") {
	$out .= $this->write;
	$donewrite = 1;
    }
#    elsif ($in->{action} eq "build") {
#	$out .= $this->scanall;
#    }
    elsif ($in->{action} eq "Build index for selected site") {
	if ($in->{section}) {
	    $out .= $this->scansite($in->{section});
	}
    }
    elsif ($in->{action} eq "Clear index for selected site") {
	if ($in->{section}) {
	    $out .= $this->scansite($in->{section});
	}
    }
    elsif ($in->{action} =~ /Manually edit/) {
	if ($in->{section}) {
	    my $p = $share{Page} || $share{PageBuf};
	    $p->location("$config{server}{CGIpath}/webdb.cgi?tab=searchurl&ptab=section&pid=$in->{section}");
	    return undef;
	}
    }

    my $a = new ExSite::Auth;
    my @sites = $a->my_real_sites;
    
    $out .= "<h1>Search Administration</h1>
<p>You have Search enabled on the following sites:</p>
<form>
<select name=\"section\">\n";
    my $nsite = 0;
    foreach my $site (@sites) {
	if ($a->license("Search",$site->{section_id})) {
	    $out .= "<option value=\"$site->{section_id}\">$site->{title}</option>\n";
	    $nsite++;
	}
    }
    if ($nsite == 0) {
	$out .= "<option value=\"0\">No search-enabled websites found!</option>\n";
    }
    $out .= "</select>
<p><input type=\"submit\" name=\"action\" value=\"Build index for selected site\">
<input type=\"submit\" name=\"action\" value=\"Build index for ALL sites\">
<input type=\"submit\" name=\"action\" value=\"Manually edit search index for selected site\"></p>\n</form>\n";
    #if (! $donewrite && $in->{section}) { $out .= $this->write; }
    return $out;
}

sub searchform {
    my ($this,$term) = @_;
    my $url = $this->link(action=>"search");
    my $out;
    if ($config{Search}{showtitle}) {
	$out .= '<h2>Search <!--$site_title--></h2>';
    }
    $out .= "\n<form method=\"POST\" action=\"$url\"><input type=\"text\" width=\"20\" name=\"searchterm\" value=\"$term\"><input type=\"submit\" value=\"Search\"></form>\n";
    return $out;
}

sub do_search {
    my ($this,$term) = @_;
    my $out .= "<h1>Search Results for \"$term\"</h1>\n";
    $out .= $this->display_results($this->search($term));
    $out .= $this->searchform($term);
    return $out;
}

# scanall : indexes every site in the system

sub scanall {
    my ($this) = @_;
    my @sites = $share{DB}->fetch_all("section");
    my $out;
    foreach my $site (@sites) {
	next if ($site->{type} =~ /template/);
	my $site = new ExSite::Section(section=>$site);
	$out .= $this->scanpages($site->get_all_pages(type=>"page",status=>"active"));
	$this->clear($site->{section_id});
    }
    $out .= $this->update;
    return $out;
}

# scanmysites : indexes sites belonging to current user

sub scanmysites {
    my ($this) = @_;
    my $a = new ExSite::Auth;
    my @sites = $a->my_real_sites;
    my $out;
    foreach my $site (@sites) {
	my $site = new ExSite::Section(section=>$site);
	$out .= $this->scanpages($site->get_all_pages(type=>"page",status=>"active"));
	$this->clear($site->{section_id});
    }
    $out .= $this->update;
    return $out;
}

# scansite : indexes every page in a site

sub scansite {
    my ($this,$id) = @_;
    my $site = new ExSite::Section(id=>$id);
    my $out .= $this->scanpages($site->get_all_pages(type=>"page",status=>"active"));
    $this->clear($id);
    $out .= $this->update;
    return $out;
}

sub scanpages {
    my ($this,@pages) = @_;
    my $out;
    foreach my $p (@pages) {
	next if (! $p->is_public);     # do not index private pages
	$out .= $this->indexpage($p);
	if ($config{Search}{scanpause}) { sleep($config{Search}{scanpause}); }
    }
    return $out;
}

# indexpage : index the metadata and content of a page

sub indexpage {
    my ($this,$page) = @_;
    $share{Page} = $page;
    $page->set_context;
    my $url = $page->get_url;
    my $html = $page->expand;
    
    my $body = $html;
    $body =~ /<body>(.+)<\/body>/si;
    $body = $1;
    if (!$body) { $body = $html; }  # in case of poorly structured document
    $body =~ s/<\/?[^>]+>/ /g;       # strip all tags


    my $filename = $page->filename;
    my $title = $page->title;
    my $descr = $page->description;
    
    $this->index($url,$this->{weight}{title},$title);
    $this->index($url,$this->{weight}{keywords},$page->keywords);
    $this->index($url,$this->{weight}{description},$descr);
    $this->index($url,$this->{weight}{body},$body);

    if ($filename !~ /page\.cgi/) { 
	$this->index($url,$this->{weight}{filename},$page->filename);
    }

    $this->{title}{$url} = $title;
    $this->{descr}{$url} = $descr;
    $this->{site}{$url} = $page->id("section");

    return "indexing $url...<br>\n";
}

# index : indexes the words in a block of plaintext

sub index {
    my ($this,$url,$weight,$text) = @_;
    return if (!$text);
    # strip endlines
    $text =~ s/\r?\n/ /g;
    # strip all non-word characters
    $text =~ s/[^\w]/ /g;
    # convert to lower case
    $text = lc($text);
    # convert to an array of words
    my @word = split /\s+/,$text;

    if (! exists $this->{index}{$url}) {
	$this->{index}{$url} = {};
    }

    # index the words found
    foreach my $word (@word) {
	next if (!$word);
	next if (grep /^$word$/,@{$this->{skipwords}});
	$this->{index}{$url}{$word} += $weight;
    }
}

# clear : clear old search data for a particular site

sub clear {
    my ($this,$section_id) = @_;
    my $db = $share{DB};
    if ($section_id) {
	my @oldurl = $db->fetch_match("searchurl",{section_id=>$section_id});
	foreach my $oldurl (@oldurl) {
	    # delete links and urls
	    $this->clear_url($oldurl->{searchurl_id});
	}
    }
}

# clear search data for a single URL

sub clear_url {
    my ($this,$url_id) = @_;
    my $db = $share{DB};
    if ($url_id) {
	$db->delete("searchterm",{searchurl_id=>$url_id});
	$db->delete("searchurl",{searchurl_id=>$url_id});
    }
}

# update : update the DB index

sub update {
    my ($this,$threshold) = @_;
    if (!$threshold) { $threshold = 1; }
    my $db = $share{DB};

    foreach my $url (keys %{$this->{index}}) {

	# clear old search data
	my @oldurl = $db->fetch_match("searchurl",{url=>$url});
	foreach my $oldurl (@oldurl) {
	    $this->clear_url($oldurl->{searchurl_id});
	}
	
	# add new data
	my $url_id = $db->insert("searchurl",{url=>$url,
					      title=>$this->{title}{$url},
					      description=>$this->{descr}{$url},
					      section_id=>$this->{site}{$url},
					      });
	foreach my $word (keys %{$this->{index}{$url}}) {
	    next if (! $word);
	    $db->insert("searchterm",{searchurl_id=>$url_id,
				      term=>$word,
				      weight=>$this->{index}{$url}{$word}});
	}
    }
}

# search the search table

sub search {
    my ($this,$terms) = @_;

    my $db = $share{DB};
    my @term = split /\s/,lc($terms);
    my $nterm = scalar @term;

    # FIXME: this search finds results across all sites in the system.
    # Hits pertaining to the current site are filtered in display_results,
    # but if the number of unrelated sites is large, this will become a
    # drag on memory and performance.  In that case the extra hit of a join 
    # with searchurl may be more appropriate.

    my @termresults;
    foreach my $term (@term) {
	push @termresults, $db->fetch_match("searchterm",{term=>$term});
    }

    # combine the results into a single result hash

    my (@url,%result);
    foreach my $termresult (@termresults) {
	$result{$termresult->{searchurl_id}} += ($termresult->{weight} / $nterm);
    }
    
    return %result;
}

# format the search results 

sub display_results {
    my ($this,%results) = @_;
    my @url_id = sort { $results{$b} <=> $results{$a}; } keys %results;
    my $out = "<dl>\n";
    my $db = $share{DB};
    my $nresult = 0;
    # only display results from this section
    my $section = $share{Page}->id("section");
    foreach my $url_id (@url_id) {
	my $url = $db->fetch("searchurl",$url_id);
	next if ($url->{section_id} != $section);  # wrong site
	$nresult++;
	my $score = int($results{$url_id}*10)/10;
	$out .= "<dt><a href=\"$url->{url}\">$url->{title}</a></dt>\n";
	$out .= "<dd>$url->{description}";
	if ($url->{description}) { $out .= "<br>\n"; }
	$out .= "<font size=1>[$url->{url}, score: $score]</font><p>\n";
	last if ($nresult == $config{Search}{maxresult});
    }
    if ($nresult == 0) {
	$out .= "<dt>Nothing relevant was found.</dt>\n";
    }
    $out .= "</dl>\n";
}

1;
