package Modules::SimpleMenu;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------
# Modules::SimpleMenu
# Author: Morgan Burke
# (c) 2003 Exware Solutions
#-----------------------------------------------------------------------
# generic menu DCM
# possible extensions to SimpleMenu
# - graphical menu buttons
# - publish menu snippets, for faster lookup on dynamic pages
# - control panel for easy menu layout/configuration
#     - manual sorting of page order using DHTML

=pod

=head1 Modules::SimpleMenu - generic menu-builder for ExSite

SimpleMenu is a multi-purpose menu-builder for ExSite websites.  It
will create context-specific menus on any web page in a site.
Context-specific means that the site navigation links will reflect the
current location in the site.

SimpleMenu can also serve as a code template for a customized menu
builder, since it contains a general purpose menu-construnction
algorithm based on the ExSite page placement parameters.

SimpleMenu can build four types of menus:

=over 4

=item vertical 

navigation links are stacked in a vertical list, with submenus
exploded within this list.  This is the default menu type.

=item horizontal 

navigation links are listed left-to-right,
with each link separated by a "|" (bar) character.  Only the top-level
links in the site are shown, and no submenus are displayed.

=item path 

the path from the site home page to the current page
is displayed in the "cookie crumb" format, ie. page1 > page2 > page3

=item sitemap 

the full site structure is shown, with child
pages indented beneath their parents.

=back

When inserting a web application tag into a block of HTML, any of the
above types may be provided as a parameter.

When manually coding a SimpleMenu dynamic content tag, the format is:

C<E<lt>!--&SimpleMenu(format)--E<gt>>

where "format" is one of the above menu types.

=head3 Sub-menus

SimpleMenu builds global menus for the whole website by default.  That
is, the "top" level pages displayed in the menu are those that have no
parent page defined.  Sub-menus show pages that have a common parent page.

SimpleMenu automatically inserts sub-menus into vertically-formatted menus
if the viewer is viewing a top-level parent page that has child pages, or 
is viewing any of those child pages.  It does descend any further 
into the menu hierarchy than this.

SimpleMenu will build a horizontal sub-menu if the "submenu" parameter
is included in the dynamic content tag, eg.

C<E<lt>!--&SimpleMenu(horizontal,submenu)--E<gt>>

In this case, the sub-menu shows the pages that are children of the 
currently-viewed page.  It can show sub-menus to any depth.

=head2 Customizing Menu Appearance

=head3 CSS

The following CSS classes are used, and may be customized in any
given site's stylesheets:

=over 4

=item a.menu

All main menu links in vertically and horizontally-formatted menus are
of the "menu" class.

=item a.submenu

All submenu links in vertically-formatted menus are of the "submenu" class.

=back

=head3 Menu Templates

Vertically-formatted menus will attempt to inline special content objects
to spice up the menu presentation:

=over 4

=item SimpleMenuTop

This content will be placed before all menu links.

=item SimpleMenuMid

This content will be placed between all menu links.

=item SimpleMenuBot

This content will be placed after all menu links.

=back

Using these content objects, a template can incorporate custom table layouts,
border graphics, and other visual tricks to lay out the menu.

=head3 Menu Spacer Strings

In addition to the above method of customizing the menu HTML, there are 
also some spacer strings that are automatically inserted at certain 
points in the menu to delineate the different links.  These are defined
in the SimpleMenu configuration file, which should be located in 
C<cgi/conf/SimpleMenu.conf>:

=over 4

=item hmenu_spacer 

inserted between menu links in a horizontal menu.  Default: C<&nbsp;|&nbsp;>

=item pmenu_spacer

inserted between menu links in a path menu.  Default: C<&nbsp;&gt;&nbsp;>

=item vmenu_spacer

inserted between menu links in a vertical menu.  Default:
C<E<lt>brE<gt>> Note that if you use display:block in your CSS for
menu links, the vmenu_spacer should be left blank to prevent extra
blank lines from being added to your menu layout.

=item submenu_prefix 

inserted in front of submenu links.  Default: C<&nbsp;&nbsp;&nbsp;&middot;&nbsp;>

=back

=cut

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Page;
use ExSite::Input;
use ExSite::Tree;
# inherit from BaseDCD class
use Modules::BaseDCD;
use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

sub write {
    my ($this, $opt) = @_;

    # setup 

    my $Page = $share{Page};
    return if (!$Page); 
    my $sid = $Page->id("section");
    if ($Page->is_page()) {
	my $version = $Page->get("page","version");
	my $s = new ExSite::Section(id=>$sid);
	$this->{map} = $s->get_mapped_pages(0,{version=>$version});
    }
    else {
	# probably a template; make a fake map
	$this->{map} = new ExSite::Tree("page_id","parent_id",
	      ({page_id=>1,parent_id=>0,label=>"Home",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>2,parent_id=>0,label=>"About",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>3,parent_id=>0,label=>"Contact",access=>0,url=>"#",section_id=>$sid},
	       {page_id=>4,parent_id=>0,label=>"Links",access=>0,url=>"#",section_id=>$sid},
	       ));
    }

    # generate a menu

    if ($opt =~ /horizontal/) { return $this->hmenu($opt); }
    if ($opt =~ /path/) { return $this->pmenu($opt); }
    if ($opt =~ /sitemap/) { return $this->sitemap($opt); }
    else { return $this->vmenu($opt); }
}

sub ioctl {
    my $this = shift;
    $_ = shift;
    if (/ModuleInfo/) {
	return "SimpleMenu is a multi-purpose menu-building system.
It automatically assembles site menus and site maps by inspecting the
relationships between your webpages.  It can build vertically-formatted
menus with exploded sub-menus, horizontally-formatted menus, and 
path menus (also known as \"cookie-crumb\" menus) that show you where
you are in the site.  It has options for customizing the formatting and
layout of menus so that they can be incorporated into most graphic designs.
SimpleMenu does not do dynamic pop-up menus, or menus with graphical 
buttons.  These options will require specialized menu plug-ins.";
    }
    elsif (/DynContentParameter/) {
	return "<p><b>Menu Type:</b>
<select name=\"param\">
<option value=\"\">(default)</option>
<option value=\"vertical\">Vertical</option>
<option value=\"horizontal\">Horizontal (top-level pages)</option>
<option value=\"horizontal-submenu\">Horizontal (submenu pages)</option>
<option value=\"path\">Path (cookie-crumb)</option>
<option value=\"sitemap\">Site Map</option>
</select></p>\n";
    }
    return undef;
}

# hmenu - return a horizontal menu (global or child pages)

sub hmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}->{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $version = $Page->get("page","version");
    my $menu_mode = $this->menu_mode;

    my ($menutop, $menumid, $menubot);
    $menutop = $Page->find("SimpleMenuTopH")->get_html() ||
	"<!-- begin menu -->\n<span class=\"menu\">\n";
    $menumid = $Page->find("SimpleMenuMidH")->get_html() ||
	$config{SimpleMenu}{hmenu_spacer};
    $menubot = $Page->find("SimpleMenuBotH")->get_html() || 
	"</span>\n<!-- end menu -->\n";

    # fetch all menu-able pages from this section
    my @pages;
    if ($opt =~ /child|submenu/) {
	# see if there are any submenu pages
	@pages = $this->{map}->get_child_data($this_page);
	if (@pages == 0) {
	    # no submenu;  rebuild the submenu we're already on
	    my $parent = $Page->get("page","parent_id");
	    if ($parent) {
		@pages = $this->{map}->get_child_data($parent);
	    }
	}
    }
    else {
	@pages = $this->{map}->get_topnodes_data();
    }

    my $access = $db->authorize;
    my @menu;
    foreach my $page (@pages) {
	my $url = $page->{url};
	if (!$url) {
	    my $p = new ExSite::Page(page=>$page);
	    $url = $p->get_url($menu_mode);
	}
	push @menu, "<a href=\"$url\" class=\"menu\">$page->{label}</a>";
    }

    # optional links for members only
    if ($access && $opt !~ /child|submenu/) {
	if (! $Page->{publish} && ! $Page->is_template) {
	    push @menu, "<a href=\"$CGIpath/logout.cgi\" class=\"menu\">Logout</a>";
	}
    }

    return $menutop.join($menumid, @menu).$menubot;
}

# pmenu - return a path menu
#
# Note that for pmenu we do not put the content into the menu class,
# because most commonly a pmenu is placed into the page body, not the
# menu bar.

sub pmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}->{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $id = my $this_page = $Page->id;
    my $version = $Page->get("page","version");
    my $pdata = $Page->get;
    my $menu_mode = $this->menu_mode;
    my $purl = $Page->get_url($menu_mode);
    my @path = "<a href=\"$purl\">$pdata->{label}</a>";
    my $out = "<!-- begin menu -->\n";

    my $toppage;
    while (my $page = $this->{map}->get_parent_data($id)) {
	$toppage = $page;
	$Page = new ExSite::Page(page=>$page);
	$purl = $Page->get_url($menu_mode);
	unshift @path, "<a href=\"$purl\">$page->{label}</a>";
	$id = $page->{parent_id};
    }
    if ($toppage->{filename} !~ /^index\./) {
	# also include the index page in the path
	my @p = $this->get_pages({section_id=>$pdata->{section_id},
				  filename=>"index.%",
				  status=>"active",
				  visibility=>"visible"},
				 undef,
				 $version,
				 );
	if (@p > 0) {
	    $Page = new ExSite::Page(page=>$p[-1]);
	    $purl = $Page->get_url($menu_mode);
	    $pdata = $Page->get;
	    unshift @path, "<a href=\"$purl\">$pdata->{label}</a>";
	}
    }

    $out .= join $config{SimpleMenu}{pmenu_spacer},@path;
    $out .= "<!-- end menu -->\n";

    return $out;
}

sub vmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}{server}$config{server}{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $pdata = $Page->get;
    my $menu_mode = $this->menu_mode;

    my $out;
    my ($menutop, $menumid, $menubot);
    $menutop = $Page->find("SimpleMenuTop")->get_html();
    if ($menutop) {
	$menumid = $Page->find("SimpleMenuMid")->get_html();
	$menubot = $Page->find("SimpleMenuBot")->get_html();
    }
    else {
	$menutop = "<!-- begin menu -->\n<span class=\"menu\">\n";
	$menumid = $config{SimpleMenu}{vmenu_spacer};
	$menubot = "</span>\n<!-- end menu -->\n";
    }

    my @items;

    my $access = $db->authorize;
    my $parent = $Page->parent;
    my $ppid = $parent->id();

    my @map = $this->{map}->get_topnodes_data();
    while (@map > 0) {
	my $p = shift @map;
	my $menuPage = new ExSite::Page(page=>$p);
	my $url = $menuPage->get_url;
	my $id = $menuPage->id();

	# test to see if we need to display the submenu
	my @submap = $this->{map}->get_child_data($id);
	my $need_submenu = ($id == $this_page); 
	if (! $need_submenu) {
	    for (my $imap = 0; $imap < scalar @submap; $imap++) {
		my $parent = $this->{map}->get_parent_data($submap[$imap]{page_id});
		if ($parent->{page_id} eq $ppid) {
		    $need_submenu = 1;
		    last;
		}
		$imap++;  # skip sub-submap
	    }
	}

	my (@submenu,$aid);
	if ($need_submenu) {
	    while (@submap > 0) {
		my $p = shift @submap;
		my $subPage = new ExSite::Page(page=>$p);
		my $url = $subPage->get_url;
		$aid = ($subPage->id() == $this_page) ? 
		    " id=\"thispage\"" : "";
		push @submenu,"$config{SimpleMenu}{submenu_prefix}<a href=\"$url\" class=\"submenu\"$aid>$p->{label}</a>";
	    }
	}
	$aid = ($id == $this_page) ? " id=\"thispage\"" : "";
	if (@submenu > 0) {
	    push @items, "<a href=\"$url\" class=\"menu\"$aid>$p->{label}</a>$config{SimpleMenu}{vmenu_spacer}\n<span class=\"submenu\">\n".join("$config{SimpleMenu}{vmenu_spacer}\n",@submenu)."\n</span>\n";
	}
	else {
	    push @items, "<a href=\"$url\" class=\"menu\"$aid>$p->{label}</a>";
	}
    }	

    # optional links for members only
    if ($db) {
	if ($Page) {
	    if (! $Page->{publish}) {
		my $level = $db->authorize;
		if ($level) {
		    push @items, "<a href=\"$CGIpath/logout.cgi\" class=\"menu\">Logout</A>";
		}
	    }
	}
    }

    return "$menutop\n".join("$menumid\n",@items)."\n$menubot";
}

sub menu_mode {
    my $this = shift;
    # always do static menus when publishing
    return 1 if ($share{Page}->{publish});
    # otherwise, use static menus only when not in CMS mode
    return &get_cookie("ExSiteCMS") ? 0 : 1;
}

sub sitemap {
    my $this = shift;
    my $out;
    my $db = $share{DB};
    my $page = $share{Page}->get;
    my @pages = $db->fetch_match("page",{section_id=>$page->{section_id},
					 type=>"page",
					 parent_id=>0,
					 visibility=>"visible"},"rank");
    $out .= "<P><UL CLASS=\"sitemap\">\n";
    foreach my $page (@pages) {
	$out .= $this->sitemap_r($page);
    }
    $out .= "</UL></P>\n";
    return $out;
}

sub sitemap_r {
    # recursive mapping routine
    my ($this,$page) = @_;
    my $db = $share{DB};
    my $pid = $page->{page_id};
    my $p = new ExSite::Page(id=>$pid);
    my $url = $p->get_url;
    my $out = "<LI><A HREF=\"$url\">$page->{label}</A>";
    if ($page->{access} ne "public") { $out .= " (restricted access)"; }
    #elsif ($page->{access} eq "public") { $out .= " ($page->{filename})"; }
    $out .= "</LI>\n";
    if ($page->{access} ne "public") { return $out; }
    my $subpage = $db->fetch_match("page",{parent_id=>$pid,
					   type=>"page",
					   visibility=>"visible"},"rank");
    if (@$subpage > 0) {
	$out .= "<UL CLASS=\"sitemap\">\n";
	foreach my $page (@$subpage) { $out .= $this->sitemap_r($page); }
	$out .= "</UL>\n";
    }
    return $out;
}

sub get_pages {
    my ($this,$match,$rank,$version) = @_;
    if ($share{Page}) {
	if ($share{Page}->is_template) {
	    # we are previewing a template; show a dummy menu
	    my $site = $share{Page}->id("section");
	    return ( 
		     { page_id=>1, label=>"Home", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"About", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"Contact", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"Links", type=>"page", access=>0, url=>"#", section_id=>$site },
		     );
	}
    }
    my $db = $share{DB};
    my @pages = $db->fetch_match("page",$match,$rank);
    if ($version) { 
	foreach my $page (@pages) {
	    my $pid = $page->{page_id};
	    my @alt = $db->fetch_match("page",{parent_id=>$pid,
					       type=>"alternate",
					       version=>$version,
					   },"page_id");
	    if (@alt > 0) {
		# alternate version of this page was found
		# swap it in place
		$page = $alt[-1];
	    }
	}
    }
    return @pages;
}

1;


