package Modules::SlideShow;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------
# SlideShow : tool to page through the images in a library
#
# Options:
#
# id=N         library id of photo album, OR
# name=ABC     name of photo album
#
# border=1|0   whether to put a border around the photos
# title=1|0    whether to title the photo album
#
# button=XXX   use button style "XXX" (black, blue, green, red) [black
#              is default if unknown style is given].  Text links are
#              used otherwise.
#
# Tip:
# <!--&&&SlideShow(...)--> uses AJAX calls to scroll through the slide
# show, resulting in a smoother refresh and faster viewing.
#-----------------------------------------------------------------------

use strict;
use ExSite::Config;
use ExSite::Input;
use ExSite::Misc;
use ExSite::Content;
use Modules::BaseDCD;

# inherit basic behaviours from BaseDCD

use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

# provide our own write() method

sub write {
    my ($this,$options) = @_;
    my $in = new ExSite::Input;
    my $q = $in->query;
    my %opt = &DecodeString($options);
    my %allopt = (%opt,%$q);
    my $db = $share{DB};
    my ($next, $curr, $prev);
    my ($album,$album_id);
    if ($allopt{id}) {
        # album passed by id#
        $album_id = $allopt{id};
        $album = $db->fetch("page",$album_id);
    }
    elsif ($allopt{name}) {
        # album passed by name
        my @lib = $db->fetch_match("page",{section_id=>$share{Page}->id("section"),
                                           filename=>$allopt{name}});
        $album = $lib[0];
        $album_id = $album->{page_id};
    }
    if (!$album) {
        # look for "images" library in current site
        my @lib = $db->fetch_match("page",{section_id=>$share{Page}->id("section"),
                                           filename=>"images"});
        $album = $lib[0];
        $album_id = $album->{page_id};
    }

    # button type
    my $button = $q->{button} || $opt{button} || $config{SlideShow}{button};

    if ($album) { 
        
        my @pics = $db->fetch_child("content","page",
                                    $album_id,"name");
        $curr = $q->{cid} || 0;
        if ($curr > $#pics || $curr < 0) { $curr = 0; }
        $next = $curr < $#pics ? $curr + 1 : 0;
        $prev = $curr > 0 ? $curr - 1 : $#pics;
        
        my $c = new ExSite::Content(content=>$pics[$curr]);
	$c->set_context();
        
        my $out = "<div class=\"SlideShow\">\n";;
        if (exists $allopt{title} ? $allopt{title} : 1) { 
            $out .= "<h1>$album->{title}</h1>\n";
        }
        $out .= "<table width=\"200\" cellspacing=\"0\" cellpadding=\"0\" border=\"0\">\n";
        if ($album->{description}) { 
            $out .= "<tr><td align=\"center\" colspan=\"3\"><p>$album->{description}</p></td></tr>\n";
        }
        $out .= "<tr><td align=\"center\" colspan=\"3\">";
        my $imgopt;
        if (exists $allopt{border} ? $allopt{border} : 1) { $imgopt = "border=\"1\""; }
        $out .= $c->get_html($imgopt);
        $out .= "</td></tr>\n<tr><td align=\"center\" colspan=\"3\"><span class=\"caption\">$pics[$curr]->{description}</span></td></tr>\n<tr><td align=\"left\" width=\"33%\">";
        if ($prev < $curr) {
            my $prevurl = $this->link(cid=>$prev);
            my $anchor = $this->anchor("prev",$button);
            $out .= "<a href=$prevurl>$anchor</a>";
        }
        $out .= "</td><td align=\"center\" width=\"33%\">\n";
        if ($curr > 0) {
            my $url = $this->link(cid=>undef);
            my $anchor = $this->anchor("back",$button);
            $out .= "<a href=$url>$anchor</a>";
        }
        $out .= "</td><td align=\"right\" width=\"33%\">";
        if ($next > $curr) {
            my $nexturl = $this->link(cid=>$next);
            my $anchor = $this->anchor("next",$button);
            if ($curr == 0) { 
                $out .= "Slide Show "; 
            }
            $out .= "<a href=$nexturl>$anchor</a>";
        }
        $out .= "</td></tr></table>\n</div>\n"; 
        return $out;
    }
    else {
        my $message = "photo album not found ($options)";
        $this->error($message);
        return "<p class=\"error\">$message</p>\n";
    }
}

sub anchor {
    my ($this,$dir,$button) = @_;
    if ($button) {
        if (-e "$config{server}{HTMLroot}$config{server}{HTMLpath}/_Modules/SlideShow/${dir}-${button}.gif") {
            return "<img src=\"$config{server}{HTMLpath}/_Modules/SlideShow/${dir}-${button}.gif\" height=\"24\" width\"24\" border=\"0\" alt=\"$dir\" align=\"middle\">";
        }
        else {
            return "<img src=\"$config{server}{HTMLpath}/_Modules/SlideShow/${dir}-black.gif\" height=\"24\" width\"24\" border=\"0\" alt=\"$dir\" align=\"middle\">";
        }
    }
    else {
        if ($dir eq "prev") { return "&lt;&lt;&nbsp;prev"; }
        elsif ($dir eq "next") { return "next&nbsp;&gt;&gt;"; }
        else { return "back to start"; }
    }
}

sub ioctl {
    my $this = shift;
    $_ = shift;
    if (/ModuleName/) {
        return "SlideShow";
    }
    elsif (/ModuleInfo/) {
        return "The SlideShow tool allows your website users to browse
through your image library, one image at a time.";
    }
}

1;
