package Modules::Trash;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------

=pod

Trash : a trashbin tool

The trashbin is a core table that stores information on all trash
calls (ie. reversible deletes).  The trashbin allows the following 
operations:

For administrators:
    - empty trash (permanently remove all trashed items)
    - restore (undelete) any trash item

For non-administrators:
    - restore (undelete) any trash item owned by the current user

When the ExSite trash methods (DB::trash_key and DB::trash_r) are 
called, the selected records are deleted, but some reconstruction
information is left behind in the trash table, allowing those
records to be re-created if necessary.  The reconstruction information
consists of a perl representation of the deleted datahashes.  These
are used to build an insert command to restore the data, if the user
so chooses.  

This technique allows us to store any number of trashed 
record(s) from any table or group of tables in a single trash record.
By default, recursive deletes of families of records are stored in a
single trash record; this allows the whole family to be restored at one
stroke, like reversing a transaction.

This also means that trash functionality will automatically extend itself
to any new tables added to the database, without any reconfiguration
needed.

By default, Trash is a simple tool which can be visited whenever the
user wishes to restore deleted records, or empty the trash to free up
database space.

=cut

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Misc;

# inherit from BaseDCD class
use Modules::BaseDCD;
use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

# write : dummy method;  this is an administrator tool only

sub write {
    my ($this,$options) = @_;
    return "";
}

# ioctl : this is a public tool

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    if (/ControlPanel/) {
	return \&ctrl_panel;
    }
    elsif (/ModuleInfo/) {
	return "The Trash manager lets you restore data that you have
previously deleted using other administration tools.  Note that the trash may
be automatically emptied periodically, so it only serves as a short-term 
emergency recovery tool, not as a long-term backup facility.";
    }
}

sub ctrl_panel {
    my $this = shift;
    my $db = $share{DB};
    my ($out,@trash);
    if ($this->{input}{undel}) {
	
	# restore a trash item

	my $trash = $db->fetch("trash",$this->{input}{undel});
	if ($trash && ($db->is_admin || $trash->{member_id} == $db->my_uid)) {
	    $db->undelete("trash",$this->{input}{undel});
	}
	else {
	    return "<h3 class=error>Permission denied</h3>\n";
	}
    }
    if ($this->{input}{empty}) {
	
	# empty the trash
	if ($db->is_admin) {
	    # ExSite doesn't let us do an unconditional
	    # delete, so we loop over each trash item individually.
	    @trash = $db->fetch_all("trash");
	    foreach my $tr (@trash) {
		$db->delete_key("trash",$tr->{trash_id});
	    }
	}
	else {
	    $db->dbdelete("trash",{member_id=>$db->my_uid()});
	}
    }

    # fetch trash items

    my $url = &relink("empty"=>1);
    my %report = ( 
		   title=>"Trash",
		   nodata=>"The trash bin is currently empty.",
		   headers=> ["Item","Deleted on","Deleted by","Options"],
		   data=>[],
		   tools=>"<a href=\"$url\">Empty the trash</a>.",
		   show_tools_if_nodata=>0,
		   );
    if ($db->is_admin) {
	@trash = $db->fetch_all("trash");
    }
    else {
	@trash = $db->fetch_match("trash",{member_id=>$db->my_uid});
    }

    # list trash items

    if (@trash > 0) {
	foreach my $tr (@trash) {
	    my @data;
	    foreach my $col qw(description ctime member_id) {
		push @data, $db->show_data("trash",$col,$tr->{$col},$tr->{trash_id});
	    }
	    my $restore_url = $this->link(undel=>$tr->{trash_id});
	    push @data, "<a href=\"$restore_url\">click here to restore this data</a>";
	    push @{$report{data}},\@data;
	}
    }
    my $r = new ExSite::ReportBuilder(%report);
    return $r->make;
}

1;

