package myConfig;
use strict;
use ExSite::Misc;
use Local;

=pod

=head1 myConfig.pm

B<myConfig.pm> - website localization.

C<myConfig.pm> contains customizations to the ExSite library for use
on a particular website.  There are four standard routines that should
be present in every instance of myConfig.pm:

=cut

# Exported Names

BEGIN {
    use Exporter;
    use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    @ISA = qw(Exporter);
    @EXPORT_OK   = ( );
    %EXPORT_TAGS = ( );
    @EXPORT      = qw(
		      &my_exsite_init &my_exsite_close
		      &my_handlers
		      &my_page_header &my_page_footer
		      );

}

use vars (@EXPORT, @EXPORT_OK);

#=====================================================================
# Non-Exported names:
#
# We cannot import %ExSite::Config::config, because it would create
# a circular reference.  Instead, we just copy it into this module
# using &my_exsite_init().  Ditto for %share.

use vars qw(%config $config $share);

#=================================================================
# my_exsite_init

=pod

=head2 my_exsite_init($)

C<my_exsite_init> is called directly from the standard ExSite
initialization routine C<exsite_init>.  Site-specific setup that needs
to be performed for all or most CGI programs may be performed here.

=cut

sub my_exsite_init {
    ($config, $share) = @_;
    %config = %$config;
    # custom setup here
}

sub my_exsite_close {
    # dummy routine - add custom cleanup here
}

#=================================================================
# my_page_header

=pod

=head2 my_page_header($$$)

C<my_page_header> generates the stock HTML that precedes the page's
core content, such as masthead, banners, menus, etc.  It is
automatically invoked by the standard ExSite page_header routine.  It
should return a simple scalar containing the HTML that comprises the
leading part of the page.

=cut

sub my_page_header {
    my ($title,$opt) = @_;
    # dummy - replace with your own page header
    return undef;
}

#=================================================================
# my_page_footer

=pod

=head2 my_page_footer($$)

C<my_page_footer> generates the stock HTML that follows the page's
core content, such as sponsors, affiliates, copyright notices, etc.
It is automatically invoked by the standard ExSite page_footer
routine.  It should return a simple scalar containing the HTML that
comprises the trailing part of the page.

=cut

sub my_page_footer {
    my ($opt) = @_;
    # dummy - replace with your own page header
    return undef;
}

#=================================================================
# handlers

=pod

=head2 my_handlers($$)

C<my_handlers> installs any special handler routines that are used by
this website to override default ExSite logic.  It is called
automatically by the database (DB), Report, and Form constructor
methods.

C<my_handlers> should contains lines of the form:

C<$db-E<gt>handler("handler_name",\&handler_routine);E<GT>>

This installs a handler, C<&handler_routine>, to be invoked when the
code checks if a handler is installed under "handler_name".
C<&handler_routine> I<must be defined and installed in myConfig.pm
following the my_handlers routine.> The handler routine itself must
conform in calling arguments and output to the expectations of the
ExSite code that invokes the handler.

The ExSite Handlers documentation should be consulted for more information.

=cut

sub my_handlers {
    my $db = $share->{DB};
    if ($db) {
	# install handlers here
	$db->handler("select_foreign_key",\&custom_select_foreign_key);
	#$db->handler("pre_update",\&custom_pre_update);
    }
}

sub custom_select_foreign_key {
    my ($this, $table, $rec, $col, $ftable) = @_;
    my $rdata;

    ### CMS handlers

    if ($this->{name} eq $config{server}{db}{name}) {
	if ($table eq "page") 
	{
	    if ($col eq "template_id") 
	    {
		# select only from templates
		return $this->fetch_match("page",{type=>"template"});
	    }
	    elsif ($col eq "parent_id") 
	    {
		my (@possible_parents,$page,@parents);
		# select only from pages in same section
		$page = $this->fetch("page",$rec);
		@parents = $this->fetch_match("page",{type=>"page",section_id=>$page->{section_id}});
		# don't allow user to select a page as its own parent
		foreach my $parent (@parents) 
		{
		    if ($parent->{page_id} ne $page->{page_id}){push @possible_parents, $parent;}
		}
		return \@possible_parents;
	    }
	}
	if ($table eq "content")
	{
	    if ($col eq "page_id")
	    {
		my ($content,$page);
		$content = $this->fetch("content",$rec);
		$page = $this->fetch("page",$content->{page_id});
		return $this->fetch_match("page",{section_id=>$page->{section_id}});
	    }
	}
	if ($table eq "section")
	{
	    if ($col eq "parent_id")
	    {
		my $sql = "select * from section where type != 'website template'";
		$sql .= ($rec =~ /^\d*$/) ? " and section_id != $rec" : "";
		return $this->custom_query($sql);
	    }
	}
    }
    return;
}
1;
