#!/usr/bin/perl
#-----------------------------------------------------------------------
# publish.pl - write public content to disk
#
# Typically run via the publish.c[gi] setuid wrapper, to obtain the 
# necessary write permissions.
#
# USAGE:
#    publish.cgi?page=xxx - publishes page id XXX
#    publish.cgi?section=xxx - publishes section id XXX
#    publish.cgi?content=xxx - publishes content id XXX
#    publish.cgi?module=xxx - publishes plug-in module XXX
#    publish.cgi?-page=xxx - unpublishes page id XXX
#    publish.cgi?-section=xxx - unpublishes section id XXX
#    publish.cgi?-content=xxx - unpublishes content id XXX
#    publish.cgi?-module=xxx - unpublishes plug-in module XXX
#
#    The parameter slient=1 causes publish.pl to run silently.
#    This is useful when invoking publish.cgi indirectly to write
#    files to disk.  Otherwise it ends up writing to the top of
#    the current page.
#
# "publish" creates files
# "unpublish" removes files (presumably ones that were previously published)
#
#  For pages and sections, all publication data is obtained from the 
#  database.  For content, publication data can be obtained from the
#  database, or from the POST data (used when files are written 
#  directly to disk, instead of being stored in the database).  For
#  Modules, publication data is determined at the discretion of the
#  module.
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

# KLUDGE!  Include local dir in path to get ExSite libraries.
# This makes the path insecure for setuid execution of other programs.
BEGIN{ if (grep ( /^\.$/, @INC ) == 0) { push @INC,"."; } }

use ExSite::Config;
use ExSite::Form;
use ExSite::Page;
use ExSite::Section;
use ExSite::Input;
use ExSite::Auth;

&exsite_init;
my $db = new ExSite::Form;
my $in = new ExSite::Input;
my $auth = new ExSite::Auth;
my %query = %{$in->query};
my $delay = 3;                           # seconds to refresh

if ($query{silent}) {
    # run silently, so we don't contaminate another page's output

    # save old streams
    open(OLDOUT, ">&STDOUT");
    open(OLDERR, ">&STDERR");
    # redirct streams to /dev/null
    open(STDOUT, '>', "/dev/null") || die "Can't redirect stdout";
    open(STDERR, ">&STDOUT")       || die "Can't dup stdout";
    # make unbuffered
    select(STDERR); $| = 1;
    select(STDOUT); $| = 1;
} 

print "Content-type: text/html\n";
if ($db->authorize < 2) {
    print "\n<p>Permission denied.</p>\n";
}
elsif ($query{page}) {
    # publish one page
    my $p = new ExSite::Page(id=>$query{page});
    # redirect to published page when done
    my $url = $p->get_url(1);
    if ($p->is_page) {
	print "Refresh: ${delay}; URL=$url\n";
    }
    print "\n";
    if ($auth->owns($p->get("page","section_id"))) {
	print "<H4>Publishing Page $query{page}</H4>\n";
	$share{Page} = $p;
	print $p->publish;
	if ($p->is_page) {
	    print "<p>(Redirecting to $url in $delay seconds...  
Click 'Reload' if you still see the old page.)</p>\n";
	}
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to publish that page.</p>\n";
    }
}
elsif ($query{'-page'}) {
    # unpublish one page
    my $p = new ExSite::Page(id=>$query{'-page'});
    if ($auth->owns($p->get("page","section_id"))) {
	print "\n<H4>Removing $p->{data}{type} $query{-page}</H4>\n";
	print $p->unpublish;
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to remove that page.</p>\n";
    }
}
elsif ($query{section}) {
    # publish whole section
    if ($auth->owns($query{section})) {
	my $s = new ExSite::Section(id=>$query{section});
	# redirect to published home page
	my $url = $s->get_url;
	print "Refresh: ${delay}; URL=$url\n\n";
	print "<H4>Publishing Site $query{section}</H4>\n";
	# section publish prints directly for better flushing of output
	$s->publish;
	print "<p>(Redirecting to $url in $delay seconds...)</p>\n";
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to publish that site.</p>\n";
    }
}
elsif ($query{'-section'}) {
    # unpublish whole section
    if ($auth->owns($query{-section})) {
	my $s = new ExSite::Section(id=>$query{'-section'});
	print "\n<H4>Removing Site $query{-section}</H4>\n";
	# section unpublish prints directly for better flushing of output
	$s->unpublish;
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to publish that site.</p>\n";
    }
}
elsif ($query{content}) {
    # publish one content object
    my $c = new ExSite::Content(id=>$query{content});
    $c->set_context;
    print "\n";
    if ($auth->owns($c->get("page","section_id"))) {
	print "<H4>Publishing Content $query{content}</H4>\n";
	print $c->publish;
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to publish that content.</p>\n";
    }
}
elsif ($query{'-content'}) {
    # unpublish one content object
    my $c = new ExSite::Content(id=>$query{'-content'});
    $c->set_context;
    print "\n";
    if ($auth->owns($c->get("page","section_id"))) {
	print "\n<H4>Removing Content $query{-content}</H4>\n";
	print $c->unpublish;
    }
    else {
	print "\n<h1>Permission Denied</h1><p>You do not have authorization to remove that content.</p>\n";
    }
}
elsif ($query{module}) {
    # publish plug-in
    print "\n";
    $query{module} =~ /^(\w+)$/;
    my $mod = $1;
    if ($mod) {
	my $stat = eval "require Modules::$mod";
	if ($stat) {
	    $dcd = eval "new Modules::$mod";
	    if ($dcd) { 
		my $publisher = $dcd->ioctl("Publish");
		if (ref $publisher eq "CODE") {
		    # check licensing
		    my $license_ok = 1;


# 	    if ($dcd->ioctl("isRestricted")) {
#		# this service requires licenses
#		my @licenses = $share{DB}->fetch_match("service",
#						       {section_id => $this->id("section"),
#							name => $modname }
#						       );
#		$license_ok = (@licenses > 0);
#		if (! $license_ok) {
#		    # log the problem
#		    $this->warn("$modname: unlicensed access from page ".$this->id("page"));
#		}
#	    }
		
		
		    if ($license_ok) {
			# publish
			print $dcd->$publisher || $@;
		    }
		    else {
			print "$mod: permission denied (no license)";
		    }
		}
		else {
		    print "$mod: this module does not publish";
		}
	    }
	    else {
		print "$mod: $@";
	    }
	}
	else {
	    # $@ is $EVAL_ERROR
	    print "<H4>Error in $mod dynamic content driver:</H4>\n<P>\n$@\n</P>";
	    $dcd->error("$mod: $@");
	}
    }
    else {
	print "$query{module}: invalid module";
    }
}
elsif ($query{'-module'}) {
    # remove plug-in files
    print "\n";
    $query{module} =~ /^(\w+)$/;
    my $mod = $1;
    if ($mod) {
	my $stat = eval "require Modules::$mod";
	if ($stat) {
	    $dcd = eval "new Modules::$mod";
	    if ($dcd) { 
		my $publisher = $dcd->ioctl("Unpublish");
		if (ref $publisher eq "CODE") {
		    # check licensing
		    my $license_ok = 1;

# 	    if ($dcd->ioctl("isRestricted")) {
#		# this service requires licenses
#		my @licenses = $share{DB}->fetch_match("service",
#						       {section_id => $this->id("section"),
#							name => $modname }
#						       );
#		$license_ok = (@licenses > 0);
#		if (! $license_ok) {
#		    # log the problem
#		    $this->warn("$modname: unlicensed access from page ".$this->id("page"));
#		}
#	    }
		
		
		    if ($license_ok) {
			# publish
			print $dcd->$publisher || $@;
		    }
		    else {
			print "$mod: permission denied (no license)";
		    }
		}
		else {
		    print "$mod: this module does not unpublish";
		}
	    }
	    else {
		print "$mod: $@";
	    }
	}
	else {
	    # $@ is $EVAL_ERROR
	    print "<H4>Error in $mod dynamic content driver:</H4>\n<P>\n$@\n</P>";
	    $dcd->error("$mod: $@");
	}
    }
    else {
	print "$query{module}: invalid module";
    }
}
else {
    print "\n<H4>Error: nothing to publish! $ENV{QUERY_STRING}</H4>\n";
}

print "<div align=center>
<table cellpadding=10 cellspacing=0 border=0><tr><td>
<a href=javascript:window.close()>Close Window</a>
</td></tr></table>
</div>\n";

my @err = $db->fetch_diagnostics("error");
my @warn = $db->fetch_diagnostics("warn");
#print $db->show_diagnostics("info","html");

if (scalar @err + scalar @warn > 0) {
    print "<h4>Diagnostic Information:</h4>\n";
    print "<p>Errors:</p>\n<ol><li>",join("\n<li>",@err),"</ol>\n";
    print "<p>Warnings:</p>\n<ol><li>",join("\n<li>",@warn),"</ol>\n";
}

if ($query{silent}) {
    # restore normal I/O streams
    close(STDOUT);
    close(STDERR);
    open(STDOUT, ">&OLDOUT");
    open(STDERR, ">&OLDERR");
}

&exsite_close;
